/*
** Module   : ThinClient.java
** Abstract : Thin Client class for generic UI driver implementation
**
** Copyright (c) 2005-2009, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050710   @21707 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 EVL 20050822   @22196 Adding code to have the client and server on
**                           the same machine.
** 003 EVL 20050824   @22239 Adding the events/triggers processing.
** 004 EVL 20050825   @22275 Adding GO and CHOOSE button handlers.
** 005 EVL 20050827   @22375 Adding usage of the ScreenBuffer to handle
**                           the events.
** 006 EVL 20050830   @22416 Adding ENTRY and LEAVE events handling.
** 007 SIY 20050831   @22519 Refactoring (see @@22394-22409) 
** 008 EVL 20050903   @22457 Adding KEY event handler. (Merged changes
**                           from EVL) 
** 009 EVL 20050907   @22598 Getting the CHARVA key code while
**                           interception of the key events.
** 010 EVL 20050908   @22617 Using single handler callback for all events.
**                           The key code is stored in ScreenBuffer for
**                           key events.
** 011 EVL 20050912   @22682 Adding the validation list and the validation
**                           method.
** 012 EVL 20050913   @22701 Adding debug info message for the case when
**                           it is not possible to identify the widget ID.
**                           Fixing focus event handling for radio-set.
** 013 EVL 20050916   @22740 Changing to use BaseDataType class as the
**                           value of the widget to be placed in screen
**                           buffer.
** 014 EVL 20050919   @22751 Modification of the key handler to be able to
**                           intercept the events from radio set,
**                           selection list and combo box.
** 015 EVL 20050920   @22763 Adding code to check whether the value of the
**                           widget to be validated is really changed or
**                           not and the validation is not necessary.
** 016 EVL 20050921   @22768 Adding the usage of the undoValue() method to
**                           roll back the widget change for failed
**                           validation.
** 017 EVL 20050922   @22815 Adding the interceptor for editing block
**                           mode.
** 018 EVL 20050926   @22856 Including the value into the screen buffer
**                           via EDITING block handling.
** 019 NVS 20051204   @24062 Added trigger and validate methods to this
**                           class that call appropriate server exports to
**                           execute triggers and validators.
** 020 SIY 20051204   @24063 Added stubs for ClientExports methods and 
**                           many other changes.    
** 021 NVS 20051229   @24064 Added consumeEvent() implementation.        
** 022 NVS 20060111   @24065 apply() now returns ScreenBuffer to server.
** 023 GES 20060113   @23901 Improvements to message (it still doesn't
**                           support any options).  Now it handles arrays
**                           of different object types including 
**                           BaseDataType.
** 024 NVS 20060114   @24066 hide() export hides widgets as well as frames
** 025 GES 20060116   @23924 Safety code in some widget/frame lookup
**                           methods.
** 026 GES 20060116   @23927 Rework and additions to putScreen() to make
**                           it functional.
** 027 GES 20060116   @23928 Fixed readKey().  It never broke out of the
**                           timing loop when it got a key, it just kept
**                           looping.
** 028 GES 20060116   @23932 view() needed to request focus otherwise the
**                           previous frame (even if hidden) still had
**                           focus.
** 029 GES 20060116   @23948 Fixed pause(0) and cleaned up signature.
** 030 NVS 20060119   @23979 Implemented correct Pause Before Hide logic
**                           for view, clear and hide operations. Other
**                           situations causing frame hide like Z-order
**                           change still need some work.
** 031 GES 20060121   @24011 Moved core apply logic to a worker that
**                           doesn't return a screen buffer since the
**                           vast majority of times that apply() is
**                           called it is from the waitFor loop which
**                           doesn't need the screen buffer returned.
**                           This way the only case where we take the
**                           hit of building the screen-buffer is the
**                           case where the server calls apply().
** 032 NVS 20060125   @24067 Implemented iterateFrame(), down(), up() and
**                           scroll() methods that manage the cursor and
**                           contents of down frame.
** 033 SIY 20060125   @24123 Support code for the frame manager. 
** 034 NVS 20060130   @24147 All client and server exports returning 
**                           screen buffer, now take screen buffer as
**                           well. These are: waitFor(), apply(),
**                           trigger() and validate.
** 035 SIY 20060130   @24152 Added pauseBeforeEnd() exported method.
** 036 SIY 20060131   @24191 Extended MESSAGE support and merged 
**                           status/help handling.
** 037 NVS 20060131   @24192 view() call should bring the frame on top.
** 038 NVS 20060201   @24217 Hide details of TempText creation in putScreen
**                           inside the TempText class.
** 039 NVS 20060202   @24218 getScreenLines() returned screen height, not
**                           the height of frame display area.
** 040 SIY 20060203   @24242 Fixes for hiding/placing frames and pause
**                           management.
** 041 NVS 20060206   @24307 Using TypeAhead class for reading events,    
**                           clearing the type ahead buffer and catching
**                           Ctrl-C (STOP).
** 042 NVS 20060209   @24371 Using Control.interrupt() to deliver Ctrl-C.
** 043 SIY 20060209   @24436 Fixes to make ComboBox, SelectionList, 
**                           RadioSet controls handle global keys.
** 044 GES 20060210   @24507 Simplified exporting of methods.
** 045 SIY 20060210   @24521 Added getSelection() method; fixed trap if
**                           typing is started before first window does 
**                           appear on the screen.
** 046 GES 20060210   @24532 frame-field attribute must return empty
**                           string when null.
** 047 GES 20060211   @24541 Added initialization for remote file system
**                           and remote stream services on the client.
** 048 GES 20060212   @24559 Added initialization for process daemon.
** 049 SIY 20060213   @24578 Support for DialogBox; fixes for the frame
**                           hiding.
** 050 NVS 20060213   @24585 Added installation of ProgressKeyboardMapper.
** 051 GES 20060212   @24588 Exposed terminal suspend/resume and pseudo-
**                           terminal process launching mechanism.
** 052 SIY 20060215   @24598 Implemented messageBox(). Some performance 
**                           improvements.
** 053 NVS 20060221   @24626 STOP key function raises STOP condition. HELP
**                           key function executes Help Application on the
**                           server.
** 054 SIY 20060220   @24645 Implemented choose() support. 
** 055 NVS 20060222   @24667 ENDKEY key action should be used for this
**                           event, not the END key.
** 056 GES 20060214   @24686 Stream redirection and some general cleanup.
** 057 NVS 20060224   @24728 ThinClient tracks the high level I/O
**                           operations for proper END-ERROR handling.
**                           Added new field "interactions". It is set to
**                           0 in blockStart() and blockIterate() and
**                           incremented by 1 on entry to waitFor(). END
**                           condition is generated if interactions equals
**                           1.
** 058 GES 20060225   @24734 Save/restore visible state when switching
**                           between interactive and redirected output
**                           modes. In redirected mode, everything on
**                           the interactive screen must be set hidden
**                           so that it doesn't draw to the stream. Also
**                           added more APIs into forced interactive mode.
** 059 GES 20060226   @24756 Header and footer support.
** 060 SIY 20060226   @24765 Fixed handling of SET/UPDATE in message().
**                           Fixed focus management in messageBox().
** 061 GES 20060227   @24774 Final header and footer support.
** 062 NVS 20060302   @24848 processEvents() export now takes and returns
**                           the ScreenBuffer.
** 063 NVS 20060302   @24872 Keyboard events are translated into
**                           respective key actions whenever possible,
**                           then actions are recognized by their codes.
** 064 NVS 20060306   @24902 Added call to ErrorManager.initClient.
** 065 NVS 20060306   @24907 WaitList does not need default events, they
**                           are recognized by ThinClient anyway.
** 066 NVS 20060308   @24969 Using new ScreenBuffer which keeps state and
**                           allows for differential exchange.
** 067 NVS 20060314   @25072 Fixed setScreenBuffer() to transfer the state
**                           to widgets for all widgets.
** 068 SIY 20060315   @25067 Yet another try to fix frame hiding.
** 069 SIY 20060316   @25142 Added support for headers.
** 070 SIY 20060228   @25143 Changed signature for the getColumns() and
**                           other methods.
** 071 NVS 20060320   @25159 Simplified validation protocol by using new
**                           class ValidationResult.
** 072 SIY 20060320   @25167 Fixed clear() on headers.
** 073 GES 20060322   @25186 Added terminal type processing methods.
** 074 NVS 20060322   @25200 Added refocusing to the specified widget in
**                           waitFor().
** 075 NVS 20060322   @25211 Implemented widget value validation during
**                           all kinds of blocking input.
** 076 SIY 20060323   @25215 Changes to support new down frame 
**                           implementation. As a side effect refactored 
**                           some encapsulation violations.
** 077 NVS 20060324   @25236 Fixed the event processing for the focus
**                           change related events. Now they are properly
**                           routed and handled. This fix enabled LEAVE
**                           and ENTER events generation and, therefore,
**                           the changed widget validation.
** 078 SIY 20060327   @25279 Pass focus events to widgets.
** 079 SIY 20060401   @25308 Access header flag directly.
** 080 SIY 20060412   @25598 Don't hide messages when frame iterates.
** 081 SIY 20060424   @25688 Fixed pausing logic and messages are set 
**                           hidden where appropriate, without breaking
**                           pausing logic.
** 082 GES 20060501   @25907 Don't return a frame marked for destruction
**                           as an active frame as this causes problems
**                           on the server.
** 083 GES 20060502   @25928 Safer pausing logic.
** 084 SIY 20060502   @25930 Mark frames as such that had input in all
**                           cases when there was a waiting for a key when
**                           this frame was a top Z-order frame.
** 085 GES 20060505   @25984 Scoped interactions member to a given block 
**                           such that enclosing block's interactions
**                           value is restored when more deeply nested
**                           blocks end. Added a mechanism to raise the  
**                           error condition when in an UPDATE and the
**                           END-ERROR key is pressed.  Previously this
**                           was hard coded to the ENDKEY condition but
**                           while the UPDATE should exit (normally it
**                           retries on error) it must also sometimes be
**                           properly delivered as an ERROR condition for
**                           the containing logic to be correct.
** 086 SIY 20060504   @25987 Another attempt to handle frame hiding
**                           properly.
** 087 SIY 20060505   @25990 Check for null frame in setScreenBuffer().
** 088 SIY 20060509   @25990 Fixed hideAll().
** 089 NVS 20060509   @26035 Partial implementation of underline() export.
**                           For regular frames, it clears the frame as it
**                           should. For down frames, it moves the cursor
**                           down without causing underlining to appear.
** 090 GES 20060510   @26047 hideAll() needs to clear all temp text
**                           widgets (put screen output is all removed).
** 091 GES 20060510   @26059 New flag passed on blockIterate().
** 092 SIY 20060510   @26063 More fixes in hiding/pausing.
** 093 SIY 20060511   @26075 Fixed hiding/pausing, hide() now clears 
**                           messages as it does in Progress.
** 094 NVS 20060511   @26087 Reimplemented conditional down() and up()
**                           calls in GenericFrame. Now block iteration 
**                           issues a conditional down() on the server and
**                           that call is entirely handled there.
** 095 SIY 20060516   @26159 hideAll() explicitly sets needPause to true.
** 096 GES 20060517   @26214 tinyInput always must have left aligned data. 
** 097 SIY 20060517   @26217 Commented out redundant call to down() in 
**                           iterateFrame() because it causes layout 
**                           issues now.
** 098 SIY 20060517   @26233 Added handling for block retry. 
** 099 GES 20060522   @26431 Ensure that any exit from waitFor() will
**                           still trip cleanup, including via exception.
** 100 GES 20060523   @26441 Refactored logic in continueEditing() to
**                           move cleanup processing into a common method
**                           stopEditingMode() which is also exposed as
**                           a client export (so it can be called directly
**                           from the server).
** 101 GES 20060523   @26448 Limited validation processing to only those
**                           widgets that are enabled.
** 102 SIY 20060524   @26458 Fixed getFrameLine() method.
** 103 SIY 20060525   @26537 Another attempt to fix pause before exit 
**                           handling.
** 104 SIY 20060526   @26547 More careful tracking of differences in 
**                           handling of F1/F4 exits in CHOOSE and 
***                          WAIT-FOR.
** 105 NVS 20060526   @26584 First attempt to display frame completes its
**                           realization. Realized frame won't be subject
**                           to layout or screen placement changes. 
** 106 GES 20060529   @26677 Added support for condition processing in
**                           pause processing (e.g. END-ERROR).
** 107 NVS 20060531   @26827 Fixed getFrameLine() method to return a right
**                           1-based value.
** 108 GES 20060531   @26810 Changes to make enable() maintain tab order
**                           and handle focus changes as needed.
** 109 NVS 20060601   @26896 Another fix to getFrameLine() to properly
**                           report 0 as the frame line before any output
**                           has occurred on the down frame.
** 110 SIY 20060601   @26908 Redesign of pause handling, many other minor 
**                           updates and fixes.
** 111 NVS 20060601   @26939 Fixed a regression in pause() caused by some
**                           new code ignoring the fact waitForKey() can
**                           return null.
** 112 NVS 20060608   @27018 Fixed a bug in button LEAVE event handling
**                           causing a class cast exception.
** 113 SIY 20060607   @27100 Changed blockIterate/blockStart to support 
**                           reliable handling of exceptions thrown in 
**                           pause processing.
** 114 SIY 20060609   @27101 Fixed another hiding issue. 
** 115 NVS 20060609   @27113 waitFor processing now differentiates the
**                           events in the WaitList as triggers vs exit
**                           events.
** 116 SIY 20060612   @27119 hide() should not hide messages, only
**                           pause() does that, if necessary.
** 117 NVS 20060612   @27142 Fixed many NullPointerExceptions in trigger
**                           handling code for ANYWHERE triggers with no
**                           frame involved.
** 118 GES 20060613   @27160 Reset interactions on block retry to ensure
**                           F4 processing is correct (otherwise it may
**                           generate an error where an end should occur).
** 119 NVS 20060613   @27172 Reimplemented getActiveFrame() based on a new
**                           method in Screen returning the top frame. 
**                           Fixed waitFor loop checking for events (now
**                           both the key codes and events are checked).
** 120 GES 20060614   @27191 Converted the status related signatures to
**                           String parameters instead of character. Fixed
**                           all status*() forms to output the latest
**                           status message to the terminal at the time
**                           of the call.
** 121 GES 20060614   @27197 Apply processing must temporarily create a
**                           default wait list if one is not already
**                           active.
** 122 SIY 20060616   @27345 Do not set pause flag if frame is not visible.
** 123 SIY 20060616   @27399 Another round of pausing/hiding fixes. 
**                           All calls to pauseBeforeHide() made exception
**                           safe.
** 124 SIY 20060617   @27421 Fixed typo in comments. 
** 125 NVS 20060616   @27431 CHOOSE event generation off of SELECTED state
**                           of ItemEvent.
** 126 GES 20060619   @27453 All screen buffers that are passed to the
**                           server (by a return value from a method call
**                           from the server or via a parameter when
**                           calling the server) must be properly matched
**                           up with the current widget being processed.
**                           The previous implementation of the
**                           getScreenBuffer() method returned the screen
**                           buffer associated with the used the currently
**                           focused widget.  This almost always worked
**                           but there were cases where apply() could be
**                           operating on a different frame than would be
**                           detected via the current focus.  This must
**                           never be allowed to happen as it corrupts
**                           the server's resulting screen buffer when
**                           the contents are merged on the server side.
** 127 NVS 20060621   @27511 Rewritten the processProgressEvent() method
**                           to straighten up the logic and implement some
**                           missing parts like CURRENT-WINDOW related
**                           triggers and exit conditions. New code makes
**                           use of the new optimized WaitListClient class
**                           methods.
** 128 NVS 20060622   @27527 Fixed a NullPointerException regression in
**                           processProgressEvent() caused by a too broad
**                           check when there is no enabled widget on the
**                           screen.
** 129 SIY 20060621   @27532 More changes in frame hiding algorithms.
** 130 GES 20060623   @27558 Return lastKey for message (if it paused),
**                           choose and pause.  Added safety code to
**                           ensure that callers of waitForWorker that
**                           need post processing use try/finally to
**                           ensure it occurs.
** 131 NVS 20060623   @27567 Fixed the focus change related event handling
**                           and provided a synchronous LEAVE event that
**                           activates a trigger and can be consumed by
**                           RETURN NO-APPLY which cancels the focus
**                           change.
** 132 NVS 20060626   @27570 Rewritten nextPrompt() method. Now it saves
**                           the widget ID for later use in waitFor(). The
**                           use is one time and only if the waitFor()
**                           does not have a specific widget to focus.
** 133 NVS 20060627   @27591 enable() call changes the focus immediately
**                           if it's disabling the widget in focus. All
**                           enabling calls calculate the new focus widget
**                           and store its ID for later use in waitFor()
**                           instead of immediate focus change. view()
**                           calls don't change focus either. Only waitFor
**                           method changes the focus for all delayed
**                           requests, including those made with calls to
**                           nextPrompt().
** 134 SIY 20060626   @27586 Shifted frame hiding from blockIterate() to
**                           placeFrameInt() to enable more granular 
**                           condition control.
** 135 GES 20060627   @27606 Signature change to pass a screen buffer
**                           to enable() to allow truly dynamic down
**                           frame sizing.
** 136 SIY 20060628   @27662 Support for dynamic down support.
** 137 NVS 20060629   @27705 Backed out the change #131 as having too much
**                           trouble with lost synchronism. Another way
**                           of controlling the focus change centrally
**                           will be implemented.
** 138 GES 20060629   @27709 Now that enable() takes a screen buffer 
**                           parameter it also needed to honor the changes 
**                           that were sent otherwise they would not be 
**                           displayed as expected (since ENABLE stmt
**                           is supposed to be a superset of VIEW). A
**                           previous bug on the server side sent the
**                           changes over multiple times, masking the
**                           problem.  When that was fixed, the bug here
**                           became visible.
** 139 GES 20060630   @27738 choose() must process a screen buffer on
**                           both input and output. Also the frameIndex
**                           and frameValue processing was made a bit
**                           safer (though they are still not handled 
**                           properly in non-CHOOSE situations).
** 140 NVS 20060629   @27769 Added static tryFocusChange() method which is
**                           a correct way of changing focus synchronously
**                           between widgets of the active frame while
**                           maintaining the Progress semantics of LEAVE
**                           and ENTER events and the ability of triggers
**                           to cancel the focus change by returning
**                           NO-APPLY. Inspected the existing requestFocus
**                           calls within this class. See details in 27728
** 141 NVS 20060703   @27804 Added static nextFocus() and previousFocus() 
**                           methods which are variations of the focus
**                           change methods providing Progress semantics
**                           through LEAVE and ENTRY events. New methods
**                           are used throughout P2J widgets code.
** 142 NVS 20060705   @27810 Fixed: initial ENTRY event handling; null
**                           ScreenBuffer in trigger() that may be returned
**                           if trigger was not recognized on the server;
**                           getScreenBuffer() method wrong buffer size.
**                           ENABLE now maintains its own last focused
**                           widget in a variable.
** 143 NVS 20060705   @27831 Fixed a regression in waitForWorker() caused
**                           by the fix #142.
** 144 SIY 20060704   @27836 Use null instead of empty string to mark
**                           unused message line.
**                           Added checking of keys before passing them 
**                           down to widgets.
**                           Added clearing of the messages before pausing
**                           new messages if needPause is set to false;
**                           Now CHOOSE does not change status line (match
**                           Progress behavior).
** 145 GES 20060706   @27863 getFrameRow() now returns a 1-based index.
** 146 GES 20060706   @27870 Explicitly disable mouse processing in
**                           CHARVA.
** 147 NVS 20060706   @27875 Fixed choose() method by defining a unique
**                           operation code for waitForWorker. Termination
**                           of CHOOSE now works properly paying attention
**                           to supplied GO-ON events.
** 148 GES 20060707   @27887 Support the implicit hide message that occurs
**                           during message processing while inside a 
**                           trigger or some code invoked by a trigger.
**                           To accommodate this, we now track the nesting
**                           (recursion) level of triggers.
** 149 GES 20060707   @27908 Honor dynamic title expressions passed via
**                           a screen buffer.  Set the title only while
**                           the frame is hidden.
** 150 GES 20060710   @27931 Fix for hide processing during triggers.
** 151 SIY 20060707   @27930 Reset cursor position on view()/enable().
** 152 GES 20060711   @27986 Clear the current focus on disable.
** 153 NVS 20060713   @28028 All client exports that cause any drawing in
**                           UI now provide invalidation rectangle control
**                           in order to optimize drawing. On entry,
**                           invalidation is turned on. On exit, it is
**                           turned off and invalidation rectangle flushed
**                           although the invalidation mode may still be
**                           on (recursion in triggers).
** 154 GES 20060714   @28059 Apply processing must properly handle a null
**                           screen buffer for certain "current-window"
**                           operations to work properly.
** 155 SIY 20060719   @28146 Simulate input status in choose to fix help
**                           processing in CHOOSE. Added support for 
**                           returning frame-value and other relevant data 
**                           in waitFor(). 
** 156 GES 20060726   @28204 Moved interactions counter maintenance to
**                           the server.  The state sync processing
**                           ensures that the latest version is always
**                           available here.
** 157 NVS 20060724   @28215 Added popTrash() method that trashes all
**                           events in the current queue and then pops it
**                           off the stack. Applied popTrash() to the
**                           hideAll() export which now clears the screen
**                           directly. Added display control flags and
**                           used them in view(), scroll() and down().
**                           Flags control the frame decorations drawing
**                           and data iteration display for down frames in
**                           both interactive and redirected mode.
** 158 GES 20060728   @28221 Moved conditional up/down processing to be
**                           driven by the state synchronization process  
**                           to avoid an extra trip to the client.
** 159 NVS 20060728   @28222 tinyInput used to leave focus control vars
**                           in a bad state. Fixed by resetting those vars
**                           inside waitForWorker instead of waitFor.
** 160 SIY 20060726   @28227 Complete rewrite of frame hiding algorithm. 
**                           Many other minor changes.
** 161 GES 20060731   @28242 Deferred processing of conditional up/down
**                           until the piggybacked client export is
**                           invoked.  Otherwise any pausing that
**                           generates a condition exception will cause
**                           the client to abend instead of forwarding
**                           the exception back to the server. Added
**                           state sync for markFrames() and
**                           destroyFrame().
** 162 SIY 20060731   @28251 Frame cleanup changes.
** 163 NVS 20060731   @28261 Removed some unnecessary repainting related
**                           code from status(), which draws directly.
** 164 NVS 20060801   @28286 Removed even more unnecessary code from
**                           status().
** 165 NVS 20060801   @28292 Removed some unnecessary code around the call
**                           to leaveChoose().
** 166 NVS 20060802   @28332 statusInput() and statusDefault() exports
**                           use deferred sync to refresh the status line
**                           thus avoiding otherwise possible partial
**                           screen updates and flashing.
** 167 SIY 20060802   @28335 More careful resetting of frame in certain
**                           cases.
** 168 NVS 20060802   @28357 putScreen() is drawn directly now instead of
**                           causing the whole tree to redraw.
** 169 EVL 20060804   @28422 The method to hide/show cursor is completely
**                           implemented. The member cursorVisible is now
**                           added to remember the cursor status. Also
**                           when we need to explicitly put the cursor at
**                           some row,column position the cursor must be
**                           turned on if it not the case.
** 170 GES 20060807   @28447 Make the cursor visible during editing
**                           operations.
** 171 GES 20060807   @28450 Moved message list content processing to
**                           server to reduce traffic over the protocol.
** 172 SIY 20060807   @28452 Updated frame hiding algorithm and other 
**                           changes.
** 173 GES 20060808   @28488 Shifted APIs to use Java primitives instead
**                           of the Progress compatible wrappers, where
**                           possible.  Moved the wrapper processing to
**                           the server from the client. Major fixes to
**                           messageBox() to properly handle initial focus
**                           and returning unknown value. Also added
**                           condition processing to the event handling.
** 174 GES 20060808   @28496 Use int instead of long for the backing 
**                           methods that support the server exports.
** 175 SIY 20060810   @28520 Added handling for explicit setting/resetting
**                           of global canHide flag.
** 176 NVS 20060811   @28554 Enabling already visible frame avoids drawing
**                           the whole frame and limits drawing to the
**                           affected widgets.
** 177 NVS 20060814   @28569 Widgets losing focus now receive a draw()
**                           call right after FOCUS_LOST event to be able
**                           to redraw their state correctly.
** 178 NVS 20060816   @28615 Fixed doubled output in redirected mode by
**                           eliminating paint event brackets in
**                           placeFrameInt() method.
** 179 NVS 20060817   @28668 Fixed headers display in down frames. Headers
**                           are drawn directly now.
** 180 GES 20060818   @28669 Editing blocks must set the focus to the
**                           widget which was deferred by the previous
**                           enable.
** 181 NVS 20060817   @28670 Headers in down frames are redrawn only when
**                           they are changed.
** 182 GES 20060821   @28673 putScreen during editing mode must leave the
**                           cursor where it was previously.
** 183 GES 20060821   @28695 Honor a special case where the apply stmt
**                           raises a stop condition. Added support for
**                           overlong message lines.
** 184 NVS 20060822   @28733 Relocated default Window creation call from
**                           the constructor to the getInstance() to avoid
**                           a circular dependency after fixes to Window.
** 185 GES 20060823   @28755 status(), message() and hideMessage() during
**                           editing mode must leave the cursor where it
**                           was previously.
** 186 GES 20060824   @28875 Make apply() safe to call with a -1 widget
**                           ID when there is no frame and/or no focused
**                           widget.
** 187 GES 20060825   @28917 Honor GO in alert box event processing. Other
**                           conditions raised will not throw an exception
**                           but instead will cause a cancel return.
** 188 NVS 20060828   @28933 Implemented direct drawing in view() when it
**                           performs display (not refresh) function.
**                           Eliminated unnecessary drawing of specific
**                           widgets in case when the whole frame is
**                           redrawn due to layout change caused by widget
**                           width change caused by new data size.
** 189 NVS 20060829   @28955 Fixed latent bugs (lack of sync) in message()
**                           and status() methods which have been masked
**                           by the keyboard thread syncing.
** 190 NVS 20060830   @29003 Fixed dialog frame painting issue caused by
**                           the dialog's inability to move to the top of
**                           Z-order.
** 191 GES 20060830   @29004 More correct implementation of H183 which
**                           should have been in readKey instead of apply.
**                           Improved resetTerminal to implement a
**                           hideAll to clear the screen.  hideAll is
**                           what sets the hadReset flag (even when a
**                           resetTerminal is not called, but the hideAll
**                           is called directly by the business logic).
**                           Apply is a NOP now in editing mode when a 
**                           reset has occurred.
** 192 GES 20060830   @29012 Fixed cursor save/restore processing to be
**                           more reliable by forcing the physical cursor
**                           to be updated before changing visibility.
** 193 NVS 20060830   @29023 enable() client export incorporates view()
**                           export functionality internally. Now it's
**                           true literally, since both methods call the
**                           shared viewWorker() method.
** 194 NVS 20060831   @29076 direct drawing of frame widgets performed in
**                           the view() method and frame headers performed
**                           in setScreenBuffer() method use new methods
**                           to control clipping rectangle. These new
**                           methods take into account the scroll viewport
** 195 NVS 20060901   @29163 setScreenBuffer method filters the character
**                           data coming from the server before assigning
**                           them to widgets. Filtering consists of looking
**                           for non-displayable characters with their
**                           replacement with spaces. Currently, all
**                           characters above \u007f and below \o0020,
**                           except '\t' and '\n', are considered
**                           non-displayable.
** 196 SIY 20060901   @29204 Support for entering data in down frames.
** 197 SIY 20060904   @29212 Fixed focus setting after application try to 
**                           set focus to inactive widget.
** 198 NVS 20060905   @29231 MESSAGE UPDATE now truncates the message text
**                           if it's too long to be accommodated on a 
**                           single message line together with a spacer 
**                           and a FILL-IN. Modified methods are message()
**                           and tinyInput().
** 199 NVS 20060911   @29444 pushScreenDefinition() method needed brackets
**                           of setInvalidate() in order to handle the
**                           dynamic widget color changes properly.
** 200 GES 20060915   @29621 Honor a screen buffer value that is passed 
**                           with an associated format string.
** 201 GES 20060921   @29923 Fix for cases where on GO, a widget fails
**                           validation and the focus was not properly
**                           reset into the field.
** 202 NVS 20060926   @29931 Added a variation of honorConditionalUpDown()
**                           method called from view() and underline().
**                           It reacts to and clears the ALL CLEAR flag
**                           of the frame. This flag controls whether the
**                           conditional UPs and DOWNs are to be honored.
** 203 NVS 20060927   @29995 pause(0) calls may be issued by the server
**                           and they are now handled properly (pending
**                           pause gets cancelled).
** 204 NVS 20060928   @29999 Fixed a regression caused by #202. Under some
**                           conditions allClear flag stayed set where it
**                           should have been cleared.
** 205 SIY 20060928   @30024 nextFocus/previousFocus should take into
**                           account TEXT group mode.
** 206 NVS 20061002   @30063 Fixed setScreenBuffer() which now bypasses
**                           static text widgets if setting UNINITIALIZED.
**                           This change fixes disappearing labels.
** 207 NVS 20061005   @30162 Refactored some common code into methods and
**                           added implementation of refreshFrameWidget().
** 208 GES 20061012   @30366 Fixed small performance regression caused by
**                           H207.
** 209 GES 20061013   @30395 Eliminated extra drawing in redirected mode
**                           during setScreenBuffer().
** 210 GES 20061017   @30561 Modifications to allow GO, ENDKEY, END-ERROR,
**                           ERROR and STOP to be honored as triggers.
** 211 NVS 20061017   @30468 Implemented ability to restart the session
**                           without exiting the process.
** 212 GES 20061020   @30560 Honor CTRL-C in suspendedPause().
** 213 NVS 20061020   @30586 Implemented recovery of frame's rectangle
**                           affected by overlay frames as well as hiding
**                           those overlay frames which are no longer in
**                           use.
** 214 GES 20061027   @30781 Added additional "marking" of a frame as
**                           no longer in scope during destroyFrame().
** 215 SIY 20061030   @30784 Added pausing after DOWN on non-DOWN frames.
** 216 GES 20061030   @30786 consume (NO-APPLY) and isGoPending flags
**                           are now safe for recursive processing.
** 217 GES 20061030   @30798 Save and restore the currently focused
**                           widget in tinyInput().  This is more
**                           correct in recursive processing. WARNING:
**                           reentrant calls to waitFor() are still
**                           broken.  An attempt to fix this was made by
**                           saving/restoring in enable/disable but this
**                           did not work properly so it was backed out.
** 218 GES 20061031   @30810 hideMessage() must be a NOP if the terminal
**                           is redirected.
** 219 SIY 20061031   @30826 Do not pause on hidden frames.
** 220 SIY 20061101   @30880 Do not apply DOWN to hidden 1-DOWN frames.
** 221 GES 20061102   @30894 Added forceDeregisterPageElement() and
**                           page element cleanup processing that does
**                           not depend upon being called while the
**                           terminal is redirected.
** 222 SIY 20061105   @30980 Changed signatures of Browse related methods.
**                           Added refreshStartRow() method.
** 223 GES 20061107   @31005 Fix to allow the drop-down box of a combo-box
**                           to ignore normal conditions/go events and to
**                           simply close the drop-down in END-ERROR.
** 224 GES 20061109   @31056 Added safety code.
** 225 SIY 20061110   @31066 Reset cursor in down frame even if no widgets
**                           are updated.
** 226 GES 20061113   @31103 Only filter unprintable chars when the
**                           terminal is interactive (not redirected).
** 227 GES 20061113   @31120 Fix for null pointer exception during page
**                           element deregistration.
** 228 GES 20061114   @31125 Enable apply to update the list of the
**                           currently active triggers.  This is
**                           especially important in the case where apply
**                           is used outside of a wait-for.  In addition,
**                           explicit widget IDs should take precedence
**                           over the session ID in checkTriggers().
** 229 GES 20061115   @31173 Provide direct access to the stream to which
**                           the terminal is redirected.  Reworked view
**                           to generate output multiple times if we are
**                           redirected and the stream had to reject the
**                           output due to crossing a page boundary.
** 230 NVS 20061115   @31200 setScreenBuffer() shouldn't attempt direct
**                           drawing on a hidden frame.
** 231 NVS 20061116   @31221 messageBox() had problems with changing focus
**                           to and from it and used to leave the cursor
**                           invisible.
** 232 NVS 20061117   @31255 Fixed NPE (regression after H231) in focus 
**                           switch in messageBox(). Don't try to switch
**                           focus where there is no focused widget!
** 233 GES 20061117   @31265 Deregistration of page elements is safe to
**                           call at any time now (a requirement since it
**                           is called from hide processing which must
**                           also act this way).
** 234 NVS 20061117   @31266 Applied a better focus change technique so it
**                           makes the H231 fix better and also eliminates
**                           the problem with H232.
** 235 SIY 20061117   @31291 Updated for support of dynamic row/column. 
** 236 GES 20061121   @31332 Added processing of dead frames during frame
**                           placement.  This eliminates a potential
**                           memory leak (frames are not otherwise
**                           removed from the client).
** 237 GES 20061121   @31342 Migrated recent changes to checkTriggers()
**                           into checkBreak() which needed to operate
**                           in the same manner.
** 238 GES 20061127   @31412 Don't mark a frame as out of scope if it is
**                           a shared frame that is exiting from a more
**                           deeply nested scope. To do this a shared
**                           flag was added to destroyFrame().
** 239 NVS 20061129   @31476 Reset down frame contents and cursor on OPEN
**                           SCOPE notification.
** 240 NVS 20061130   @31502 Full frame refresh was not always performed
**                           when required. 
** 241 NVS 20061130   @31511 stopEditingMode() didn't reset currentWait-
**                           List variable causing malfunctioning of some
**                           triggers after editing block completion.
** 242 NVS 20061130   @31512 Fixed a painting regression caused by the fix
**                           H240.
** 243 NVS 20061204   @31544 Reimplemented LEAVE and ENTRY event handling
**                           model to make APPLY ENTRY work correctly as
**                           well as any triggers associated with LEAVE or
**                           ENTRY regardless of the event origin.
** 244 GES 20061206   @31597 Added promptFor() helper to provide a higher
**                           performance interface than multiple
**                           independent calls (enable then wait-for...)
**                           could provide.
** 245 SIY 20061207   @31611 Fixed handling of the nested waitFor() calls.
** 246 SIY 20061208   @31626 Deeper preservation of state during trigger
**                           invocation.
** 247 SIY 20061208   @31630 Retain cursor state during trigger 
**                           invocation.
** 248 GES 20061208   @31637 Made the cursor visibility change and restore
**                           unconditional on every entry and exit from 
**                           input mode (as opposed to only doing this on
**                           the "top" level input mode). In the case of
**                           multiple nested input mode entries, the
**                           original cursor visibility is stored in a
**                           stack, allowing the proper restore at every
**                           exit from a nested input mode, including at
**                           the top level.
** 249 GES 20061211   @31662 Added versions of client exports which take
**                           a stream ID and which implement an inline
**                           switch of the output destination to that ID
**                           and then a restoration of the previous
**                           destination back before returning.  This
**                           eliminates a pair of bracketed switchOutput()
**                           calls from the server in these cases. The
**                           objective is to improve performance of
**                           redirected output.
** 250 SIY 20061214   @31679 Force cursor position before waiting for user
**                           input.
** 251 NVS 20061214   @31695 Fixed viewWorker() to properly draw
**                           all uninitialized widgets of the current
**                           iteration of the down frame. Fixed
**                           setScreenBuffer() that used to reset
**                           uninitialized state of widgets excessively.
** 252 NVS 20070111   @31789 Fixed viewWorker() to show the frame if
**                           it's hidden by the time the method is called
**                           to display a widget in it.
** 253 GES 20070110   @31805 Eliminated unnecessary interface, simplified
**                           initialization and moved UI-specific init
**                           here from non-UI packages.
** 254 SIY 20070116   @31841 Triggering removal of already hidden out of
**                           scope frames.
** 255 SIY 20070176   @31847 Remove frame from the component list to 
**                           prevent ID conflicts.
** 256 NVS 20070118   @31891 Using new trigger registry management based 
**                           on EventList class. WaitList class is no 
**                           longer in use.
** 257 SIY 20070119   @31898 Removed separate handling of shared frames.
** 258 SIY 20070122   @31908 Call setScreenBuffer() AFTER placing frame on
**                           the screen.
** 259 SIY 20070125   @31970 Removed unused removeFrame() method.
** 260 SIY 20070126   @32018 Fixed drawing of frame title.
** 261 SIY 20070202   @32068 Fixed access to missing frame in 
**                           honorConditionalUpDown(boolean).
** 262 SIY 20070215   @32132 Prevent accessing non-existent frame.
** 263 SIY 20070221   @32200 Fixed handling of pause in hide().
** 264 SIY 20070301   @32280 Changed handling of frame removal.
** 265 NVS 20070308   @32293 Added detailed description of the process
**                           launch method since its semantics have
**                           changed.
** 266 SIY 20070309   @32383 Added special handling of frame hiding in 
**                           promptFor. Removed unnecessary call to 
**                           cleanupDeadFrames.
** 267 SIY 20070314   @32393 Added support for frame activeness tracking.
** 268 NVS 20070328   @32633 help() request from the client should report
**                           the lastKey variable.
** 269 SIY 20070328   @32643 Fixed handling of focus/frame activation in
**                           triggers. Pause flag in frames now is reset 
**                           after each input key in order to properly 
**                           handle new frames activated inside waitFor()
**                           loop from triggers. 
** 270 SIY 20070329   @32657 Removed temporary workaround.
** 271 NVS 20070329   @32667 Blocking recursive HELP calls. The terminal
**                           beeps instead.
** 272 NVS 20070404   @32763 Rewrote help() method implementation. Now it
**                           preserves frames visibility, hides frames
**                           created by help code and clears the messages
**                           (excluded the message that is part of INPUT).
** 273 NVS 20070417   @33093 Universal event recognition logic moved to
**                           EventList class.
** 274 NVS 20070417   @33257 Fixed tab order processing in enable() method
**                           ignoring KEEP-TAB-ORDER attribute of frames.
**                           Fixed message cleanup in recursive triggers.
**                           Fixed focus restoration after triggers which
**                           create frames.
**                           Added new exports: a variation of enable()
**                           and isEnabled() in support of SENSITIVE
**                           attribute, nextTabItem().
** 275 SIY 20070430   @33295 Added handling for some special cases for 
**                           getActiveFrame().
** 276 SIY 20070504   @33421 Updated getRowCount().
** 277 NVS 20070523   @33725 Fixed redirected output of DOWNs. down()
**                           method maintains a counter and data iteration
**                           display produces the required number of empty
**                           lines. Conditional DOWNs don't count.
** 278 SIY 20070517   @33748 New focus handling approach.
**                           Fixed down frame painting after scroll().
** 279 SIY 20070527   @33781 Post FOCUS_GAINED when widget is forced to 
**                           be focused. Fixed case when first enabled 
**                           widget is not focus traversable.
** 280 SIY 20070528   @33784 Fixed handling of status line and message 
**                           clearing after choose processing.
** 281 SIY 20070531   @33883 More focus and messagBox() fixes. 
** 282 EVL 20070609   @34005 Adding explicit import of the class
**                           com.goldencode.p2j.net.Queue to eliminate
**                           conflict with the same class from java.util
**                           package to be able to compile for Java 6
** 283 SIY 20070611   @34007 Rolled back overlay frame handling in frame
**                           hiding. It does cause regression.
** 284 NVS 20070611   @34009 Fixed deregistration of page elements. Fixed
**                           the page headers output, when it's caused by
**                           drawing a frame having no visible widgets.
** 285 NVS 20070614   @34034 Fixed output of page top frames produced by
**                           DISPLAY statement. They produced duplicates
**                           every time the frame was the first one on the
**                           page.
** 286 SIY 20070614   @34038 Fixed focus handling for ComboBox.
** 287 NVS 20070618   @34082 Standardized event processing in 
**                           refreshStartRow() export.
** 288 SIY 20070619   @34097 Fixed pausing in rare case of iterated single 
**                           frame display without explicit enter/pause.
** 289 SIY 20070622   @34203 Separate flag for message pausing handling. 
** 290 SIY 20070706   @34389 Fixed restoring focus after showing frame.
**                           Fixed processing of nested event lists. 
** 291 SIY 20070712   @34460 Fixed CLEAR/CLEAR ALL for down frames.
** 292 NVS 20070716   @34541 Massive changes to the redirected output mode
**                           applied to multiple methods, dictated by the
**                           need to support partial displays and other
**                           features.
** 293 SIY 20070717   @34560 Finally cleaned up #283. Do not print stack
**                           trace for IllegalComponentStateException.
** 294 NVS 20070724   @34674 Another take on redirected output mode. All
**                           output remains buffered until the target 
**                           frame changes. At this moment the buffer is
**                           flushed.
** 295 NVS 20070730   @34735 Fixed down frame headers output in redirected
**                           mode and blocked extra output for page top
**                           frames caused by their false visibility
**                           status.
** 296 NVS 20070801   @34749 Added notifyAdvance() method. It gets called
**                           whenever page advance is performed on a
**                           stream. The purpose is to flush the terminal
**                           buffer.
** 297 SIY 20070802   @34770 Fixed SET STATUS DEFAULT behavior inside input
**                           loop. Ref: 34768
** 298 NVS 20070802   @34779 Added notifyFlush() method. It gets called
**                           whenever flush operation is performed on a
**                           stream. The purpose is to flush the terminal
**                           buffer.
** 299 NVS 20070807   @34798 DOWN for a non-down frame in redirected mode
**                           is a meaningful operation. It flushes the
**                           terminal buffer and inserts the required
**                           number of new lines into the stream.
** 300 SIY 20070807   @34812 Handling for width of redirected output
**                           terminal. Some refactoring and cleanup.
** 301 NVS 20070808   @34819 Temporary redirection is provided when needed
**                           in notifyClose(), notifyAdvance() and
**                           notifyFlush() methods.
** 302 NVS 20070808   @34820 Fixed double display of page element frames
**                           that are DISPLAYed vs VIEWed by implementing
**                           deferred registration of the frame as page
**                           element. Registration should not happen until
**                           the DISPLAY output is flushed.
** 303 NVS 20070808   @34821 Deferred registration should be done for page
**                           top elements only.
** 304 NVS 20070809   @34830 Fixed forced deregistration of page elements
**                           for frames going out of scope.
** 305 NVS 20070810   @34839 Fixed pause() method. In redirected mode,
**                           PAUSE MESSAGE should temporarily switch to
**                           the interactive terminal to display the
**                           message. 
** 306 SIY 20070814   @34845 Track frame redirection.
** 307 NVS 20070815   @34851 Frame visibility in redirected mode is
**                           controlled by setVisibilityNoPaint() method.
**                           Frames are set to visible only for the time
**                           of the view() call. Only the first view()
**                           call in a chain of calls for the same frame
**                           makes the frame visible in the traditional
**                           way, which is required for the frame to draw
**                           decorations and be placed properly.
** 308 NVS 20070816   @34852 Page advance listener should flush redirected
**                           terminal *and* reset the owning frame ID.
** 309 NVS 20070817   @34881 Quick return for no-op call of redirectOutput
** 310 NVS 20070829   @34929 Fixes to the conditional down() processing. 
** 311 NVS 20070830   @34977 down() calls to the client now pass the flag
**                           that tells conditional downs. Down exports
**                           reimplemented.
**                           Numerous changes to the view() and down()
**                           exports to support stream switching for
**                           frames.
** 312 SIY 20070904   @34940 Fixed status message drawing. Ref: #34941.
** 313 NVS 20070906   @34993 switchStreams() should bypass flushing to a
**                           stream which is no longer around.
** 314 NVS 20070907   @35006 switchStreams() should be called from the
**                           promptFor() export as well.
** 315 NVS 20070907   @35013 Clearing of streamed frames switching back to
**                           interactive terminal is done via a new method
**                           clearCurrent() in FrameHelper.
** 316 NVS 20070925   @35217 Redirected frames are marked as visible while
**                           terminal is redirected to avoid duplicates of
**                           frame headers or partially drawn headers.
**                           Hiding and showing of the redirected frames
**                           is done when switching to and from the
**                           redirected mode in Window.java
** 317 ECF 20070928   @35271 Added browse-specific event dispatching. We
**                           must give Browse an opportunity to handle
**                           (and consume) END-ERROR event under certain
**                           conditions. Browse has an internal undo mode
**                           for edits, which is triggered by END-ERROR.
** 318 NVS 20071001   @35310 Down may cause page break. The corresponding
**                           listener makes the frame invisible so that
**                           the very first use of the frame on the new
**                           page redraws the frame decorations. But the
**                           same event caused also the current iteration
**                           of the down frame to be redrawn since the
**                           frame wasn't cleared yet. Now the frame gets
**                           cleared before flushing the buffers.
** 319 NVS 20071011   @35431 Pausing of streamed frames in interactive
**                           mode is implemented. Minor fix to downWorker.
** 320 NVS 20071025   @35522 Added new method getFrameIdForStream() which
**                           searches the list of frames for one connected
**                           to the specified stream. Added new method
**                           drawFrameHeaders() which outputs frame
**                           headers for the last frame used with the
**                           specified stream.
** 321 EVL 20071025   @35514 Adding switchOutput(null) to restore output
**                           device to be the real terminal for the case
**                           we need restore from Ctrl-C sequence (#35515).
** 322 SIY 20071105   @35712 Support for correct SKIP handling in 
**                           redirected mode.
** 323 NVS 20071107   @35725 HELP event processing should not take any
**                           precedence over exit conditions or GO-ON
**                           events.
** 324 CA  20071109   @35715 Removed the status restoration in method
**                           invokeTrigger.
** 325 NVS 20071109   @35765 Fixed a NPE in the view() method that could
**                           be raised in redirected mode when the stream
**                           gets finalized.
** 326 NVS 20071113   @35801 Prevented drawing of the current row when
**                           drawing the frame headers after page break.
** 327 SIY 20071116   @35822 Explicit deregistration of page element 
**                           must be more thorough to avoid reappearing of
**                           frame in the output. Some minor refactoring
**                           and cleanup.
**                           Ref: #35800.
** 328 NVS 20071116   @35830 drawFrameHeaders() bypasses frames that are
**                           registered as page elements.
** 329 SVL 20071120   @35844 Previously displayed status messages hide
**                           even if PAUSE 0 was called.
** 330 SVL 20071120   @35846 The information about last focused element 
**                           is saved on frame deactivation in focusId.
** 331 ECF 20071120   @35913 Change to match net package refactoring.
**                           Replaced uses of Queue with Session.
** 332 SIY 20071120   @35917 Do not redraw regular frames on page break.
**                           Ref: #35824
** 333 NVS 20071121   @35925 Merging in multiple parallel updates.
** 334 SIY 20071122   @35958 Fixed redundant pause on frame down() when 
**                           redirection is not active anymore.
**                           Ref: #35957
** 335 CA  20071126   @36016 Down statements with down frames redirected
**                           and paged which land outside current page 
**                           will add NL's to fill this page and only one 
**                           NL in the next page, after the frame header. 
** 336 SIY 20071126   @36023 Less aggressive resetting pause in hide().
**                           Ref: #36022
** 337 CA  20071127   @36032 Down statements with redirected, paged, 
**                           non-boxed and down frames which land outside 
**                           current page will add only one NL to the 
**                           second page, after the header. Ref: #36031
** 338 SIY 20071126   @36036 Added support for frame DOWN maintenance 
**                           using StateSynchronizer.
**                           Ref: #36034
** 339 SVL 20071128   @36044 When frame is enabled, the focusId cannot be
**                           set to a widget that is missing in the 
**                           visible widgets list.
** 340 SIY 20071128   @36057 Restored lost #322 fix.
** 341 SVL 20071206   @36228 screenCurrentInitial tab order restored when
**                           ENABLE ALL is called for frame with no
**                           keep-tab-order specified.
** 342 GES 20071207   @36261 Added safety code to getChanges() to allow
**                           it to be called when the UI has never been
**                           used (may be the case with batch processes).
** 343 SVL 20071207   @36255 Corrections to #339 fix - all enable widgets
**                           on frame are considered.
** 344 NVS 20071212   @36318 view() call for DISPLAY now tries to freeze
**                           the list of page headers for the target
**                           stream in redirected mode.
** 345 GES 20071213   @36331 Copy the frame line into the screen buffer
**                           whenever a screen buffer is returned or
**                           passed to the server. This allows frame line
**                           requests on the server to be serviced without
**                           a call to the client (improves performance).
**                           It also allows the screen buffer caching
**                           implementation to be kept in sync with
**                           iteration changes that happen only on the
**                           client side (e.g. during choose).
** 346 GES 20071213   @36340 During CHOOSE, bypass default processing of
**                           an event when a match to a GO-ON event is
**                           found.
** 347 SVL 20071214   @36356 focus events in waitForWorker() are processed
**                           immediately
** 348 GES 20071214   @36367 Only send frame-line updates in the screen
**                           buffer if the frame is a down frame and it
**                           is not currently redirected. Frame-line in
**                           these non-down or redirected cases is a
**                           constant and doesn't need to be sent.
** 349 GES 20071220   @36507 Safety code to rebalance invalidation state
**                           when down worker causes a pause and the
**                           user presses end-error.
** 350 GES 20071227   @36589 Focus changes caused by triggers are honored
**                           just after the trigger returns to ensure
**                           that subsequent event processing (for the
**                           event that caused the trigger) will "see"
**                           the focus changes.
** 351 SVL 20071225   @36568 focusable variable in enable() is now
**                           assigned correctly in case widgetId==null 
** 352 SVL 20071221   @36538 focusChange() and tryFocusChange() functions
**                           do not process Progress leave/entry events
**                           if focus remains on the same widget.
** 353 CA  20071219   @36544 The downWorker method must freeze the 
**                           headers before making the frame visible.
**                           The notifyClose method must freeze the 
**                           headers before sending the data to the stream
**                           The viewWorker must freeze the headers if the 
**                           frames have switched.
** 354 NVS 20071229   @36614 view_() method implements new redirected mode
**                           logic as described in the method javadoc.
**                           Removed some obsolete code: drawFrameHeaders
**                           and getFrameIdForStream methods.
** 355 GES 20080103   @36661 Eliminated use of freezeHeaders(). This was
**                           an invalid approach to processing page
**                           headers. It has been replaced with code in
**                           the redirected terminal to force page headers
**                           on first output.
** 356 SIY 20080107   @36701 Fixed frame displaying when its size is 
**                           changed by form headers.
**                           Ref: #36700
** 357 NVS 20080110   @36770 Fixed viewWorker() method. Down frames getting
**                           continuing output are now drawn properly and
**                           optimally in redirected mode, using
**                           repaintCurrent() helper method in FrameHelper
** 358 NVS 20080114   @36805 Delayed registration of page elements made by
**                           DISPLAY statement is no longer needed with
**                           the new implementation of the redirected mode
**                           output. Registration is done right in the
**                           view() export for both page top and page
**                           bottom elements.   
** 359 SIY 20080115   @36813 Missing pause on hideAll()
** 360 SIY 20080117   @36820 Fixed centering in redirected mode when 
**                           terminal width is other than 80 columns.
** 361 GES 20080116   @36828 downWorker in redirected mode was completely
**                           reworked to cleanup problems and to properly
**                           handle cases where the down crosses a page
**                           boundary. view_() has bugs fixed in cases
**                           dealing with page boundaries. Stream flushing
**                           now bypasses page header processing when
**                           the flushed output is from the redirected
**                           terminal. This avoids duplicate page headers
**                           since the first write to a fresh redirected
**                           terminal already handles page headers.
**                           Backed out H350 since this causes a stack
**                           overflow in certain cases.
** 362 SIY 20080121   @36874 Support for postponed event processing.
**                           Ref: #36871
** 363 CA  20080123   @36898 For shared frames, the nesting level is 
**                           maintained on client too. When the frame 
**                           closes it's root nesting level or performs a 
**                           down on its root nesting level the flag 
**                           delayedRepaint will be set to true. On the
**                           next display/view stmt, viewWorker() will 
**                           trigger a repaint if delayedRepaint is true.
**                           Ref: #36895
** 364 SIY 20080121   @36933 Do not clear redirected terminal to prevent
**                           unexpected clearing of terminal. Otherwise
**                           after launch of application first operation
**                           which updates part of screen will trigger 
**                           wiping out of rest of screen. 
**                           Ref: #36932
** 365 NVS 20080212   @37034 Fixed launching of interactive processes.
**                           Now NCURSES is properly suspended and resumed
** 366 ECF 20080212   @37036 Added getSelectedRowIndex(). Gets the row
**                           index of a selected row in a browse widget.
**                           Delegates to Browse.
** 367 SIY 20080207   @36971 Paging should not be triggered if there are 
**                           still some space at the page and non-down
**                           frame is displayed unless frame is larger
**                           than the page itself.
**                           Ref: #36968
** 368 SIY 20080211   @36988 Repaint terminal after resume() if 
**                           redirection is not active.
**                           Ref: #36987 
** 369 SIY 20080218   @37114 Simplified redirected mode frame layout and 
**                           size logic.
**                           Ref: #37107
** 370 SIY 20080218   @37139 Rolled back #367. 
** 371 NVS 20080221   @37143  hideAll() export should switch temporarily to
**                           interactive mode if redirected, since it acts
**                           on the contents of the interactive terminal
**                           only.
** 372 CA  20080303   @37234 If redirected to an unnamed stream, 
**                           message stmt sends the data to the stream too;
**                           else, the message is sent only to the terminal.
**                           Ref: #37225
** 373 SIY 20080304   @37292 Restore Z-order after displaying page headers.
**                           Ref: #37288
** 374 EVL 20080306   @37312 The insert mode indicator should be repainted
**                           as a part of update - promptFor execution.
** 375 SIY 20080305   @37317 Reset start row and trigger header processing
**                           at the beginning of the page.
**                           Ref: #37313
** 376 NVS 20080307   @37412 Added setLocation() for dynamic ROW and COL
**                           widget attributes support. For now this is a
**                           placeholder. The implementation will be added
**                           when an application shows the need for it.
** 377 SIY 20080310   @37422 Tell Screen.hideFrames() the mode (redirected
**                           or interactive) to let it choose frames for
**                           hiding.
**                           Ref: #37420
** 378 SIY 20080317   @37458 Added helper methods for terminal stream.
**                           Ref: #37453
** 379 SIY 20080320   @37607 Fixed handling of the SCREEN-VALUE and data
**                           value.
** 380 NVS 20080320   @37622 A major overhauling of the focus management
**                           with changes to enable/disable, next-prompt,
**                           wait-for and choose implementations. New
**                           implementation is based on the findings
**                           described in the "Progress Focus Management"
**                           document.
** 381 SIY 20080321   @37625 Postpone generation of conditions until all
**                           event lists are checked.
** 382 SIY 20080326   @37682 Fixed handling of pausing in messages.
** 383 NVS 20080326   @37688 Finished overhauling of the focus management
**                           for editing blocks. Allows indirect recursion
**                           of editing blocks through WAIT-FOR.
** 384 NVS 20080327   @37694 Target widget for keystrokes made when there
**                           is no valid focus is CURRENT-WINDOW now.
**                           This fixes the trigger lookup issue in GSO186
** 385 GES 20080327   @37700 Removed filtering code from setChangedValue
**                           method. This was needed when the CHARVA layer
**                           did not properly handle the extended ASCII
**                           range.  Now that that range is properly
**                           handled, the filtering logic is actually
**                           harmful since it removes any I18N characters
**                           that use the range above 7-bit ASCII.
** 386 SIY 20080328   @37723 Send LEAVE message at the end of WAIT-FOR.
** 387 NVS 20080328   @37727 List of displayed frames is no longer 
**                           maintained.
** 388 SIY 20080331   @37735 Improved #386 - event should be generated only
**                           in limited number of cases.
**                           Ref: #37734 
** 389 NVS 20080401   @37786 Fixed multiple inconsistencies when looking
**                           up frames for focusable widgets. Added widget
**                           filtering by focusability in enable() export.
** 390 NVS 20080402   @37790 Setting initial focus in waitForWorker() now
**                           takes into account dialog box option.
** 391 SIY 20080403   @37801 Pass ENTRY event to the widget which receives
**                           focus unless it is consumed by trigger.
** 392 NVS 20080403   @37827 Fixed sending the final LEAVE event on exit
**                           from wait-for. The event list was restored
**                           too early.
** 393 SIY 20080404   @37844 Fixed initialization of editing mode for 
**                           down frames.
** 394 SIY 20080408   @37870 Flush stream when switching from non-terminal
**                           to terminal stream.
**                           Ref: #37868
** 395 SIY 20080409   @37885 Fixed NPE when widget which should receive 
**                           focus has ho parent.
**                           Ref: #37887
** 396 SIY 20080410   @37900 Fixed "sticky focus" behavior to limit it to
**                           cases of source FILL-IN widget with events
**                           other than VALUE-CHANGED.
**                           Ref: #37882
** 397 SIY 20080414   @37975 Added complete ENTRY generation/processing 
**                           on repeated activation of the same widget.
**                           Ref: #37973
** 398 NVS 20080417   @38024 Eliminated excessive drawing when disabling
**                           displayed widgets.
** 399 GES 20080421   @38069 Added setFormat() client export.
** 400 NVS 20080423   @38099 Based on SIY's submission, when exiting event
**                           processing loop with a widget in focus, that
**                           widget is sent LEAVE event. Thus, LEAVE 
**                           trigger, if any, gets its chance to run.
**                           NO-APPLY return is honored by consuming GO
**                           event.
**                           Ref: #38083  
** 401 NVS 20080423   @38104 Major reimplementation of CHOOSE in field
**                           mode to support triggers and regular focus
**                           change.
** 402 NVS 20080428   @38126 Fixed few cursor positioning issues with the
**                           WaitForKey() method, which were responsible 
**                           for a few recent regressions, including extra
**                           blank lines in redirected reports. 
** 403 GES 20080501   @38189 Merged endOutput calls into other client
**                           calls to avoid the extra trip.
** 404 NVS 20080501   @38192 Fixed few not so recent regressions with UI
**                           found in MAJIC training menu.Roster report.
**                           Fixed a focus restoration bug in
**                           waitForWorker().
** 405 NVS 20080502   @38197 Fixed NPE in WaitForWorker() when dealing
**                           with null ScreenBuffer. 
** 406 NVS 20080502   @38198 Eliminated duplicate LEAVE events in both
**                           known cases: when setting initial focus in
**                           new wait-for and when GO generates LEAVE.
** 407 NVS 20080503   @38201 Based on the update submitted by SIY.
**                           Additional code cleanup for GO event handling
**                           Do not generate LEAVE on GO for WAIT-FOR.
**                           Moved code which generates LEAVE on GO before
**                           all calls to invokeTriggers() in order to 
**                           preserve correct order of events as they are 
**                           seen by triggers (must be LEAVE then GO, not
**                           GO then LEAVE). Also, if trigger on LEAVE
**                           returns NO-APPLY, there must be no GO event.
**                           GO event does not cause exit from WAIT-FOR if
**                           it is not in the list of events.   
** 408 NVS 20080503   @38214 Based on the update submitted by CA. Extended
**                           LEAVE event generation on exit for any kind
**                           of exit condition, not simply GO.
** 409 CA  20080506   @38215 Fixed regression introduced by H403 (#38189): 
**                           ThinClient.endOutput() must append a NL to 
**                           the stream, if needed (i.e. must call stream.
**                           endOutput(true)).
** 410 CA  20080507   @38227 screenCurrent is initialized when the class
*                            is initialized. Ref: #38226
** 411 NVS 20080508   @38239 Removed checks like screenCurrent == null
**                           They are redundant since the H410 and should
**                           have been removed as part of H410 indeed.
** 412 SIY 20080509   @38248 Enable frame before staring event processing
**                           loop in choose() and restore state when
**                           processing is finished.
**                           Ref: #38247
** 413 SIY 20080509   @38249 Cleanup message area in hideMessage() if there
**                           was some content.
** 414 SIY 20080512   @38256 Allow Browse receive 4GL events even in
**                           disabled state.
** 415 NVS 20080512   @38261 Message box should not interfere with the
**                           focus management. Now current focus is
**                           preserved across the message box activity.
** 416 NVS 20080512   @38262 Help text should be saved and restored around
**                           trigger invocation.
** 417 SIY 20080516   @38288 If tinyInput() detects that message is too 
**                           long, it should replace redundant characters  
**                           with spaces, otherwise garbage left from full
**                           message may remain on screen.
** 418 SVL 20080520   @38333 hideAll() does nothing if the terminal is
**                           redirected. 
** 419 SIY 20080522   @38369 Frame scope opening resets regular frames too.
**                           Ref: #38365
** 420 SIY 20080522   @38372 Fixed regression caused by #419.
** 421 NVS 20080528   @38428 Frame as a focus target of wait-for should be
**                           simply ignored and handled as if there were
**                           no focus clause.
** 422 SIY 20080527   @38404 Fixed missing pause and redundant message
**                           clearing when MESSAGE is invoked from two
**                           sequential triggers with no other intermediate
**                           interactive statements.
** 423 SIY 20080530   @38461 Enable/disable special handling of DELETE key
**                           for VT220 terminal.
**                           Ref: #38459
** 424 SIY 20080530   @38465 Call to AlertBox.finished() should be 
**                           bracketed with push()/pop(), otherwise hiding
**                           is not handled properly.
**                           Ref: #38464 
** 425 NVS 20080530   @38467 STATUS INPUT issued from an editing block
**                           should not override the field's help text.
** 426 SIY 20080530   @38470 Prepare ScreenBuffered before focus state is
**                           restored to get correct value for FRAME-FIELD
**                           and FRAME-VALUE (see Frame.getFrameValue()).
** 427 NVS 20080603   @38506 Eliminated unnecessary drawing causing PUT
**                           SCREEN output erasure in pushScreenDefinition
**                           seen in GSO 214.
** 428 NVS 20080605   @38586 Fix 427 should allow drawing for hidden frame
**                           that is being made visible through config
**                           change.
** 429 NVS 20080606   @38611 Fixed setHelp() method which now properly
**                           reverts the status text to the default text
**                           when called for a widget having no own help
**                           text.
** 430 SIY 20080609   @38651 Repaint areas which belong to frames whose
**                           Z-order is about to be changed.
** 431 SIY 20080610   @38674 Make #430 a bit less aggressive - repaint only
**                           frames which overlap current. Also, when there
**                           are overlapping frames, force frame placement
**                           since this affects pausing.
** 432 NVS 20080611   @38689 message() worker saves and restores cursor
**                           position always as it can be called for error
**                           message display internally.
**                           waitForWorker() undefines focus for high
**                           level operations if it was undefined on entry
** 433 NVS 20080611   @38711 Fixed focus issues after END-ERROR issued in
**                           previous frame.
** 434 NVS 20080613   @38763 Focus should be restored in cases of CHOOSE
**                           nested within WAIT-FOR or editing blocks.
**                           Added missing type-ahead buffer reset to
**                           pause().
** 435 NVS 20080527   @38403 Implemented conditional targets for keystroke
**                           which may be widget, frame or current window.
**                           Implemented LEAVE event for frames received
**                           ENTRY.
** 436 NVS 20080616   @38790 NEXT PROMPT should be ignored if it points 
**                           outside of the widget list for current op.
** 437 NVS 20080619   @38873 Fixed help text display for CHOOSE in row
**                           mode. Fixed PUT SCREEN quirk for the status
**                           line: no output when it would touch either of
**                           the two rightmost character position (depends
**                           on the terminal window width).
** 438 GES 20080613   @38881 Shifted to using ClientState for state sync
**                           and added a hadPause flag to notify the
**                           server whenever an explicit or implicit
**                           pause occurs.  It is used to conditionally
**                           control infinite loop control in certain
**                           cases.
** 439 NVS 20080625   @38921 Fixed status line display in WAIT-FOR when
**                           there is no focused widget.
**                           CA: fixed a potential NPE in waitForWorker().
** 440 NVS 20080626   @38947 Backed out H416. Now waitForWorker() does a
**                           better job of managing focus and H416 causes
**                           a regression.
** 441 SIY 20080630   @38989 Protect existing frames from removal when help 
**                           application is invoked.
**                           Ref: #38987
** 442 SIY 20080711   @39101 Use event source instead of current focus when
**                           making decision about consuming event. 
**                           Otherwise trigger which activated/deactivated
**                           some other widget may prevent event from
**                           passing to source widget.
** 443 SIY 20080731   @39260 Added handling of pause for frames which are
**                           cleared because of repeated entering/leaving 
**                           frame scope.
** 444 SIY 20080808   @39345 Added handling of forced frame refreshing
**                           in VIEW after interrupted pause on cursor move
**                           in DOWN frames. 
** 445 NVS 20080812   @39372 Frames with editors in redirected mode are
**                           subjects of widget rearrangement before
**                           drawing and restoration right after it. This
**                           is required to allow editors output all of
**                           their contents regardless of the visible
**                           widget's height.  
** 446 NVS 20080818   @39437 view() call may cause pausing due to frame
**                           hiding and F4 pressed while in that pause
**                           should be properly propagated to the business
**                           logic interrupting frame display. Fixed
**                           placeFrameInt so it bypasses frame display
**                           in case of END_ERROR_CONDITION and viewWorker
**                           to propagate the condition up. Based on CA's
**                           findings.
** 447 NVS 20080820   @39440 waitForWorker() saves and restores leaveSent
**                           flag so that it's preserved across nested
**                           waits. This guarantees that LEAVE event is
**                           sent no more than once per WAIT-FOR, even if
**                           caused by APPLY LEAVE from a business logic
**                           of a trigger.
** 448 SIY 20080826   @39573 Processing of message box should reset pause
**                           just like other kinds of pauses do.
** 449 NVS 20080826   @39598 Modified refreshScreen() method to use 
**                           invalidation region technique instead of 
**                           invalidation rectangle.
** 450 SIY 20080901   @39642 Added ANY-KEY event support.
** 451 SIY 20080903   @39697 Fixed order of checks in ANY-KEY support.
** 452 CA  20080904   @39749 WAIT-FOR will keep a list of frames which are
**                           waiting for events. When another frame is 
**                           viewed, it will not hide the dialogs which
**                           are still waiting for evens. Added the getter
**                           for FRAME:VISIBLE attribute. If the frame was
**                           moved, we must hide the frame before applying
**                           the new configuration in pushScreenDefinition.
** 453 ECF 20080911   @39795 Partially rolled back #452 (@39749). The loop
**                           at the top of the finally block in the
**                           viewWorker() method was commented out in
**                           response to a ConcurrentModificationException
**                           in Window.draw(). Further investigation is
**                           needed.
** 454 SIY 20080915   @39810 Rethrow exception in viewWorker() regardless
**                           from the redrawing/refreshing modes.
** 455 CA  20080909   @39750 getFrameLine was changed to consider the
**                           pendingUpDown value.
** 456 SVL 20080914   @39801 Added lastActionCode field. Some changes in
**                           the insert mode logic.
** 457 NVS 20080917   @39842 While in editing block, readKey() should NOT
**                           reset "message needs pause" flag.
** 458 SVL 20080919   @39860 Auto-zap is reseted when the focus moves
**                           between fill-ins (in some cases). 
** 459 NVS 20080922   @39902 hideFrame() for dialog frame checks if there 
**                           is another dialog and reactivates it
**                           immediately. viewWorker() brings frames on
**                           top conditionally - dialog frame comes on top
**                           always whereas regular frame is placed below
**                           first visible dialog, if any. 
** 460 SIY 20080918   @39900 Draw labels in side-labeled DOWN frame.
** 461 CA  20080926   @39953 Avoid duplicate LEAVE events to be sent by
**                           waitForWorker.
** 462 NVS 20080922   @39997 Fixed handling of interactive conditional
**                           DOWNs when previous output was redirected.
**                           Fixed unconditional DOWN N in redirected mode
**                           to produce the specified number of iterations 
**                           not just the blank lines. 
**                           Fixed forced page break to produce iteration
**                           not just a single blank line. 
** 463 SVL 20081001   @39978 Call to frame.draw() removed from scroll(..).
** 464 SIY 20081003   @40018 Added support for MarkEntry.UNHIDE. 
** 465 NVS 20081008   @40072 Partially backed out #462 due to regressions.
**                           Redone iteration based calculations for the
**                           number of blank lines in downWorker().
** 466 SIY 20081013   @40084 Do not sent LEAVE on exit from WAIT-FOR if
**                           exit was performed during processing of 
**                           consumed condition (the condition was consumed
**                           but other (possibly artificial) event caused
**                           end of WAIT-FOR loop).
** 467 CA  20080930   @39981 Added accessors for FRAME:HEIGHT/WIDTH-CHARS
**                           FRAME:COL/ROW attributes.
** 468 NVS 20081015   @40093 Reimplemented conditional vs unconditional
**                           DOWN processing for redirected mode.
**                           Interactive conditional DOWN causes flushing
**                           of the previous stream and that affects the
**                           logic of the unconditional DOWN to the stream
**                           which may follow.
** 469 SIY 20081016   @40100 Slightly changed clearWorker() signature and
**                           allowed it pass through value returned by
**                           FrameHelper.clearWorker(). 
** 470 NVS 20081016   @40101 Fixed a regression in MAJIC caused by #468
**                           and the last known regression in redirected
**                           reports suite.
** 471 SIY 20081017   @40157 Fixed missing pause after breaking pause on
**                           frame clearing.
** 472 SIY 20081020   @40171 Improved #471 and now it also hides frames
**                           which can be hidden.
** 473 SIY 20081022   @40198 Do not clear messageNeedPause when hiding 
**                           frame without pauses. 
**                           Ref: #40196
** 474 SIY 20081023   @40204 Adopted changes in message text handling API,
**                           removed redundant bracketing.
** 475 CA  20081023   @40163 Fixed cursor reset when NEXT-PROMPT is used
**                           and message handling within a EDITING block.
**                           Modified by NVS to use continueEditing()
**                           export, which already existed.
** 476 CA  20081024   @40222 On a new client instance, clear the toolkit
**                           window list.
** 477 SIY 20081007   @40234 Added support for postponed cursor wrapping
**                           for DOWN frames.
**                           Ref: #40235
** 478 SIY 20081029   @40264 Do not clear status line when WAIT-FOR loop is
**                           interrupted by condition exception.
**                           Ref: #40263 
** 479 SIY 20081031   @40283 Reworked #478 to avoid regressions.
** 480 SIY 20081031   @40285 In some cases, when focus target is frame with  
**                           browse, we should choose browse as target.
**                           Ref: #40284
** 481 GES 20081106   @40360 Avoid treating editing blocks as a wait-for
**                           when processing END-ERROR even though the
**                           opcode is set as WAIT_FOR.
** 482 SIY 20081107   @40372 Replaced #480 with generic fix.
**                           Ref: #40367
** 483 SVL 20081107   @40391 getFrameLine() does not return 0 for hidden
**                           and redirected frames any more. Current
**                           iteration is cleared for non-down frames
**                           into downWorker().
** 484 CA  20081114   @40447 Changes due to new clipping implementation.
** 485 SIY 20081117   @40507 Do not redraw hidden widgets in enable().
**                           Hide overlay frames which intersects with one
**                           which starts the input loop.
** 486 SIY 20081120   @40598 Fixed regression caused by #457. Fixed issue
**                           for which #457 was introduced (actual cause - 
**                           difference in message auto-clearing behavior
**                           inside EDITING block).
**                           Ref: #40597
** 487 CA  20081124   @40684 Only frames marked with canHide=true can hide
**                           other frames, on a VIEW/DISPLAY call.
** 488 SIY 20081125   @40693 Added missing drawing bracket on early exit in
**                           waitForWorker().
** 489 NVS 20081126   @40698 Added a state flag that tracks the duration
**                           of the viewWorker() method activity. Required
**                           for fixing temp text deletion when frame
**                           hiding is caused by view.
** 490 SIY 20081127   @40709 Fixed more focused frame Z-order issues 
**                           (related to second part of #485), common code
**                           is refactored in separate method.
** 491 CA  20081127   @40706 Modified H487 to use the 'canHideOthers' flag;
**                           this flags tracks if a certain frame can hide
**                           other frames, on its first DISPLAY/VIEW call.
** 492 SVL 20081127   @40713 lastActionCode in not modified if an alert 
**                           box is displayed.
** 493 CA  20081128   @40717 After a WAIT-FOR time-out, no leave event will
**                           be sent.
** 494 SIY 20081201   @40755 Fixed pausing regressions caused by changed
**                           'can hide' flag maintenance and frame hiding 
**                           behavior.
** 495 CA  20081203   @40758 Browse widgets will clear the temp-texts from
**                           the message area when they are viewed or 
**                           repositioned.
** 496 SIY 20081203   @40771 Set widgets values early if possible. This is
**                           necessary for correct frame size calculation.
**                           Ref: #40772.
** 497 NVS 20081204   @40789 CHOOSE processing shouldn't try and restore
**                           focus when exiting event processing loop in
**                           waitForWorker().
**                           When iterating editing blocks, focus change
**                           may cause needed events, so now it's enclosed
**                           in push()/pop(). It restores the HELP string
**                           after triggers correctly.  
** 498 CA  20081205   @40798 Fixed regression caused by H495: frames which 
**                           intersect the message area need to be redrawn,
**                           if temp-text was hidden.
** 499 NVS 20081205   @40809 Standardized the use of getAndResetBitmap()
**                           and getAndSetBitmap().
** 500 SIY 20081205   @40804 Partially backed out #494.
** 501 CA  20081208   @40822 Changes needed for the ROW-DISPLAY event.
** 502 SVL 20081208   @40827 Operation code is set to UPDATE or SET during
**                           "message ... update/set ..." actions.
** 503 SVL 20081209   @40849 Added code for the exceptional conditions
**                           handling into processProgressEvent().
**                           isGoPending changed to int in order to count
**                           the number of conditions which cause GO
**                           event. Added immediateException concept.
** 504 SIY 20081209   @40855 Reworked frame hiding implementation.
** 505 CA  20081210   @40846 Implemented CLOSE QUERY statement - on close,
**                           the browse widget will be notified to reset
**                           its state.
** 506 SVL 20081211   @40862 immediateException is thrown into
**                           processEventsWorker() only if we are in the
**                           WAIT-FOR statement.
** 507 NVS 20081211   @40882 Backed out the problematic fixes ##503,506.
** 508 NVS 20081215   @40911 APPLY adjusts focus instead of applying event
**                           if the focus was not valid at the time. This
**                           is a partial fix as no event filtering occurs
**                           yet.
** 509 SVL 20081216   @40920 Reapplied #503, added triggerNestingByWaitFor
**                           and lastAppliedCode.
** 510 NVS 20081218   @40941 Backed out #508 due to regressions.
** 511 SIY 20081218   @40957 Added static accessor for Screen instance.
** 512 SVL 20081224   @40981 Temp text is cleared in the message area
**                           before (for the specific line) and after the
**                           MESSAGE ... SET/UPDATE ... action.
** 513 SIY 20081224   @40989 Added hideWindow() method. Added fix for
**                           frame Z-order in promptFor() (replacement for
**                           Window.java #044 and #046).
** 514 SIY 20081225   @40994 Hide dialog immediately when it leaves scope.
** 515 NVS 20081226   @40998 Added private method highlight() and used it
**                           in waitForWorker() to control highlighted
**                           state of the focused widget. Added highlight
**                           control when sending ENTRY and LEAVE events. 
** 516 CA  20090108   @41024 Re-display frame if it was hidden during 
**                           trigger invocation and it was not an explicit
**                           HIDE call. Added more highlight state fixes.
**                           ENABLE calls will make all targeted hidden 
**                           widgets visible. Fixed VIEW FRAME calls - all
**                           hidden widgets remain hidden.
** 517 CA  20090109   @41071 On event processing, any ConditionException's
**                           must be raised and re-thrown. Ref @41070
** 518 SIY 20090112   @41094 hideAll() disables placement quirk mode just 
**                           like explicit frame hiding.
** 519 CA  20090114   @41124 Fix focus change when current focus is 
**                           invalidated during trigger invocation. The
**                           cause for this invalidation is focus parent
**                           frame visibility change from not-visible to
**                           visible, via an explicit VIEW.
** 520 NVS 20090116   @41150 Fixed direct drawing issue caused by enable()
**                           called from promptFor() for a visible frame
**                           which was obscured by an overlay frame.
**                           Just the frame and widget's visibility check
**                           is not enough: widget should not be even
**                           partially obscured by overlay frames to be
**                           eligible for direct drawing.
**                           Added isAllShowing() static helper method.
**                           TO-DO: check whether all other cases of the
**                           direct drawing are susceptible to the same
**                           problem.
** 521 SVL 20090120   @41182 Added inEnable flag.
** 522 CA  20090121   @41185 Fixed CLEAR FRAME stmt for down frames - 
**                           clearWorker(boolean) should handle this stmt 
**                           by calling FrameHelper.clearCurrent instead 
**                           of trying to clear the entire frame (calling 
**                           FrameHelper.clearWorker(boolean)).
** 523 NVS 20090122   @41195 Fixed a regression caused by #520: ENABLE or
**                           DISABLE referencing BrowseColumn would throw
**                           an exception. Optimized enable() to bypass
**                           shadowed state check if no frames are above.
** 524 SIY 20090128   @41233 Fixed order of setEnabled()/view() calls which
**                           may affect visible appearance of the widgets
**                           if enabled widget belongs to iterated frame.
** 525 SIY 20090130   @41246 Remove highlighting from widget which looses 
**                           focus.
** 526 SIY 20090203   @41255 Trigger call should transfer highlighted state
**                           from widget which has focus before trigger  
**                           invocation to widget which receives focus 
**                           after trigger invocation.
** 527 NVS 20090209   @41283 Installed double buffered output driver as
**                           a standard interactive output driver.
**                           Reimplemented putScreen() and putCursor()
**                           exports using direct output driver.
** 528 SIY 20090212   @41293 Refresh wiped out area in hideAll().
** 529 SIY 20090213   @41310 Added 'force' parameter into refreshStartRow
**                           signature.
** 530 NVS 20090213   @41318 Fixed few regressions in message and status
**                           processing caused by transition to the double
**                           buffering output driver. Code cleanup.
** 531 NVS 20090225   @41363 Added static getScreenSize() method.
** 532 SIY 20090225   @41378 Refresh message area during implicit clear
**                           if there was a message.
** 533 CA  20090304   @41410 ENTRY event must not adjust focus before its
**                           trigger is fired; reworked focus handling when
**                           explicit or implicit focus or first enabled 
**                           widget on screen get disabled/hidden during
**                           trigger invocation; consume'd events do not
**                           specify that the trigger returned no-apply -
**                           the last trigger result for the current level
**                           will be kept on a stack.
** 534 CA  20090306   @41424 Fixed a focus related regression introduced 
**                           by H533. Fixed editing block behavior: send  
**                           an ENTRY to the requested widget, when 
**                           editing mode starts. Fixed a issue related to 
**                           UPDATE: when a widget list is used, first  
**                           focusable widget from the list must be used.
** 535 SVL 20090305   @41392 Insert indicator states are saved on entering
**                           and restored on leaving of prompting blocks.
** 536 SVL 20090311   @41504 Enable auto-zap and reset insert mode for
**                           fill-in which validation has failed when
**                           it is currently focused.
** 537 CA  20090310   @41455 Implemented focus handling when implicit or
**                           explicit focus are hidden/disabled before 
**                           trigger invocation. A MSG ... UPDATE stmt 
**                           will adjust focus after execution, if it is
**                           previously undefined. Fixed READKEY behavior
**                           related to CHOOSE event.
** 538 CA  20090312   @41516 Implemented focus handling rules when in not
**                           in nested trigger mode.
** 539 SIY 20090311   @41545 Added widget drawing brackets for conditional
**                           widget refresh.
** 540 CA  20090317   @41623 Fixed issues related to out-of-scope pending
**                           focus. Fixed a regression related to APPLY,
**                           introduced by H538.
** 541 SIY 20090319   @41647 Added conditionalRefresh(Rectangle) method.
** 542 SVL 20090320   @41658 If leave action for a widget has returned
**                           no-apply, then insert mode will be reseted
**                           and zap will be enabled for the focused
**                           fill-in.
** 543 CA  20090323   @41663 Added missed rule about READKEY and CHOOSE
**                           trigger: if a GO event was trigger, no more
**                           key is read by current READKEY op. Fixed more
**                           issues related to APPLY, ENTRY/LEAVE events 
**                           and focus search. Added missed rule about
**                           focused undef explicit for wait-for.
** 544 CA  20090324   @41669 Fixed focus restoration for nested UPDATE.
**                           Focus must not be adjusted if a trigger 
**                           generated a GO event. In focus undef explicit
**                           case, operation must end if no widget accepts
**                           entry and the explicit focus is disabled or
**                           hidden.
** 545 CA  20090324   @41671 Fixed issues related to APPLY stmt. Fixed  
**                           possible NPE when combo widget is used.
** 546 CA  20090327   @41694 Fixed NPE in processEventsWorkers caused by
**                           H545: locateFrame may return NULL if a window
**                           component is passed as a parameter.
** 547 CA  20090330   @41695 Fixed regression introduced by H546, related
**                           to APPLY stmt, when target is invalid.
** 548 SVL 20090402   @41716 Insert mode is saved/restored on triggers.
**                           Insert mode saving/restoration refactored.
**                           Fixed: unnecessary insert indicator clearing
**                           with HIDE ALL, zap on WAIT-FOR PAUSE, insert
**                           mode on focus restoration after WAIT-FOR,
**                           insert mode accepting if there are no focused
**                           widgets. 
** 549 SIY 20090326   @41710 Implemented status saving/restoring logic.
** 550 CA  20090402   @41722 On frame destroy, all widgets get disabled, 
**                           if frame is still visible.
** 551 CA  20090403   @41726 Fixed regression introduced by H550 - disable
**                           frame before removing it.
** 552 CA  20090402   @41725 Reworked APPLY focus handling (before and
**                           after trigger execution).
** 553 SIY 20090407   @41729 Do not disable widgets in frames which don't
**                           belong to interactive terminal.
** 554 SIY 20090407   @41731 Changed approach of disabling widgets during
**                           frame destroying in order to avoid regressions
**                           in UNDERLINE handling.
** 555 CA  20090408   @41735 On LEAVE, fill-in cursor must be set on first
**                           position. No trigger should be invoked when
**                           a combo's drop-down list is in focus. Fixed 
**                           focus adjustment after wait-for iteration,
**                           when focus got disabled. Fixed frame:sensitive
**                           flag. Fixed issues related to WINDOW and FRAME
**                           events. WAIT-FOR will generate a STOP condition
**                           if it doesn't wait for global events and there
**                           no enabled widgets between ones registered for
**                           exit events.
** 556 CA  20090413   @41782 Fixed a couple of focus issues: adjustFocus  
**                           must not skip focus adjustment if focus is  
**                           disabled and there was a wait-for loop  
**                           executed by the previous trigger; after a 
**                           wait-for iteration, focus must be adjusted if 
**                           current state is undefined. Fixed SENSITIVE/
**                           VISIBLE/HIDDEN flags behavior for widgets and 
**                           frames. Widgets which had their HIDDEN or 
**                           VISIBLE flag changes will send the new state
**                           back to the server, for synchronization.
** 557 SIY 20090414   @41786 Prevent NPE in popStatus().
** 558 CA  20090414   @41785 Made collectFlag package private, to collect
**                           flags for frame header widgets, when a frame 
**                           has its VISIBLE flag changed.
** 559 CA  20090414   @41790 Focus must be restored after an wait-for loop
**                           when executed inside an editing block. If it
**                           needs adjustment, will be adjusted regardless
**                           of operation type. When executing first 
**                           wait-for iteration, frame is not used as  
**                           target if focus exists and is not visible.
** 560 CA  20090415   @41795 Fixed another regression related to focus do
**                           not use frame initial target, if focus does 
**                           not exist; fixed a regression related to an
**                           APPLY stmt executed inside an EDITING block.
** 561 CA  20090416   @41804 Fix for an abend generated by UPDATE calls 
**                           for down frames - related to initial focus.
** 562 CA  20090416   @41805 UPDATE/SET/PROMPT-FOR with EDITING block must
**                           restore focus to the widget in focus prior to
**                           this call (the same way simple UPDATE statements
**                           work).
** 563 CA  20090416   @41806 Fixed a case about initial hidden focus frame
**                           when wait-for has explicit target.
** 564 CA  20090417   @41808 Fix abend caused by H562 - Dequeue does not 
**                           allow null values on stack, use Stack instead
**                           as type for the editingFocus field.
** 565 CA  20090421   @41820 No triggers should be invoked when for an APPLY 
**                           with implicit target there is no focused widget.
**                           CHOOSE in row mode must use frame as target, 
**                           even if P2J doesn't set it in focus.
** 566 SIY 20090421   @41834 Wipe screen when frame can hide others.
** 567 CA  20090422   @41836 The implicitApplyNoTrigger flag must be restored 
**                           in TC.apply in any situation, even if a condition
**                           is raised (due to F4) - this fixes a regression
**                           introduced by H565.
** 568 SIY 20090422   @41844 Limited wiping screen to only specially marked
**                           frames.
** 569 CA  20090422   @41840 Added handling of NEXT-FRAME and PREV-FRAME 
**                           events.
** 570 CA  20090423   @41903 Fixed a special case about read-only EDITORs: if
**                           a trigger had a DISPLAY with editor widgets as
**                           target, then caret (not focus) is moved to last
**                           targeted editor, in tab-order.
** 571 SVL 20090424   @41922 Addition to #542 - do not reset insert mode and
**                           enable zap if focus has remained on the same
**                           fill-in. Skip insert mode restoration if an
**                           editor is in focus.
** 572 GES 20090424   @41967 Import change.
** 573 SIY 20090429   @42073 Save/restore status when invoking message box.
** 574 SIY 20090430   @42085 Status saving/restoring in PROMPT-FOR and 
**                           MESSAGE ... UPDATE... .
** 575 SIY 20090501   @42086 Rolled back #573 because it causes regressions
**                           and relevant cases are fixed by #574.
** 576 SIY 20090501   @42089 Added notification of frames about stream end of 
**                           life.
** 577 SIY 20090519   @42263 Fixed widget zap state on widget activation when  
**                           ENTRY event is not sent.
** 578 SVL 20090515   @42185 Removed unnecessary insert mode adjustment into
**                           waitForWorker(). Do not redraw insert indicator
**                           after HIDE ALL if it is into ambiguous state.
** 579 SVL 20090521   @42422 Empty help string allows default input message to 
**                           be used.
** 580 SVL 20090527   @42472 For the APPLY "ENTRY" statement, we should enable
**                           zap and reset insert mode for the ENTRY target
**                           (fill-in), or to the currently focused fill-in if
**                           ENTRY has returned NO-APPLY. When the focus is
**                           restored after a trigger call, zap should be 
**                           enabled for the focus target. Rolled back #542.
** 581 SIY 20090528   @42477 Avoid redundant repainting in redirected mode if
**                           entire iteration of DOWN frame is displayed.
**                           Ref: #42476.
** 582 SIY 20090529   @42500 Limit #581 to redirected mode only.
** 583 SVL 20090529   @42487 After external procedures restore only type of
**                           the status rather than statuses themselves.
** 584 SIY 20090601   @42506 Removed unused inner class.
** 585 SIY 20090602   @42513 Fixed NPE in popStatus().
** 586 SIY 20090609   @42619 inputActive should be reset on new procedure entry
**                           and restored upon exit.
** 587 SIY 20090610   @42637 Added support for authentication operation mode.
**                           In this mode Ctrl-C is processed as regular key
**                           which triggers StopConditionException.
** 588 SIY 20090611   @42658 Reset all pause flags if Ctrl-C is observed.
** 589 SIY 20090612   @42668 inputActive is increased for waitFor() with 
**                           explicit focus.
** 590 SVL 20090612   @42681 Fixed NPE in message(). Notify BufferSizeManager
**                           about blocking operations. "Attempt to write
**                           to closed stream" may be raised into view_().
**                           Added resetBufferSizeManager() method. Raise STOP
**                           instead of ERROR into switchStreams(),
**                           flushStream() and flushRedirected(). 
** 591 SIY 20090623   @42782 Clear keyboard buffer in pause() only if non-zero
**                           time interval is passed.
** 592 GES 20090623   @42792 Any implicit or explicit pause has a special
**                           behavior in processCondition() that it does not
**                           change the interactions counter BUT if it executes
**                           when the interactions counter is 1 (but not when it
**                           is 0), then an END-ERROR will be translated into
**                           ERROR (as if the real interactions counter was 2
**                           or more) instead of into an END.
** 593 SIY 20090623   @42794 Added protection and logging for attempts to pop
**                           from empty cursor stack.
** 594 SIY 20090624   @42886 Rolled back #593 and fixed trap during redundant
**                           stopEditingMode() call.
** 595 SIY 20090701   @43022 Added client side part of format checking code.
** 596 SIY 20090707   @43069 Do not force frame z-order if it is displayed
**                           inside editing block and has no enabled widgets.
** 597 SIY 20090708   @43105 Added one more case of wiping screen from 
**                           PUT SCREEN artifacts.
** 598 GES 20090701   @43028 Prevent event brackets from getting out of balance.
**                           Also added more complete state clearing on CTRL-C.
** 599 SIY 20090709   @43124 Less aggressive screen wiping (fix for #597).
** 600 NVS 20090709   @43130 The fix #598 has left an unbalanced setInvalidate
**                           call in popStatus() method. Fixed.
** 601 SIY 20090713   @43147 Fixed handling of focus restoring after trigger 
**                           when source of the event is Button.
** 602 GES 20090720   @43249 NPE protection for highlight().
** 603 SIY 20090723   @43348 Added event/drawing brackets for focus restoring.
** 604 SIY 20090724   @43388 Fixed terminal width adjustment in redirected mode
**                           for frames which have initial column specified.
** 605 SIY 20090804   @43510 Added support for NO-PAUSE in handling of 
**                           conditional UP/DOWN.
** 606 CA  20090805   @43520 Ensure the event queue is popped in pop(), even 
**                           if a condition is raised during event processing. 
**                           This will solve the client-side OOME caused by 
**                           event queue leaks.
** 607 SIY 20090806   @43541 Fixed paging of DOWN frames with iterations which
**                           have height > 1 but not end with SKIP widgets.
** 608 SIY 20090810   @43580 Fixed message hiding in some special cases (when
**                           placed frame hides other frame at the same row).
** 609 SIY 20090811   @43595 Made more narrow conditions which trigger message
**                           hiding in #608.
** 610 SIY 20090813   @43611 Even more narrow conditions which trigger message
**                           hiding in #608.
** 611 SIY 20090814   @43617 Reworked handling of DOWN paging for frames which 
**                           have iteration height greater than 1. Fixed 
**                           handling of leading/trailing SKIP widgets in 
**                           iteration.
** 612 SIY 20090820   @43697 Fixed handling of Ctrl-C during pause. 
** 613 SIY 20090821   @43715 Fixed handling focus in cases when frame is 
**                           completely disabled in trigger.
** 614 SIY 20090824   @43731 Use hasAnyEnabled() instead of hasEnabled() in 
**                           hideFramesInt(). Fixes regressions caused by 
**                           recent changes in hasEnabled().
** 615 SIY 20090824   @43792 Added proper handling for 'forced message hiding'
**                           mode which can be triggered in some circumstances
**                           during PROMPT-FOR initialization.
** 616 SIY 20090901   @43807 Added event bracket for widget hiding loop 
**                           in hideAll() method.
** 617 SIY 20090902   @43826 Fixed handling focus in certain scenarios of
**                           button-browse-trigger interactions.
** 618 SIY 20090906   @43830 Do not process events from components in removed
**                           frames.
** 619 SIY 20090907   @43835 Fixed restoring of redirection in message() in 
**                           case when user presses END-KEY during pause.
** 620 CA  20090902   @43823 Fixed a deadlock in initializePrep - if CTRL-C is
**                           pressed at a certain time, the watchCtrl call 
**                           will block, as it wants to acquire class-level 
**                           lock both on TC and Window classes.
** 621 SIY 20090903   @43838 Rolled back #618.
** 622 SIY 20090904   @43839 Partially rolled back #617.
** 623 SIY 20090909   @43850 Protection from NPE in iterateEditingMode().
** 624 SIY 20090910   @43856 getFocus() should not return widget in hidden 
**                           frame as current focus.
** 625 SIY 20090924   @44008 Skip tinyInput() frame when preparing batch for
**                           server.
** 626 SIY 20091015   @44136 Refresh frame, if it is about to be edited and 
**                           there are other frames at the screen.
** 627 SIY 20091025   @44161 Fix for regressions caused by #626. It should be
**                           activated only if last frame which was made 
**                           visible is not the same as one which is about to 
**                           be edited.  
** 628 SIY 20091028   @44274 Disable automated removal of frame created by 
**                           tinyInput.
** 629 SIY 20091118   @44429 Made tinyInput() re-entrant.
** 630 CA  20091111   @44329 Added a statusDefault(String, boolean) and 
**                           status(String, booolean) which disable 63-char 
**                           trimming, if the noLengthLimit flag is set.
*/

package com.goldencode.p2j.ui.chui;

import java.io.*;
import java.util.*;

import charva.awt.*;
import charva.awt.event.*;
import charvax.swing.*;
import com.goldencode.util.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;

/**
 * Thin Client generic CHUI driver.
 */
public class ThinClient
implements ClientExports,
           StateSynchronizer,
           Watcher,
           CloseListener,
           AdvanceListener,
           FlushListener
{
   /** Local frame ID. */
   private static final int INPUT_FRAME_ID = 1;

   /** Default pause before hide message text. */
   private static final String PBH_DEFAULT_TEXT = "Press space bar to " +
                                                  "continue.";
   
   /** Default pause before end message text. */
   private static final String PBE_DEFAULT_TEXT = "Procedure complete. " +
                                                  PBH_DEFAULT_TEXT;
   
   /** Exception message used in END-ERROR key processing. */
   private static final String END_ERR_MSG = "END-ERROR key function";
   
   /** Default status message displayed when input operation is active. */
   private static final String INPUT_DEFAULT_TEXT = 
      "Enter data or press " + Keyboard.kbLabel("END-ERROR") + " to end.";

   /** State of the focus on entry to waitFor: undefined */
   private static final int FOCUS_UNDEF = 0; 

   /** State of the focus on entry to waitFor: invalid */
   private static final int FOCUS_INVALID = 1; 

   /** State of the focus on entry to waitFor: valid */
   private static final int FOCUS_VALID = 2; 

   /** 
    * State of the focus on entry to waitFor: hidden, as in not visible.
    * This is not the HIDDEN flag state.
    */
   private static final int FOCUS_HIDDEN = 3; 

   /** Lookup mode for focusable widgets in frames: no wrap */
   private static final boolean FOCUS_NOWRAP = false;

   /** Lookup mode for focusable widgets in frames: wrap */
   private static final boolean FOCUS_WRAP   = true;

   /** Singleton. */
   private static ThinClient instance = null;
   
   /** Screen dimensions. */
   private static Dimension screenSize = null;

   /** Postponed VALUE-CHANGED event from BROWSE. */
   private KeyInput postponedValueChanged = null;
   
   /** This view refreshes the frame including decorations */
   boolean viewRefresh = false;
   
   /** This view displays frame data */
   boolean viewDisplay = false;
   
   /** This flag is on for the duration of view operation */
   boolean viewInProgress = false;

   /** 
    *  <code>false</code> has no effect.
    *  <code>true</code> forces full frame redrawing, including decorations. 
    */
   boolean nextViewRefresh = false;
   
   /** 
    *  <code>false</code> has no effect.
    *  <code>true</code> forces data only output.
    *  Normally, only one of {nextViewRefresh, nextViewDisplay} flags should be
    *  set to <code>true</code> or else the requirements are in contradiction.
    *  If that happens nonetheless, nextViewRefresh takes precedence. 
    */
   boolean nextViewDisplay = false;
   
   /** Currently active screen. */
   private Screen screenCurrent = new Screen();
   
   /** Reference to <code>Toolkit</code> instance. */
   private Toolkit tk = null;
   
   /** Server exports */
   private ServerExports server = null;

   /** Validation message from last failed validation operation. */
   private String validationError = null;

   /** Pause before hide state flag. <code>true</code> means pause is due. */
   private boolean needPause = true;

   /** 
    * Separate pause before hide flag for messages. <code>true</code> means 
    * pause is due. 
    */
   private boolean messageNeedPause = false;
   
   /**
    * Flag to force message clearing on the next MESSAGE stmt.
    */
   private boolean clearMessage = false;
   
   /** Pause before hide time for automatic frame hiding. */
   private int timePBH = -1;

   /** Flag that determines if pause before hide mode is active. */
   private boolean honorPBH = true;

   /** Text displayed while processing pause before hide. */
   private String textPBH = PBH_DEFAULT_TEXT;

   /** Status message displayed when application is running */
   private String defaultMessage = "";

   /** Status message displayed when input operation is active. */
   private String inputMessage = INPUT_DEFAULT_TEXT;

   /** STATUS DEFAULT */
   public static final int DEFAULT = 0;
   /** STATUS INPUT */
   public static final int INPUT   = 1;
   /** Widget HELP string. */
   public static final int HELP    = 2;
   
   /** Current active wait list. */
   private EventList currentEventList = null;
   
   /** Current registered ROW-DISPLAY events. */
   private Map<Integer, Integer> rowDisplayEvents = null;

   /** Flag indicating if a ROW-DISPLAY event is being executed. */
   private boolean inRowDisplayTrigger = false;

   /** Number of times when we've found that the GO event is pending. */
   private int isGoPending = 0;

   /**
    * An exception which causes all trigger processing recursions to be
    * terminated.
    */
   private ConditionException immediateException = null;

   /**
    * Contains levels of trigger nesting relative to active WAIT-FOR cycles.
    */
   private Stack<Integer> triggerNestingByWaitFor = new Stack<Integer>();

   /**
    * Contains true if the last trigger result (on this level) returned 
    * no-apply.
    */
   private Stack<Boolean> triggerResultNoApply = new Stack<Boolean>();

   /**
    * Contains true if the last trigger (on this level) did execute a nested
    * wait-for loop.
    */
   private Stack<Boolean> triggerHadWaitFor = new Stack<Boolean>();

   /** The action code currently being applied by the apply() method. */
   private int lastAppliedCode = KeyEvent.CHAR_UNDEFINED;
   
   /** Last key received from the keyboard. */
   private int lastKey = -1;
   
   /** Last help text. */
   private String helpText = null;

   /** Track input mode. */
   private int inputActive = 0;

   /** File system daemon. */
   @SuppressWarnings("unused")
   private FileSystemDaemon fsd = null;
   
   /** Stream daemon (provides access to streams by id). */
   private StreamDaemon sd = null;

   /** Process daemon. */
   @SuppressWarnings("unused")
   private ProcessDaemon pd = null;
   
   /** Tracks pseudo-terminals used for redirecting output to a stream. */
   private Map<Integer, RedirectedTerminal> redirections = 
      new HashMap<Integer, RedirectedTerminal>();
   
   /** Tracks pseudo-terminals which need deferred cleanup processing. */
   private Set<Integer> deferredCleanup = new HashSet<Integer>();
   
   /** Tracks the id of the current output stream (-1 for interactive). */
   private int currentStream = -1;
   
   /** Tracks the number of I/O interactions within the current block */
   private int interactions = 0;
   
   /** Tracks all page elements which are active. */
   private Map<Integer,PageElement> pageElements = 
      new HashMap<Integer, PageElement>();
   
   /** Tracks the streams in which a given page element is registered. */
   private Map<PageElement, Set<Stream>> streamMap = 
      new HashMap<PageElement, Set<Stream>>();

   /** Level of nesting (recursion) that exists with trigger processing. */
   private int triggerNesting = 0;
   
   /**
    * <code>true</code> if automatic clearing occurred during trigger 
    * processing.
    */
   private boolean autoCleared = false;

   /** Tracks any conditional up or down that must be applied. */
   private int conditionalUpDown = 0;
   
   /** Tracks the frame ID associated with the conditional up or down. */
   private int conditionalFrameId = 0;
   
   /** Cache of the cursor visibility. */
   private boolean cursorVisible = true;

   /** Stores the state of the cursor at the time input mode was activated. */
   private Stack<Boolean> oldcursor = new Stack<Boolean>();
   
   /** Track editing block usage. */
   private boolean inEditingBlock = false;

   /** Track APPLY with implicit target usage and there is no focus. */
   private boolean implicitApplyNoTrigger = false;
   
   /** Track resetTerminal usage in editing blocks. */
   private boolean hadReset = false;
   
   /** List of frames which should be deleted at server. */
   private ArrayList<Integer> removedFrames = new ArrayList<Integer>();

   /** List of DOWN frames which were wrapped. */
   private ArrayList<Integer> wrappedFrames = new ArrayList<Integer>();
   
   /** Track HELP function activity */
   private boolean helpActive = false;
   
   /** message line number where tiny input frame is placed */
   private int tinyInputLine = -1;
   
   /** Last frame which was displayed. */
   private Frame lastViewed = null;

   /** Counter of nested (just in case) alert boxes. */
   private int alertBoxActive = 0;

   /** Force automatic clearing. */
   private boolean afterChoose = false;
   
   /** Do not adjust focus in WAIT-FOR if set */
   private boolean noFocusAdjustment = false;
   
   /** Tracks the LEAVE event sent status */
   private boolean leaveSent = false;

   /** Flags if the current operation had an implicit or explicit pause. */
   private boolean hadPause = false;

   /** Stack of frames targeted by WAIT-FOR statements. */
   private Deque<Frame> waitForFrames = new ArrayDeque<Frame>();
   
   /** Stack of frames targeted by editing blocks. */
   private Deque<Frame> editingFrames = new ArrayDeque<Frame>();
   
   /** Stack of prior focus before an editing block started. */
   private Stack<Component> editingFocus = new Stack<Component>();

   /** Contains the action code of the last key input event. */
   private int lastActionCode = 0;
   
   /** Set to <code>true</code> when condition exception may be thrown. */
   private boolean condExit = false;

   /**
    * If <code>true</code> then <code>enable(..)</code> function is being run. 
    */
   private boolean inEnable = false;
   
   /** The direct terminal output driver. */
   private OutputPrimitives directOutput = null;

   /**
    * The stack of saved insert indicator states. -1 value indicates that the
    * saved state has been cleared (mostly because it has been applied to a
    * fill-in). The current states are pushed at the beginning of SET/UPDATE/
    * PROMPT-FOR/MESSSAGE UPDATE/MESSAGE SET (including EDITING) blocks, and
    * popped and their end. Note that this stack always keeps at least one
    * value (in order to restore insert indicator state after leaving
    * first-level prompting block).
    */
   private Stack<Integer> savedInsStates = new Stack<Integer>();

   /**
    * The stack of saved insert indicator states. Unlike savedInsState, saved
    * values cannot be cleared, and current states are pushed/popped only
    * for EDITING blocks.
    */
   private Stack<Integer> savedInsStateInitials = new Stack<Integer>();

   /** Status message stack. */
   private Stack<StatusData> statusStack = new Stack<StatusData>(); 
   
   /** 
    * Save pending focus (to which will be adjusted to on the next wait-for 
    * loop.
    */
   private Component pendingFocus = null;

   /** External procedure call nesting. */
   private int nesting = 0;

   /** Current status type. */
   private int statusType = DEFAULT;

   /** 
    * Save pending focus, when focus is disabled or hidden. In some case, it
    * will be used by the APPLY statement.
    */
   private Component pendingFocusApply = null; 

   /**
    * Determines whether any output has been performed using PUT SCREEN
    * statement after the last call of HIDE ALL. 
    */
   private boolean putScreenPerformed = false;

   /** Map of (widgetId, state) pairs, for which the HIDDEN flag changed */
   private Map<Integer, Integer> hiddenWidgets = null;

   /** Map of (widgetId, state) pairs, for which the VISIBLE flag changed */
   private Map<Integer, Integer> visibleWidgets = null;
   
   /** Track the last displayed and enabled editor widget. */
   private Editor lastEditorDisplayed = null;

   /** 
    * Flag indicating if caret draw is skipped on next cursor positioning, 
    * after a wait-for loop iteration. 
    */
   private boolean skipCaretDraw = false;

   /** Authentication mode flag. */
   private boolean authMode = false;

   /** Control editing mode processing. */
   private boolean ignoreNextStopEditing = false;

   /** 
    * Force wiping of the screen from PUT SCREEN artifacts on next frame 
    * hiding. 
    */
   private boolean forceWipeScreen = false;

   protected boolean messageClearQuirk = false;

   /** Synchronization object for pause/Ctrl-C handling. */
   private Object cleanupSynchronizer = new Object();
   /** Pause cleanup block. */
   private Runnable cleanupBlock = null;

   /** Last frame, which was made visible */
   private Frame lastExposedFrame = null;
   
   /**
    * Default constructor.
    */
   private ThinClient()
   {
      triggerResultNoApply.push(false);
      triggerHadWaitFor.push(false);
      triggerNestingByWaitFor.push(0);
      savedInsStates.push(-1);

      tk = Toolkit.getDefaultToolkit();

      // ensure there is no other window in the window list
      tk.clearWindowList();
      
      ProgressKeyboardMapper mapper = new ProgressKeyboardMapper();
      
      String type = Toolkit.getTermType();
      
      if (type != null && type.equalsIgnoreCase("vt220"))
      {
         mapper.setDelWorkaroundActive(true);
      }
      
      tk.setKeyboardMapper(mapper);
      tk.setMouseDisabled(true);
      screenSize = tk.getScreenSize();
      TypeAhead.init(this);

      // force export of other client-side services
      fsd = new FileSystemDaemon();
      sd  = new StreamDaemon(this);
      pd  = new ProcessDaemon(this, sd);
      
      // export our interfaces
      RemoteObject.registerNetworkServer(ClientExports.class, this);

      // install direct output driver with shadow buffer
      tk.setDirectOutput(new DirectPrimitives(tk.getDirectOutput()));
      
      // install the double buffered terminal driver
      DoubleBufferedTerminal dbt = new DoubleBufferedTerminal();
      tk.setInteractiveOutput(dbt);

      directOutput = tk.getDirectOutput();
   }

   /**
    * Get the singleton instance of this class.
    * 
    * @return   <code>null</code> if {@link #initializePrep} has not yet
    *           been called. Otherwise, the singleton instance of this class
    *           will be returned.
    */
   public static synchronized ThinClient getInstance()
   {
      return instance;
   }
   
   /**
    * Get the screen size.
    * 
    * @return   screen <code>Dimension</code>
    */
   public static Dimension getScreenSize()
   {
      return new Dimension(screenSize);
   }

   /**
    * Initialize core UI services and create or re-create the singleton
    * instance of this class.  This MUST only be called after the security
    * manager and dispatcher have been initialized BUT BEFORE any call to
    * {@link #getInstance}.
    */
   public static synchronized void initializePrep()
   {
      // this must be syncronized on the Window.class object, as CTRL-C may
      // appear during initialization (right after the thin client was 
      // constructed but before the Window was initialized), resulting in a
      // deadlock
      synchronized (Window.class)
      {
         // initialize the thin client and export its APIs
         if (instance == null)
         {
            // the process is just starting up
            Toolkit.initCharva();
            instance = new ThinClient();
            Window.getDefaultWindow();
         }
         else
         {
            // reset our state because we are going to initialize again 
            Toolkit tk = Toolkit.getNewToolkit();
   
            // reset the current primitives (it might be redirected)
            tk.setBitmap(null);
            tk.switchOutput(null);
   
            // reset other state
            ErrorManager.resetClient();
            Window.resetWindow();
            instance = new ThinClient();
         }
      }
   }
   
   /**
    * Initialize network services (must only be called after an authenticated
    * network connection/session has been obtained AND after 
    * {@link #initializePrep} has been called).
    * 
    * @param    session
    *           The connection to the server.
    */
   public static void initializePost(Session session)
   {
      ThinClient tc = getInstance();
      
      tc.server = (ServerExports)
         RemoteObject.obtainNetworkInstance(ServerExports.class, session);

      RemoteErrorData errorServer = (RemoteErrorData)
         RemoteObject.obtainNetworkInstance(RemoteErrorData.class, session);
         
      ErrorManager.initClient(errorServer);
      
      session.registerSynchronizer(tc);
   }
   
   /**
    * Returns <code>true</code> if the whole widget is not even partially
    * obscured by any frame from the specified array.
    *
    * @param   comp
    *          widget to test
    *
    * @param   above
    *          array of visible frames to test against; normally obtained by
    *          Screen.getFramesAbove() helper
    * @return  <code>true</code> if it is not obscured
    */ 
   public static boolean isAllShowing(Component comp, Frame[] above)
   {
      // get the widget's rectangle expressed in screen coordinates
      // some service widgets like BrowseColumn do not have parents and do not
      // take screen space and have to be excluded from the check
      if (comp.getParent() == null)
      {
         return false;
      }

      Point origin = comp.getLocationOnScreen();
      Rectangle bounds = new Rectangle(origin, comp.getSize());

      for (int i = 0; i < above.length; i ++) 
      {
         // frame rectangle is (practically) in screen coordinates
         Rectangle rect = above[i].getBounds();
         if (rect.intersection(bounds) != null)
         {
            return false;
         }
      }
      return true; 
   }

   /**
    * Clears all content from the terminal.
    * This is a direct output operation that clears the screen through
    * ncurses.
    */
   private void clearTerminal()
   {
      if (tk.isRedirected())
         return;

      // these are direct output calls      
      tk.clear();
      directOutput.setCursor(0, 0);
      directOutput.sync();
   }

   /**
    * Temporarily halts the client UI processing (key reading, normal drawing
    * mode)  After entering suspend mode, the current contents of the
    * terminal are cleared.  Use {@link #resume} to return to normal mode.
    */
   public void suspend()
   {
      TypeAhead.suspend();
      Window.suspend();
      clearTerminal();
      
      Toolkit.getDefaultToolkit().suspend();
   }

   /**
    * Re-enables client UI processing (key reading, normal drawing mode).
    * Before resuming the current contents of the terminal are cleared.  This
    * is the mechanism to return to normal processing mode after having
    * called {@link #suspend}. 
    */
   public void resume()
   {
      Toolkit.getDefaultToolkit().resume();
      Window.resume();
      TypeAhead.resume();
      refreshScreen();
   }
   
   /**
    * Redraw everything after suspending terminal or after use of terminal
    * stream.
    */
   public void refreshScreen()
   {

      // if terminal is not redirected, redraw everything because
      // we have cleared terminal in suspend().
      // WARNING!!!
      // PUT SCREENs will not survive this redrawing!
      // whether this constitutes a problem or a solution remains to be seen
      // with a testcase. 
      if (!tk.isRedirected())
      {
         tk.refresh(new Rectangle(new Point(0, 0), screenSize));
         tk.sync();
      }
   }
   
   /**
    * Indefinitely pause and display the default message. This method works
    * when in suspend mode and it will not trigger a full redraw of the
    * entire screen.
    *
    * @return   The key code of the key pressed to clear the pause or -1
    *           on any failure. 
    */
   public int suspendedPause()
   {
      Toolkit.getDefaultToolkit().resetMode();

      setNeedPause(false);      
      
      // honor pause before-hide setting
      if (honorPBH && timePBH != 0)
      {
         AWTEvent evt = null;
         KeyInput key = null;
         
         // the pause itself (reads keys directly)
         while ((evt = tk.checkForKeystroke()) == null);
         
         // maintain lastKey properly
         if (evt instanceof KeyInput)
         {
            key = (KeyInput) evt;
            lastKey = key.getKeyCode();
         }
         else
         {
            lastKey = -1;
         }
         
         try
         {
            if (key != null)
            {
               // honor CTRL-C processing
               if (lastKey == KeyEvent.VK_CTRL_C)
               {
                  lastKey = -1;
                  throw new StopConditionException("CTRL-C");
               }

               // honor the STOP, ENDKEY and END-ERROR actions
               processCondition(key.getActionCode(), -1, true); 
            }
         }
         
         catch (ConditionException ce)
         {
            // rebalance the invalidate mode
            tk.resetInvalidate();
            
            // let the original exception propagate
            throw ce;
         }
      }

      return lastKey;
   }
   

   /**
    * Creates a child process with all or some of its stdio handles redirected
    * to the pipes and the rest of them left unchanged.
    * Unchanged stdio handles are inherited from the calling process and
    * normally are connected to the process' terminal.
    * <p>
    * The handles redirected to the pipes can be used to communicate between
    * processes.
    * <p>
    * In Progress, there are following use cases for the redirected I/O:
    * <table style="text-align: left; width: 100%;" border="1" cellspacing="2"
    *  cellpadding="2">
    *   <tbody>
    *     <tr>
    *       <td style="vertical-align: top;">stdio\\use<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;
    *           font-weight: bold;">Interactive Process Launch<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;
    *           font-weight: bold;">INPUT THROUGH<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;
    *           font-weight: bold;">INPUT-OUTPUT THROUGH<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;"><span
    *           style="font-weight: bold;">OUTPUT THROUGH</span><br>
    *       </td>
    *     </tr>
    *     <tr>
    *       <td style="vertical-align: top;">stdin<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">TTY<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">TTY<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">pipe<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">pipe<br>
    *       </td>
    *     </tr>
    *     <tr>
    *       <td style="vertical-align: top;">stdout<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">TTY<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">pipe<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">pipe<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">TTY<br>
    *       </td>
    *     </tr>
    *     <tr>
    *       <td style="vertical-align: top;">stderr<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">TTY<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">pipe<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">pipe<br>
    *       </td>
    *       <td style="vertical-align: top; text-align: center;">TTY<br>
    *       </td>
    *     </tr>
    *   </tbody>
    * </table>
    *
    * @param   cmdline
    *          The command line to execute. The first element must be the
    *          program name.
    * @param   in
    *          File descriptor to be associated with the pipe that redirects
    *          stdin, or <code>null</code> if no redirection required.
    * @param   out
    *          File descriptor to be associated with the pipe that redirects
    *          stdout, or <code>null</code> if no redirection required.
    * @param   err
    *          File descriptor to be associated with the pipe that redirects
    *          stderr, or <code>null</code> if no redirection required.
    * @param   silent
    *          If <code>true</code> then no additional messages will be
    *          displayed.
    *
    * @return  The ID of the launched child process.
    */
   public int pseudoTerminalLaunch(String[]       cmdline,
                                   FileDescriptor in,
                                   FileDescriptor out,
                                   FileDescriptor err,
                                   boolean silent)
   {
      return Toolkit.pseudoTerminalLaunch(cmdline, in, out, err, silent);
   }
   
   /**
    * Wait for a specific child process (which was previously launched
    * using {@link #pseudoTerminalLaunch}) to exit.
    *
    * @param    pid
    *           The child process to launch.
    */
   public void pseudoTerminalWait(int pid)
   {
      Toolkit.pseudoTerminalWait(pid, true);
   }
   
   /**
    * Wait for a specific child process (which was previously launched using
    * {@link #pseudoTerminalLaunch}) to exit.
    * 
    * @param   pid
    *          The child process to launch.
    * @param   silent
    *          If <code>true</code> then no additional messages will be
    *          displayed.
    */
   public void pseudoTerminalWait(int pid, boolean silent)
   {
      Toolkit.pseudoTerminalWait(pid, silent);
   }
   
   /**
    * Renders the specified message into the message line.
    * 
    * @param   message
    *          The message body to be rendered.
    */
   public void message(String message)
   {
      message(message, ColorTable.MESSAGES, false, true);
      tk.sync();
   }

   /**
    * Renders the specified message into the message line.
    *
    * @param   message
    *          The message body to be rendered.
    * @param   cs
    *          color specification.
    *
    * @return  Message line number used to display message.
    */
   public int message(String message, int cs)
   {
      return message(message, cs, false, true);
   }
   
   /**
    * Renders the specified message into the message line.
    * 
    * @param   message
    *          The message body to be rendered.
    * @param   cs
    *          color specification.
    * @param   clear
    *          If <code>true</code> then the line(s) on which the message will
    *          be displayed, will be cleared first.
    * @param   toTerm
    *          <code>true</code> if the message should be sent to the terminal
    *          too.
    * 
    * @return  Message line number used to display message.
    */
   public int message(String message, int cs, boolean clear, boolean toTerm)
   {
      // check if the message is too large for 1 message line
      if (screenSize.width < message.length())
      {
         String[] lines = StringHelper.nolossWrapText(message,
                                                      screenSize.width);
         int num = -1;
         
         for (int i = 0; i < lines.length; i++)
         {
            // all lines except the first have leading whitespace removed
            if (i > 0)
               lines[i] = StringHelper.safeTrimLeading(lines[i]);
            
            // 1-word lines that are longer than the screen must be split
            while (screenSize.width < lines[i].length())
            {
               // output the front of the line
               message(lines[i].substring(0, screenSize.width),
                       cs, 
                       clear, 
                       toTerm);
               
               // shorten our line by the part that was output
               lines[i] = lines[i].substring(screenSize.width);
            }
            
            // use recursion to process each line
            num = message(lines[i], cs, clear, toTerm);
         }
         
         return num;
      }

      // query redirection
      boolean switched = isRedirected();
      int redir = -1;
      
      // bypass normal message line processing in redirected mode
      if (switched)
      {
         Stream out = streamFromId(currentStream);         
         
         // if the terminal is redirected, then send this message to the  
         // stream
         if (out != null && out.isUnnamed())
         {         
            // assumes that we are in a reset state
            tk.addString(message, 0, 0);
            flushStream(out);
         }
         
         // the message is displayed on the terminal only if the input is not
         // read from a null device, like /dev/null
         
         // switch back to interactive mode temporarily, to allow the message
         // to be displayed on the terminal
         if (toTerm)
         {
            redir = switchOutput(-1, false);
         }
      }
      
      if (toTerm == false)
      {
         return 0;
      }

      int i = 0;
      
      // these are the implicit conditions that cause the message lines to
      // be cleared before output (without any pauses)

      if (clearMessage      ||
          afterChoose       || 
          (triggerNesting > 0 && !autoCleared) ||
          (!messageNeedPause && !inEditingBlock))
      {
         // if there was something - refresh message area
         if (!Window.isMessageNull())
            clearPutScreenInMessageArea(-1);
         
         Window.setMessageText(null, null);
         autoCleared = true;
         afterChoose = false;
         clearMessage = false;
      }
      
      String[] content = Window.getMessageText();
      
      try
      {
         if (content[0] == null)
         {
            // 'clear' flag is no longer used
            content[0] = message;
            Window.setMessageText(content, new int[] {cs, -1});
            clearPutScreenInMessageArea(0);
         }
         else if (content[1] == null)
         {
            i = 1;

            // 'clear' flag is no longer used
            content[1] = message;
            Window.setMessageText(content, new int[] {-1, cs});
            clearPutScreenInMessageArea(1);
         }
         else
         {
            try
            {
               if (messageNeedPause)
                  pauseBeforeHide();
            }
            finally
            {
               // 'clear' flag is no longer used
               Window.setMessageText(new String[] {message, null}, 
                                     new int[] {cs, -1});
               clearPutScreenInMessageArea(-1);
            }
         }
      }
      finally
      {
         // if currently redirected, restore the redirection
         if (switched)
         {
            switchOutput(redir, false);
         }
         
         messageNeedPause = true;
      }
      
      return i;
   }

   /**
    * Renders the specified message into the status line.
    * 
    * @param   message
    *          The message body to be rendered.
    */
   public void status(final String message)
   {
      status(message, false);
   }
   
   /**
    * Renders the specified message into the status line. The message 
    * is trimmed only if the <code>noLengthLimit</code> flag is not set.
    * 
    * @param   message
    *          The message body to be rendered.
    * @param   noLengthLimit
    *          flag indicating the status message should not be trimmed
    */
   public void status(final String message, final boolean noLengthLimit)
   {
      int tempTerm = forceInteractive();
      
      Point   curpos = null;
      boolean curvis = true;
      
      if (inputActive > 0)
      {
         // the cursor is active in an edit operation, remember the current
         // cursor position (and make the cursor invisible to avoid flashing)
         curvis = tk.setCursorStatus(false);
         curpos = tk.getCursor();
      }
         
      // safely ignore any events that may be produced by processHide() since
      // it only may remove temp texts over the status line which is being
      // redrawn anyway.
      independentEventDrawingBracket(new Runnable()
      {
         public void run()
         {
            String txt = message;
            
            if (message == null)
            {
               if (inputActive > 0)
               {
                  if (helpText != null)
                  {
                     txt        = helpText;
                     statusType = HELP;
                  }
                  else
                  {
                     txt        = inputMessage;
                     statusType = INPUT;
                  }
               }
               else
               {
                  txt        = defaultMessage;
                  statusType = DEFAULT;
               }
            }
      
            // the whole status line will lose PUT SCREENs
            tk.refresh(Window.getStatusLineRectangle());
            Window.status(txt, noLengthLimit);
         }
      });
      
      helpText = null;
      
      if (curpos != null)
      {
         restoreCursor(curpos, curvis);
      }
      
      restoreRedirection(tempTerm);
   }
   
   /**
    * Process pause before hide as requested.
    */
   public void pauseBeforeHide()
   {
      if (honorPBH)
         pause(timePBH, textPBH, false);
   }
 
   /**
    * Process pause after page for terminal stream. 
    */
   public void pauseAfterPage()
   {
      pause(timePBH, textPBH, false);
   }
 
   /**
    * Applies the encoded event to the widget currently in focus in the given
    * frame, or directly to the specified widget.
    * 
    * @param    widgetId
    *           ID of frame or a contained widget.
    * @param    eventCode
    *           Progress-compatible event code.
    * @param    inbuf
    *           The screen buffer for the related frame if the widgetId is 
    *           from a widget that is a frame or is contained within a frame.
    *           May be <code>null</code> if the widget is not associated with
    *           any frame (e.g. the session window).
    * @param    eventList
    *           The list of the latest triggers (which may be more recent 
    *           than that used in any previous <code>WAIT-FOR</code>).
    * @param    explicit
    *           <code>true</code> if this APPLY call has an explicit target.
    *
    * @return   Updated data for the associated frame representing the current
    *           screen values of all widgets in that frame.  May be
    *           <code>null</code> if the frame has no data containing widgets
    *           or if the input screen buffer is <code>null</code> or is
    *           otherwise invalid.
    */
   public ScreenBuffer apply(int          widgetId,
                             int          eventCode,
                             ScreenBuffer inbuf,
                             EventList    eventList,
                             boolean      explicit)
   {  
      // when a reset occurs inside an editing block, apply is a NOP
      if (inEditingBlock && hadReset)
         return null;

      Component saveFocus  = getCurrentFocus();
      Component focusFrame = locateFrame(saveFocus);
      int focusState       = getFocusState(saveFocus);
      if (!(focusFrame instanceof Frame))
      {
         focusFrame = null;
      }

      // check if this is the NEXT-FRAME or PREV-FRAME event
      boolean frameChangeEvent = eventCode == Keyboard.SE_NEXT_FRAME ||
                                 eventCode == Keyboard.SE_PREV_FRAME;

      // determine pending focus - pending from hidden/disabled widget has
      // priority over pending from hidden frame
      Component pending = pendingFocus == null ? pendingFocusApply 
                                               : pendingFocus;

      boolean isFocusFrameHidden = focusFrame != null && 
                                   !focusFrame.isVisible();
      boolean isFocusInvalid     = focusState == FOCUS_INVALID && 
                                   !isFocusFrameHidden;
      boolean isFocusHidden  = focusState == FOCUS_HIDDEN || isFocusFrameHidden;
      boolean isFocusValid   = !isFocusHidden && !isFocusInvalid;
      boolean isPendingValid = false;

      // NEXT-FRAME and PREV-FRAME have a special behavior - no focus adjustment
      // takes place prior to event processing
      if (triggerNesting > 0 && !frameChangeEvent)
      {
         if (pending != null)
         {
            Component pendingFrame = locateFrame(pending);
            if (pendingFrame instanceof Frame && 
                ((Frame) pendingFrame).isAlive())
            {
               // determine if pending focus is valid - only if its enabled
               // and parent frame is visible. hidden widget state is OK
               isPendingValid = pending.isEnabled() && 
                                pendingFrame.isVisible();
            }
         }

         if (!explicit && saveFocus != null && focusFrame != null)
         {
            // must adjust focus if focus is disabled and implicit target is 
            // used
            if (isFocusInvalid)
            {
               // if current focus is invalid, search for focus
               searchForFocusWithoutEntry(saveFocus);
            }
            else
            {
               // if pending focus is valid and focus is hidden, set focus on
               // found pending
               if (isPendingValid && isFocusHidden)
               {
                  setFocusOn(pending);
                  
                  // reset all pending
                  pendingFocus = pendingFocusApply = null;
                  isPendingValid = false;
               }
            }
         }
      }
      
      // null input buffer just silently returns here
      setScreenBuffer(inbuf);
      
      Component inFocus = getWidget(widgetId);

      if (inFocus == null && widgetId != -1)
      {
         inFocus = getFrame(widgetId);
      }

      if (inFocus == null)
      {
         // the source of the event can be one of the dynamically
         // generated widgets
         inFocus = getCurrentFocus();
      }
      
      if (inFocus == null)
      {
         // OK, we have no frames and/or no focused widgets so use this as
         // the last resort (it will always exist), otherwise the super
         // constructor of KeyInput() will fail badly (on null) 
         inFocus = Window.getDefaultWindow();
      }
      
      Component tmpFrame = locateFrame(inFocus);
      
      if (tmpFrame instanceof Frame)
         inFocus = ((Frame) tmpFrame).getFrameHelper().getField(inFocus);
      
      if (triggerNesting > 0              &&
          inFocus instanceof P2JComponent &&
          inFocus.isVisible()             &&
          inFocus.isEnabled()             &&
          (tmpFrame == null || 
           tmpFrame.isVisible()))
      {
         // no pending is used after trigger invocation if target is valid 
         // before trigger invocation
         pending = null;
         isPendingValid = false;
      }
      
      KeyInput event = new KeyInput(eventCode, KeyEvent.KEY_TYPED, inFocus);

      // setup event list chain
      EventList save = currentEventList;
      eventList.setChain(currentEventList);

      currentEventList = eventList;
      lastAppliedCode = eventCode;

      // if this is an APPLY with implicit target and target is not a valid 
      // widget, then no trigger should be invoked
      boolean saveImplicit = implicitApplyNoTrigger;
      
      try
      {
         implicitApplyNoTrigger = !explicit && widgetId == -1;

         int tempTerm = forceInteractive();

         try
         {
            // TODO: not sure this one is true - might have worked with H519
            // only
            tk.setInvalidate(true);
            applyWorker(event);
            tk.setInvalidate(false, true);
         }
         
         finally
         {         
            restoreRedirection(tempTerm);
         }

         boolean targetValid = inFocus instanceof P2JComponent &&
                               inFocus.isVisible()             &&
                               inFocus.isEnabled()             &&
                               (tmpFrame == null || 
                                tmpFrame.isVisible());
         if (triggerNesting > 0 && !frameChangeEvent)
         {
            if (explicit)
            {
               if (!targetValid)
               {
                  // focus must be adjusted if we are in trigger nesting mode,
                  // and:
                  // 1. target is not valid
                  // 2. there was a valid pending focus before trigger
                  if (isPendingValid)
                  {
                     setFocusOn(pending);
      
                     // reset all pending
                     pendingFocus = pendingFocusApply = null;
                     
                     // if no valid pending, leave focus where it is
                  }
               }
               else
               {
                  // else, if target is valid, adjust focus only if focus was 
                  // not valid before trigger invocation
                  if (!isFocusValid)
                  {
                     if (tmpFrame != focusFrame && !isFocusHidden)
                     {
                        // adjust to target only if is not in the same frame 
                        // as the previous focus and the focus was not hidden 
                        // before trigger invocation
                        setFocusOn(inFocus);
                     }
                     else
                     {
                        // else, restore focus to previous one, only if this 
                        // is not the ENTRY event
                        if (event.getActionCode() != Keyboard.SE_ENTRY)
                        {
                           setFocusOn(saveFocus);
                        }
                     }
                  }
               }
            }
            else
            {
               // implicit case, refocus target, if its still valid
               // when executed inside an editing block, do not readjust focus
               if (targetValid && !inEditingBlock && widgetId != -1)
               {
                  setFocusOn(inFocus);
               }
            }
         }
      }
      finally
      {
         // restore flag, even if a condition is raised
         implicitApplyNoTrigger = saveImplicit;
      }

      if (eventCode == Keyboard.SE_ENTRY)
      {
         // For the APPLY "ENTRY" statement, we should enable zap and reset
         // insert mode for the ENTRY target (fill-in), or to the currently
         // focused fill-in if ENTRY has returned NO-APPLY.

         FillIn fin = null;
         Component focus = getCurrentFocus();

         if (triggerResultNoApply.peek() && focus != null)
         {
            if (focus instanceof FillIn)
            {
               fin = (FillIn) focus;
            }
         }
         else if (inFocus != null && inFocus instanceof FillIn)
         {
            fin = (FillIn) inFocus;
         }

         if (fin != null)
         {
            fin.setInsertMode(false);
            fin.enableZapMode();
         }
      }

      lastAppliedCode = KeyEvent.CHAR_UNDEFINED;
      // restore event list chain to its previous state
      currentEventList = save;
      
      return inbuf == null ? null 
                           : getScreenBuffer(inbuf.getBaseId(), inbuf.size());
   }
   
   /**
    * Rings a bell on the terminal.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * BELL
    * </pre>
    */
   public void bell()
   {
      if (!isRedirected())
         tk.beep();
   }
   
   /**
    * Performs CHOOSE processing.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>CHOOSE</b>
    * 
    * @param    frameId
    *           ID of frame this request came for
    * @param    c
    *           CHOOSE statement options
    * @param    wl
    *           Event list.
    *            
    * @return   result of CHOOSE processing.
    */
   public ChooseData choose(int frameId, final ChooseData c, EventList wl)
   {
      // must be called first!
      honorConditionalUpDown(false);
      
      final Frame frame = getFrame(frameId);

      // honor the outstanding next-prompt request if any
      int toFocusId = frame.getFocusId();
      frame.setFocusId(-1);

      Component[] list = null;
      
      if (c.isFieldMode())
      {
         int[] widgetIds = c.getFields();
         list = new Component[widgetIds.length];
         
         final Component[] comps = list;
         
         // access and enable all involved fields unconditionally
         eventBracket(true, new Runnable()
         {
            public void run()
            {
               for (int i = 0; i < comps.length; i++)
               {
                  comps[i] = getWidget(c.getField(i));
                  comps[i].setEnabled(true);
               }
            }
         });

         // honor the outstanding next-prompt request if any
         if (toFocusId != -1)
         {
            // validate the widget
            toFocusId = validateWidget(toFocusId, widgetIds);
            Component widget = getWidget(toFocusId);
            
            if (widget instanceof FillIn)
            {
               ((FillIn) widget).resetCaret();
            }
         }
      }
      else
      {
         list = new Component[1];
         
         list[0] = getWidget(c.getRowId());

         // set row mode help text
         setHelp(c.getRowHelp());
      }
      
      final Component[] comps   = list;
      final Integer[]   focusId = new Integer[1];
      
      focusId[0] = toFocusId;
      
      eventDrawingBracket(new Runnable()
      {
         public void run()
         {
            focusId[0] = frame.enterChoose(comps, c, focusId[0]);
         }
      });
      
      toFocusId = focusId[0];

      setNeedPause(false);
      
      ChooseData cdret = null;
      
      ScreenBuffer sb  = c.getScreenBuffer();
      ScreenBuffer sbo = null;
      
      lastViewed = frame;
      
      boolean saveEnabled = frame.isEnabled();
      
      try
      {
         // enable frame if necessary
         if (!saveEnabled)
            frame.setEnabled(true);

         BufferSizeManager.getInstance()
                          .notifyBlockingOperation(BlockingOperation.CHOOSE);

         // note that in CHOOSE WAIT-FOR does not set status line
         sbo = waitForWorker(wl,
                             toFocusId,
                             c.getPause(),
                             sb,
                             true,
                             false,
                             false);
      }
      
      catch (ConditionException ce)
      {
         // rebalance the invalidate mode
         tk.resetInvalidate();
         
         // let the original exception propagate
         throw ce;
      }
      
      finally
      {
         // restore enabled state
         if (saveEnabled != frame.isEnabled())
            frame.setEnabled(saveEnabled);

         if (c.isFieldMode())
         {
            // disable all involved fields unconditionally
            eventBracket(true, new Runnable()
            {
               public void run()
               {
                  for (int i = 0; i < comps.length; i++)
                  {
                     comps[i].setEnabled(false);
                  }
               }
            });
         }

         cdret = frame.getChooseData();
         cdret.setScreenBuffer(sbo);
         lastViewed = frame;
      }

      if (frame.isInternalGo())
         setNeedPause(true);
      
      // force automatic clearing of messages in any case.
      afterChoose = true;
      
      return cdret;
   }

   /**
    * Clears the data and colors for all fill-in fields of the frame. For
    * down frames, also clears side labels.
    * <p>
    * This method corresponds to this Progress statement:<br>
    * <b>CLEAR</b>
    * 
    * @param    frameId
    *           The ID of the frame to clear.
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used (no pause will
    *           occur even if the client detects that the user has had no
    *           chance to review the screen's contents).
    */
   public void clear(int frameId, boolean noPause)
   {
      clearWorker(getFrame(frameId), noPause, false);
   }
   
   /**
    * Clears all occurrences for the down frame and resets the current display
    * position to the top of the frame.
    * <p>
    * This method corresponds to this Progress statement:<br>
    * <b>CLEAR ALL</b>
    * 
    * @param    frameId
    *           The ID of the frame to clear.
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used (no pause will
    *           occur even if the client detects that the user has had no
    *           chance to review the screen's contents).
    */
   public void clearAll(int frameId, boolean noPause)
   {
      tk.setInvalidate(true);
      clearWorker(getFrame(frameId), noPause, true);
      getFrame(frameId).getFrameHelper().clear();
      tk.setInvalidate(false, true);
      getFrame(frameId).setAllClear(true);
   }

   /**
    * Implements the <b>DESELECT-ROWS() </b> <b>BROWSE </b> widget method.
    * 
    * @param   browseId
    *          ID of the browse widget this request came for
    */
   public void deselectRows(final int browseId)
   {
      eventDrawingBracket(new Runnable()
      {
         public void run()
         {
            Browse browse = (Browse) getWidget(browseId);
            
            if (browse != null)
               browse.deselectAll();
         }
      });
   }

   /**
    * Enables/disables input for all specified field-level widgets of the
    * frame.
    * If the enabled status changes, the widgets redraw themselves with the
    * appropriate visual cue added or removed. This redrawing is limited to
    * the widget rectangle, however, and should preserve any PUT SCREEN output
    * that may have appeared anywhere else on the frame.
    * <p>
    * This method maintains the tab order of the specified frame. Every widget
    * listed for enabling is moved to the end of the tab order
    * retaining the same order as specified in the list.  If a
    * frame with x, y and z widgets gets an enable call for y and
    * x (in that order), then the tab order will change from the
    * default <code>x y z</code> to <code>z y x</code>. In the case
    * where all widgets are enabled, no change to the tab order
    * occurs.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DISABLE ALL</b> <br>
    * <b>DISABLE ALL EXCEPT</b> <br>
    * <b>ENABLE ALL</b> <br>
    * <b>ENABLE ALL EXCEPT</b>
    * 
    * @param    frameId
    *           ID of frame this request came for
    * @param    widgetId
    *           array of field level widget IDs to be enabled or
    *           <code>null</code>, which means all of them.
    * @param    sb
    *           The current frame's screen buffer.
    * @param    enabled
    *           New state for widgets.
    */
   public void enable(int          frameId,
                      int[]        widgetId,
                      ScreenBuffer sb,
                      boolean      enabled) 
   {
      enable(frameId, widgetId, sb, enabled, false);
   }

   /**
    * Enables/disables input for all specified field-level widgets of the
    * frame.
    * If the enabled status changes, the widgets redraw themselves with the
    * appropriate visual cue added or removed. This redrawing is limited to
    * the widget rectangle, however, and should preserve any PUT SCREEN output
    * that may have appeared anywhere else on the frame.
    * <p>
    * This method maintains the tab order of the specified frame. Every widget
    * listed for enabling is moved to the end of the tab order
    * retaining the same order as specified in the list.  If a
    * frame with x, y and z widgets gets an enable call for y and
    * x (in that order), then the tab order will change from the
    * default <code>x y z</code> to <code>z y x</code>. In the case
    * where all widgets are enabled, no change to the tab order
    * occurs.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DISABLE ALL</b> <br>
    * <b>DISABLE ALL EXCEPT</b> <br>
    * <b>ENABLE ALL</b> <br>
    * <b>ENABLE ALL EXCEPT</b>
    * 
    * @param    frameId
    *           ID of frame this request came for
    * @param    widgetId
    *           array of field level widget IDs to be enabled or
    *           <code>null</code>, which means all of them.
    * @param    sb
    *           The current frame's screen buffer.
    * @param    enabled
    *           New state for widgets.
    */
   public void enable(int          frameId,
                      int[]        widgetId,
                      ScreenBuffer sb,
                      boolean      enabled,
                      boolean      noView) 
   {
      inEnable = true;
      Component focus = getCurrentFocus();
      
      // must be called first (should only be needed in prompt-for, set and
      // update situations but not in a normal enable language stmt)!
      honorConditionalUpDown(false);
      
      int tempTerm = forceInteractive();
      
      try
      {
         final Frame frame = getFrame(frameId);
   
         // quick exit if this a FRAME:SENSITIVE call
         if (widgetId != null     && 
             widgetId.length == 1 && 
             frameId == widgetId[0])
         {
            frame.setEnabled(enabled);
            return;
         }
   
         // check if frame will be reset or paused during view()
         boolean preview = (frame.isNeedReset() && 
                            !frame.getFrameHelper().isDown());
         
         if (enabled && preview)
         {
            // this is a view call
            viewWorker(frame, sb, widgetId, true);
         }
         
         startWidgetOutput();
         tk.setInvalidate(true);
         
         try
         {
            // list of affected widgets
            Component[] affw = null;
            int naffw = 0;
            
            push();
            
            try
            {
               FrameConfig cfg = (FrameConfig) frame.getRef();
               boolean seen = frame.isVisible() && enabled;
         
               if (!enabled && sb != null)
               {
                  // honor any changes in our data
                  setScreenBuffer(sb);
                  
                  // adjust DOWN value
                  frame.adjustDown(sb.getDown());
                  frame.setDynRow(sb.getRow());
                  frame.setDynCol(sb.getColumn());
               }
         
               // get list of all widgets in this frame
               Component[] list = frame.getContentPane().getComponents();
         
               // core enable/disable processing
               if (widgetId == null)
               {
                  // no widget list specified
                  
                  // enumerate all widgets in the frame and enable/disable all
                  // field level widgets
                  Component comp = null;
         
                  for (int i = 0; i < list.length; i++)
                  {
                     comp = frame.getFrameHelper().getField(list[i]);
                     
                     if ((comp instanceof AttributeContainer) &&
                        ((AttributeContainer)comp).isFocusable())
                     {
                        comp.setEnabled(enabled);
                        if (comp == focus && !enabled)
                        {
                           // save the next enabled widget, which may be used
                           // by APPLY
                           pendingFocusApply = getNextEnabledWidget(frame, 
                                                                    comp,
                                                                    FOCUS_WRAP);
                        }
                        naffw++;
         
                        if (enabled)
                        {
                           if (!comp.isVisible())
                           {
                              collectFlag(comp, true, true);
                              comp.setVisible(true);
                           }
         
                           collectFlag(comp, false, false);
                           comp.setHidden(false);
                        }
                     }
                  }
         
                  // maintain tab order
                  if (enabled && !cfg.isKeepTabOrder())
                  {
                     frame.restoreInitialTabOrder();
                  }
         
                  // copy the affected widgets into the list
                  if (naffw > 0)
                  {
                     affw = new Component[naffw];
                     int j = 0;
                     for (int i = 0; i < list.length; i++)
                     {
                        comp = frame.getFrameHelper().getField(list[i]);
                        
                        if ((comp instanceof AttributeContainer) &&
                           ((AttributeContainer)comp).isFocusable())
                        {
                           affw[j] = comp;
                           j++;
                        }
                     }
                  }
               }
               else
               {
                  // a widget list specified
                  naffw = widgetId.length;
                  if (naffw > 0)
                  {
                     affw = new Component[naffw];
                  }
                  Set<Integer> ids = new HashSet<Integer>(naffw);
         
                  // enable/disable all widgets in the given list
                  Component comp = null;
         
                  for (int i = 0; i < widgetId.length; i++)
                  {
                     comp = getWidget(widgetId[i]);
                     
                     comp = frame.getFrameHelper().getField(comp);
         
                     if ((comp instanceof AttributeContainer) &&
                        ((AttributeContainer)comp).isFocusable())
                     {
                        comp.setEnabled(enabled);
                        if (comp == focus && !enabled)
                        {
                           // save the next enabled widget, which may be used
                           // by APPLY
                           pendingFocusApply = getNextEnabledWidget(frame, 
                                                                    comp, 
                                                                    FOCUS_WRAP);
                        }
                     }
         
                     if (enabled)
                     {
                        if (!comp.isVisible() && 
                            (frame.isVisible() || 
                             !frame.isHidden()))
                        {
                           collectFlag(comp, true, true);
                           comp.setVisible(true);
                        }
         
                        collectFlag(comp, false, false);
                        comp.setHidden(false);
                     }
         
                     // maintain tab order
                     if (enabled                    &&
                         comp.isFocusTraversable()  &&
                         !cfg.isKeepTabOrder())
                     {
                        frame.getContentPane().moveToTop(comp);
                     }
         
                     // copy the affected widgets into the list
                     affw[i] = comp;
                     ids.add(widgetId[i]);
                  }
         
                  // for ENABLE stmt, if any other widgets remain, they are 
                  // moved last in tab-order
                  if (enabled && naffw > list.length && !cfg.isKeepTabOrder())
                  {
                     for (int i = 0; i < list.length; i++)
                     {
                        comp = list[i];
                        int wid = getWidgetId(comp);
                        if (ids.contains(wid))
                        {
                           continue;
                        }
                        
                        if (comp.isFocusTraversable())
                        {
                           frame.getContentPane().moveToTop(comp);
                        }
                     }
                  }
               }
               
               if (enabled)
                  frame.getFrameHelper().needClearRow();
               
               if (seen)
                  frame.setPaintable(false);
               
               // FRAME:SENSITIVE flag can only be set to ON by an ENABLE call
               // or by explicit FRAME:SENSITIVE = true
               if (enabled)
               {
                  frame.setEnabled(enabled);
               }
               
               if (seen)
                  frame.setPaintable(true);
            }
            
            finally
            {
               if (frame.isVisible())
               {
                  popTrash();
               }
               else
               {
                  pop();
               }
            }
            
            if (frame.isVisible())
            {
               // some widgets may be obscured by overlay frames
               // prepare for selective direct drawing
               final Frame[] above = screenCurrent.getFramesAbove(frame);
               
               final int         numAffw    = naffw;
               final Component[] affWidgets = affw;
               
               // draw widgets directly
               eventBracket(true, new Runnable()
               {
                  public void run()
                  {
                     ScreenBitmap bitmap = widgetDrawingOn(frame);
                     
                     for (int i = 0; i < numAffw; i++)
                     {
                        if (affWidgets[i].isVisible() &&
                            (above.length == 0 || 
                             isAllShowing(affWidgets[i], above)))
                           affWidgets[i].draw();
                     }
                     
                     widgetDrawingOff(bitmap);
                  }
               });
            }
            
            // ENABLE calls for widgets within a hidden frame will not make
            // any widgets visible
            if (enabled && !preview && !frame.isHidden())
            {
               // if frame is not visible, then force to view all widgets
               int[] wids = !frame.isVisible() ? null : widgetId;
               // this is a view call
               viewWorker(frame, sb, wids, true);
            }
         }
            
         finally
         {
            tk.setInvalidate(false, true);
            endWidgetOutput();
         }
      }
      
      finally
      {
         restoreRedirection(tempTerm);
         inEnable = false;
      }
   }

   /**
    * Enables/disables input for the specified field-level widget. This method
    * preserves the current tab order of the frame.
    * <p>
    * This method corresponds to these Progress statement: <br>
    * <b>widget:SENSITIVE=true/false</b>
    * 
    * @param   frameId
    *          ID of frame this request came for
    * @param   widgetId
    *          widget ID to be enabled or disabled
    * @param   sb
    *          The current frame's screen buffer.
    * @param   enabled
    *          New state of the widget
    */
   public void enable(int          frameId,
                      int          widgetId,
                      ScreenBuffer sb,
                      boolean      enabled)
   {
      Frame frame = getFrame(frameId);
      if (frame == null)
      {
         return;
      }
      FrameConfig cfg = (FrameConfig) frame.getRef();

      boolean save_kto = cfg.isKeepTabOrder();
      cfg.setKeepTabOrder(true);
      
      enable(frameId, new int[] {widgetId}, sb, enabled);
      
      if (!save_kto)
      {
         cfg.setKeepTabOrder(false);
      }
   }
   
   /**
    * Queries the enabled status of the widget.
    * 
    * @param    frameId
    *           ID of frame this request came for
    * @param    widgetId
    *           widget ID to be enabled or disabled
    * @return   <code>TRUE</code> if the widget is enabled
    */
   public boolean isEnabled(int frameId, int widgetId)
   {
      Frame frame = getFrame(frameId);
      if (frame == null)
      {
         return false;
      }
      
      Component comp = getWidget(widgetId);
      comp = frame.getFrameHelper().getField(comp);

      return comp.isEnabled();
   }

   /**
    * Queries the visible attribute of the frame.
    * 
    * @param    frameId
    *           ID of frame this request came for
    * @return   <code>TRUE</code> if the attribute is ON
    */
   public boolean isVisible(int frameId)
   {
      Frame frame = getFrame(frameId);
      if (frame == null)
      {
         return false;
      }
      
      return frame.getFrameHelper().config.isVisible();
   }

   /**
    * Underlines a set of widgets of the specified frame. If the target
    * frame is not a down frame, the frame is simply cleared.
    * 
    * @param    frameId
    *           The frame ID this request came for.
    * @param    widgetIds
    *           The array of field level widget IDs to be underlined.
    */
   public void underline(final int frameId, final int[] widgetIds)
   {
      // must be called first!
      honorConditionalUpDown();
      
      if (widgetIds == null)
         return;
      
      eventDrawingBracket(new Runnable()
      {
         public void run()
         {
            Frame frame = getFrame(frameId);
            FrameHelper fh = frame.getFrameHelper();
      
            if (!fh.isDown())
            {
               // clear the regular frame
               frame.getFrameHelper().clearWorker(true);
               frame.repaint();
            }
            else
           {
               Component[] list = new Component[widgetIds.length];
               for (int i = 0; i < widgetIds.length; i ++)
               {
                  list[i] = getWidget(widgetIds[i]);
               }
               fh.underline(list);
            }
         }
      });
   }

   /**
    * Positions the cursor on a new line in a down frame.  When 
    * <code>lines</code> is negative, this operates as a move of the cursor
    * <code>up</code>.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DOWN n </b>
    * <b>UP n </b>
    * 
    * @param    frameId
    *           ID of frame this request came for
    * @param    lines
    *           Number and direction of iterations of the data to move. If
    *           positive, the move is <code>down</code>, if negative the
    *           move is <code>up</code>.  0 is ignored.
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    * @param    sid
    *           The stream for which pending PUT or EXPORT output must be
    *           ended and written to the stream. -1 means that this is not
    *           necessary.
    */
   public void down(int frameId, int lines, boolean cond, int sid)
   {
      // do this regardless of anything else (even if we will be returning
      // immediately)
      endOutput(sid);
      
      if (lines == 0)
         return;
      
      Frame frame = (Frame) getWidget(frameId);
      
      if (frame == null)
         return;
      
      // check the frame for possible stream switching
      boolean newStream = switchStreams(frame);

      downWorker(frameId, lines, cond, newStream, false);
   }
   
   /**
    * Positions the cursor on a new line in a down frame.  When
    * <code>lines</code> is negative, this operates as a move of the cursor
    * <code>up</code>.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DOWN n </b>
    * <b>UP n </b>
    *
    * @param    frameId
    *           ID of frame this request came for
    * @param    steps
    *           Number and direction of iterations of the data to move. If
    *           positive, the move is <code>down</code>, if negative the
    *           move is <code>up</code>.  0 is ignored.
    * @param    conditional
    *           <code>true</code> when this call is for conditional up/down.
    * @param    newStream
    *           <code>true</code> when stream switch has just occurred
    * @param    noPause
    *           if <code>true</code> then no pause will be generated on cursor
    *           wrap.
    *           
    * @return   <code>true</code> if wrap is occurred. 
    */
   private boolean downWorker(int frameId, int steps, boolean conditional,
                              boolean newStream, final boolean noPause)
   {
      if (steps == 0)
         return false;
      
      Frame frame = (Frame) getWidget(frameId);
      
      if (frame == null)
         return false;
      
      final FrameHelper fh = frame.getFrameHelper();
      boolean down = fh.isDown();
      boolean redir = isRedirected();

      // process redirected down, conditional or not
      if (redir)
      {
         RedirectedTerminal rt     = redirections.get(currentStream);
         Stream             stream = sd.streamFromId(currentStream);
         boolean reducedDown = false;
         
         if (stream == null)
         {
            throw new IllegalStateException("Invalid stream id " +
                                            currentStream + ".");
         }
         
         // for redirected down frames, switching streams has flushed
         // the stream by now, which is equivalent to DOWN 1
         // detect and save this condition
         if (down      &&                 // it's a down frame
             newStream &&                 // switching streams has flushed
             !fh.isCurrentTouched())      // no output since last down
         {
            reducedDown = true;
         }

         // express the initial DOWN size in terms of newlines
         int lines = steps * fh.getIterationHeight();

         int prevLineNum = -1;

         boolean flushed = rt.flush();
         
         // if there was pending output in the redirected terminal, we flush
         // now
         if (flushed)
         {
            // DOWN X handles page breaks with special logic below so the
            // stream must not do any of the default page break processing
            // TO-DO: should the check below be for steps or lines?
            if (steps /*lines*/ > 1)
            {
               stream.setContiguous(true);
            }
            
            // following call may cause page break
            prevLineNum = flushRedirected(stream, true);
         }

         // clear the frame buffer early enough so that any views of the
         // frame (which may occur in the down processing below) won't
         // redraw the current iteration but rather one that is uninitialized
         if (down)
         {
            clearAll(frameId, true);
         }
         else
         {
            if (conditional)
            {
               eventBracket(new Runnable()
               {
                  public void run()
                  {
                     fh.clearWorker(true);
                  }
               });
            }
         }
         
         // end of conditional down processing
         if (conditional)
            return false;

         boolean deferView = false;
         
         // if an unconditional down is the first output for a given frame on
         // a page, it needs to draw the frame (headers etc...), this must
         // occur BEFORE down output for DOWN frames and AFTER the output
         // for non-DOWN frames
         if (!fh.isWithHeaders(stream))
         {
            if (down)
            {
               forceFrameDraw(rt, frameId, currentStream);
            }
            else
            {
               deferView = true;
            }
         }
         
         boolean forceNextPage = false;
         boolean forceNewLine  = false;
         
         int rawSize     = stream.rawPageSize();
         int pageSize    = stream.getEffectivePageSize();
         int nextLineNum = stream.rawNextLineNum();
         
         // output was flushed 
         boolean prevFlush = prevLineNum != -1 && prevLineNum != nextLineNum;

         // check here for possible page spill with one full iteration
         if (steps == 1 && lines > 1 && /* fh.isTrailingSkips() && */
             (nextLineNum + fh.getIterationHeight() > pageSize))
         {
            int delta = fh.getLeadingSkipSize() + fh.getTrailingSkipSize();
            
            delta += fh.getIterationHeight() - delta - 1;
            
            if (fh.getLeadingSkipSize() == -1 || 
                delta > (pageSize - nextLineNum))
            {
               // the page would spill, so: force page eject, bypass regular
               // down processing and force full iteration of blank lines
               steps = 0;
               forceNextPage = true;
            }
         }
 
         if (!flushed && !reducedDown)
         {
            stream.writeControlled("");
         }
         
         // DOWN 1
         if (steps == 1)
         {
            // flush the pending new line that we just added
            if (!flushed && !reducedDown)
            {
               // the rest of the full iteration
               for(int i = 1; i < lines; i++)
               {
                  stream.writeControlled("");
               }

               nextLineNum = flushRedirected(stream, false);
            }
         
            // only paged streams get additional processing below
            if (rawSize > 0)
            {
               // the previous flushing at the top of this worker may have
               // caused a page break, we detect that case here
               if (prevFlush)
               {
                  // if the previous line number was greater than the end
                  // of the writable portion of the page
                  if (prevLineNum > pageSize)
                  {
                     // output a single new line on the next page
                     forceNewLine = true;
                  }
               }
               else
               {
                  // no page break was caused by the previous flush, but
                  // would one be caused by the our next flush?
                  if (nextLineNum > pageSize)
                  {
                     // force a page break and then output a new line on the
                     // next page
                     forceNextPage = true;
                  }
               }
            }
         }
         
         // DOWN X
         if (steps > 1)
         {
            // by default output our lines 
            int numNL = lines - 1;
          
            // TO-DO: the DOWN reduction below is not well tested
            if (reducedDown)
            {
               numNL -= fh.getIterationHeight();
            }

            // paged streams get special/bizarre behavior with overflows
            if (rawSize > 0)
            {
               // if we are starting PAST the end of the writable page
               if (nextLineNum > pageSize)
               {
                  // only output 1 line on this page
                  // TO-DO: line or iteration?
                  numNL = 1;
                  
                  // then force the page break and output one more line 
                  forceNextPage = true;
               }
               else
               {
                  // we are starting BEFORE or AT the writable limit of the
                  // page, calculate where we are going to end our output
                  int estEnd = nextLineNum + lines - 1;
                  
                  // check if we are crossing the page boundary
                  if (estEnd > pageSize)
                  {
                     // calculate how many new lines can fit "legally"
                     int fin = pageSize - nextLineNum;
                     
                     // calculate how many extra newlines we have been told
                     // to output (these are "illegal" ones)
                     int extra = estEnd - pageSize;
                     
                     // only a maximum of 2 of the illegal newlines will be
                     // output
                     extra = extra <= 2 ? extra : 2;
                     
                     // calculate the actual number of legal + illegal
                     // newlines to be output
                     numNL = fin + extra;
                     
                     // force the page break and one more line on the next
                     // page
                     forceNextPage = true;
                  }
               }
               
               // explicitly forcing the page break means that the output
               // of our NL chars must not trigger the page break processing
               if (forceNextPage)
               {
                  stream.setContiguous(true);
               }
            }
            
            // output our new lines
            for(int i = 0; i < numNL; i++)
            {
               stream.writeControlled("");
            }

            flushRedirected(stream, false);
         }
         
         // manually force the page break if deferred above
         if (forceNextPage)
         {
            stream.forcePageBreak(false);
         }
         
         // output the frame and then a NL char if deferred above
         if (forceNextPage || forceNewLine || deferView)
         {
            // don't duplicate the footer/header processing since on a small
            // page (that is smaller than the size of headers + footers) this
            // processing can cause inappropriate recursion
            stream.setContiguous(true);
            
            // draw the frame
            forceFrameDraw(rt, frameId, currentStream);
            
            // push the output to the stream
            boolean bypass = rt.flush();
            
            // add single empty iteration and flush
            for(int i = 0; i < fh.getIterationHeight(); i++)
            {
               stream.writeControlled("");
            }

            flushRedirected(stream, bypass);
         }
         
         // end of redirected down processing
         return false;
      }

      // process interactive down, conditional or not
      if (frame.isStreamed())
      {
         // clearing of a streamed frame in interactive mode is required
         // this action should cause pausing on the terminal if redirection 
         // is still active   
         if (redirections.get(frame.getStreamId()) != null)
         {
            pause(-1, null, false); 
         }

         frame.getFrameHelper().clearCurrent();
         // no cursor advance needed in this case; we're done
         return false;
      }

      // advance the frame cursor
      if (down)
      {
         final int       num  = steps;
         final boolean[] wrap = new boolean[1];
         
         // interactive down is meaningful for the down frames only
         try
         {
            eventDrawingBracket(new Runnable()
            {
               public void run()
               {
                  wrap[0] = false;
                  
                  try
                  {
                     // This call can trigger pause if we are moving cursor
                     // over the edge of the frame. During the pause a 
                     // condition can be raised which will cause this block
                     // to exit abnormally.
                     wrap[0] = fh.moveCursor(num, noPause);
                  }
                  
                  finally
                  {
                     if (wrap[0])
                     {
                        viewRefresh = true;
                        viewDisplay = false;
                     }
                  }
               }
            });
         }
         
         finally
         {
            if (wrap[0])
               nextViewRefresh = true;
         }
         
         return wrap[0];
      }
      
      fh.clearCurrent();
      
      return false;
   }

   /**
    * Positions the cursor on a new line in a down frame.  When 
    * <code>lines</code> is negative, this operates as a move of the cursor
    * <code>up</code>.
    * <p>
    * The terminal will be temporarily redirected to the given stream for the
    * duration of this method call.  This allows the caller to avoid 
    * bracketing this call with {@link #switchOutput} invocations, as the
    * same result is achieved internal to this method's operation.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DOWN n </b>
    * <b>UP n </b>
    * 
    * @param    frameId
    *           ID of frame this request came for
    * @param    lines
    *           Number and direction of iterations of the data to move. If
    *           positive, the move is <code>down</code>, if negative the
    *           move is <code>up</code>.  0 is ignored.
    * @param    streamId
    *           The stream to be used as the temporary terminal redirection
    *           (only for the duration of this call).
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    * @param    endOut                
    *           <code>true</code> to force any pending PUT or EXPORT output
    *           to be ended for the given stream (so long as that stream is
    *           not the terminal).
    */
   public void down(int     frameId,
                    int     lines,
                    int     streamId,
                    boolean cond,
                    boolean endOut)
   {
      // do this regardless of whether we are about to return immediately
      // or not
      if (endOut)
      {
         endOutput(streamId);
      }
      
      if (lines == 0)
         return;
      
      boolean switched = currentStream == streamId;
      int tempOut = -1;

      if (!switched)
      {
         tempOut = switchOutput(streamId, true);
      }
      
      // check the frame for possible stream switching
      Frame frame = (Frame) getWidget(frameId);
      
      if (frame == null)
         return;

      boolean newStream = switchStreams(frame);

      try
      {
         downWorker(frameId, lines, cond, newStream, false);
      }
      
      finally
      {
         if (!switched)
         {
            switchOutput(tempOut, false);
         }
      }
   }
   
   /**
    * Get ID of the widget which has the input focus.
    * 
    * @return   ID of the focused widget.
    */
   public int getFocus()
   {
      Component focus = getCurrentFocus();
      Component frame = locateFrame(focus);
      
      if (frame instanceof Frame && !frame.isVisible())
         return -1;
      
      return ComponentConfig.getWidgetId(focus);
   }

   /**
    * Register the frame as a header or footer from the stream that is the
    * current redirected terminal output destination.  The terminal must be
    * non-interactive for this method to process properly.  Subsequent
    * pagination operations on the stream will cause the header or footer to
    * render as needed, without any further explicit request from the server.
    *
    * @param   frameId
    *          The id of the frame for this request.
    * @param   header
    *          <code>true</code> to register as a header, <code>false</code>
    *          to register as a footer.
    */
   public void registerPageElement(int frameId, boolean header)
   {
      registerPageElement(frameId, header, currentStream, false);
   }
   
   /**
    * Register the frame as a header or footer with the given stream as the
    * redirected terminal output destination.  Subsequent pagination
    * operations on the stream will cause the header or footer to render
    * as needed, without any further explicit request from the server.
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    * @param    streamId
    *           The stream to associate with the frame as a page element.
    */
   public void registerPageElement(int frameId, boolean header, int streamId)
   {
      registerPageElement(frameId, header, streamId, true);
   }
   
   /**
    * Register the frame as a header or footer with the given stream as the
    * redirected terminal output destination.  Subsequent pagination
    * operations on the stream will cause the header or footer to render
    * as needed, without any further explicit request from the server.
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    * @param    streamId
    *           The stream to associate with the frame as a page element.
    * @param    endOut
    *           <code>true</code> to force any pending PUT or EXPORT output
    *           to be ended for the given stream (so long as that stream is
    *           not the terminal).
    */
   public void registerPageElement(int     frameId,
                                   boolean header,
                                   int     streamId,
                                   boolean endOut)
   {
      if (endOut)
         endOutput(streamId);
      
      PageElement pe  = null;
      Integer     key = new Integer(frameId);

      if (streamId == -1)
      {
         throw new IllegalStateException("The given stream is the " +
                                         "interactive terminal!");
      }

      if (pageElements.containsKey(key))
      {
         pe = pageElements.get(key);
      }
      else
      {
         pe = new PageElement(this, frameId);
         pageElements.put(key, pe);
      }

      Stream s = sd.streamFromId(streamId);

      if (s == null)
      {
         throw new IllegalStateException("Invalid stream ID.");
      }
      
      // save off the mapping to ensure that a forced deregistration
      // can occur later
      if (streamMap.containsKey(pe))
      {
         streamMap.get(pe).add(s);
      }
      else
      {
         Set<Stream> streams = new HashSet<Stream>();
         streams.add(s);
         streamMap.put(pe, streams);
      }

      if (header)
      {
         s.registerHeader(pe);
      }
      else
      {
         s.registerFooter(pe);
      }
   }

   /**
    * Deregister (remove) the frame as a header or footer from the stream 
    * that is the current redirected terminal output destination.  The
    * terminal must be non-interactive for this method to process properly.
    * <p>
    * This is called (among other places) during hide processing on the
    * server, which must be safe to be called at any time.  This is true,
    * even if we aren't redirected, if the frame is not visible and/or if
    * there was no page element ever registered for this frame (because it
    * was never viewed).
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    */
   public void deregisterPageElement(int frameId, boolean header)
   {
      deregisterPageElement(frameId, header, currentStream);
   }
   
   /**
    * Deregister (remove) the frame as a header or footer from the given
    * stream.
    * <p>
    * This is called (among other places) during hide processing on the
    * server, which must be safe to be called at any time.  This is true,
    * even if we aren't redirected, if the frame is not visible and/or if
    * there was no page element ever registered for this frame (because it
    * was never viewed).
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    * @param    streamId
    *           The stream to associate with the frame as a page element.
    */
   public void deregisterPageElement(int     frameId,
                                     boolean header,
                                     int     streamId)
   {
      // Clear references to frame from terminal and from frame to stream. 
      Frame frame = (Frame) getWidget(frameId);
      
      if (frame != null)
      {
         RedirectedTerminal rt = redirections.get(frame.getStreamId());
         
         if (rt != null)
            rt.resetLastId(frameId);
         
         frame.setRedirected(false, -1);
      }

      // deregister page elements
      Integer     key = new Integer(frameId);
      PageElement pe  = pageElements.get(key);

      if (pe == null)
      {
         return;
      }

      if (streamId == -1)
      {
         return;
      }

      pageElements.remove(key);

      Stream s = sd.streamFromId(streamId);

      if (s == null)
      {
         throw new IllegalStateException("Invalid stream ID.");
      }

      // remove this stream from our list of mapped streams
      Set<Stream> streams = streamMap.get(pe);

      if (streams != null)
      {
         streams.remove(s);
         
         // remove the entry completely when no longer needed
         if (streams.isEmpty())
         {
            streamMap.remove(pe);
         }
      }
      
      if (header)
      {
         s.deregisterHeader(pe);
      }
      else
      {
         s.deregisterFooter(pe);
      }
   }

   /**
    * Deregister (remove) the frame as a header or footer from all streams
    * in which it is registered.
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    */
   public void forceDeregisterPageElement(int frameId, boolean header)
   {
      Integer     key = new Integer(frameId);
      PageElement pe  = pageElements.get(key);

      if (pe == null)
      {
         // we must never have been registered or we were cleanly
         // deregistered so we are done
         return;
      }

      pageElements.remove(key);

      Set<Stream> streams = streamMap.get(pe);
      
      if (streams != null)
      {
         for (Stream s : streams)
         {
            if (header)
               s.deregisterHeader(pe);
            else
               s.deregisterFooter(pe);
         }

         // cleanup
         streamMap.remove(pe);
      }
   }

   /**
    * Deregister (remove) the frame either as a header or footer from all 
    * streams in which it is registered.
    *
    * @param    frameId
    *           The id of the frame for this request.
    */
   public void forceDeregisterPageElement(int frameId)
   {
      Integer     key = new Integer(frameId);
      PageElement pe  = pageElements.get(key);

      if (pe == null)
      {
         // we must never have been registered or we were cleanly
         // deregistered so we are done
         return;
      }

      Set<Stream> streams = streamMap.get(pe);
      
      if (streams != null)
      {
         for (Stream s : streams)
         {
            s.deregisterHeader(pe);
            s.deregisterFooter(pe);
         }
         
         // cleanup
         streamMap.remove(pe);
      }
   }

   /**
    * Get the component which is in focus.
    * 
    * @return   <code>Component</code> instance which is in focus or 
    *           <code>null</code> if there is no focused widget.
    */
   private static Component getCurrentFocus()
   {
      return Window.getDefaultWindow().getCurrentFocus();
   }
   
   /**
    * Force the given frame to output to the given stream in redirected mode.
    *
    * @param    rt
    *           The current redirected terminal for the stream.
    * @param    frameId
    *           The frame to output.
    * @param    streamId
    *           The output destination.
    */
   private void forceFrameDraw(RedirectedTerminal rt,
                               int                frameId,
                               int                streamId)
   {
      rt.setSyncable(true);
      view(frameId, 
           null,                  //   frame buffer
           new int[] {frameId},   //   list of widgets
           streamId,              //   stream id
           false,                 //   register
           false,                 //   page top
           true);                 //   is view
      rt.setSyncable(false);
   }
   
   /**
    * Returns the column position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-COL
    * </pre>
    * 
    * @param    frameId
    *           The ID of the frame for this request.
    * 
    * @return   The column position for the upper-left corner of the frame.
    */
   public double getFrameCol(int frameId)
   {
      Frame frame = getFrame(frameId);
      int result = 0;
      
      if (frame != null && frame.isVisible() && !frame.isRedirected()) 
      {
         result = frame.getRef().getColumn() + 1;
      }
      
      return result;
   }

   /**
    * Returns the column position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress <code>FRAME:COLUMN</code> 
    * attribute.
    * 
    * @param    frameId
    *           The ID of the frame for this request.
    * 
    * @return   The column position for the upper-left corner of the frame.
    */
   public int getColumn(int frameId)
   {
      Frame frame = getFrame(frameId);
      int col = 0;
      
      if (frame != null) 
      {
         col = frame.getRef().getColumn();
         
         if (col == BaseConfig.INV_COORD)
         {
            col = 0;
         }
         
         col = col + 1;
      }
      
      return col;
   }
   
   /**
    * Returns the row position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress <code>FRAME:ROW</code>
    * attribute.
    * 
    * @param    frameId
    *           The ID of the frame for this request.
    * 
    * @return   The row position for the upper-left corner of the frame.
    */
   public int getRow(int frameId)
   {
      Frame frame = getFrame(frameId);
      
      FrameHelper fh = frame.getFrameHelper();
      FrameConfig config = (FrameConfig) fh.getRef();
      int row;
      if (frame.isRedirected())
      {
         row = config.getRowAttribute();
      }
      else
      {
         row = config.getRow();
      }
      
      if (row == BaseConfig.INV_COORD)
      {
         row = 0;
      }
      
      if (conditionalFrameId == frame.getFrameHelper().config.getID())
      {
         row += conditionalUpDown;
      }
      
      return row + 1;
   }

   /**
    * Returns the width of the specified frame (the <code>FRAME:WIDTH-CHARS
    * </code> attribute).
    * 
    * @param    frameId
    *           The ID of the frame for this request.
    * 
    * @return   The width of the frame.
    */
   public int getWidth(int frameId)
   {
      Frame frame = getFrame(frameId);
      int width = 0;
      
      if (frame != null) 
      {
         width = frame.getRef().getWidthChars();
         
         if (width < 0)
         {
            width = 0;
         }
      }
      
      return width;
   }

   /**
    * Returns the height of the specified frame (the <code>FRAME:HEIGHT-CHARS
    * </code> attribute).
    * 
    * @param    frameId
    *           The ID of the frame for this request.
    * 
    * @return   The height of the frame.
    */
   public int getHeight(int frameId)
   {
      Frame frame = getFrame(frameId);
      int height = 0;
      
      if (frame != null) 
      {
         height = frame.getRef().getHeightChars();
         
         if (height < 0)
         {
            height = 0;
         }
      }
      
      return height;
   }

   /**
    * Returns the number of iterations in a down frame.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-DOWN
    * </pre>
    * 
    * @param    frameId
    *           The ID of the frame for this request.
    * 
    * @return   The number of data iterations in the frame.
    */
   public int getFrameDown(int frameId)
   {
      return getFrame(frameId).getDown();
   }

   /**
    * Returns the current logical line number in a frame.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-LINE
    * </pre>
    *
    * @param    frameId
    *           The ID of the frame for this request.
    * @param    pendingUpDown
    *           The pending down value for this frame.
    * 
    * @return   1-based ordinal of the current data iteration of the frame.
    *           If the frame is not a down frame then 1 will be returned.
    */
   public int getFrameLine(int frameId, int pendingUpDown)
   {
      Frame frame = getFrame(frameId);
      return getFrameLine(frame, pendingUpDown);
   }

   /**
    * Returns the row position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-ROW
    * </pre>
    * 
    * @param    frameId
    *           The ID of the frame for this request.
    * 
    * @return   The row position for the upper-left corner of the frame.
    */
   public double getFrameRow(int frameId)
   {
      Frame frame = getFrame(frameId);
      int result = 0;
      
      if (frame != null && frame.isVisible()) 
      {
         result = frame.getRef().getRow() + 1;
      }
      
      return result;
   }
   
   /**
    * Returns the number of screen lines available for message display.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * MESSAGE-LINES
    * </pre>
    *
    * @return   The number of available lines in the message area.
    */
   public int getMessageLines()
   {
      return Window.RESERVED_LINES - 1;
   }

   /**
    * Getter for the <b>NUM-SELECTED-ROWS </b> attribute.
    * 
    * @param   browseId
    *          ID of the browse widget this request came for
    * 
    * @return  current value of NUM-SELECTED-ROWS attribute
    */
   public int getNumSelectedRows(int browseId)
   {
      Browse browse = (Browse) getWidget(browseId);
      
      if (browse != null)
         return browse.getRef().getNumSelectedRows();

      return 0;
   }

   /**
    * Retrieve the 0-based row index of the <i>n</i>th selected row in the
    * given browse widget.
    * 
    * @param   browseId
    *          ID of the browse widget associated with the request.
    * @param   n
    *          0-based index into the browse's list of currently selected
    *          rows.
    *          
    * @return  Row index associated with the <i>n</i>th selected row in the
    *          browse, or -1 if the associated Browse widget could not be
    *          identified.
    * 
    * @throws  IndexOutOfBoundsException
    *          if <code>selIndex</code> is less than 0 or greater than or
    *          equal to the number of selected rows.
    */
   public int getSelectedRowIndex(int browseId, int n)
   {
      Browse browse = (Browse) getWidget(browseId);
      
      if (browse != null)
         return browse.getRef().getSelectedRowIndex(n);

      return -1;
   }

   /**
    * Returns the number of screen lines available for frame display (does
    * not include the message or status lines).
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * SCREEN-LINES
    * </pre>
    *
    * @return   The number of available screen lines.
    */
   public int getScreenLines()
   {
      return Window.getScreenDimension().height;
   }

   /**
    * Returns the terminal type.  This usually corresponds to the value
    * set as the <code>TERM</code> environment variable on UNIX or Linux
    * unless set via {@link #setTerminalType}.  On Windows it depends upon
    * whether the terminal is in graphical or character mode.  If it is
    * in character mode, the result depends on the monitor type.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * TERMINAL
    * <pre>
    * 
    * @return   Terminal type for in this session.
    */
   public String getTerminalType()
   {
      return Toolkit.getTermType();
   }
   
   /**
    * Makes the specified frame or widget invisible.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>HIDE widget </b>
    * 
    * @param   widgetId
    *          ID of frame or widget this request came for
    * @param   noPause
    *          <code>true</code> if NO-PAUSE option was used
    */
   public void hide(final int widgetId, final boolean noPause)
   {
      eventDrawingBracket(new Runnable()
      {
         public void run()
         {
            try
            {
               Component comp = getWidget(widgetId);
               
               if (comp instanceof Frame)
               {
                  Frame frame = (Frame) comp;
                  // reset frame visibility tracking
                  frame.setVisibilityChanged(false);
      
                  if (frame.isVisible())
                  {
                     // for visible frames, set the explicit hide flag
                     frame.setExplicitHide(true);
      
                     if (pendingFocus == null &&
                         triggerNesting > 0   &&
                         locateFrame(getCurrentFocus()) == frame)
                     {
                        // if the focus frame is hidden, then first enabled
                        // widget on screen must be saved as pending focus;
                        // this is only for nested trigger mode
                        pendingFocus = getFirstEnabledWidget();
                     }
      
                     // reset VISIBLE flag for all widgets
                     Component[] comps = frame.getFrameHelper()
                                              .getContentPane()
                                              .getComponents();
                                              
                     for (int i = 0; i < comps.length; i++)
                     {
                        if (comps[i].isVisible())
                        {
                           collectFlag(comps[i], false, true);
                           comps[i].setVisible(false);
                           comps[i].setHighlighted(false);
                        }
                     }
                  }
                  
                  if (!frame.isDialog() && 
                      frame.isNeedPause() && 
                      !frame.getFrameHelper().isDown())
                  {
                     messageClearQuirk  = true;
                  }
                  
                  screenCurrent.hideFrame(frame, (noPause == false));
      
                  if (frame.isDialog())
                  {
                     // check to see whether another dialog has to be
                     // reactivated
                     Container parent = frame.getParent();
                     Component lst[]  = parent.getComponents();
                     
                     int nlst = lst.length;
                     int i = 0;
               
                     for (i = nlst - 1; i >= 0; i--)
                     {
                        if (lst[i].isVisible() && lst[i] instanceof Frame &&
                           (((Frame)lst[i]).isDialog()))
                        {
                           ((Frame)lst[i]).moveToTop();
                           lst[i].repaint();
                           break;
                        }
                     }
                  }
               }
               else if (comp instanceof AttributeContainer)
               {
                  if (((AttributeContainer) comp).isFocusable())
                  {
                     Frame frame = (Frame) locateFrame(comp);
                     // set as pending focus the next enabled widget
                     pendingFocusApply = getNextEnabledWidget(frame, 
                                                              comp, 
                                                              FOCUS_WRAP);
                  }
                  collectFlag(comp, false, true);
                  ((AttributeContainer) comp).getRef().setVisible(false);
                  
                  collectFlag(comp, true, false);
                  ((AttributeContainer) comp).getRef().setHidden(true);
   
                  // once hidden, its highlight state must be lost
                  comp.setHighlighted(false);
               }
            }
            
            finally
            {
               if (noPause)
               {
                  needPause = false;
               }
            }
         }
      });
   }

   /**
    * Hides all widgets and messages displayed in the message area. If pausing
    * is in effect, the standard message "Press space bar to continue" is
    * displayed on the status line and the user is required to press a key.
    * <p>
    * This method operates on the contents of the interactive terminal no
    * matter what redirection stream is in effect.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>HIDE ALL [NO_PAUSE] </b>
    * 
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding.
    */
   public void hideAll(final boolean noPause)
   {
      // do nothing if the terminal is redirected 
      if (isRedirected())
      {
         return;
      }

      if (inEditingBlock)
         hadReset = true;

      final boolean[] framesToHide = {false};
      
      eventBracket(true, new Runnable()
      {
         public void run()
         {
            // set explicitHide flag for all visible frames
            for (Component comp : screenCurrent.getTopLevelComponents())
            {
               if (!(comp instanceof Frame))
                  continue;
               
               Frame frame = (Frame) comp;

               // reset frame visibility tracking
               frame.setVisibilityChanged(false);

               if (frame.isVisible())
               {
                  frame.setExplicitHide(true);
                  frame.resetPlacementQuirk(true);
                  framesToHide[0] = true;

                  for (Component widget : frame.getFrameHelper()
                                               .getContentPane()
                                               .getComponents())
                  {
                     if (widget.isVisible())
                     {
                        collectFlag(widget, false, true);
                        widget.setVisible(false);
                     }
                  }
               }
            }
         }
      });
      
      try
      {
         hideMessage(noPause);
      }
      
      finally
      {
         status(null);
         
         try
         {
            eventBracket(true, new Runnable()
            {
               public void run()
               {
                  screenCurrent.hideAllFrames(!noPause);
               }
            });
         }
         
         finally
         {
            // clear entire screen directly
            ComponentConfig config = Window.getDefaultWindow().getRef();
            
            P2JColor color = ColorTable.getColor(((WindowConfig)config).
                                                  getDcolor(null));
            tk.blankBox(new Point(0, 0), screenSize,
                        color.getAttribute(), color.getColor());
      
            tk.refresh(new Rectangle(new Point(0, 0), screenSize));
            tk.sync();

            if (!framesToHide[0] &&
                !putScreenPerformed &&
                !Window.isInsertInAmbiguousState())
            {
               // redraw insert indicator
               Window.setInsertMode(Window.getInsertMode(), true);
            }
            else if (Window.getInsertMode() == Window.INSERT_MODE_INS)
            {
               // insert is ON and we didn't redraw it
               Window.setInsertInAmbiguousState();
            }
            
            putScreenPerformed = false;
         }
      }
   }

   /**
    * Hides all messages displayed in the message area. If pausing is in
    * effect, the standard message "Press space bar to continue" is displayed
    * on the status line and the user is required to press a key.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>HIDE MESSAGE [NO_PAUSE] </b>
    * 
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding
    *          messages
    */
   public void hideMessage(boolean noPause)
   {
      if (Window.isMessageNull() || isRedirected())
      {
         // nothing to do, avoids pause (which doesn't happen in Progress)
         // and extra work when the message is already empty
         return;
      }
      
      Point   curpos = null;
      boolean curvis = true;

      if (inputActive > 0)
      {
         // the cursor is active in an edit operation, remember the current
         // cursor position (and make the cursor invisible to avoid flashing)
         curvis = tk.setCursorStatus(false);
         curpos = tk.getCursor();
      }
         
      try
      {
         if (!noPause && messageNeedPause &&
             (inEditingBlock == false ||
             (inEditingBlock && clearMessage == false)))
         {
            pauseBeforeHide();
         }
      }
      finally
      {
         // force cleanup of message area
         Window.setMessageText(null, null, true);
         clearPutScreenInMessageArea(-1);
         
         if (curpos != null)
         {
            restoreCursor(curpos, curvis);
         }
         
         autoCleared = false;
      }
   }

   /**
    * Clears the type-ahead buffer if the input source is the keyboard.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>INPUT CLEAR </b>
    */
   public void inputClear()
   {
      TypeAhead.clear();
   }
   
   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    * 
    * @param   text
    *          Text to be displayed.
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE,
    *          not used if var parameter is <code>null</code>.
    * @param   var
    *          The variable to set or update.  <code>null</code> means no
    *          set or update will occur.
    * @param   auto
    *          <code>true</code> if the AUTO-RETURN option is specified. This
    *          is only valid if <code>var</code> is not <code>null</code>.
    * @param   format
    *          The format specification to use for the set or update.  This
    *          is only valid if <code>var</code> is not <code>null</code>.
    * @param   cs
    *          The color specification to use for the text.
    * @param   isInput
    *          <code>true</code> if the input is redirected to an unnamed
    *          stream.
    * @param   echo
    *          <code>true</code> if either the input or output stream is
    *          marked as echo.
    * 
    * @return  The result of the input operation if input parameter
    *          <code>var</code> is not <code>null</code> and operation
    *          finished successfully and/or if any key was pressed during
    *          editing/pausing.   If no pause occurred and there was no input
    *          (editing) in process, then <code>null</code> will be returned.
    */
   public MessageReturnValue message(String       text,
                                     boolean      set,
                                     BaseDataType var,
                                     boolean      auto,
                                     String       format,
                                     int          cs,
                                     boolean      isInput,
                                     boolean      echo)
   {
      MessageReturnValue mrv = null;
      Component focused      = getCurrentFocus();

      Point   curpos = tk.getCursor();
      boolean curvis = tk.setCursorStatus(false);
         
      int prevKey = lastKey;
      if (var != null && text.length() > screenSize.width)
         text = text.substring(0, screenSize.width);
      
      if (inEditingBlock && var != null)
      {
         clearMessage = false;
      }
      
      // query redirection
      boolean switched = isRedirected();
      int redir = -1;
      
      // switch to interactive mode to allow variable reading
      if (switched)
      {
         redir = switchOutput(-1, false);
      }

      boolean outputTerminal = true;
      boolean outputStream = true;
      
      if (var == null)
      {
         if (switched)
         {
            if (isInput)
            {
               outputTerminal = false;
               outputStream  = true;
            }
            else
            {
               outputTerminal = true;
               outputStream = true;
            }
         }
         else
         {
            outputTerminal = true;
            outputStream  = false;
         }
      }
      else
      {
         if (switched)
         {
            if (isInput)
            {
               outputTerminal = false;
               outputStream = echo;
            }
            else
            {
               outputTerminal = true;
               outputStream = true;
            }
         }
         else
         {
            if (isInput)
            {
               outputTerminal = false;
               outputStream = false;
            }
            else
            {
               // TODO: something is wrong here
               outputTerminal = true;
               outputStream = true;               
            }
         }
      }
      
      int row = outputTerminal ? message(text, cs, true, true) : -1;

      // detect if a key was pressed (because of pausing)
      if (prevKey != lastKey)
      {
         mrv = new MessageReturnValue();
         mrv.keycode = lastKey;
      }
      
      try
      {
         if (var != null)
         {            
            if (outputTerminal)
            {
               // handle SET/UPDATE modes
               mrv = tinyInput(row, var, format, auto, set);
            }
            else
            {
               mrv = new MessageReturnValue();
               mrv.var = var;
               mrv.keycode = -1;
               mrv.formattedVar = viewFillInVariable(var, format, auto);
            }

            // TODO: this method was removed at a certain time; check if 
            // needed to be replaced with other code
            // clearTempTextInMessageArea();
         }
      }
      
      finally
      {
         if (triggerNesting > 0 && var != null && focused == null)
         {
            // adjust focus to the first focusable widget in the topmost frame
            // only if there is trigger nesting
            Frame frame = screenCurrent.getTopFrame();
            if (frame != null)
            {
               Component first = getFirstEnabledWidget(frame);
               if (first != null)
               {
                  setFocusOn(first);
               }
            }
         }

         restoreCursor(curpos, curvis);   // syncs internally

         // if currently redirected, restore the redirection
         if (switched)
         {
            switchOutput(redir, false);
            // output the message to the stream only if there is no variable
            // or the var input executed succefully.
            if (outputStream && (var == null || (var != null && mrv != null)))
            {
               if (var != null && mrv != null)
               {
                  StringBuilder sb = new StringBuilder(text);
                  sb.append(" ");
                  sb.append(mrv.formattedVar);
                  text = sb.toString();
               }

               message(text, cs, false, false);
            }
         }
         
         if (mrv != null)
         {
            mrv.formattedVar = null;
         }
       }
            
      return mrv;
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statements:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX
    * MESSAGE list VIEW-AS ALERT-BOX SET/UPDATE var
    * ... and all other variants that include options such as title, buttons
    * and colors
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    color
    *           Color scheme to use.
    * 
    * @return   The edited result which can be mapped into a specified 
    *           button.
    */
   public logical messageBox(final Object[] list,
                                   logical  var,
                                   int      alertType,
                             final int      buttons,
                                   String   title,
                             final int      color)
   {
      final AlertBox[] alrtBox = new AlertBox[1];
      
      if (var == null)
         var = new logical(true);

      if (title == null)
      {
         switch (alertType)
         {
            case LogicalTerminal.ALERT_QUESTION:
               title = " Question ";
               break;
            case LogicalTerminal.ALERT_INFO:
               title = " Information ";
               break;
            case LogicalTerminal.ALERT_ERROR:
               title = " Error ";
               break;
            case LogicalTerminal.ALERT_MESSAGE:
               title = " Message ";
               break;
            case LogicalTerminal.ALERT_WARNING:
               title = " Warning ";
               break;
         }
      }
      
      // save and check focus
      final Component saveFocus  = getCurrentFocus();
      final int       focusState = getFocusState(saveFocus);

      EventList save = currentEventList;
      currentEventList = null;

      tk.setInvalidate(true);

      final String  ttle = title;
      final logical bool = var;
      
      eventBracket(new Runnable()
      {
         public void run()
         {
            alrtBox[0] = new AlertBox(list, buttons, bool, ttle, color);
         }
      });

      final Container pane = Window.getDefaultWindow().getContentPane();
      
      eventBracket(new Runnable()
      {
         public void run()
         {
            pane.add(alrtBox[0]);
            alrtBox[0].setFocus(alrtBox[0].getFocus());
            Toolkit.postEvent(new FocusEvent(AWTEvent.FOCUS_GAINED,
                                             alrtBox[0].getFocus()));
            alrtBox[0].repaint();
         }
      });
      
      tk.setInvalidate(false, true);
      
      try
      {
         alertBoxActive++;
         
         while (!alrtBox[0].inputDone())
         {
            AWTEvent evt = null; 
               
            // Progress does not maintain LASTKEY in a VIEW-AS ALERT-BOX so
            // this "direct" keyboard reading is safe (otherwise we would
            // need to call waitForKey() which does maintain LASTKEY)
            while (evt == null)
               evt = TypeAhead.getKeystroke(1000);

            if (evt instanceof KeyInput)
            {
               setNeedPause(false);
               KeyInput key = (KeyInput) evt;
               
               int action = key.getActionCode();
               
               // honor the GO, STOP, ERROR, ENDKEY and END-ERROR actions
               // (they DO NOT raise the associated condition but rather they
               // have alert-box specific functions)
               
               if (action == Keyboard.KA_GO)
               {
                  // dismiss the alert box with by selecting the current 
                  // button
                  key.setKeyCode(KeyEvent.VK_ENTER);
               }
               else if (action == Keyboard.KA_STOP)
               {
                  tk.beep();
                  continue;
               }
               else if (action == Keyboard.KA_ENDKEY ||
                        action == Keyboard.KA_ERROR  ||
                        action == Keyboard.KA_END_ERROR)
               {
                  // dismiss the alert box with an "unknown" result
                  eventDrawingBracket(new Runnable()
                  {
                     public void run()
                     {
                        alrtBox[0].finished(new logical());
                     }
                  });
               }
            }
            
            final AWTEvent nextEvt = evt;
            
            // let the button and alert box handle the event 
            eventDrawingBracket(new Runnable()
            {
               public void run()
               {
                  Toolkit.postEvent(nextEvt);
               }
            });
         }
      }

      finally
      {
         alertBoxActive--;
         
         eventDrawingBracket(new Runnable()
         {
            public void run()
            {
               if (alrtBox[0].isVisible())
                  alrtBox[0].setVisible(false);
               pane.remove(alrtBox[0]);
      
               switch (focusState)
               {
                  case FOCUS_UNDEF:
                     pane.setFocus(null);
                     break;
         
                  case FOCUS_VALID:
                  case FOCUS_INVALID:
                     // focus has to be restored no matter enabled or not
                     setFocusOn(saveFocus);
               }
            }
         });
         
         tk.setCursorStatus(true);
         
         currentEventList = save;
      }
      
      return alrtBox[0].getResult();
   }

   /**
    * Specifies the field to be put in focus during the next input operation
    * which involves this field of the frame.
    * <p>
    * If the very next operation does not involve the field, this request will
    * be disregarded.
    * <p>
    * The next high level input operation clears this information for the 
    * involved frame only. WAIT-FOR does not clear it, however.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>NEXT-PROMPT </b>
    * 
    * @param   frameId
    *          ID of frame this request came for
    * @param   widgetId
    *          widget to be put in focus
    */
   public void nextPrompt(int frameId, int widgetId)
   {
      Frame frame = getFrame(frameId);
      
      if (frame != null)
      {
         frame.setFocusId(widgetId);
      }
   }

   /**
    * Suspends processing for the specified number of seconds or until the
    * user presses any key, whichever comes first. The type-ahead buffer is
    * cleared. The specified (or no) message appears on the status line.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>PAUSE n MESSAGE/NO-MESSAGE </b>
    * 
    * @param   seconds
    *          Pause duration in seconds (0 means immediate return, -1 means
    *          infinite wait).
    * @param   text
    *          Either <code>null</code> if NO-MESSAGE is in effect, or the
    *          message text.
    * @param   noMessage
    *          <code>true</code> if NO-MESSAGE is in effect. This suppresses
    *          the message display, even if an explicit message text was
    *          passed.
    *
    * @return   The key code of the key pressed to clear the pause or -1
    *           if a timeout occurred (no key was pressed) or -2 if this
    *           was an immediate return due to a timeout of 0.
    */
   public int pause(int seconds, String text, boolean noMessage)
   {
      // reset need pause condition
      setNeedPause(false);

      // pause immediately returns when called with a 0 timeout
      if (seconds == 0)
      {
         //hide previously displayed status messages even if PAUSE 0 called
         if(!noMessage)
         {
            Window.setMessageText(null, null);
         }

         // in this case the lastkey should not be changed
         return -2;
      }

      // clear type-ahead buffer
      TypeAhead.clear();
      
      if (isRedirected() && !noMessage)
      {
         // if MESSAGE mode is in effect but the terminal is redirected,
         // switch output back to the terminal temporarily

         synchronized (cleanupSynchronizer)
         {
            cleanupBlock = new Runnable()
            {
               private int tempTerm = forceInteractive();
               
               @Override
               public void run()
               {
                  // take down temporary interactive terminal, restore redirection
                  switchOutput(tempTerm, false);
               }
            };
         }
      }

      // display temporary status message
      if (!noMessage)
      {
         status((text == null) ? makePauseText(seconds) : text);
      }
      
      // track that a pause did occur
      hadPause = true;

      BufferSizeManager.getInstance().notifyBlockingOperation(BlockingOperation.PAUSE);
      
      // pause (this will maintain LASTKEY too, which is required!)
      // target widget is not important
      KeyInput key = waitForKey(seconds, false);
      
      try
      {
         // honor the STOP, ENDKEY and END-ERROR actions
         if (key != null)
            processCondition(key.getActionCode(), -1, true);
      }
      finally
      {
         // erase status message
         if (!noMessage)
         {
            status(null);
         
            // hide messages
            Window.setMessageText(null, null);
         }

         pauseCleanup();
      }
      
      return lastKey;
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts the specified number of seconds or until
    * the user presses a key, whichever comes first. The specified (or no)
    * message appears on the status line.
    * <p>
    * This method corresponds to Progress statements:
    * <p>
    * <pre>
    * PAUSE seconds BEFORE-HIDE MESSAGE text
    * PAUSE seconds BEFORE-HIDE NO-MESSAGE
    * </pre>
    * 
    * @param    seconds
    *           Pause duration in seconds.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect or the
    *           message text to display.
    */
   public void pauseBeforeHide(int seconds, String text)
   {
      // we must store the new values even if honorPBH is false!
      timePBH = seconds;      
      textPBH = (text == null) ? PBH_DEFAULT_TEXT : text;
   }
   
   /**
    * Processes all outstanding events without blocking for user input.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>PROCESS EVENTS </b>
    *
    * @param   inbuf
    *          current screen buffer known to the server
    * 
    * @return  <code>ScreenBuffer</code> with the current values
    */
   public ScreenBuffer processEvents(ScreenBuffer inbuf)
   {
      // setup 
      setScreenBuffer(inbuf);

      EventList save = currentEventList;
      currentEventList = new EventList(true);

      tk.setInvalidate(true);
      processEventsWorker();      
      tk.setInvalidate(false, true);

      // cleanup
      currentEventList = save;
      return getScreenBuffer(inbuf.getBaseId(), inbuf.size());
   }
   
   /**
    * Register the specified ROW-DISPLAY events.
    *
    * @param   wids
    *          The id of the browse widget for which there are ROW-DISPLAY
    *          triggers.
    * @param   triggerIds
    *          The corresponding trigger ID for each widget.
    */
   public void registerRowDisplayEvents(int[] wids, int[] triggerIds)
   {
      if (rowDisplayEvents == null)
      {
         rowDisplayEvents = new HashMap<Integer, Integer>(wids.length);
      }
      for (int i = 0; i < wids.length; i++)
      {
         rowDisplayEvents.put(wids[i], triggerIds[i]);
      }
   }

   /**
    * Remove the ROW-DISPLAY triggers from the specified list.
    *
    * @param   triggers
    *          The triggers to be removed, if they are associated to a
    *          ROW-DISPLAY event.
    */
   public void removeRowDisplayEvents(int[] triggers)
   {
      if (rowDisplayEvents == null)
      {
         return;
      }

      Collection<Integer> values = rowDisplayEvents.values();
      Set<Integer> keys = rowDisplayEvents.keySet();
      
      for (int i = 0; i < triggers.length; i++)
      {
         if (values.contains(triggers[i]))
         {
            for (int wid : keys)
            {
               if (rowDisplayEvents.get(wid) == triggers[i])
               {
                  keys.remove(wid);
               }
            }
         }
      }
   }
   
   /**
    * Return the frame which contains the specified widget.
    * 
    * @param    widget
    *           Widget for which to find the frame.
    *
    * @return   The frame which contains specified widget or <code>null</code>
    *           if <code>null</code> is passed as the parameter.
    */
   public static Component locateFrame(Component widget)
   {
      while (widget != null && !(widget instanceof Frame))
      {
         widget = widget.getParent();
         if (widget instanceof ComboBox.DropDown)
         {
            ComboBox.DropDown dd = (ComboBox.DropDown) widget;
            widget = dd.getComboBox();
         }
      }
      
      return widget;
   }

   /**
    * Process array of frame/widget definitions provided by server.
    * 
    * @param   sd
    *          Array of screen definitions.
    */
   public void pushScreenDefinition(final ScreenDefinition[] sd)
   {
      // must be called before the main processing, should only be needed
      // during color stmt processing, otherwise this will be a NOP
      honorConditionalUpDown(false);
      
      for (int i = 0; i < sd.length; i ++)
      {
         final int             idx   = i;
         final int             fid   = sd[i].getRootFrameID();
         final ComponentConfig cc    = sd[i].getConfig(fid);
         final Frame           frame = (Frame) getWidget(fid);

         // suppress any drawing which may be caused by this screen definition
         // if the frame is invisible, unless the frame is to become visible.
         // Frame.postprocess() addresses this issue from the other end.
         // If there is a need to change conditions in which painting
         // events should be thrown away, appropriate change should be
         // done in the Frame.postprocess() as well.

         boolean ignorePaint = false;
         
         if (frame == null || cc == null || inRowDisplayTrigger)
         {
            ignorePaint = true;
         }
         else
         {
            ignorePaint = !frame.isVisible() && !cc.isVisible();
         }

         eventDrawingBracket(ignorePaint, true, new Runnable()
         {
            public void run()
            {
               // if the position was changed, before doing anything else,
               // we must hide the frame from the old position.
               if (frame != null && frame.isVisible())
               {
                  FrameConfig fc = (FrameConfig) cc;
                  if (!frame.isRedirected())
                  {
                     if (fc.isPositionChanged() || fc.isSizeChanged())
                     {
                        frame.setVisible(false);
                        frame.repaint();
                        frame.setVisible(true);
                     }
                  }
                  else
                  {
                     int toCol   = fc.getColumn();
                     int fromCol = frame.getFrameHelper().getRef().getColumn();
                     
                     if (fc.isPositionChanged() && toCol != fromCol && toCol >= 0)
                     {
                        // all the previous output must be moved to the new column
                        // if the column was modified
                        int sid = frame.getStreamId();
                        RedirectedTerminal term = redirections.get(sid);
                        term.moveOutputToColumn(fromCol, toCol);
                     }               
                  }
               }
      
               screenCurrent.pushDefinition(new ScreenDefinition[] { sd[idx] });
            }
         });
      }
   }
   
   /**
    * Process window configuration provided by server.
    * 
    * @param   config
    *          instance of <code>WindowConfig</code>
    */
   public void pushWindow(final WindowConfig config)
   {
      eventBracket(new Runnable()
      {
         public void run()
         {
            Window.pushConfig(config);
         }
      });
   }

   /**
    * Sets new location for a widget within its frame.
    * 
    * @param    frameId
    *           frame where the widget is placed
    *
    * @param    widgetId
    *           target widget which is being moved
    *
    * @param    row
    *           new row 
    * @param    col
    *           new column
    */
   public void setLocation(int frameId, int widgetId, int row, int col)
   {
      // TO-DO: implementation
      // today's server code is not aware of the assigned positions of widgets
      // withing frames, so that part has to be fixed first
      // See row_column.p testcase for how it should work
   }

   /**
    * Returns the ID of the currently active frame.
    * 
    * @return   The frame ID or -1 if there is no active frame (or if the
                active frame has been marked for destruction).
    */
   public int getActiveFrame()
   {
      Frame frame = getActiveFrameInt();

      if (frame == null)
         return -1;

      return ((AttributeContainer) frame).getRef().getID();
   }
   
   /**
    * Turns cursor visibility on or off.
    * <p>
    * This method corresponds to the Progress statements <br>
    * <b>PUT CURSOR </b> <br>
    * <b>PUT CURSOR OFF </b>
    * 
    * @param    onOff
    *           <code>true</code> if the cursor display is being turned on.
    */
   public void putCursor(boolean onOff)
   {
      try
      {
         tk.setCursorStatus(onOff);
         cursorVisible = onOff;
      }
      catch (IllegalStateException ise)
      {
         // TODO: log this
         ise.printStackTrace(System.err);
      }
   }

   /**
    * Puts cursor to the specified screen position.  If the cursor visibility
    * is OFF, this method will first make the cursor visible before moving
    * it.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>PUT CURSOR ROW COLUMN</b>
    * 
    * @param    row
    *           1-based screen row number
    * @param    column
    *           1-based screen column number
    */
   public void putCursor(double row, double column)
   {
      int tempTerm = forceInteractive();
      
      if(!cursorVisible)
      {
         putCursor(true);
      }

      int x = (int)column; 
      int y = (int)row; 
      
      x = x < 1 || x > screenSize.width  ? screenSize.width  - 2 : x - 1;
      y = y < 1 || y > screenSize.height ? screenSize.height - 1 : y - 1;

      tk.setCursorStay(x, y);
      
      restoreRedirection(tempTerm);                        
   }

   /**
    * Puts text on the screen at the specified position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    * @param    color
    *           The color specification or -1 for the default color.
    */
   public void putScreen(String text, double row, double column, int color)
   {
      putScreenPerformed = true;

      // trivial case first
      int textLen = text.length();
      if (textLen == 0)
      {
         return;
      }

      // isolate the peculiar case of putting text on the status line:
      // any output that would touch either of the two rightmost positions
      // is silently rejected
      Rectangle stat = Window.getStatusLineRectangle();
      int statLine = stat.getTop();
      int screenChars  = stat.getRight() + 1;
      int effRow = (int)(row - 1);
      int effCol = (int)(column - 1);

      if (effRow == statLine)
      {
         if (effCol + textLen > screenChars - 2)
         {
            return;
         }
      }

      // OK, there will be some output here
      putScreenDirect(text, row, column, color);
   }

   /**
    * Reads and returns one keystroke from the keyboard type-ahead buffer.
    * <p>
    * This method corresponds to the Progress statements <br>
    * <b>READKEY PAUSE n READKEY </b>
    * 
    * @param   seconds
    *          -1 means indefinite wait, 0 means no wait and positive values
    *          mean the number of seconds to wait
    * @return  -1 if timed out, otherwise the event code
    */
   public int readKey(int seconds)
   {
      // reset need pause condition
      setNeedPause(false);

      // detect special case where we must raise a stop condition
      if (inEditingBlock && hadReset)
      {
         try
         {
            stopNoAvailableWidgets();
         }
         
         finally
         {
            hadReset = false;
         }
      }
      
      // TO-DO: prove 100% it's the right place for this logic!
      // continueEditing() is a candidate for this action
      if (inEditingBlock)
      {
         // if a message is output after a READKEY is executed, all messages
         // must be cleared
         clearMessage = true;
      }

      BufferSizeManager.getInstance().notifyBlockingOperation(BlockingOperation.READKEY);
      
      // this will maintain LASTKEY too, which is required!
      // target widget is not important
      KeyInput key = waitForKey(seconds, false);

      if (inEditingBlock && 
          key != null    &&
          (key.getKeyChar() == ' ' ||
           key.getKeyCode() == KeyEvent.VK_ENTER))
      {
         // check if this key can fire the CHOOSE event for the current focus
         // if it can, and CHOOSE is a valid event for the focus widget, 
         // trigger CHOOSE and read another key before returning
         Component focused = getCurrentFocus();
         if (focused != null && focused instanceof Button)
         {
            Frame frame = (Frame) locateFrame(focused);
            int focusId = getWidgetId(focused);
            int frameId = getWidgetId(frame);
            
            // check if there is a trigger for this specific key
            int anyTriggerId = currentEventList.lookup(frameId, 
                                                       focusId, 
                                                       key.getKeyCode(), 
                                                       true, 
                                                       true);
            if (anyTriggerId == -1)
            {
               // check if there is a trigger for CHOOSE event
               int triggerId = currentEventList.lookup(frameId, 
                                                       focusId, 
                                                       Keyboard.SE_CHOOSE, 
                                                       true, 
                                                       true);
               if (anyTriggerId == -1 && triggerId != -1)
               {
                  // invoke trigger
                  invokeTriggers(focused, Keyboard.SE_CHOOSE, true);               
               }
            
               if (!isGoPending())
               {
                  // read another key only if trigger didn't generate a GO
                  // event
                  return readKey(seconds);
               }
            }
         }
      }
      
      return key == null ? -1 : key.getKeyCode(); 
   }

   /**
    * Implements the <b>REFRESH() </b> <b>BROWSE </b> widget method.
    * 
    * @param   browseId
    *          ID of the browse widget this request came for
    */
   public void refresh(int browseId)
   {
      tk.setInvalidate(true);

      Browse browse = (Browse) getWidget(browseId);
      
      if (browse != null)
         browse.refresh();
      cleanupQueue();      

      tk.setInvalidate(false, true);
   }

   /**
    * Defines a specific action for the key.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>ON key-label key-function </b>
    * 
    * @param   keyLabel
    *          the label of the key
    * @param   keyFunction
    *          the action to take when the user presses the key
    */
   public void remapKey(String keyLabel, String keyFunction)
   {
      Keyboard.remapKey(keyLabel, keyFunction);
   }

   /**
    * Scrolls the lines of the down frame up or down.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>SCROLL </b>
    * 
    * @param   frameId
    *          ID of frame this request came for
    * @param   fromCurrent
    *          <code>true</code> if FROM-CURRENT option was used
    * @param   up
    *          <code>true</code> if scrolling up
    */
   public void scroll(final int     frameId,
                      final boolean fromCurrent,
                      final boolean up)
   {
      // must be called first!
      honorConditionalUpDown(false);
      
      // reset view flags
      viewRefresh = true;
      viewDisplay = false;
      
      eventDrawingBracket(new Runnable()
      {
         public void run()
         {
            Frame frame = getFrame(frameId);
            FrameHelper fh = frame.getFrameHelper();
            fh.scroll(fromCurrent, up);
         }
      });
   }

   /**
    * Implements the <b>SELECT-ALL() </b> <b>BROWSE </b> widget method.
    * 
    * @param   browseId
    *          ID of the browse widget this request came for
    */
   public void selectAll(int browseId)
   {
      tk.setInvalidate(true);
      
      Browse browse = (Browse) getWidget(browseId);
      
      if (browse != null)
         browse.selectAll();
      cleanupQueue();      

      tk.setInvalidate(false, true);
   }

   /**
    * Force clearing of any temp-texts inside the message area.
    */
   public void clearPutScreenInMessageArea()
   {
      clearPutScreenInMessageArea(-1);
   }

   /**
    * Force clearing of any temp-texts inside the message area or on the
    * specific line in the message area.
    *
    * @param line
    *        Message line specification: either 0-based line number for
    *        specific line or -1 for the whole message area.
    */
   private void clearPutScreenInMessageArea(int line)
   {
      Rectangle bounds = Window.getMessageLineRectangle(line);
      tk.refresh(bounds);
      tk.sync();
   }
   
   /**
    * Implements the <b>SELECT-FOCUSED-ROW() </b> <b>BROWSE </b> widget
    * method.
    * 
    * @param   browseId
    *          ID of the browse widget this request came for
    */
   public void selectFocusedRow(int browseId)
   {
      tk.setInvalidate(true);
      
      Browse browse = (Browse) getWidget(browseId);
      
      if (browse != null)
         browse.selectFocusedRow();
      cleanupQueue();      

      tk.setInvalidate(false, true);
   }

   /**
    * Implements the <b>SET-REPOSITIONED-ROW() </b> <b>BROWSE </b> widget
    * method.
    * 
    * @param   browseId
    *          ID of the browse widget this request came for
    * @param   n
    *          1-based row number where the new record is displayed
    * @param   conditional
    *          if <code>true</code>, tries to reposition to the browse
    *          viewport existing row first
    */
   public void selectRepositionedRow(int browseId, int n, boolean conditional)
   {
      tk.setInvalidate(true);

      Browse browse = (Browse) getWidget(browseId);
      
      if (browse != null)
         browse.selectRepositionedRow(n - 1, conditional);
      cleanupQueue();      

      tk.setInvalidate(false, true);
   }

   /**
    * Notify browse widget at client side that query was repositioned.
    * 
    * @param   browseId
    *          ID of the browse widget this request came for
    * @param   force
    *          Force refreshing.
    */
   public void refreshStartRow(final int browseId, final boolean force)
   {
      eventDrawingBracket(new Runnable()
      {
         public void run()
         {
            Browse browse = (Browse) getWidget(browseId);
            
            if (browse != null)
               browse.refreshStartRow(force);
         }
      });

      clearPutScreenInMessageArea();
   }
   
   /**
    * Notify browse widget at client side that query was closed.
    * 
    * @param   browseId
    *          ID of the browse widget this request came for
    */
   public void queryClosed(final int browseId)
   {
      eventDrawingBracket(new Runnable()
      {
         public void run()
         {
            Browse browse = (Browse) getWidget(browseId);
            
            if (browse != null)
               browse.queryClosed();
         }
      });

      clearPutScreenInMessageArea(-1);
   }
   
   /**
    * Stores the value in a field of the active frame, which is in focus.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>FRAME-VALUE = expression </b>
    * 
    * @param   value
    *          the value to store in the frame field
    */
   public void setFrameValue(final BaseDataType value)
   {
      eventDrawingBracket(new Runnable()
      {
         public void run()
         {
            Component comp = getCurrentFocus();
      
            if (comp instanceof AttributeContainer)
            {
               ((AttributeContainer) comp).getRef().setValue(value);
            }
         }
      });
   }

   /**
    * Sets the terminal type.  This will cause the terminal to potentially
    * change its configuration if the new type is different from the old
    * type.
    * <p>
    * This processing also clears the input buffer and resets any key
    * mappings to their default values.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>TERMINAL</b>
    * 
    * @param   term
    *          Terminal type name.  If <code>null</code>, this resets the
    *          terminal without changing the type.
    */
   public void setTerminalType(String term)
   {
      // restore default key mappings (must be done on both server and client)
      Keyboard.resetKeyboard();
      
      // clear typeahead
      inputClear();
      
      if (term != null)
      {
         Toolkit.setTermType(term);
      }
      else
      {
         // the extra hide message here is needed because this one is
         // supposed to pause (but the one in hideAll won't)
         hideMessage(false);
         hideAll(true);
         
         // calling resetTermType causes catastrophic behavior so let's avoid
         // it (in Progress it seems to make things bad too but in ncurses
         // things just stop working altogether)
         // Toolkit.resetTermType();
      }
   }
   
   /**
    * Specifies a new default status message for the running procedure when
    * not in INPUT mode (when editing).
    * <p>
    * This method corresponds to the Progress statement 
    * <code>STATUS DEFAULT text</code>.
    * 
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    */
   public void statusDefault(String text)
   {
      statusDefault(text, false);
   }
   
   /**
    * Specifies a new default status message for the running procedure when
    * not in INPUT mode (when editing).
    * <p>
    * This method is used to display a status message, which is trimmed to
    * 63 chars only if the <code>noLengthLimit</code> flag is not set.
    * 
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    * @param    noLengthLimit
    *           flag indicating the status message should not be trimmed
    */
   public void statusDefault(String text, boolean noLengthLimit)
   {
      statusType = DEFAULT;
      defaultMessage = (text == null) ? "" : text;
      status(text, noLengthLimit);
   }

   /**
    * Specifies a new status message used when the user interface is blocked
    * on input.
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT text</code>.
    * 
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    */
   public void statusInput(String text)
   {
      if (text != null && text.equals(inputMessage))
         return;
      
      inputMessage = (text == null) ? "" : text;
      status(null);
   }

   /**
    * Reverts the input status message to the default.
    */
   public void statusInputRevert()
   {
      statusInput(INPUT_DEFAULT_TEXT);
   }

   /**
    * Disables display of the input status message.
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT OFF</code>.
    */
   public void statusInputOff()
   {
      statusInput(null);
   }
   
   /**
    * Shows all fields of the specified frame on screen.
    * 
    * @param    widgetId
    *           The frame ID or widget this request came for.
    * @param    frameBuf
    *           The screen buffer with current values of widgets.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    isView
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    */
   public void view(int          widgetId,
                    ScreenBuffer frameBuf,
                    int[]        widgetIds,
                    boolean      isView)
   {
      view_(widgetId, frameBuf, widgetIds, isView, true);
   }
   
   /**
    * Shows all fields of the specified frame on screen.
    * <p>
    * Notes on the logic of view/display calls in redirected mode:
    * <p>
    * Listed below are steps taken in redirected mode before the common worker
    * viewWorker() is called.
    * <nl>
    *  <li>
    *     Check the redirected terminal (RT) ownership. If it's owned by
    *     another frame, do frame switching and set SWITCHED flag.
    *     otherwise clear SWITCHED flag and proceed with the next step.<br>
    *     Frame switching flushes the RT buffer to the stream S and sets 
    *     the RT ownership to F. The RT buffer becomes fresh.
    *  <li>
    *     Check if the RT buffer is fresh. If it is not, all the remaining 
    *     steps are skipped.
    *  <li>
    *     The RT buffer is fresh and owned by the frame F. This is the only 
    *     moment when we can determine whether:
    *     <ul>
    *       <li> the frame has to be drawn with or without headers;
    *       <li> the current page needs to be ejected;
    *       <li> the current page needs to be contiguous.
    *     </ul>
    *     The following steps do these determinations.
    *  <li>
    *     Query the stream S and obtain the following values:
    *     <ul> 
    *        <li> a flag FRESH-PAGE, telling whether the page is fresh 
    *             or contains some output (page headers are not considered
    *             as output here) 
    *        <li> page size PS;
    *        <li> remaining lines on the page RL. This value accounts for 
    *             the size of the current output within S and all defined  
    *             page footers.
    *     </ul>
    *  <li>
    *     Query the frame F and obtain the following values:
    *     <ul> 
    *        <li> a flag DOWN-FRAME which is true for the down frame;
    *        <li> a flag PAGE-LIST which is true if DOWN-FRAME is true and
    *             when F is not in pagedList for the stream S (pagedList gets
    *             reset with page break notifications as well as during frame
    *             switching and reflects the fact that this down frame has
    *             contributed some output to the current page with frame 
    *             headers already in place)
    *        <li> size of frame headers HS;
    *        <li> size of frame body BS
    *     </ul>
    *  <li>
    *     Determine whether the frame F needs headers and set NEED-HDR flag
    *     accordingly.<br>
    *     Headers are needed if any condition below is true:
    *     <ul> 
    *        <li> frame switch has happened, SWITCHED flag is set;
    *        <li> fresh page flag FRESH-PAGE is set;
    *        <li> this is a 1-down frame (DOWN-FRAME flag is reset);
    *        <li> PAGE-LIST flag is set.
    *     </ul>
    *  <li>
    *     Calculate the number of lines the frame F needs on the page as<br>
    *     <code>FS = (NH ? HS : 0) + BS</code>
    *     <br>
    *  <li>
    *     Handle the large output case. The large output means the frame size  
    *     FS is greater than the page size PS and output is paged. If this is
    *     the case:
    *     <ul> 
    *        <li> check whether we are on the fresh page, and if we are not, 
    *             call the stream S and force page break;  
    *        <li> now we are on the fresh page; call the stream S and set
    *             contiguous mode.  
    *     </ul>
    *     The contiguous mode gets reset with the next flush and page break
    *     operation.<br>
    *     In contiguous mode the rest of steps is skipped.
    *  <li>
    *     Determine whether the page break is required. If the output is paged 
    *     and the frame size FS is greater than the remaining lines RL, call 
    *     the stream S and force page break.<br>
    *     Page break outputs Form Feed character, resets page state variables
    *     and outputs page headers. During page headers drawing, the RT will 
    *     be used for drawing, but this is safe since at this time the RT is 
    *     fresh and the page break processing will return it back fresh.
    * </nl>
    * Now the current view/display operation can proceed with drawing into 
    * the RT buffer with or without frame headers according to
    * the NEED-HDR flag. 
    * 
    * @param    widgetId
    *           The frame ID or widget this request came for.
    * @param    frameBuf
    *           The screen buffer with current values of widgets.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    isView
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    * @param    markTarget
    *           <code>true</code> if the call is for regular frame,
    *            <code>false</code> for page elements rendering
    */
   private void view_(int          widgetId,
                      ScreenBuffer frameBuf,
                      int[]        widgetIds,
                      boolean      isView,
                      boolean      markTarget)
   {
      Component widget = getWidget(widgetId);
      
      // nothing to do?
      if (widget == null)
         return;

      // check the frame for possible stream switching
      int   frameId = WidgetId.frameIdFromWidgetId(widgetId);
      Frame frame   = (Frame) getWidget(frameId);

      if (frame != null)
      {
         switchStreams(frame);
      }

      if (!isRedirected())
      {
         FrameHelper fh = (frame != null) ? frame.getFrameHelper() : null;

         if (fh != null && fh.isDown())
            fh.prepareIteration(widget, frameBuf, widgetIds, isView);
         
         // must be called first!
         honorConditionalUpDown();

         // do the drawing
         viewWorker(widget, frameBuf, widgetIds, isView);
         return;
      }

      // the rest of the code is for redirected mode only

      Stream             s  = sd.streamFromId(currentStream);

      if (s == null)
      {
         ErrorManager.recordOrThrowError(1387, "Attempt to write to closed stream");
         return;
      }

      RedirectedTerminal rt = redirections.get(currentStream);

      // filter out the case of utility call to render a page element
      if (markTarget)
      {
         // check the redirected terminal's ownership and
         // notify the redirected terminal about the new target frame
         boolean freshRT     = false;
         boolean switched    = switchFrames(currentStream, frameId);
         boolean needHeaders = false;

         // figure out the redirected terminal's buffer state
         if (switched)
         {
            freshRT = true;
         }
         else
         {
            freshRT = rt.isFresh();
         }

         // check to see if the redirected terminal's buffer is fresh
         if (freshRT)
         {
            // this is the moment for establishing the place of the frame
            // with the stream

            // query the stream and collect all state variables we may need
            boolean freshPage = s.isFreshPage();
            int     pageSize  = s.getPageSize().intValue();
            int     remLines  = s.getAvailableLines();

            // query the frame and collect all state variables we may need
            FrameHelper fh        = frame.getFrameHelper();
            boolean     downFrame = fh.isDown();
            int         headSize  = fh.getFixedHeight();
            boolean     noHeaders = !fh.isWithHeaders(s);
            boolean     box       = ((FrameConfig)fh.getRef()).isBox();
            int         bodySize  = fh.getIterationHeight();

            // determine whether this frame needs to be drawn with headers
            needHeaders = (switched | freshPage | !downFrame | noHeaders);

            // the remaining steps are needed for the paged output only
            if (pageSize > 0)
            {
               // calculate the frame size on the output
               int frameSize = (needHeaders ? headSize : 0) + bodySize;

               // boxed frames that have to draw their headers must account
               // for the top border line (in redirected mode only the top
               // border draws)
               if (box && needHeaders)
                  frameSize++;
               
               // handle the large output case where the frame is bigger than
               // there is available room
               if (frameSize > remLines)
               {
                  if (!freshPage)
                  {
                     // break the page now (early)
                     s.forcePageBreak();
                     
                     // ensure that the full frame is redrawn
                     needHeaders = true;
                  }
   
                  // ensure uninterruptible output for this whole frame
                  s.setContiguous(true);
               }
            }
         }

         // register event listeners
         if (s != null)
         {
            s.registerCloseListener(this);
            s.registerAdvanceListener(this);
            s.registerFlushListener(this);
         }

         // the final preparation for the frame drawing
         if (needHeaders)
         {
            // force drawing of headers and iteration
            nextViewRefresh = true;
            nextViewDisplay = false;
         }
         else
         {
            // force drawing of iteration only
            nextViewRefresh = false;
            nextViewDisplay = true;
         }
      }

      // must be called first!
      honorConditionalUpDown();

      // do the drawing
      viewWorker(widget, frameBuf, widgetIds, isView);
   }
   
   /**
    * Shows all fields of the specified frame on screen.
    * <p>
    * The terminal will be temporarily redirected to the given stream for the
    * duration of this method call.  This allows the caller to avoid 
    * bracketing this call with {@link #switchOutput} invocations, as the
    * same result is achieved internal to this method's operation.
    * 
    * @param    widgetId
    *           The frame ID or widget this request came for.
    * @param    frameBuf
    *           The screen buffer with current values of widgets.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    streamId
    *           The stream to be used as the temporary terminal redirection
    *           (only for the duration of this call). If this is not -1,
    *           then the any pending PUT or EXPORT output on this stream will
    *           be ended/forced out.
    * @param    register
    *           <code>true</code> if the frame associated with this request
    *           must be registered as a page element with the given stream. 
    * @param    pageTop
    *           <code>true</code> if the page element registration is for
    *           a page top element, <code>false</code> for a page bottom
    *           element.  This parameter is ignored if <code>register</code>
    *           is <code>false</code>.
    * @param    isView
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    */
   public void view(int          widgetId,
                    ScreenBuffer frameBuf,
                    int[]        widgetIds,
                    int          streamId,
                    boolean      register,
                    boolean      pageTop,
                    boolean      isView)
   {
      // end any PUT or EXPORT output that is pending and write it to the
      // stream
      endOutput(streamId);
      
      boolean switched = currentStream == streamId;
      int tempOut = -1;

      if (!switched)
      {
         tempOut = switchOutput(streamId, true);
      }
      
      if (register)
      {
         // this deregistration removes the current registration for this
         // page element for the following reasons:
         // - the view_ call produces a single copy of frame anyway
         // - the registration that follows keeps the right order of elements
         deregisterPageElement(frameBuf.getBaseId(), pageTop, streamId);
      }
      
      try
      {
         view_(widgetId, frameBuf, widgetIds, isView, true);
      }
      
      finally
      {
         // (re-)register page element now
         if (register)
         {
            registerPageElement(frameBuf.getBaseId(), pageTop, streamId);
         }

         if (!switched)
         {
            switchOutput(tempOut, false);
         }
      }
   }
   
   /**
    * Provides a composite helper to implement a single call to the client
    * which does an enable, wait-for, disable and hide message (in that
    * order).  This single call is more efficient than executing the calls
    * separately since the client has all the state needed AND since it
    * avoids extra round trips.
    *
    * @param    frameId
    *           ID of frame this request came for
    * @param    widgetIds
    *           array of field level widget IDs to be enabled or
    *           <code>null</code>, which means all of them.
    * @param    sb
    *           The current frame's screen buffer.
    * @param    seconds
    *           Time to wait for event.
    * @param    el
    *           List of events to wait for.
    * @param    doIO
    *           tells to perform the requested operation if <code>true</code>.
    *           otherwise prepares the frame for editing session only.
    * @param    isInput
    *           <code>true</code> if the input is redirected to an unnamed
    *           stream.
    *
    * @return  <code>ScreenBuffer</code> with the current values
    */
   public ScreenBuffer promptFor(int          frameId,
                                 int[]        widgetIds,
                                 ScreenBuffer sb,
                                 int          seconds,
                                 EventList    el,
                                 boolean      doIO,
                                 boolean      isInput)
   {
      boolean outputTerminal = false;
      boolean outputStream   = false;
      boolean switched       = isRedirected();
      if (switched)
      {
         if (isInput)
         {
            outputTerminal = false;
            outputStream = false;
         }
         else
         {
            outputTerminal = true;
            outputStream = false;
         }
      }
      else
      {
         if (isInput)
         {
            outputTerminal = false;
            outputStream = false;
         }
         else
         {
            // TODO: something is wrong here
            outputTerminal = true;
            outputStream = true;               
         }
      }
      // TO-DO (eventually):
      // In Progress, the active editing block can't be reentered with any
      // other IO operation like SET, PROMPT-FOR or UPDATE issued against
      // the same frame. WAIT-FOR is allowed, however. Currently, we don't
      // check this limitation and simply stack the state for recursion.

      // save and reset editing mode flag for possible recursion
      boolean saveInEditingBlock = inEditingBlock;
      inEditingBlock = false;

      int savedInsState = saveInsState();

      // this call guarantees frame is not null
      Frame frame = getFrame(frameId);

      int redir = -1;
      // switch to interactive mode to allow variable reading, if terminal
      // output is needed
      if (switched && outputTerminal)
      {
         redir = switchOutput(-1, false);
      }
      
      boolean exception = true;
      boolean saveCHO   = frame.isCanHideOthers();
      ignoreNextStopEditing = false;
      
      try
      {
         enable(frameId, widgetIds, sb, true, outputTerminal == false);
         
         if (lastExposedFrame != frame &&
             !Frame.containsFrames(frame, new Rectangle(new Point(0, 0), 
                                                        screenSize)))
         {
            tk.refresh(frame.calculateFrameRect(false));
         }
         
         exception = false;
      }
      finally
      {
         if (!doIO && exception)
         {
            // We were about to enter editing mode but user had pressed F4. 
            // Cleanup code at server is not aware about real reason and 
            // should not make assumptions about clients' state, so it will
            // invoke stopEditinMode() anyway. Since we haven't actually 
            // entered editing mode, we should ignore next stopEditingMode call.  
            ignoreNextStopEditing = true;
            frame.setForceWipeScreen(saveCHO);
            frame.setCanHideOthers(saveCHO);
         }
      }

      if (!doIO)
      {
         // save insert indicator state in order to restore it after leaving 
         // editing block
         savedInsStateInitials.push(savedInsState);
      }
      
      sb.resetChanged();

      ScreenBuffer result = null;

      // honor the outstanding next-prompt request if any
      int focusWidgetId = frame.getFocusId();
      frame.setFocusId(-1);

      if (focusWidgetId != -1)
      {
         // validate the widget
         focusWidgetId = validateWidget(focusWidgetId, widgetIds);
         Component widget = getWidget(focusWidgetId);
         if (widget instanceof FillIn)
         {
            ((FillIn) widget).resetCaret();
         }
      }

      // if no next-prompt, use the first widget from the list or the first
      // enabled widget of the frame
      if (focusWidgetId == -1)
      {
         if (widgetIds != null && widgetIds.length != 0)
         {
            // here the first focusable widget must be used, from all ids
            for (int i = 0; i < widgetIds.length; i++)
            {
               Component widget = getWidget(widgetIds[i]);
               // get the actual field, in a down frame
               widget = frame.getFrameHelper().getField(widget);

               if (widget.isFocusTraversable())
               {
                  focusWidgetId = widgetIds[i];
                  break;
               }
            }
         }
         else
         {
            focusWidgetId = getFirstFocusableId(frame);
         } 
      }

      // check if it is editing block preparation only
      if (!doIO)
      {
         // this call sets the editing mode flag so we leave it as is
         startEditingMode(focusWidgetId, el);
         result = getScreenBuffer(sb.getBaseId(), sb.size(), false);
         if (switched && outputTerminal)
         {
            switchOutput(redir, false);
         }
         return result;
      }

      final Component focus = getCurrentFocus();
      
      boolean condition = false;
      String status     = Window.getStatusText();
      int type          = statusType;
      forceWipeScreen   = false;
      
      try
      {
         if (focus != null)
         {
            highlight(focus, false);
         }

         // if in trigger nesting mode, save previous focus
         if (triggerNesting > 0)
         {
            pendingFocus = focus;
         }

         BufferSizeManager.getInstance().notifyBlockingOperation(BlockingOperation.PROMPT_FOR);

         // old focus has to be undefined
         Window.getDefaultWindow().setFocus(null);
         result = waitFor(el, focusWidgetId, seconds, sb);
      }
      
      catch (ConditionException ce) 
      {
         condition = true;
         
         throw ce;
      }

      finally
      {
         statusType = type;
         status(status);
         
         if (condition && focus != null)
         {
            // restore focus and highlight it, only if a condition was raised
            // focus is restored only if within a trigger
            if (triggerNesting > 0)
            {
               eventDrawingBracket(false, false, new Runnable()
               {
                  public void run()
                  {
                     focus.requestFocus();
                  }
               });
	         }
            highlight(focus, true);
         }

         // reset the pending focus
         pendingFocus = null;

         hideMessage(false);
         // fix Z-order
         frame.moveToTop();
         

         // make sure we disable, even on an ERROR/END condition
         enable(frameId, widgetIds, null, outputTerminal == false);
         restoreInsState(savedInsState, true);

         forceWipeScreen = true;

         // if there was terminal output and stream output is needed too, 
         // output the frame
         if (switched)
         {
            switchOutput(redir, false);
            
            if (outputStream)
            {
               viewWorker(frame, result, null, true);
            }
         }
      }

      // restore editing mode flag
      inEditingBlock = saveInEditingBlock;

      return result;
   }

   /**
    * Refreshes the specified frame and widget with a new screen value.
    * <p>
    * The frame is not necessarily on top of Z-order and the widget is not
    * necessarily fully visible. This operation does not change the Z-order.
    * 
    * @param   fid
    *          frame to refresh
    * @param   wid
    *          widget to refresh
    * @param   value
    *          new widget value
    * @return  string representation of the value
    */
   public String refreshFrameWidget(int fid, int wid, BaseDataType value)
   {
      // must be called first!
      honorConditionalUpDown();
      
      Component widget = getWidget(fid);
      
      // nothing to do?
      if (widget == null || !(widget instanceof Frame))
         return null;

      Frame frame = (Frame) widget;
      widget = getWidget(wid);

      if (widget == null)
         return null;
      
      // update the screen value
      ComponentConfig config = ((AttributeContainer) widget).getRef();
      setChangedValue(frame, widget, config, value, null);
      
      final Component widgetToPaint = widget;
      
      // repaint the affected rectangle
      eventDrawingBracket(new Runnable()
      {
         public void run()
         {
            widgetToPaint.repaint();
         }
      });

      // return the string representation of the widget value
      Frame.FrameData dt = frame.getFrameValue(widget);
      
      if (dt != null && dt.getValue() != null)
         return dt.getValue();

      return "";
   }
   
   /**
    * This method is called on each start of an EDITING block iteration.
    */
   private void iterateEditingBlock()
   {
      Frame frame = editingFrames.peek();

      // Stack can be empty if there are no active widgets in the widget list
      // (see second parameter in BlockManager.updateEditing() methods) and 
      // no frame was pushed in the stack (and no frames will be popped).
      if (frame == null)
         return;
      
      // honor the outstanding next-prompt request if any
      int focusWidgetId = frame.getFocusId();
      frame.setFocusId(-1);

      if (focusWidgetId != -1)
      {
         final Component widget = getWidget(focusWidgetId);
         
         if (widget instanceof FillIn)
         {
            ((FillIn) widget).resetCaret();
         }

         int currentFocus = getFocus();
         
         if (focusWidgetId != currentFocus)
         {
            // change the focus; this action may trigger focus gained event
            // that would notify listeners and change the status line with
            // the correct help string
            eventBracket(new Runnable()
            {
               public void run()
               {
                  widget.requestFocus();
               }
            });
         }
      }
   }
   
   /**
    * Externally accessible static wrapper for
    * {@link #viewWorker(Component, ScreenBuffer, int[], boolean)}.
    *
    * @param   widget
    *          The frame or widget for this request.
    * @param   frameBuf
    *          The current values of widgets.
    * @param   widgetIds
    *          Array of field level widget IDs to be shown or
    *          <code>null</code>, which means show all contained widgets
    *          (for frame only).
    * @param   isView
    *          <code>true</code> if the call is for VIEW, <code>false</code>
    *          for DISPLAY.
    */
   public static void view(Component    widget,
                           ScreenBuffer frameBuf,
                           int[]        widgetIds,
                           boolean      isView)
   {
      getInstance().viewWorker(widget, frameBuf, widgetIds, isView);
   }

   /**
    * Force the visibility and drawing of all specified fields of the 
    * given frame.  This method may be safely called while processing in 
    * the {@link #view} method.
    * 
    * @param   widget
    *          The frame or widget for this request.
    * @param   frameBuf
    *          The current values of widgets.
    * @param   widgetIds
    *          Array of field level widget IDs to be shown or
    *          <code>null</code>, which means show all contained
    *          widgets (for frame only).
    * @param   isView
    *          <code>true</code> if the call is for VIEW, <code>false</code>
    *          for DISPLAY.
    */   
   private void viewWorker(final Component    widget,
                           final ScreenBuffer frameBuf,
                           final int[]        widgetIds,
                           final boolean      isView)
   {
      // mark the view operation in progress
      boolean saveViewInProgress = viewInProgress;
      viewInProgress = true;

      // inspect and reset forced view flags
      boolean forcedRefresh = nextViewRefresh;
      boolean forcedDisplay = nextViewDisplay;
      
      if (forcedRefresh)
      {
         forcedDisplay = false;
      }
      
      nextViewRefresh = false;
      nextViewDisplay = false;
      
      viewRefresh = forcedRefresh;
      viewDisplay = forcedDisplay;

      if (!(widget instanceof Frame))
      {
         eventDrawingBracket(new Runnable()
         {
            public void run()
            {
               // update our backing data to the latest version
               if (frameBuf != null)
               {
                  setScreenBuffer(frameBuf, true);
               }
               
               // handle widget view
               viewDisplay = true;         // always?
               collectFlag(widget, true, true);
               widget.setVisible(true);
   
               // change the hidden flag for widget
               collectFlag(widget, false, false);
               widget.setHidden(false);
               
               // ensure the frame is visible
               Frame frame = (Frame) getWidget(frameBuf.getBaseId());
               
               if (frame != null && !frame.isVisible())
               {
                  viewWorker(frame, null, null, isView);
               }
            }
         });

         viewInProgress = saveViewInProgress;

         // at this time, the frame's explicitHide flag is already reset by
         // the above viewWorker recursive call, so nothing more is
         // needed
         return;
      }

      // access the frame
      final Frame       frame = (Frame) widget;
      final FrameHelper fh    = frame.getFrameHelper();
      
      final boolean down          = fh.isDown();
      final boolean toBeDisplayed = !frame.isVisible();
      final boolean redir         = isRedirected();
      final boolean editors       = frame.hasEditors();

      if (!frame.isVisible())
      {
         // if the frame is not visible when this explicit VIEW is called, it 
         // must be marked. 
         // if this VIEW call was executed by a trigger (called via an APPLY
         // statement), then focus will be set to the APPLY target widget, if
         // the focused widget's parent frame change its visibility from
         // not-visible to visible.
         frame.setVisibilityChanged(true);
         
         if (!redir || isView)
            lastExposedFrame  = frame;
      }

      if (redir && !isView)
      {
         frame.markAsStreamed();
      }

      lastViewed = frame;
      
      // create list of widgets to set visible
      Component[] wids = null;

      if (widgetIds == null ||
          (widgetIds.length == 1 && getWidget(widgetIds[0]) == widget))
      {
         // this is a view frame call
         if (!forcedDisplay)
         {
            viewRefresh = true;
         }
         
         // all widgets
         Component[] comps = fh.getContentPane().getComponents();
         ArrayList<Component> visibleComps = 
            new ArrayList<Component>(comps.length);
         
         // remove hidden widgets from list
         for (int i = 0; i < comps.length; i++)
         {
            if (!comps[i].isHidden())
            {
               visibleComps.add(comps[i]); 
            }
         }
         
         wids = new Component[visibleComps.size()];
         visibleComps.toArray(wids);
      }
      else
      {
         // this is a data display call
         if (!forcedRefresh)
         {
            viewDisplay = true;
         }

         frame.setAllClear(false);
         
         // explicit widget list

         if (!frame.isVisible())
         {
            // collect all non-hidden widgets, if frame is not visible
            Component[] comps = fh.getContentPane().getComponents();
            HashSet<Component> visibleComps = 
               new HashSet<Component>(comps.length);
               
            for (int i = 0; i < comps.length; i++)
            {
               if (!comps[i].isHidden())
               {
                  if (down)
                  {
                     comps[i] = fh.getField(comps[i]);
                  }
                  visibleComps.add(comps[i]);
               }
            }

            // add the explicit components, regardless of their state
            for (int i = 0; i < widgetIds.length; i++)
            {
               Component comp = getWidget(widgetIds[i]);
               if (down)
               {
                  comp = fh.getField(comp);
               }
               visibleComps.add(comp);
            }

            wids = new Component[visibleComps.size()];
            visibleComps.toArray(wids);
         }
         else
         {
            // frame is visible, use only explicit widgets
            wids = new Component[widgetIds.length];

            for (int i = 0; i < wids.length; i++)
            {
               wids[i] = getWidget(widgetIds[i]);

               if (down)
               {
                  wids[i] = fh.getField(wids[i]);
               }
            }
         }
      }
      
      final Component[] list = wids;
      
      // handle frame view
      boolean copyRefresh = viewRefresh;
      boolean copyDisplay = viewDisplay;
      
      startWidgetOutput();
      
      boolean frameRepaint = false;
      
      // runtime exception saver 
      final RuntimeException[] rtx = new RuntimeException[1];
      
      final Boolean[] allDisplayed = new Boolean[1];
      allDisplayed[0] = false;
      
      try
      {
         frameRepaint = eventDrawingBracket(new Runnable()
         {
            public void run()
            {
               if (fh.isForcedRefresh())
               {
                  frame.repaint();
                  fh.resetForcedRefresh();
               }
               
               if (frameBuf != null && down && redir && 
                   (!fh.isSideLabels() || !fh.isLabels()))
               {
                  Set<Integer> iteration = fh.getInterationIds();
                  
                  Iterator<Integer>  iter = frameBuf.getWidgetIDIterator();
                  
                  while (iter.hasNext())
                  {
                     iteration.remove(iter.next());
                  }
                  
                  allDisplayed[0] = (iteration.size() == 0);
               }
               
               if (frame.isNeedReset() && !down)
               {
                  try
                  {
                     if (frame.isVisible())
                     {
                        // force immediate frame hiding
                        if (!hideFramesInt(frame))
                        {
                           if (frame.isNeedPause())
                              pauseBeforeHide();
                        }
                        frame.setCanHide(false);
                     }
                  }
                  
                  finally
                  {
                     fh.clearWorker(true);
                     frame.setNeedReset(false);
                  }
               }
               else if (down && frame.isForcePause())
               {
                  boolean normalExit = false;
                  
                  try
                  {
                     frame.setForcePause(false);
                     pauseBeforeHide();
                     normalExit = true;
                  }
                  
                  finally
                  {
                     if (!normalExit && !redir)
                     {
                        fh.clearWorker(true);
                        frame.setCanHideOthers(true);
                        hideFramesInt(frame);
                        frame.repaint();
                        frame.setNeedReset(false);
                        frame.setNeedPause(true);
                     }
                  }
               }
      
               if (frame.isDelayedRepaint() && frame.isVisible())
               {
                  fh.clearWorker(true);
               }
               
               if (frame.isCanHide() && frame.isVisible() && frame.isNeedPause())
               {
                  if (needPause)
                  {
                     pauseBeforeHide();
                  }
               }
               
               // update dynamic properties, such as headers, dynamic down, row, etc.
               
               ArrayList<Integer> updatedHeaders = null;
               
               if (frameBuf != null)
               {
                  // adjust DOWN value
                  if (down)
                     frame.adjustDown(frameBuf.getDown());
                  
                  frame.setDynRow(frameBuf.getRow());
                  frame.setDynCol(frameBuf.getColumn());
                  
                  // update headers
                  updatedHeaders = updateHeaders(frameBuf);
               }
               
               // display of a regular frame hides all dialogs
               if (!frame.isDialog())
               {
                  screenCurrent.hideDialogs(waitForFrames);
               }
               
               if (!viewRefresh)
               {
                  viewRefresh = toBeDisplayed;
               }
            
               if (redir)
               {
                  RedirectedTerminal rt = redirections.get(currentStream);
      
                  if (rt.isFresh() && rt.needsHeadersOutput())
                  {
                     screenCurrent.resetStartRow();
                     rt.processHeaders();
                  }
               }
               
               boolean bufferSet = false;
      
               // try set buffer early to be sure that  
               // all widgets reach their real size
               if (frameBuf != null && !down)
               {
                  setScreenBuffer(frameBuf, !isRedirected());
                  bufferSet = true;
               }
      
               placeFrameInt(frame, false);
               
               if (down && redir)
               {
                  fh.arrangeToStream();
               }
               
               // assign values AFTER placing frame on the screen
               if (frameBuf != null)
               {
                  // update our backing data to the latest version (only force
                  // direct drawing/refresh if we are not redirected)
                  if (!bufferSet)
                     setScreenBuffer(frameBuf, !isRedirected());
      
                  if (redir && editors)
                  {
                     fh.rearrangeEditors();
                  }
      
                  // redraw updated headers now
                  if (!isRedirected() && 
                      updatedHeaders != null && 
                      !updatedHeaders.isEmpty())
                  {
                     for (Integer hid : updatedHeaders)
                        screenCurrent.getComponent(hid.intValue()).draw();
                  }
                  
                  needPause = true;
               }
               
               if (frame.isDelayedRepaint())
               {
                  frame.repaint();
                  frame.setDelayedRepaint(false);
               }
      
               if (toBeDisplayed)
               {
                  for (int i = 0; i < list.length; i++)
                  {
                     if (list[i] != null)
                     {
                        collectFlag(list[i], true, true);
                        list[i].setVisible(true);
      
                        collectFlag(list[i], false, false);
                        list[i].setHidden(false);
                     }
                  }
               }
            }
         });
      }
      
      catch (RuntimeException xc)
      {
         rtx[0] = xc;
      }
      
      finally
      {
         if (down && redir)
         {
            fh.arrangeToInteract();
         }

         if (isRedirected() && frameRepaint && frame.getHeight() == 0)
         {
            Stream s = sd.streamFromId(currentStream);
            
            if (s == null)
            {
               viewInProgress = saveViewInProgress;

               throw new IllegalStateException("Invalid stream id " +
                                               currentStream + ".");
            }
            
            s.pulseOutput();
         }

         if (rtx[0] == null)
         {
            // check if any of the widgets is a browse; if yes, hide any 
            // temp texts inside message area
            for (int i = 0; i < list.length; i++)
            {
               if (list[i] instanceof Browse)
               {
                  // if found, clear the message area
                  clearPutScreenInMessageArea(-1);
                  break;
               }
            }
         }
         
         if (toBeDisplayed || frameRepaint || allDisplayed[0])
         {
             viewRefresh = copyRefresh;
             viewDisplay = copyDisplay;
 
             if (redir && editors)
             {
                fh.restoreEditors();
             }
 
             if (rtx[0] == null)
             {
                viewInProgress = saveViewInProgress;
                return;
             }
         }
         
         // rethrow exception regardless from the mode
         if (rtx[0] != null)
         {
            viewInProgress = saveViewInProgress;
            throw rtx[0];
         }
      }
      
      // down frames receiving continuous output aren't redrawn by the code
      // above since they are visible before the call
      // this extra drawing cycle below solves this by explicitly repainting
      // the current iteration.

      if (redir && down)
      {
         eventDrawingBracket(new Runnable()
         {
            public void run()
            {
               fh.repaintCurrent();
            }
         });

         viewRefresh = copyRefresh;
         viewDisplay = copyDisplay;
         
         if (editors)
         {
            fh.restoreEditors();
         }

         viewInProgress = saveViewInProgress;
         return;
      }

      viewRefresh = copyRefresh;
      viewDisplay = copyDisplay;

      ScreenBitmap bitmap = widgetDrawingOn(frame);
      
      for (int i = 0; i < list.length; i++)
      {
         if (list[i] != null)
         {
            if (!list[i].isVisible())
            {
               final int idx = i;
               
               eventBracket(true, new Runnable()
               {
                  public void run()
                  {
                     collectFlag(list[idx], true, true);
                     list[idx].setVisible(true);
                     
                     // change hidden flag too
                     collectFlag(list[idx], false, false);
                     list[idx].setHidden(false);
                  }
               });
            }
            list[i].draw();
            
            // redraw side labels if any
            if (down && fh.isSideLabels())
            {
               if (list[i] instanceof LabeledWidget)
               {
                  Label label = ((LabeledWidget) list[i]).getLabelInstance();

                  if (label != null)
                     label.draw();
               }
            }
         }
      }

      if (down)
      {
         fh.drawUninitialized();
      }

      widgetDrawingOff(bitmap);
      endWidgetOutput();

      if (redir && editors)
      {
         fh.restoreEditors();
      }

      // find the last enabled editor, in tab order, targeted by this display 
      // call
      if (frameBuf != null)
      {
         Set<Editor> edset = new HashSet<Editor>(3);
         for (int i = 0; i < list.length; i++)
         {
            if (list[i].isEnabled() && list[i] instanceof Editor)
            {
               // the editor must have at least some text, too
               Editor editor = (Editor) list[i];
               int eid = editor.config.getID();
               if (frameBuf.getState(eid) == ScreenBuffer.CHANGED &&
                   editor.getText() != null                       && 
                   editor.getText().length() > 0)
               {
                  edset.add(editor);
               }
            }
         }
         
         // get all widgets of this frame and search for last targeted editor,
         // in tab order
         if (edset.size() > 0)
         {
            Component[] widgets = frame.getContentPane().getComponents();
            for (int i = widgets.length - 1; i >= 0; i--)
            {
               if (widgets[i] instanceof Editor && edset.contains(widgets[i]))
               {
                  // found the last editor, in tab order
                  lastEditorDisplayed = (Editor) widgets[i];
                  break;
               }
            }
         }
      }

      viewInProgress = saveViewInProgress;
   }
   
   /**
    * Mark end of widget output.
    */
   void endWidgetOutput()
   {
      OutputPrimitives output = tk.getInteractiveOutput(); 
      if (output instanceof DoubleBufferedTerminal)
         ((DoubleBufferedTerminal) output).conditionalRefresh();
   }
   
   /**
    * Mark start of widget drawing.
    */
   void startWidgetOutput()
   {
      OutputPrimitives output = tk.getInteractiveOutput();
      
      if (output instanceof DoubleBufferedTerminal)
         ((DoubleBufferedTerminal) output).reset();
   }
   
   /**
    * Conditionally refresh specified rectangle not occupied by frames.
    * 
    * @param    rect
    *           Rectangle to refresh.
    */
   void conditionalRefresh(Rectangle rect)
   {
      OutputPrimitives output = tk.getInteractiveOutput();
      
      if (output instanceof DoubleBufferedTerminal)
         ((DoubleBufferedTerminal) output).conditionalRefresh(rect);
   }
   
   /**
    * Hide frames if necessary to prepare space for specified frame and then
    * place and realize current frame.
    *
    * @param   frame
    *          Frame which will be displayed.
    * @param   force
    *          <code>true</code> means forced frame placement even if
    *          the frame is currently visible; useful for recovering partially
    *          hidden frames.
    */
   private void placeFrameInt(Frame frame, boolean force)
   {
      hideFramesInt(frame);
      
      if (frame.isVisible() && !isRedirected() && !frame.isRedirected())
      {
         Frame[] frames = screenCurrent.getFramesAbove(frame);
         
         if (frames.length > 0)
         {
            force = true;
            // Code below has following effect: all frames above current
            // are not repainted, but invalidated rectangle is calculated 
            // instead. Then current frame is then moved to top (see code
            // inside 'finally' block) and later calculated invalidated 
            // rectangle is repainted.
            // Improved: only frames which overlap current are taken into 
            // account.
            
            for (Frame frm : frames)
            {
               frm.repaint();   
            }
         }
      }

      if (!frame.isVisible() || force)
      {
         if (isRedirected())
         {
            // set terminal logical width to let frame fit terminal 
            int width      = Window.getScreenDimension().width;
            int frameWidth = frame.getSavedDimension().width;
            
            if (frame.getLocation().x != BaseConfig.INV_COORD)
               frameWidth += frame.getLocation().x; 
            
            RedirectedTerminal rt = redirections.get(currentStream);
            
            rt.setLogicalWidth(Math.max(frameWidth, width));
            
            // relayout frame because it can be formatted for wrong width
            frame.relayout();
         }
         
         screenCurrent.placeFrame(frame);
         collectFlag(frame, true, true);
         frame.setVisible(true);

         // fix focus in root window
         if (frame.getCurrentFocus() != null && 
             frame.getParent().getCurrentFocus() == null)
         {
            frame.getParent().setFocus(frame);
         }
      }

      if (editingFrames.isEmpty() || 
          editingFrames.peekFirst() == frame || 
          frame.hasEnabled())
      {
         frame.placeAtTop();   
      }
      
      frame.setCanHide(false);
      frame.setCanWipeScreen(false);
      frame.setCanHideOthers(false);
      frame.setAlive(true);
      frame.setNeedPause(true);
      frame.placed();
   }

   /**
    * Internal method which hides frames if necessary and maintains global
    * flag.
    * 
    * @param   frame
    *          Frame which is about to be displayed.
    *          
    * @return  <code>true</code> if pause was triggered.         
    */
   private boolean hideFramesInt(Frame frame)
   {
      FrameConfig cfg = (FrameConfig) frame.getRef();
      boolean[] res = new boolean[] {false, false};
      
      if ((cfg.getID() != INPUT_FRAME_ID || helpActive) && 
          frame.isCanHideOthers())
      {
         res = screenCurrent.hideFrames(frame, (currentStream != -1));
         frame.setCanHideOthers(false);
         
         // trigger screen refresh in following cases:
         // 1. Frame is explicitly marked as such that can wipe screen and
         //    there was at least one other frame hidden.
         // 2. Frame was hidden and has no enabled widgets and forceWipeScreen
         //    flag is set.
         // 3. There was at least one other frame hidden and among hidden 
         //    frames there is a frame which has the same row as given one.
         
         if ((!tk.isRedirected() && frame.isForceWipeScreen()) ||
             (!tk.isRedirected() && frame.isCanWipeScreen() && res[0]) || 
             (forceWipeScreen && !frame.isVisible() && !frame.hasAnyEnabled()) ||
             (res[0] == true && res[1] == true && messageClearQuirk))
         {
            tk.refresh(new Rectangle(new Point(0, 0), screenSize));
            
            if (res[0] == true && res[1] == true && messageClearQuirk)
            {
               // Replace messages with spaces.
               // Note that hiding of messages is not adequate, because
               // affects pausing.
               String[] msg = Window.getMessageText();
               int[] cs = new int[] {-1, -1};    
               
               for (int i = 0; i < msg.length; i++)
               {
                  if (msg[i] != null)
                  {
                     msg[i] = StringHelper.repeatChar(' ', msg[i].length());
                     cs[i] = ColorTable.NORMAL;
                  }
               }
               
               Window.setMessageText(msg, cs);
            }
            
            if (frame.isForceWipeScreen())
            {
               Window.setMessageText(null, null);
               messageNeedPause = false;
            }
         }
         
         forceWipeScreen = false;
         frame.setForceWipeScreen(false);
      }
      
      messageClearQuirk = false;
      
      return res[0];
   }

   /**
    * Perform operation on array of frames.
    * 
    * @param   me
    *          <code>MarkEntry</code> container with operation and array of
    *          frames to process.
    */
   public void markFrames(MarkEntry me)
   {
      for (int i = 0; i < me.frames.length; i++)
      {
         if (me.frames[i]  == -1)
            continue;

         Frame frame = (Frame) getWidget(me.frames[i]);
         
         if (frame == null)
            continue;

         switch(me.mode)
         {
            case MarkEntry.RESET_FRAME:
               frame.setNeedReset(true);
               break;
               
            case MarkEntry.DESTROY:
               destroyFrame(me.frames[i]);
               break;
               
            case MarkEntry.HIDE_NO_PAUSE:
               if (frame.isVisible())
               {
                  frame.setNeedPause(false);
                  frame.setCanHide(true);
               }
               break;
               
            case MarkEntry.CAN_HIDE_OTHERS:
            case MarkEntry.CAN_HIDE_OTHERS_EXT:
               {
                  frame.setCanHideOthers(true);
                  
                  if (me.mode == MarkEntry.CAN_HIDE_OTHERS_EXT)
                     frame.setCanWipeScreen(true);
               }
               break;
               
            case MarkEntry.HIDE:
               if (frame.isVisible() && !frame.isNoHide())
               {
                  frame.setCanHide(true);
               }
               break;

            case MarkEntry.UNHIDE:
               if (frame.isVisible())
               {
                  frame.setCanHide(false);
               }
               break;
               
            case MarkEntry.OPEN_SCOPE:
               int scopeNesting = frame.getScopeNesting();
               frame.setScopeNesting(scopeNesting + 1);

               if (scopeNesting == 0)
               {
                  if (frame.getFrameHelper().isDown())
                  {
                     if (frame.getFrameHelper().getCursor() != -1 &&
                         frame.isNeedPause())
                     {
                        frame.setForcePause(true);
                     }
                     
                     frame.setNeedReset(true);
                     frame.reset();
                  }
                  else
                  {
                     frame.getFrameHelper().resetAll();  
                  }
                  frame.setInScope(true);
               }
               break;
               
            case MarkEntry.CLOSE_SCOPE:
               scopeNesting = frame.getScopeNesting() - 1;
               frame.setScopeNesting(scopeNesting);
               
               if (scopeNesting == 0)
               {
                  if (lastViewed == frame)
                     lastViewed = null;
                  frame.setInScope(false);
                  forceDeregisterPageElement(me.frames[i]);

                  // the repaint will be postponed until the next display/view
                  // statement is executed with this frame (when the frame is
                  // inside a loop)
                  // for shared frames, the repaint will be forced only if 
                  // this is the root nested scope
                  frame.setDelayedRepaint(true);
                  
                  if (frame.isDialog())
                  {
                     tk.setInvalidate(true);
                     frame.setVisible(false);
                     tk.setInvalidate(false);
                  }
               }               
               break;
         }
      }
   }

   /**
    * Wait for specified list of events. <br>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget PAUSE n </b>
    * 
    * @param   list
    *          List of events to wait for.
    * @param   focusWidgetId
    *          Widget ID of a widget receiving focus first, or -1.
    * @param   seconds
    *          Time to wait for event.
    * @param   inbuf
    *          current screen buffer known to the server
    * 
    * @return  <code>ScreenBuffer</code> with the current values
    */
   public ScreenBuffer waitFor(EventList    list,
                               int          focusWidgetId, 
                               int          seconds,
                               ScreenBuffer inbuf)
   {
      // must be called first!
      honorConditionalUpDown(false);
      
      // save and reset editing mode flag for possible recursion
      boolean saveInEditingBlock = inEditingBlock;
      inEditingBlock = false;
      
      if (focusWidgetId != -1)
         inputActive++;

      // do the work
      ScreenBuffer result = waitForWorker(list, focusWidgetId, seconds, inbuf,
                                          false, true, true);

      // restore editing mode flag
      inEditingBlock = saveInEditingBlock;
      
      if (focusWidgetId != -1)
         inputActive--;

      return result;
   }
   
   /**
    * Main WAIT-FOR worker routine.
    * <p>
    * It manages the focus scopes, sets the initial focus and provides a
    * message processing loop. The focus validity is constantly monitored
    * and when it needs adjustment and the conditions are appropriate, the
    * focus gets adjusted. 
    * <p>
    * 
    * 
    * @param    list
    *           List of events to wait for.
    * @param    focusWidgetId
    *           Widget ID of a widget receiving focus first, or -1.
    * @param    seconds
    *           The time to wait for a given key event.
    * @param    inbuf
    *           The current screen buffer known to the server.
    * @param    force
    *           If <code>true</code>, on output, copy the current data into
    *           the returned screen buffer without regard to the changed 
    *           flag.
    * @param    status
    *           If <code>true</code> method will change status line.
    * @param    entryWindow
    *           <code>true</code> if ENTRY for window should be sent.
    *
    * @return   <code>ScreenBuffer</code> with the current values.
    */
   private ScreenBuffer waitForWorker(EventList          list,
                                      int                focusWidgetId, 
                                      int                seconds,
                                      final ScreenBuffer inbuf,
                                      boolean            force, 
                                      boolean            status,
                                      boolean            entryWindow)
   {
      // save cursor state
      boolean curvis = tk.setCursorStatus(true);

      // save and check focus
      Component saveFocus = getCurrentFocus();
      int focusState = getFocusState(saveFocus);

      // save pending focus - UPDATE operation may restore focus to it
      Component savePending = pendingFocus;
      int pendingState = getFocusState(pendingFocus);

      // save the global events flag and the widget ids registered for 
      // exit events
      boolean hasGlobalEvents = checkEventList(list);
      int[] exitWidgets = list.getExitWidgetIds();

      // save off the previous value of go pending and reset to false
      // this should be done early enough to allow normal event processing
      // for LEAVE and ENTRY that may be generated on entry
      int oldGoPending = isGoPending;
      isGoPending = 0;
      immediateException = null;

      // determine the initial frame and focused widget
      Component startWidget = null;
      boolean explicit = false;
      boolean pendingExplicit = false;

      if (focusWidgetId == -1     &&
          triggerNesting == 0     && 
          pendingFocus != null    &&
          pendingFocus.isFocusTraversable())
      {
         // if there is no explicit widget and we are on root level, then 
         // force explicit widget to be the previous widget which had ENTRY
         // this is usable only for wait-for, as UPDATE always has explicit
         // focus
         Frame pendingFrame = (Frame) locateFrame(pendingFocus);
         if (pendingFrame != null && pendingFrame.isAlive())
         {
            // force explicit focus only if the frame is not out of scope
            pendingExplicit = true;
            focusWidgetId = getWidgetId(pendingFocus);
         }
         else
         {
            pendingFocus = null;
         }
      }

      if (focusWidgetId != -1)
      {
         Component target = getWidget(focusWidgetId);
         
         Frame sFrame = (Frame) locateFrame(saveFocus);
         
         // in some cases frame specified as a target may attract focus 
         // and break regular focus handling rules
         if (target instanceof Frame && sFrame != null)
         {
            Frame tFrame = (Frame) target;
            
            if (inputActive > 0 &&
                sFrame.getBounds().intersects(tFrame.getBounds()) &&
                tFrame.isFocusAttractor())
            {
               Component first = getFirstEnabledWidget(tFrame);
               
               if (first != null)
                  target = first;
            }
            
            // reset the flag
            tFrame.setFocusAttractor(false);
         }
         
         // ignore frame as focus target
         if (target == null || !(target instanceof Frame))
         {
            explicit   = true;
            startWidget = target;
         }
         else
         {
            focusWidgetId = -1;
         }
      }
      else
      {
         startWidget = saveFocus;
      }
      
      // set the screen buffer and refresh the visual appearance if needed
      tk.setInvalidate(true);

      eventBracket(new Runnable()
      {
         public void run()
         {
            setScreenBuffer(inbuf);
         }
      });
      
      // save off old wait list and override with the specified list
      boolean saveLeaveSent = leaveSent;
      leaveSent = false;
      EventList save = currentEventList;
      currentEventList = list;
      
      // set the initial focus
      boolean frameAsTarget = true;
      boolean immedExit = false;
      int cop = currentEventList.getOperation();
      
      // a dialog box needs special attention
      boolean dialog = false;
      Frame framed = screenCurrent.getFirstVisibleFrame();

      if (framed != null && framed.isDialog())
      {
         dialog = true;
      }

      // row mode CHOOSE needs special attention
      boolean rowChoose = false;

      if (cop == EventList.CHOOSE)
      {
         Frame chFrame = (Frame) getWidget(inbuf.getBaseId());
         if (!chFrame.getChooseData().isFieldMode())
         {
            rowChoose = true;

            // no focus for row mode CHOOSE
            chFrame.setFocus(null);
         }
      }

      if (rowChoose)
      {
         // no focus for row mode CHOOSE
      }
      else if (!explicit)
      {
         switch (focusState)
         {
            case FOCUS_VALID:
               if (dialog)
               {
                  // treat as undefined, implicit
                  immedExit = !focusUndefImplicit(list, false);
               }
               else
               {
                  // valid, implicit
                  immedExit = !focusValidImplicit(saveFocus);
               }
               break;

            case FOCUS_UNDEF:
               // undefined, implicit
               immedExit = !focusUndefImplicit(list, false);
               break;

            case FOCUS_INVALID:
               // invalid, implicit
               immedExit = !focusInvalidImplicit(saveFocus);
               break;

            case FOCUS_HIDDEN:
               // hidden, implicit
               immedExit = !focusUndefImplicit(list, true);
               break;
         }
      }
      else
      {
         switch(focusState)
         {
            case FOCUS_VALID:
               // valid, explicit
               immedExit = !focusValidExplicit(saveFocus, 
                                               startWidget,
                                               list, 
                                               cop, 
                                               false);
               break;

            case FOCUS_UNDEF:
               // undefined, explicit
               immedExit = !focusUndefExplicit(startWidget,
                                               list,
                                               cop,
                                               !pendingExplicit);
               break;

            case FOCUS_INVALID:
               // invalid, explicit
               immedExit = !focusInvalidExplicit(saveFocus, startWidget, list);
               break;

            case FOCUS_HIDDEN:
               // hidden, explicit
               immedExit = !focusValidExplicit(saveFocus, 
                                               startWidget,
                                               list, 
                                               cop, 
                                               true);
               break;
         }
      }

      if (immedExit)
      {
         // close drawing bracket
         tk.setInvalidate(false);
         
         // immediate return
         currentEventList = save;
         leaveSent = saveLeaveSent;
         return inbuf;
      }

      // prepare to the event processing loop
      boolean canAdjustFocus = !noFocusAdjustment;
      Component nowFocused = getCurrentFocus();
      int focusStateNow = getFocusState(nowFocused);

      if (rowChoose)
      {
         canAdjustFocus = false;
         focusStateNow = FOCUS_UNDEF;
      }

      // if no focus, park cursor right after the insert mode indicator
      // and disallow frame as event target
      boolean statusSet = false;

      if (focusStateNow != FOCUS_VALID)
      {
         Window.setInsertMode(Window.INSERT_MODE_OFF, true);
         // if focus is invalid, then frame is used as target. 
         // the flag will be adjusted on the next wait-for iteration.
         // when in row-choose mode, frame must be used as target so the
         // frame will accept events (even P2J doesn't set it in focus)
         if (!rowChoose && 
             (focusState == FOCUS_HIDDEN ||
              focusState == FOCUS_UNDEF))
         {
            frameAsTarget = false;
         }
      }
      // or else generate frame ENTRY event, only if previous focus is not 
      // part of this frame
      else
      {
         Frame frame = (Frame)locateFrame(getCurrentFocus());
         if (saveFocus == null || frame != locateFrame(saveFocus))
         {
            sendEntry(frame);

            if (triggerNesting == 0 && entryWindow)
            {
               // send entry to the window
               sendEntry(Window.getDefaultWindow());
            }
         }

         // set status line if there is a focused widget
         if (status)
         {
            startInputStatus(true);
            statusSet = true;
         }

         // highlight the currently focused widget
         highlight(getCurrentFocus(), true);
      }
      
      fixFocusedFrameZOrder();

      tk.setInvalidate(false);

      KeyInput key = null;
      
      // clear GO pending flag just in case some event triggered it again
      isGoPending = 0;
      immediateException = null;
      
      // focus adjustments has been completed, so we can reset saved insert
      // indicator state
      clearSavedInsState();

      // if we have a WAIT-FOR PAUSE cycle and a fill-in receives the focus
      // upon initial focus adjustment, then auto-zap state becomes OFF for
      // this fill-in
      if (seconds >= 0)
      {
         Component focus = getCurrentFocus();
         if (focus != null &&
             focus instanceof FillIn &&
             focus.isEnabled() &&
             focus.isVisible())
         {
            ((FillIn) focus).setPendingNoZap();
         }
      }

      triggerNestingByWaitFor.push(0);

      boolean saveCondExit = condExit;
      boolean condition = false;
      
      // event processing loop
      try
      {
         boolean fixFocus = false;
         boolean saveFrameAsTarget = frameAsTarget;
         do
         {
            processPostponedValueChanged();
            Object source = null;
            boolean sourceAvailable = false;
            
            try
            {
               BufferSizeManager.getInstance().notifyBlockingOperation(BlockingOperation.WAIT_FOR);

               // this will maintain LASTKEY too, which is required!
               key = waitForKey(seconds, frameAsTarget);
               
               condExit = false;
               
               if (key != null)
               {
                  source = key.getSource();

                  if (source instanceof Editor)
                  {
                     lastEditorDisplayed = (Editor) source;
                  }

                  Component prevFocus = getCurrentFocus();
                  int prevState = getFocusState(prevFocus);

                  if (source instanceof Component)
                  {
                     source = locateFrame((Component) source);
                  }
                  
                  sourceAvailable = source != null;
                  
                  if (sourceAvailable)
                  {
                     waitForFrames.push((Frame) source);
                  }
                  
                  setNeedPause(false);

                  int action = key.getActionCode();
                  Component src = (Component) key.getSource();

                  if((action == Keyboard.KA_STOP   ||
                      action == Keyboard.KA_ENDKEY ||
                      action == Keyboard.KA_ERROR  ||
                      action == Keyboard.KA_END_ERROR) &&
                     checkBreak(src, action, false) &&
                     checkBreak(src, Keyboard.SE_LEAVE, false))
                  {
                     // The original event constitutes an exceptional exit
                     // condition which is used in the wait-for list as
                     // a regular exit condition and the LEAVE event is also
                     // specified in the wait list. No regular event
                     // application is taken (no call to applyWorker()), but
                     // a special event processing is performed.

                     Component cur = prevFocus;
                     if (cur != null && !cur.isEnabled())
                     {
                        cur = null;
                     }

                     if (src instanceof Frame || src instanceof Window)
                     {
                        cur = src;
                     }

                     // Immediately consume the original event.
                     key.consume();

                     // Generate and apply LEAVE event.
                     if (cur != null)
                     {
                        if (sendLeave(cur))
                        {
                           // Exit the processing loop.
                           isGoPending++;
                        }
                     }
                  }
                  else
                  {
                     try
                     {
                        // event application
                        applyWorker(key);
                     }

                     finally
                     {
                        if (source != null)
                        {
                           Frame frame = (Frame) source;
                           // if the frame was not hidden via an explicit
                           // HIDE call, re-display it
                           if (!frame.isVisible() && !frame.isExplicitHide())
                           {
                              // restore focus
                              if (prevFocus != null)
                              {
                                 prevFocus.requestFocus();
                                 prevFocus.setHighlighted(true);
                              }

                              // repaint this frame without the hidden widgets
                              view(frame.getRef().getID(), null, null, true);
                           }

                           frame.setExplicitHide(false);
                        }
                     }
                  }

                  // determine the state of the focus after event application
                  // if either undefined or invalid, focus needs adjustment
                  Component focus = getCurrentFocus();
                  int state = getFocusState(focus);
                  fixFocus = state != FOCUS_VALID &&
                             (focus != prevFocus ||
                              state != prevState ||
                              state == FOCUS_UNDEF);
                  // restore flag to value as before wait-for loop, only if
                  // there is a focused widget
                  if (state != FOCUS_UNDEF)
                  {
                     frameAsTarget = saveFrameAsTarget;
                  }

                  // in case a GO event was generated by this key, then no 
                  // focus adjustment is needed
                  if (isGoPending == 0 && canAdjustFocus && fixFocus)
                  {
                     adjustFocus();
                     // if no focus, then frame as target must not be used
                     frameAsTarget = getCurrentFocus() != null;
                  }

                  // if no GO was generated 
                  if (isGoPending == 0 && !hasGlobalEvents)
                  {
                     // if the wait-for is not waiting for window events,
                     // check if at least on widget is still enabled and 
                     // visible
                     if (exitWidgets != null)
                     {
                        boolean noWidgets = true;
                        for (int i = 0; i < exitWidgets.length; i++)
                        {
                           Component widget = getWidget(exitWidgets[i]);
                           if (widget.isEnabled() && widget.isVisible())
                           {
                              noWidgets = false;
                              break;
                           }
                        }
                        
                        if (noWidgets)
                        {
                           // if no enabled or visible widgets exist, then
                           // processing must be stopped
                           stopNoAvailableWidgets();
                        }
                     }
                  }

                  // a special case for read-only editor widget
                  focus = getCurrentFocus();
                  if (lastEditorDisplayed != null &&
                      focus instanceof Editor     &&
                      ((Editor) focus).config.isReadOnly())
                  {
                     // if focus is a read-only editor, then check the last
                     // editor which had buffer update. if found, set 
                     // cursor on it
                     lastEditorDisplayed.drawCaretUnconditionally();
                     
                     // get the cursor position and draw real cursor
                     Point p = new Point(screenSize.width - 2, 
                                         screenSize.height - 1);
                     tk.getCursor(p);
                     tk.setCursorStay(p.x, p.y);
                     
                     // do not re-draw caret on next key wait
                     skipCaretDraw = true;
                  }
                  // clear the saved editor
                  lastEditorDisplayed = null;
               }
            }

            finally
            {
               if (sourceAvailable)
               {
                  waitForFrames.pop();
               }

               fixFocusedFrameZOrder();
            }
         }
         while (isGoPending == 0 && key != null);

         // Usually when condition is thrown we just don't reach this point
         // and down't send LEAVE and this is consistent with 4GL. But in case
         // when condition was intercepted by trigger and it exits with
         // NO-APPLY then we also should not send LEAVE to be consistent with
         // 4GL behavior.
         // if no key was pressed, it means the wait-for timed-out; so, no 
         // leave event will be sent.
         if (!condExit && key != null)
         {
            // generate LEAVE event for the last focused widget
            if (!leaveSent && cop != EventList.CHOOSE)
            {
               currentEventList = list;
               sendLeave(getCurrentFocus());
               currentEventList = save;
            }
         }
         else
         {
            // remove the highlight from the widget
            highlight(getCurrentFocus(), false);
         }
      }

      catch (ConditionException ce)
      {
         // rebalance the invalidate mode
         tk.resetInvalidate();
         
         // set condition flag
         condition = true;
         
         // let the original exception propagate
         throw ce;
      }
      
      finally
      {
         // cleanup
         currentEventList = save;
         isGoPending      = oldGoPending;
         leaveSent        = saveLeaveSent;
         condExit         = saveCondExit;
         immediateException = null;

         Component focused = getCurrentFocus();
         if (focused != null && condition)
         {
            Frame frame = (Frame) locateFrame(focused);
            if (focused.isVisible() && frame.isVisible())
            {
               highlight(focused, false);
            }
            else
            {
               focused.setHighlighted(false);
            }
         }

         triggerNestingByWaitFor.pop();
         
         if (statusSet)
         {
            stopInputStatus(false);
         }

         screenCurrent.resetNeedPause();
         
         // leave choose processing is done here because it needs to occur
         // *before* the getScreenBuffer() below, that is why it is here
         // instead of in the more logical place in choose()
         if (list.getOperation() == EventList.CHOOSE)
         {
            final Frame frame = (Frame) getWidget(inbuf.getBaseId());
            
            if (frame != null)
            {
               // leaveChoose should not produce any useful events
               eventBracket(true, new Runnable()
               {
                  public void run()
                  {
                     frame.leaveChoose();
                  }
               });
            }
         }
         
         tk.setCursorStatus(curvis);

         // set this trigger as it had wait-for
         triggerHadWaitFor.pop();
         triggerHadWaitFor.push(true);
      }

      // grab screen buffer before focus will be restored, because
      // focus state is relevant for FRAME-FIELD/FRAME-VALUE
      ScreenBuffer result = null;
      
      // where there is an WAIT-FOR ... ANYWHERE , there might be no active
      // frame
      if (inbuf != null)
      {
         result = getScreenBuffer(inbuf.getBaseId(), inbuf.size(), force);
      }
      
      // restore focus in the previous scope, if inside an editing block or
      // inside a trigger
      if (triggerNesting == 0 && !inEditingBlock)
      {
         // always set focus to undefined, if on root nesting level
         Window.getDefaultWindow().setFocus(null);
      }
      else
      {
         // for UPDATE operation, focus must be restored to previous focus - 
         // get it from the saved pending focus
         if (triggerNesting > 0 && 
             (cop == EventList.UPDATE || 
              cop == EventList.SET    || 
              cop == EventList.PROMPT_FOR))
         {
            focusState = pendingState;
            saveFocus = savePending;
         }

         switch(focusState)
         {
            case FOCUS_UNDEF:
               if (cop == EventList.WAIT_FOR)
               {
                  // focus has to be left wherever it is now
               }
               else
               {
                  // focus has to be undefined
                  Window.getDefaultWindow().setFocus(null);
               }
               break;
   
            case FOCUS_VALID:
            case FOCUS_INVALID:
               // focus has to be restored no matter enabled or not, and if
               // a fill-in receives it, we should use the current insert
               // mode for it rather than a stored one
               if (saveFocus != null &&
                   saveFocus instanceof FillIn)
               {
                  ((FillIn) saveFocus).setInsertMode(
                     Window.getInsertMode() == Window.INSERT_MODE_INS);
               }
               setFocusOn(saveFocus);
         }
      }

      // return the final screen buffer
      return result;
   }

   /**
    * Fix possible Z-ordering issues.
    */
   private void fixFocusedFrameZOrder()
   {
      if (getCurrentFocus() == null)
         return;
      
      final Frame frame = (Frame) locateFrame(getCurrentFocus());

      if (frame == null || !frame.hasEnabled())
         return;

      Frame[] flist = screenCurrent.getFramesAbove(frame);

      for (Frame ovl : flist)
      {
         FrameConfig config = (FrameConfig) ovl.getRef();

         if (!config.isOverlay())
            continue;
         
         eventDrawingBracket(false, false, new Runnable()
         {
            public void run()
            {
               frame.placeAtTop();
               frame.repaint();
            }
         });
         
         return;
      }
   }

   /**
    * Sets initial focus for the WAIT-FOR with valid focus on entry
    * and no explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @return <code>true</code> always to indicate success
    */
   private int getFocusState(Component inFocus)
   {
      if (inFocus == null)
      {
         return FOCUS_UNDEF;
      }
      
      if (inFocus.isVisible() == false)
      {
         return FOCUS_HIDDEN;
      }

      if (inFocus.isEnabled())
      {
         return FOCUS_VALID;
      }

      return FOCUS_INVALID;
   } 

   /**
    * Sets initial focus for the WAIT-FOR with valid focus on entry
    * and no explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @return <code>true</code> always to indicate success
    */
   private boolean focusValidImplicit(Component inFocus)
   {
      noFocusAdjustment = false;
      boolean state = inFocus.isFocusTraversable();

      // currently focused widget receives ENTRY event
      // NO-APPLY, if returned by a trigger, is ignored
      sendEntry(inFocus);

      if (state && inFocus.isFocusTraversable() != state)
      {
         // if it was disabled or hidden by the trigger, readjust focus,
         // starting from current frame
         if (searchForFocusWithoutEntry(inFocus) == false)
         {
            // focus remains undefined
            Window.getDefaultWindow().setFocus(null);
         }
      }

      // currently focused widget remains in focus or focus is undefined
      return true;
   }
   
   /**
    * If the toFocus widget lost its focusability, then:
    * <ul>
    *    <li>if the inFocus and toFocus widgets are both in the same frame,
    *        search for focus starting from:
    *          <ul>
    *             <li>toFocus, if inFocus is invalid</li>
    *             <li>inFocus, if inFocus is valid</li>
    *          </ul>
    *    </li>
    *    <li>else, focus is undefined if:
    *       <ul>
    *          <li>inFocus is invalid</li>
    *          <li>inFocus is valid and still focusable</li>
    *          <li>inFocus is valid, not focusable: focus first enabled and
    *              visible widget on screen, without ENTRY. if none found,
    *              focus is undefined.
    *          </li>
    *       </ul>
    *    </li>
    * </ul>
    * 
    * @param    inFocus
    *           the focus before the ENTRY event was triggered
    * @param    toFocus
    *           the widget which had ENTRY event
    * @param    state
    *           previous state of this widget (focusable or not)
    * @param    focusState
    *           the state of the focused widget, as before the trigger was
    *           invoked (valid or invalid).
    * 
    * @return   <code>true</code> if this widget lost focusability
    */
   private boolean widgetLostFocusability(Component inFocus,
                                          Component toFocus,
                                          boolean   state, 
                                          int       focusState)
   {
      if (state && toFocus.isFocusTraversable() != state)
      {
         boolean undefined = false;
         
         if (locateFrame(inFocus) == locateFrame(toFocus))
         {
            // Y was disabled, search for focus without ENTRY
            Component start = focusState == FOCUS_VALID ? inFocus : toFocus;
            undefined = searchForFocusWithoutEntry(start) == false;
         }
         else
         {
            // if X is still enabled and visible or no other widget to
            // set focus to, focus is undefined
            undefined = focusState == FOCUS_VALID && 
                         (inFocus.isFocusTraversable() ||
                          searchForFocusWithoutEntry(inFocus) == false);
         }
         
         if (undefined)
         {
            // focus remains undefined
            Window.getDefaultWindow().setFocus(null);
         }

         return true;
      }
      
      return false;
   }

   /**
    * Search for focus with ENTRY, starting from the specified widget,
    * without wrapping. If no enabled and visible widget which accepts entry
    * found in current frame, focus is undefined.
    * 
    * @param    start
    *           the widget from which focus search start
    * 
    * @return   <code>true</code> if a widget was found
    */
   private boolean searchForFocusWithEntry(Component start, Component toFocus)
   {
      Frame startFrame = (Frame) locateFrame(start);
      
      // send ENTRY to the specified widget
      boolean check = sendEntry(start);

      if (check == false)
      {
         // the specified widget did not accept entry
         // send entry to next widgets, in tab-order
         Component next = getNextEnabledWidget(startFrame, 
                                               start, 
                                               FOCUS_NOWRAP);
         while (next != null)
         {
            if (next == toFocus)
            {
               // reached explicit focus, it must receive focus before sending
               // ENTRY to it
               setFocusOn(next);

               return true;
            }

            // send ENTRY event
            check = sendEntry(next);
            if (check)
            {
               // focus found
               break;
            }
            
            next = getNextEnabledWidget(startFrame, next, FOCUS_NOWRAP);
         }
         
         if (next == null)
         {
            // no widget accepts ENTRY in firstFrame, focus is undefined
            Window.getDefaultWindow().setFocus(null);
         }
      }
      
      return false;
   }

   /**
    * Search for focus without ENTRY, starting from the specified widget,
    * with wrapping in its frame. If no enabled and visible widget found in
    * current frame, search for focus starting with first frame on screen.
    * 
    * @param    start
    *           the widget from which focus search start
    * 
    * @return   <code>true</code> if a widget was found
    */
   private boolean searchForFocusWithoutEntry(Component start)
   {
      if (start == null)
      {
         return false;
      }
      
      Frame frame = (Frame) locateFrame(start);
      
      if (frame.isVisible())
      {
         Component next = start;
         
         do
         {
            
            if (next.isFocusTraversable())
            {
               // widget was found, request focus
               setFocusOn(next);
               return true;
            }
            
            next = getNextEnabledWidget(frame, next, FOCUS_WRAP);
         }
         while (next != null);
      }
      
      Component first = getFirstEnabledWidget();
      if (first != null)
      {
         setFocusOn(first);
         return true;
      }

      // no widget was found
      return false;
   }

   /**
    * Sets initial focus for the WAIT-FOR with undefined or hidden focus on 
    * entry and no explicit FOCUS specification.
    *
    * @param    list
    *           list of event to wait for 
    * @param    hidden
    *           <code>true</code> if this is the case of hidden initial focus
    * 
    * @return   <code>true</code> if WAIT-FOR can proceed
    */
   private boolean focusUndefImplicit(EventList list, boolean hidden)
   {
      noFocusAdjustment = false;

      // find the very first enabled widget
      Component first = getFirstEnabledWidget();
      if (first == null)
      {
         // no enabled widgets; check event list for global events
         if (!checkEventList(list))
         {
            // no global events; should exit immediately
            return false;
         }
      }

      // found an enabled widget; figure out the owning frame
      Frame frame = (Frame)locateFrame(first);

      // find the first enabled widget of the frame which accepts ENTRY
      while (first != null)
      {
         noFocusAdjustment = true;
         boolean state = first.isFocusTraversable();

         // send ENTRY event and see the response
         boolean check = sendEntry(first);
         
         if (check)
         {
            if (state && first.isFocusTraversable() != state)
            {
               if (hidden == false || 
                   searchForFocusWithoutEntry(first) == false)
               {
                  // focus was disabled during ENTRY, focus remains undefined
                  Window.getDefaultWindow().setFocus(null);
               }

               return true;
            }

            // ENTRY accepted
            noFocusAdjustment = false;            
            break;
         }
         else
         {
            // ENTRY not accepted, remove highlight
            first.setHighlighted(false);
         }

         first = getNextEnabledWidget(frame, first, FOCUS_NOWRAP);
      }

      // end of search; either found or hit end of frame
      if (first == null)
      {
         // no enabled widgets; check event list for global events
         if (!checkEventList(list))
         {
            // no global events; should exit immediately
            return false;
         }
         
         // focus remains undefined
         return true;
      }

      // set focus on the found widget
      setFocusOn(first);

      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with invalid focus on entry
    * and no explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @return <code>true</code> if WAIT-FOR can proceed
    */
   private boolean focusInvalidImplicit(Component inFocus)
   {
      noFocusAdjustment = false;

      // figure out the frame
      Frame frame = (Frame)locateFrame(inFocus);

      // find the first enabled widget that follows
      Component first = getNextEnabledWidget(frame, inFocus, FOCUS_WRAP);

      if (first == null)
      {
         // no enabled widgets within the same frame; check others
         first = getFirstEnabledWidget();

         if (first == null)
         {
            // no enabled widgets at all; focus becomes undefined
            Window.getDefaultWindow().setFocus(null);

            return true;
         }
      }

      // set focus on the found widget
      setFocusOn(first);

      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with valid focus on entry
    * and an explicit FOCUS specification.
    *
    * @param    inFocus
    *           widget in focus
    * @param    toFocus
    *           candidate widget for receiving focus
    * @param    list
    *           list of event to wait for 
    * @param    cop
    *           currently executed operation
    * @param    hidden
    *           <code>true</code> if this is the case of hidden focus
    *         
    * @return   <code>true</code> if WAIT-FOR can proceed
    */
   private boolean focusValidExplicit(Component inFocus, 
                                      Component toFocus,
                                      EventList list, 
                                      int       cop,
                                      boolean   hidden)
   {
      noFocusAdjustment = false;

      // check the target's enabled status
      if (!toFocus.isFocusTraversable())
      {
         // leave the focus where it is
         setFocusOn(inFocus);

         return true;
      }
      
      Component inFrame = locateFrame(inFocus);
      Frame toFrame = (Frame) locateFrame(toFocus);
      boolean sameFrame = inFrame == toFrame;
      boolean fastExit = false;

      boolean isUpdate = cop == EventList.UPDATE || 
                         cop == EventList.SET    || 
                         cop == EventList.PROMPT_FOR;
      boolean isWaitFor = cop == EventList.WAIT_FOR;

      if (isWaitFor && hidden)
      {
         Component first = getFirstEnabledWidget();
         Frame firstFrame = (Frame) locateFrame(first);

         // if previous focus is hidden and explicit focus is the first 
         // focusable widget on screen, adjust focus to it
         if (toFocus == first)
         {
            setFocusOn(toFocus);
         }
         else
         {
            if (inFrame.isVisible() == false && inFrame == firstFrame)
            {
               // if the previous focus frame is hidden, focus is set to
               // undefined
                Window.getDefaultWindow().setFocus(null);
            }
            // else, search for focus. if explicit focus was reached, adjust
            // focus to it, but do not send ENTRY - it will be sent later
            fastExit = searchForFocusWithEntry(first, toFocus);
         }
      }
      
      // remove highlight from the previous focus
      highlight(inFocus, false);

      Component next = toFocus;

      // scan enabled widgets of the frame and chose the first accepting ENTRY
      while (next != null)
      {
         noFocusAdjustment = true;
         boolean state = next.isFocusTraversable();

         // send ENTRY event and see the response
         boolean check = sendEntry(next);

         if (check || (sameFrame && next == inFocus) || fastExit)
         {
            // no-apply must be ignored if X and Y are in the same frame and
            // X is reached
            // no-apply must be ignored if it was focused due to hidden 
            // initial focus

            if (widgetLostFocusability(inFocus, next, state, FOCUS_VALID))
            {
               return true;
            }

            // ENTRY accepted
            noFocusAdjustment = false;
            break;
         }
         else
         {
            // ENTRY not accepted, remove highlight
            next.setHighlighted(false);
         }

         // look up the next candidate
         next = getNextEnabledWidget(toFrame, next, FOCUS_NOWRAP);
      }

      if (next == null && sameFrame == false) 
      {
         // if focus was not found and X and Y are not in the same frame, then:
         // 1. if event list waits for Y frame events, end, sending leave to X
         // 2. else, focus another widget without ENTRY, starting from X
         if (isUpdate                                     ||
             list.anyEventForWidget(getWidgetId(toFocus)) ||
             list.anyEventForWidget(getWidgetId(toFrame)))
         {
            sendLeave(inFocus);

            return false;
         }
         else
         {
            if (!searchForFocusWithoutEntry(inFocus))
            {
               // focus remains null if no widget found
               Window.getDefaultWindow().setFocus(null);
            }
         }

         return true;
      }

      // check the success
      if (next == null)
      {
         // no widget wants focus; check event list for global events
         if (!checkEventList(list))
         {
            // no global events; should exit after sending LEAVE to inFocus
            sendLeave(inFocus);

            return false;
         }
         else
         {
            // leave the focus where it is
            setFocusOn(inFocus);

            return true;
         }
      }

      // a good candidate is found; set focus on it
      setFocusOn(next);

      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with undefined focus on entry
    * and an explicit FOCUS specification.
    *
    * @param  toFocus
    *         candidate widget for receiving focus
    * @param  list
    *         list of event to wait for
    * @param  cop
    *         currently executed operation
    * @param  explicit
    *         when <code>true</code>, it marks a real explicit focus; else, it
    *         marks explicit focus generated due to previous pending focus, in
    *         non-nested trigger mode.
    * 
    * @return <code>true</code> if WAIT-FOR can proceed
    */
   private boolean focusUndefExplicit(Component toFocus, 
                                      EventList list, 
                                      int cop,
                                      boolean explicit)
   {
      noFocusAdjustment = false;
      // if focus candidate is not the very first enabled widget on the screen
      // ENTRY may be sent to the latter first
      Component first  = getFirstEnabledWidget();
      Frame firstFrame = (Frame) locateFrame(first);
      Frame toFrame = (Frame) locateFrame(toFocus);

      boolean isUpdate = cop == EventList.UPDATE || 
                         cop == EventList.SET    || 
                         cop == EventList.PROMPT_FOR;
      boolean isWaitFor = cop == EventList.WAIT_FOR;

      // check the target's enabled status
      if (!toFocus.isFocusTraversable())
      {
         // find the very first enabled widget
         if (first == null)
         {
            // no enabled widgets; focus remains undefined
            return true;
         }

         if (isWaitFor)
         {
            
            // search for focus starting from the first widget sending ENTRY
            searchForFocusWithEntry(first, null);
            Component focused = getCurrentFocus();
            
            if (focused == first && !first.isFocusTraversable())
            {
               // if first enabled widget on screen accepted ENTRY and got  
               // disabled, focus is undefined
               Window.getDefaultWindow().setFocus(null);
            }
            // if no focus was found, then current operation must end
            if (focused == null)  
            {
               return false;
            }
         }
         else
         {
            // set focus on the found widget without sending ENTRY, if not
            // WAIT-FOR
            setFocusOn(first);
         }

         return true;
      }
      

      // send entry to the first enabled widget on screen W if this is the
      // wait-for operation and the target focus is not W
      boolean entryFirst = cop == EventList.WAIT_FOR && 
                           first != null             && 
                           first != toFocus;
      boolean fastExit = false;

      if (isUpdate && triggerNesting == 0 && pendingFocus != null)
      {
         Frame pendingFrame = (Frame) locateFrame(pendingFocus);

         // before sending ENTRY to any widget, adjust focus to last widget
         // which had ENTRY only if the frame is not out of scope
         if (pendingFrame != null && pendingFrame.isAlive())
         {
            if (pendingFocus.isFocusTraversable())
            {
               setFocusOn(pendingFocus);
            }
            // in this case, if explicit focus is disabled during ENTRY, focus
            // search starts from this 'implicit' focus
            first = pendingFocus;
         }
         
         // the pending focus is consumed on first usage
         pendingFocus = null;
      }
      
      if (entryFirst)
      {
         fastExit = searchForFocusWithEntry(first, toFocus);
      }

      // scan enabled widgets of the frame and chose the first accepting ENTRY
      Component next = toFocus;

      if (next != null && entryFirst)
      {
         // at least one more ENTRY event will be sent, remove highlight from
         // the widget
         first.setHighlighted(false);
      }
      
      while (next != null)
      {
         noFocusAdjustment = true;
         boolean state = next.isFocusTraversable();
         
         // send ENTRY event and see the response
         boolean check = sendEntry(next);

         if (check || fastExit)
         {
            // ignore NO-APPLY if there is a fast exit

            if (state && next.isFocusTraversable() != state)
            {
               boolean undefined = false;
               Component nextFrame = locateFrame(next);

               if (isUpdate || (isWaitFor && firstFrame == nextFrame))
               {
                  undefined = searchForFocusWithoutEntry(first) == false;
               }
               
               if (isWaitFor && firstFrame != nextFrame)
               {
                  undefined = (first != null && 
                               first.isFocusTraversable()) || 
                              searchForFocusWithoutEntry(first) == false;
               }
               
               if (undefined)
               {
                  Window.getDefaultWindow().setFocus(null);
               }
               
               return true;
            }

            // ENTRY accepted
            noFocusAdjustment = false;
            break;
         }
         else
         {
            // ENTRY not accepted, remove highlight
            next.setHighlighted(false);
         }

         // look up the next candidate
         next = getNextEnabledWidget(toFrame, next, FOCUS_NOWRAP);
      }

      // check the success
      if (next == null)
      {
         // if event list has events for toFocus or toFocus frame, 
         // wait-for ends, only if explicit is not caused by pending focus
         if (explicit &&
             (isUpdate                                     ||
              list.anyEventForWidget(getWidgetId(toFocus)) ||
              list.anyEventForWidget(getWidgetId(toFrame))))
         {
            if (pendingFocus != null)
            {
               // send LEAVE, ignoring NO-APPLY
               sendLeave(pendingFocus);

               // set focus to undefined
               Window.getDefaultWindow().setFocus(null);
               // 'consume' the focus
               pendingFocus = null;
            }
            return false;
         }
         
         // no widget wants focus; set focus on first enabled widget on screen
         // or undefined, if none found
         next = getFirstEnabledWidget();
         if (next == null)
         {
            Window.getDefaultWindow().setFocus(null);
         }
         else
         {
            setFocusOn(next);
            // although ENTRY was not accepted, focus was changed
            noFocusAdjustment = false;
         }

         return true;
      }

      // a good candidate is found; set focus on it
      setFocusOn(next);

      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with invalid focus on entry
    * and an explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @param  toFocus
    *         candidate widget for receiving focus
    * @param  list
    *         list of event to wait for
    *
    * @return <code>true</code> if WAIT-FOR can proceed
    */
   private boolean focusInvalidExplicit(Component inFocus, Component toFocus,
                                        EventList list)
   {
      noFocusAdjustment = false;

      Frame frame = null;
      Component next = null;

      // check the target's enabled status
      if (!toFocus.isFocusTraversable())
      {
         if (searchForFocusWithoutEntry(inFocus) == false)
         {
            // no enabled widgets at all; focus becomes undefined
            Window.getDefaultWindow().setFocus(null);
         }

         return true;
      }

      // scan enabled widgets of the frame and chose the first accepting ENTRY
      frame = (Frame)locateFrame(toFocus);
      next = toFocus;

      while (next != null)
      {
         noFocusAdjustment = true;
         boolean state = next.isFocusTraversable();

         // send ENTRY event and see the response
         boolean check = sendEntry(next);

         if (check)
         {
            // check if the widget was disabled during trigger
            if (widgetLostFocusability(inFocus, next, state, FOCUS_INVALID))
            {
               return true;
            }

            // ENTRY accepted
            noFocusAdjustment = false;
            break;
         }
         else
         {
            // ENTRY not accepted, remove highlight
            next.setHighlighted(false);
         }

         // look up the next candidate
         next = getNextEnabledWidget(frame, next, FOCUS_NOWRAP);
      }

      // check the success
      if (next == null)
      {
         // no widget wants focus; check event list for global events
         if (!checkEventList(list))
         {
            // no global events; should exit immediately
            return false;
         }
         else
         {
            // set focus on toFocus ignoring NO-APPLY 
            setFocusOn(toFocus);

            return true;
         }
      }

      // a good candidate is found; set focus on it
      setFocusOn(next);

      return true;
   }

   /**
    * Sends ENTRY event to the specified widget and returns a boolean result
    * of the possible trigger invocation.
    *
    * @param   widget 
    *          target widget for sending ENTRY event
    * @return  <code>true</code> if no trigger or the trigger has returned
    *                            normally<br>
    *          <code>false</code> if the trigger has returned no-apply
    */
   private boolean sendEntry(Component widget)
   {
      // create an ENTRY event which does not require matching LEAVE
      KeyInput event = new KeyInput(Keyboard.SE_ENTRY, KeyEvent.KEY_TYPED,
                                    widget, true);

      tk.setInvalidate(true);
      applyWorker(event);
      tk.setInvalidate(false, true);

      return triggerResultNoApply.peek() == false;
   }

   /**
    * Sends LEAVE event to the specified widget.
    *
    * @param   widget 
    *          target widget for sending LEAVE event
    *          
    * @return  <code>true</code> if no trigger or the trigger has returned
    *                            normally<br>
    *          <code>false</code> if the trigger has returned no-apply
    */
   private boolean sendLeave(Component widget)
   {
      leaveSent=true;

      if (widget == null || widget == Window.getDefaultWindow())
      {
         // no leave is sent if no widget or the widget is the default window
         return true;
      }

      KeyInput event = new KeyInput(Keyboard.SE_LEAVE, KeyEvent.KEY_TYPED,
                                    widget);

      tk.setInvalidate(true);
      applyWorker(event);
      tk.setInvalidate(false, true);
      
      if (widget instanceof FillIn)
      {
         // fill-in's always reset the cursor on the first position, after the
         // widget is left
         ((FillIn) widget).setCursorOffset(0);
      }

      return !event.isConsumed();
   }

   /**
    * Looks up the very first enabled widget among all displayed frames.
    *
    * @return the first enabled widget or <code>null</code>
    */
   private Component getFirstEnabledWidget()
   {
      // loop through the displayed frames
      Frame frame = screenCurrent.getFirstVisibleFrame();

      while (frame != null)
      {
         if (frame.isAlive())
         {
            // find the first enabled widget of this frame, searching only for
            // in-scope frames
            Component[] widgets = frame.getContentPane().getComponents();
            Component comp = null;
   
            for (int i = 0; i < widgets.length; i++)
            {
               comp = frame.getFrameHelper().getField(widgets[i]);
               if ((comp instanceof AttributeContainer) &&
                    comp.isFocusTraversable())
               {
                  return comp;
               }
            }
         }

         // try next frame
         frame = screenCurrent.getNextVisibleFrame(frame);
      }

      // no enabled widgets in all frames
      return null; 
   }

   /**
    * Looks up the very first enabled widget in specified frames.
    * 
    * @param   frame
    *          The frame where widget will be searched.
    *          
    * @return  the first enabled widget or <code>null</code>
    */
   private Component getFirstEnabledWidget(Frame frame)
   {
      if (frame == null)
         return null;
      
      // find the first enabled widget of this frame
      Component[] widgets = frame.getContentPane().getComponents();
      Component comp = null;

      for (int i = 0; i < widgets.length; i++)
      {
         comp = frame.getFrameHelper().getField(widgets[i]);
         
         if ((comp instanceof AttributeContainer) && comp.isFocusTraversable())
         {
            return comp;
         }
      }

      // no enabled widgets in all frames
      return null; 
   }
   
   
   /**
    * Looks up the next enabled and visible widget in the tab order of the 
    * specified frame.
    *
    * @param  frame
    *         frame to search 
    * @param  widget
    *         widget to start search after 
    * @param  wrap
    *         if <code>true</code> the search wraps to the first widget
    *         of the frame when the last widget fails
    *
    * @return the next enabled and visible widget or <code>null</code>
    */
   private Component getNextEnabledWidget(Frame frame, Component widget,
                                          boolean wrap)
   {
      // get all widgets of this frame
      Component[] widgets = frame.getContentPane().getComponents();

      // find the reference point
      int nref = 0;
      for (nref = 0; nref < widgets.length; nref++)
      {
         if (widgets[nref] == widget)
         {
            break;
         }
      }

      // if no reference point, assume no result
      if (nref == widgets.length)
      {
         return null;
      }

      // find the next enabled widget with no wrapping first
      Component comp = null;

      for (int i = nref + 1; i < widgets.length; i++)
      {
         comp = frame.getFrameHelper().getField(widgets[i]);
         if ((comp instanceof AttributeContainer) &&
              comp.isFocusTraversable())
         {
            return comp;
         }
      }

      // if wrapping is required, check the first section of the list
      if (wrap && nref > 0)
      {
         for (int i = 0; i < nref; i++)
         {
            comp = frame.getFrameHelper().getField(widgets[i]);
            if ((comp instanceof AttributeContainer) &&
                 comp.isFocusTraversable())
            {
               return comp;
            }
         }
      }

      // no enabled widgets in this frame
      return null; 
   }

   /**
    * Checks the given event list for global events.
    *
    * @param  list
    *         event list to check 
    * @return <code>true</code> if the list contains at least one global event
    */
   private boolean checkEventList(EventList list)
   {
      return list.anyEventForWidget(WindowConfig.SESSION_WINDOW_ID);
   }

   /**
    * Sets focus on the specified widget silently and unconditionally.
    *
    * @param  widget
    *         widget to set focus on; it should be a valid widget, but
    *         can be disabled  
    */
   private void setFocusOn(final Component widget)
   {
      Component oldFocus = getCurrentFocus();
      boolean ins = Window.getInsertMode() == Window.INSERT_MODE_INS;

      eventDrawingBracket(false, false, new Runnable()
      {
         public void run()
         {
            setInitialFocus(widget);
         }
      });

      Component newFocus = getCurrentFocus();
      if (newFocus != null &&
          newFocus instanceof FillIn)
      {
         FillIn newFillIn = (FillIn) newFocus;

         if (oldFocus == null)
         {
            newFillIn.setInsertMode(ins);
         }
         else if (newFocus != oldFocus &&
                  (!oldFocus.isEnabled() || !oldFocus.isVisible()))
         {
            if (oldFocus instanceof Editor)
            {
               newFillIn.setInsertMode(false);
            }
            else
            {
               newFillIn.setInsertMode(ins);

               if (oldFocus instanceof FillIn &&
                   !((FillIn) oldFocus).isZapActive() &&
                   !newFillIn.isLinkedToBrowse())
               {
                  newFillIn.setPendingNoZap();
               }
            }
         }
      }
   }

   /**
    * Adjusts focus if it is either undefined or invalid and a trigger has
    * enabled at least one widget.
    */
   private void adjustFocus()
   {
      // get the current focus
      Component focus = getCurrentFocus();
      Component next  = null;

      // get the focus state: either undefined or invalid
      int focusState = getFocusState(focus);

      // if focus is invalid, the next enabled widget in the frame gets it
      if (focusState == FOCUS_INVALID || focusState == FOCUS_HIDDEN)
      {
         Frame frame = (Frame)locateFrame(focus);
         // if invalid, search for next enabled widget, with wrapping
         next = getNextEnabledWidget(frame, focus, FOCUS_WRAP);

         if (next != null)
         {
            // set focus on the next enabled widget silently
            setFocusOn(next);

            return;
         }
         
         // if focus is disabled and there is no other enabled widget, 
         // then focus is not adjusted if other widget type beside combo or 
         // fill-in is in focus and the last trigger returned normal
         if (!triggerResultNoApply.peek() &&
             !triggerHadWaitFor.peek()    &&
             !(focus instanceof FillIn || focus instanceof ComboBox))
         {
            // this is only if there was no nested wait-for loop executed
            // by the trigger
            
            if (frame.hasEnabled())
               return;
         }
      }

      // if focus is undefined, the very first enabled widget gets it
      next = getFirstEnabledWidget();

      if (next == null)
      {
         // still nothing
         // set focus to undefined, if no enabled widget found
         Window.getDefaultWindow().setFocus(null);
         // set the status to default if there is no enabled widget
         status(null);
         return;
      }

      // set focus on the first enabled widget silently
      setFocusOn(next);

      // adjust zap and insert state as if ENTRY was sent
      if (focus instanceof FillIn && 
          next instanceof FillIn  && 
          focusState == FOCUS_HIDDEN)
      {
         if (((FillIn) focus).isZapActive())
         {
            FillIn fNext  = (FillIn) next;
            
            fNext.enableZapMode();
            fNext.setInsertMode(false);
         }
      }
   }

   /**
    * Checks the current focus to see whether it has to be adjusted.
    *
    */
   private boolean focusNeedsAdjustment()
   {
      // get the current focus
      Component focus = getCurrentFocus();
      if (focus == null)
      {
         return true;
      }

      // locate the frame in focus
      Component frame = ThinClient.locateFrame(focus);
      if (frame == null)
      {
         return true;
      }

      // check if the ficus is valid
      if (getFocusState(focus) != FOCUS_VALID)
      {
         return true;
      }

      // check if the widget's frame is visible
      if (!frame.isVisible())
      {
         return true;
      }

      // focus needs no adjustment
      return false;
   }

   /**
    * Handle postponed VALUE-CHANGED if any.
    */
   private void processPostponedValueChanged()
   {
      while (postponedValueChanged != null)
      {
         KeyInput key = postponedValueChanged;
         postponedValueChanged = null;
         
         Component frame = ThinClient.locateFrame((Component) key.getSource());
         
         if (frame != null && frame.getParent() != null)
            applyWorker(key);
      }
   }
         
   /**
    * Activate input mode.
    *
    * @param    cursor
    *           New state of the cursor. <code>true</code> for the cursor
    *           to be forced into visibility and <code>false</code> to hide
    *           the cursor.
    */
   private void startInputStatus(boolean cursor)
   {
      // force the new cursor mode and save the original mode
      oldcursor.push(new Boolean(tk.setCursorStatus(cursor)));
      inputActive++;
      
      statusType = HELP;
      status(helpText);
   }
   
   /**
    * Deactivate input mode.
    */
   private void stopInputStatus()
   {
      stopInputStatus(true);
   }
   
   /**
    * Deactivate input mode.
    * 
    * @param   changeStatus
    *          Allow changing status line. 
    */
   private void stopInputStatus(boolean changeStatus)
   {
      // reset cursor to previous state
      tk.setCursorStatus(oldcursor.pop().booleanValue());

      inputActive--;

      if (inputActive == 0 && changeStatus)
      {
         // reset (blank) the status line
         status(null);
      }
   }

   /**
    * Signals a trigger to the application.  If a trigger is registered for
    * this event and source widget combination, then it will be executed.
    * 
    * @param    triggerId
    *           The trigger to be called.
    * @param    eventId
    *           The event which was generated.
    * @param    sourceId
    *           The widget which generated the event.
    * @param    sb
    *           The associated frame's current <code>ScreenBuffer</code>
    *
    * @return   <code>true</code> if the event must be consumed (this is
    *           equivalent to <code>RETURN NO-APPLY</code>.
    */
   private boolean trigger(int          triggerId,
                           int          eventId,
                           int          sourceId,
                           ScreenBuffer sb)
   {
      boolean saveLeaveSent = leaveSent;
      int saveGoPending = isGoPending;
      Component savePending = pendingFocus;
      leaveSent = false;
      pendingFocus = null;

      // a special case for EDITOR - clear saved editor
      lastEditorDisplayed = null;

      // save the frame for the current focus
      Component focus = getCurrentFocus();
      Frame frame = null;
      boolean saveVisibilityChanged = false;
      boolean saveHighlighted = false;
      
      if (focus != null)
      {
         frame = (Frame) locateFrame(focus);
         // save visibility-changed flag, so it can be restored after trigger
         // invocation
         saveVisibilityChanged = frame.isVisibilityChanged();
         frame.setVisibilityChanged(false);
         saveHighlighted = focus.isHighlighted();
      }
      
      int savedInsState = saveInsState();

      int saveLevel = tk.saveInvalidate();

      int val = triggerNestingByWaitFor.pop();
      triggerNestingByWaitFor.push(val + 1);
      
      triggerNesting++;

      TriggerReturnValue trv = null;
      
      try
      {
         trv = server.trigger(triggerId, eventId, sourceId, sb);
      }
      finally
      {
         triggerNesting--;

         val = triggerNestingByWaitFor.pop();
         triggerNestingByWaitFor.push(val - 1);

         tk.restoreInvalidate(saveLevel);
         pendingFocus = savePending;

         if (trv != null && !trv.consume)
         {
            restoreInsState(savedInsState, false);

            // fix for the case when trigger starts and ends up into the same
            // fill-in
            if (focus != null &&
                focus == getCurrentFocus() &&
                focus instanceof FillIn)
            {
               ((FillIn) focus).setInsertMode(savedInsState == Window.INSERT_MODE_INS);
               clearSavedInsState();
            }
         }
         else
         {
            // trigger returned NO-APPLY, leave current insert mode 
            restoreInsState(-1, false);
         }

         if (isGoPending == saveGoPending)
         {
            // restore flag only if no GO was generated during trigger - this 
            // will block duplicate LEAVE events
            leaveSent = saveLeaveSent;
         }

         // check if focused frame changed its visibility (from not-visible to
         // visible)
         if (frame != null)
         {
            // move focus to the targeted widget
            if (frame.isVisibilityChanged() && !frame.isFocusReset())
            {
               // old focus loses highlight
               if (focus != null)
                  focus.setHighlighted(false);

               Component comp = getWidget(sourceId);
               // the target widget receives focus
               comp.requestFocus();

               highlight(comp, saveHighlighted);
            }

            // restore the flag
            frame.setVisibilityChanged(saveVisibilityChanged);
         }
      }

      if (trv != null)
      {
         setScreenBuffer(trv.buffer);
         return trv.consume;
      }
      
      return false;
   }

   /**
    * Signals a validation event to the application.
    * <p>
    * This signal is used to trigger validation execution on the server.
    * 
    * @param   sourceId
    *          widget ID of the widget generated the trigger.
    * @param   sb
    *          current <code>ScreenBuffer</code>
    * 
    * @return  result of validation check.
    *  
    */
   public boolean validate(int sourceId, ScreenBuffer sb)
   {
      ValidationResult vr = server.validate(sourceId, sb);
      setScreenBuffer(vr.getScreenBuffer());
      
      boolean res = vr.isValidated();
      if (res == false)
      {
         // retrieve validation error message
         validationError = vr.getError();
      }
      
      return res;
   }

   /**
    * Returns the last validation error message.
    * <p>
    *
    * @return  error message generated during the last validation call or
    *          <code>null</code> if the last validation succeeded.
    */
   public String getValidationError()
   {
      return validationError;
   }

   /**
    * Return information about columns.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * 
    * @return  A meta information about columns.
    */
   public BrowseColumnConfig[] getColumns(int browseId)
   {
      return server.getColumns(browseId);
   }

   /**
    * Get number of available rows starting from the given one.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * 
    * @param   startRow
    *          Initial row to start count rows from. 
    * 
    * @return  Number of rows in browse.
    */
   public int getRowCount(int browseId, int startRow)
   {
      return server.getRowCount(browseId, startRow);
   }
   
   /**
    * Retrieves several subsequent rows from the database.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   startRowIndex
    *          Index of the first row to retrieve.
    * @param   count
    *          Maximum count of rows to return.
    * 
    * @return  Array of rows. Each row is array of cell values.
    */
   public BaseDataType[][] getRows(int browseId, int startRowIndex, int count)
   {
      return server.getRows(browseId, startRowIndex, count);
   }
   
   /**
    * Updates selected cells in the given row.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          Index of the row to update.
    * @param   cellIndices
    *          Cells for update indexes.
    * @param   cellValues
    *          Cells for update new values.
    * 
    * @throws  DataValidationException
    *          New values don't passed database validation rules.
    */
   public void updateRow(int            browseId,
                         int            rowIndex,
                         int[]          cellIndices, 
                         BaseDataType[] cellValues)
   throws DataValidationException
   {
      try
      {
         server.updateRow(browseId, rowIndex, cellIndices, cellValues);
      }
      catch (DataValidationException e)
      {
         validationError = e.getMessage();
         throw e;
      }
      cleanupQueue();      
   }

   /**
    * Returns <code>true</code> if specified row is new and is not present
    * in the data base.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          Index of the row to return information about.
    * @return  <code>true</code> if row is new and <code>false</code>
    *          otherwise.
    */
   public boolean isNewRow(int browseId, int rowIndex)
   {
      return server.isNewRow(browseId, rowIndex);
   }
   
   /**
    * Inform server about current row in the browse.
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          Index of the current row.
    *          
    * @return  New current row.
    */
   public int setCurrentRow(int browseId, int rowIndex)
   {
      return server.setCurrentRow(browseId, rowIndex);
   }

   /**
    * Get current row in the query.
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    *          
    * @return  The current row index.
    */
   public int getCurrentRow(int browseId)
   {
      return server.getCurrentRow(browseId);
   }
   
   /**
    * Checks to see whether a pending GO event is about to end the editing
    * block.
    * 
    * @return  <code>true</code> if there is no pending GO and the editing
    *          mode continues
    */
   public boolean continueEditing()
   {
      putCursor(true);
      // handle EOF too
      boolean flag = isGoPending > 0 || lastActionCode == -2;

      if (flag)
      {
         stopEditingMode();
      }
      else
      {
         iterateEditingBlock();
      }

      return !flag;
   }

   /**
    * Marks the beginning of the editing block and passes a customized list of
    * GO-ON events that define termination condition for this I/O operation.
    * <p>
    * This call is responsible for setting the initial focus to the first
    * enabled widget or NEXT-PROMPT hint.
    * 
    * @param   focusWidgetId
    *          ID of the target frame for this editing session
    * @param   eventList
    *          Information about events to wait for.
    */
   private void startEditingMode(int focusWidgetId, EventList eventList)
   {
      currentEventList = eventList;
      inEditingBlock  = true;
      isGoPending     = 0;
      hadReset        = false;
      forceWipeScreen = false;
      Component comp =  getWidget(focusWidgetId);
      Component frame = locateFrame(comp);
      // translate widget 
      if (frame instanceof Frame)
         comp = ((Frame) frame).getFrameHelper().getField(comp);
      
      // this should be before the focus change to leave the cursor in
      // the right place
      startInputStatus(true);

      // save current focus, to be restored when editing block finishes
      editingFocus.push(getCurrentFocus());

      // ENTRY must be sent to the requested widget before the EDITING loop 
      // starts executing
      focusUndefExplicit(comp, eventList, eventList.getOperation(), true);
      
      editingFrames.push((Frame)frame);
   }

   /**
    * Marks the end of the editing block when triggered by any abnormal exit
    * from the block on the server.  This method does NOT need to be called
    * if the exit is naturally triggered via {@link #continueEditing}. This
    * method allows the client's state to be cleaned up in the abnormal
    * termination case.
    */
   public void stopEditingMode()
   {
      if (ignoreNextStopEditing)
      {
         ignoreNextStopEditing = false;
         return;
      }
      
      stopInputStatus();
      isGoPending      = 0;      
      inEditingBlock   = false;
      hadReset         = false;
      currentEventList = null;
      editingFrames.pop();

      if (isRedirected())
      {
         // if the frame is redirected, perform a view so the frame will
         // output to the stream
         int frameId = lastViewed.getFrameHelper().config.getID();
         ScreenBuffer sb = server.refresh(frameId);
         view(frameId, sb, null, true);
      }

      // restore focus to the widget which had focus before editing mode 
      // started, if in nested trigger mode
      Component focus = editingFocus.pop();
      if (triggerNesting > 0 && focus != null)
      {
         setFocusOn(focus);
      }
      else
      {
         Window.getDefaultWindow().setFocus(null);
      }

      // restore insert indicator state
      restoreInsState(savedInsStateInitials.pop(), true);
      forceWipeScreen = true;
   }
   
   /**
    * Conditionally blocks until a next UI event is available.
    */
   public void waitForNextKey()
   {
      TypeAhead.waitForInput();
   }
   
   /**
    * Called by the protocol driver to obtain the latest state that should
    * be sent to the other side of the connection.  This is called just
    * before a "real" message is sent and any changes provided will be
    * piggybacked into that message.
    * <p>
    * To enable incremental changes, it is guaranteed that this method will
    * be called every time any message is sent to the other side of the
    * connection.  In addition, this will only be called once per message.
    * <p>
    * In a system in which control flow is "sent" to the other side of the
    * connection (a type of "conversation" or single-threaded synchronous
    * mode), it is further guaranteed that any changes that are made at the
    * other side after this is called will result in a call to this side's
    * {@link #applyChanges} method before this method is ever called again
    * (when control flow returns to this side of the connection). This allows
    * one to ensure that on the call to this method, any "dirty" flags can be
    * safely cleared and the state of this side will be updated with
    * the results of the other side's changes before this method must ever
    * again check to see if there are changes to send.
    *
    * @return   The payload of state changes to be sent to the other side or
    *           <code>null</code> if no changes need to be sent.
    */
   public Serializable getChanges()
   {
      ClientState cs = new ClientState();

      if (removedFrames.size() > 0)
      {
         cs.deregister = Utils.integerCollectionToPrimitive(removedFrames);
         removedFrames.clear();
      }
      
      if (wrappedFrames.size() > 0)
      {
         cs.wrapped = Utils.integerCollectionToPrimitive(wrappedFrames);
         wrappedFrames.clear();
      }

      ArrayList<Integer> ids  = new ArrayList<Integer>();
      ArrayList<Integer> down = new ArrayList<Integer>();

      int[] list = screenCurrent.getFrameIDs();
      
      for (int i : list)
      {
         // skip tinyInput frame
         if (i == INPUT_FRAME_ID)
            continue;
         
         Frame frame = getFrame(i);

         if (!frame.getFrameHelper().isDown() || frame.getDown() == 0)
            continue;

         ids.add(new Integer(i));
         down.add(new Integer(frame.getDown()));
      }

      cs.downIDs = Utils.integerCollectionToPrimitive(ids);
      cs.down    = Utils.integerCollectionToPrimitive(down);

      // was there a pause in this last operation?
      if (hadPause)
      {
         // notify the server
         cs.hadPause = true;
         
         // clear our state
         hadPause = false;
      }
      
      // send the widgets with changed HIDDEN flag to the server
      if (hiddenWidgets != null)
      {
         cs.hiddenWidgets = Utils.integerMapToPrimitive(hiddenWidgets);
         // clear the list, no longer needed
         hiddenWidgets = null;
      }
      
      // send the widgets with changed VISIBLE flag to the server
      if (visibleWidgets != null)
      {
         cs.visibleWidgets = Utils.integerMapToPrimitive(visibleWidgets);      
         // clear the list, no longer needed
         visibleWidgets = null;
      }

      return cs;
   }

   /**
    * Called by the protocol driver upon reading a message that contains
    * state changes that need to be applied to this side of the connection.
    * This is called just after a new message is read from the connection
    * and only if that message had piggybacked state changes attached. This
    * is guaranteed to be called before the containing message is enqueued
    * for further "real" processing on this node.
    *
    * @param    changes
    *           The payload of state changes that must be locally applied.
    */
   public void applyChanges(Serializable changes)
   {
      // WARNING: DO NOT call any processing from here which may generate
      //          an exception (especially any of the ConditionExceptions).
      //          Any pausing logic or other event processing must be
      //          avoided.  This is because this method is called during
      //          protocol-driver processing, before the proper environment
      //          is setup to deliver any condition exceptions back to the
      //          server as a valid response.  Anything thrown at this time
      //          will cause the client to abend!

      ServerState sstate = (ServerState) changes;
      interactions = sstate.interactions;
      
      if (sstate.nesting > nesting)
      {
         pushStatus(sstate.nesting - nesting);
      }
      else if (sstate.nesting < nesting)
      {
         popStatus(nesting - sstate.nesting);
      }
      
      nesting = sstate.nesting;

      // cache any conditional up/down value for the specified frame (it will
      // be honored in the actual ClientExports call that follows)
      if (sstate.conditionalFrameId != 0)
      {
         this.conditionalUpDown = sstate.conditionalUpDown;
         this.conditionalFrameId = sstate.conditionalFrameId;
         
         Component frame = getWidget(conditionalFrameId);
         
         if (frame instanceof Frame && 
             !((Frame) frame).getFrameHelper().isDown() &&
             frame.isVisible())
         {
            Frame f = (Frame)frame;

            f.setNeedReset(true); 

            // only frames on the root nesting level are
            // forced a repaint
            if (f.getScopeNesting() == 1)
            {
               f.setDelayedRepaint(true);
            }
         }
      }

      if (sstate.formatCheckId > 0)
      {
         Component frame = getWidget(sstate.formatCheckId);
         
         if (frame instanceof Frame && frame.isVisible())
         {
            ((Frame) frame).setFormatCheck(true);
         }
      }
      
      // implements deferred hide/pause MARKING (only state changes, no
      // actual hiding or pausing occurs)
      if (sstate.markList != null)
      {
         for (MarkEntry item : sstate.markList)
         {
            markFrames(item);
         }
      }
   }

   /**
    * Checks whether the CHUI is up and running on the client.
    * 
    * @return  Always <code>true</code>.
    */
   public boolean isChui()
   {
      return true;
   }
   
   /**
    * Destroy the frame and free associated resources as needed. If the frame
    * is a dialog, it will be hidden immediately, otherwise it will be marked
    * as out of scope and valid for being hidden and resources will be
    * released when frame will be hidden.
    * 
    * @param    frameId
    *           The frame requiring cleanup.
    */
   public void destroyFrame(int frameId)
   {
      final Frame frame = (Frame) getWidget(frameId);

      if (frame == null)
         return;
      
      // cleanup any traces of this frame in our page elements list
      forceDeregisterPageElement(frameId);
      
      // mark the frame as one that can be destroyed (it is inactive)
      frame.setAlive(false);
      
      // dialog boxes MUST actually hide right now
      if (frame.isDialog() && frame.isVisible())
      {
         eventDrawingBracket(new Runnable()
         {
            public void run()
            {
               // this also triggers frame removal
               frame.setVisible(false);
            }
         });
      }
      else
      {
         // hidden frames removed immediately.
         if (!frame.isVisible())
         {
            removeFrame(frame);
         }
         else
         {
            // redirected visible frames should be removed immediately
            // safety net, should not happen normally
            if (frame.isRedirected() || frame.isPlacedForRedirected())
            {
               removeFrame(frame);
            }
            else
            {
               // at this time, all enabled widgets must be disabled
               // but only for frames which were nor redirected nor
               // placed for redirected mode (i.e. this is only for
               // interactive mode frames!)

               selectiveDisable(frame);
            }

            // regular visible frames will be removed when setVisible(false)
            // is called on them.
         }
      }
   }
   
   /**
    * Selectively disable and redraw widgets.
    * 
    * @param    frame
    *           Frame where widgets should be disabled.
    */
   private void selectiveDisable(final Frame frame)
   {
      final ArrayList<Component> widgets = new ArrayList<Component>();
      
      final Frame[] above = screenCurrent.getFramesAbove(frame);
      
      eventBracket(true, new Runnable()
      {
         public void run()
         {
            ScreenBitmap bitmap = widgetDrawingOn(frame);
      
            Component[] list = frame.getContentPane().getComponents();
      
            // collect enabled widgets
            for (Component comp : list)
            {
               if (comp.isEnabled() && comp instanceof AttributeContainer)
               {
                  comp.setEnabled(false);
                  widgets.add(comp);
               }
            }
      
            // redraw those which are visible and were actually disabled
            for (Component comp : widgets)
            {
               if (comp.isVisible() && 
                   (above.length == 0 || isAllShowing(comp, above)))
               {
                  comp.draw();
               }
            }
            
            widgetDrawingOff(bitmap);
         }
      });
   }

   /**
    * Checks whether a GO event is pending for this editing operation
    * (either an <code>editing</code> block or a <code>wait-for</code>).
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>GO-PENDING </b>
    * 
    * @return   <code>true</code> if an event has been applied that caused
    *           pending GO for the current editing operation.
    */
   public boolean isGoPending()
   {
      return isGoPending > 0;
   }
   
   /**
    * Get frame widget from the given frame ID.
    * 
    * @param   frameId
    *          Frame ID.
    * @return  Reference to Frame instance.
    */
   public Frame getFrame(int frameId)
   {
      Frame frame = (Frame) getWidget(frameId);
      
      if (frame == null)
         throw new RuntimeException("Invalid frame ID " + frameId);
      
      return frame;
   }
   
   /**
    * Returns the current logical line number in a frame.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-LINE
    * </pre>
    *
    * @param    frame
    *           The frame for this request.  Must NOT be <code>null</code>.
    * @param    pendingUpDown
    *           The pending down value for this frame.
    * 
    * @return   1-based ordinal of the current data iteration of the frame.
    *           If the frame is not a down frame then 1 will be returned.
    */
   private int getFrameLine(Frame frame, int pendingUpDown)
   {
      FrameHelper fh = frame.getFrameHelper();
      int cursor = fh.getCursor() + pendingUpDown;

      return cursor + 1;
   }

   /**
    * Process events internally if there are no interactive requests from the
    * server.
    */
   private void cleanupQueue()
   {
      if (currentEventList == null)
         processEventsWorker();
   }

   /**
    * Force the cursor to the specified location and make it visible again.
    *
    * @param    curpos
    *           The 0-based location to which the cursor must be forced.
    * @param    curvis
    *           The new cursor visibility state (<code>true</code> means
    *           set the cursor visible).
    *
    * @return   The previous cursor visibility state.
    */
   private boolean restoreCursor(Point curpos, boolean curvis)
   {
      // move the cursor
      tk.setCursor(curpos);
      
      // the physical cursor position on the terminal is not moved until a
      // sync, until then the cursor is only moved in a logical operation
      // and setCursorStatus will re-display the current physical cursor 
      // without a sync so the cursor will appear at an unexpected location
      // if we don't call sync here
      tk.sync();
      
      // set the cursor visibility
      return tk.setCursorStatus(curvis);
   }
      
   /**
    * Force the initial focus to a specific widget when entering an
    * interactive/editing mode.  This will generate <code>FOCUS_GAINED</code>
    * but it will not generate a <code>LEAVE</code> or <code>FOCUS_LOST</code>
    * since it is the initial focus set.
    * <p>
    * This processing will cause a repaint so it must be bracketed by
    * invalidation processing in the caller.
    *
    * @param    widget
    *           The widget to receive focus (must not be <code>null</code>).
    */
   private void setInitialFocus(Component widget)
   {
      // unconditional focus change since it's the initial set and should
      // not generate a LEAVE event for the previous widget

      Toolkit.postEvent(new FocusEvent(AWTEvent.FOCUS_GAINED, widget));
      
      if (widget.getParent() != null)
         widget.getParent().setFocus(widget);
      
      if (inputActive > 0)
      {
         widget.setHighlighted(true);
      }

      widget.repaint();
      
      Frame frame = (Frame) locateFrame(widget);
      
      if (frame != null && getActiveFrameInt() != frame)
         lastViewed = frame;
   }
   
   /**
    * Honor the deferred processing for a conditional up or down.
    * 
    * @param    noPause
    *           Make no pause if frame will need it during DOWN processing.
    *           
    * @return   <code>true</code> if down frame was wrapped during processing of
    *           conditional UP/DOWN.
    */
   private boolean honorConditionalUpDown(boolean noPause)
   {
      boolean res = false;
      
      if (conditionalFrameId != 0)
      {
         int tmp = conditionalFrameId;
         conditionalFrameId = 0;
         res = downWorker(tmp, conditionalUpDown, true, false, noPause);
         conditionalUpDown  = 0;
      }
      
      return res;
   }
   
   /**
    * Honor the deferred processing for a conditional up or down.
    */
   private void honorConditionalUpDown()
   {
      if (conditionalFrameId != 0)
      {
         Frame frame = (Frame) getWidget(conditionalFrameId);
         
         if (frame != null)
         {
            if (frame.isAllClear())
            {
               frame.setAllClear(false);
               conditionalUpDown  = 0;
               conditionalFrameId = 0;
               return;
            }
            int tmp = conditionalFrameId;
            conditionalFrameId = 0;
            downWorker(tmp, conditionalUpDown, true, false, false);
         }
         conditionalUpDown  = 0;
         conditionalFrameId = 0;
      }
   }
   
   /**
    * Get <code>Component</code> instance for a given ID.
    * 
    * @param   componentId
    *          ID of the requested <code>Component</code> instance.
    * @return  reference to <code>Component</code> instance with given ID.
    */
   public Component getWidget(int componentId)
   {
      return screenCurrent.getComponent(componentId);

   }
   
   /**
    * Sets the format specification for the specified widget and redraws the
    * widget.
    *
    * @param    widget
    *           The widget to modify.
    * @param    format
    *           The format specification.
    */
   public void setFormat(final int widget, final String format)
   {
      eventDrawingBracket(new Runnable()
      {
         public void run()
         {
            Component comp = getWidget(widget);
            
            if (comp != null)
            {
               int         fid   = WidgetId.frameIdFromWidgetId(widget);
               Frame       frame = getFrame(fid);
               FrameHelper fh    = frame.getFrameHelper(); 
               
               if (fh.isDown())
               {
                  comp = fh.getField(comp);
               }
            
               // TODO: this really should raise an error condition if the
               //       new format string specifies a larger size than can 
               //       already fit in the widget (the frame should NOT relayout
               //       but it does in our implementation)
               if (comp instanceof AttributeContainer)
                  ((AttributeContainer) comp).getRef().setFormat(format);
            }
         }
      });
   }

   /**
    * Returns the action code of the last key input event. 
    *
    * @return The action code of the last key input event.
    */
   int getLastActionCode()
   {
      return lastActionCode;
   }

   /**
    * Prepare message text for specified time to wait.
    * 
    * @param   seconds
    *          Seconds to wait for an event. Negative value means wait
    *          indefinitely and number of seconds is not displayed.
    * 
    * @return  prepared text message.
    */
   private String makePauseText(int seconds)
   {
      StringBuffer buf = new StringBuffer();
      
      if (seconds > 0)
      {
         buf.append("Pausing for ");
         buf.append(seconds);
         buf.append(" seconds. ");
      }
      
      buf.append(PBH_DEFAULT_TEXT);

      return buf.toString();
   }
   
   /**
    * Wait for key press indefinitely or for specified number of seconds.
    * 
    * @param   seconds
    *          -1 means indefinite wait, 0 means no wait and positive values
    *          mean the number of seconds to wait
    * @param   frameAsTarget
    *          if <code>true</code>, then the active frame is the target of
    *          events if nothing is enabled and in focus, otherwise
    *          the current window
    * @return  key event code or -1 if no key was received during specified
    *          time.
    */
   private KeyInput waitForKey(int seconds, boolean frameAsTarget)
   {
      // get and check the focus - the widget in focus may be disabled
      Component comp = getCurrentFocus();
      if (comp != null && !comp.isEnabled())
      {
         comp = null;
      }

      // reposition the cursor before blocking (interactive mode only)
      positionCursor();

      // determine the target for this keystroke
      // If there is no focused and enabled widget, the target should be
      // either the last used frame or the current window.
      // This implementation leaves the decision to the caller. Check the
      // waitForWorker() method, where this decision is made. That logic
      // should allow CHOOSE to always direct the keystrokes to the frame

      if (comp == null || comp instanceof Window)
      {
         Frame frame = getActiveFrameInt();

         if (frameAsTarget && frame != null)
         {
            comp = frame;
         }
         else
         {
            comp = Window.getDefaultWindow();
         }
      }
      
      AWTEvent evt = TypeAhead.getKeystroke(seconds * 1000);
      
      KeyInput key = null;
      
      if (evt instanceof KeyEvent)
      {
         if (comp == null)
            comp = (Component) evt.getSource();
         
         boolean real = (evt instanceof KeyInput) ? 
                         ((KeyInput) evt).isRealKey() : false; 
         
         key = new KeyInput(((KeyInput) evt).getKeyCode(), 
                            evt.getID(), 
                            comp,
                            false,
                            real);
         lastKey = key.getKeyCode();
      }
      else
      {
         lastKey = -1;
      }
      
      return key;
   }
   

   /**
    * Run pause cleanup block if any.
    */
   private void pauseCleanup()
   {
      synchronized (cleanupSynchronizer)
      {
         if (cleanupBlock != null)
            cleanupBlock.run();
         cleanupBlock = null;
      }
   }

   /**
    * Generate a copy of the <code>ScreenBuffer</code> instance for the given
    * frame or for the frame in which the given widget is contained.
    *
    * @param    widget
    *           A frame or a widget in the frame whose screen buffer must be
    *           returned.
    *
    * @return   <code>ScreenBuffer</code> instance.
    */
   private ScreenBuffer getScreenBuffer(Component widget)
   {
      // find the associated frame 
      Component frame = locateFrame(widget);
      
      if (frame == null)
      {
         // nothing to do
         return null;
      }
      
      int frameId = ((AttributeContainer) frame).getRef().getID();
      
      return getScreenBuffer(frameId, -1, false);
   }
         
   /**
    * Generate a copy of the <code>ScreenBuffer</code> instance for the given
    * frame.
    *
    * @param    frameId
    *           The frame whose screen buffer must be returned.
    * @param    size
    *           The number of widgets in the screen buffer in total or -1
    *           to allow this value to be calculated.
    *
    * @return   <code>ScreenBuffer</code> instance.
    */
   private ScreenBuffer getScreenBuffer(int frameId, int size)
   {
      return getScreenBuffer(frameId, size, false);
   }  
      
   /**
    * Generate a copy of the <code>ScreenBuffer</code> instance for the given
    * frame.
    *
    * @param    frameId
    *           The frame whose screen buffer must be returned.
    * @param    size
    *           The number of widgets in the screen buffer in total or -1
    *           to allow this value to be calculated.
    * @param    force
    *           <code>true</code> to copy the current data into the screen
    *           buffer without regard to the changed flag.
    *
    * @return   <code>ScreenBuffer</code> instance.
    */
   private ScreenBuffer getScreenBuffer(int frameId, int size, boolean force)
   {      
      if (frameId == -1)
      {
         // nothing to do
         return null;
      }
      
      // get the list of all widgets
      int[] list = screenCurrent.getIDs();
      
      // find out the maximum number of widgets, if needed
      if (size == -1)
      {
         size = 0;
         
         for (int i = 0; i < list.length; i++)
         {
            if (list[i] <= frameId || list[i] >= frameId + 1000)
               continue;
            size ++;
         }
      }
      
      // create the ScreenBuffer that big
      Component       comp   = null; 
      ComponentConfig config = null;
      Frame           frame  = (Frame) getWidget(frameId);
      
      ScreenBuffer buffer = new ScreenBuffer(frameId, size);
      
      for (int i = 0; i < list.length; i++)
      {
         if (list[i] <= frameId || list[i] >= frameId + 1000)
            continue;
            
         comp = screenCurrent.getComponent(list[i]); 
         
         // WARNING: short-cut taken here! Since the only caller that uses
         //          force flag today is CHOOSE, we use this flag for a
         //          2nd purpose. If more users are made, then this would
         //          need to change at that time.
         
         if (force)
         {
            // CHOOSE does not maintain the down frame cursor in ROW mode
            // so we need to find the actual widget that is at the selected
            // index position using the base widget to identify the column.
            comp = frame.getSelectedField(comp);
         }
         else
            comp = frame.getFrameHelper().getField(comp);
         
         if (comp instanceof DataContainer)
         {
            config = ((AttributeContainer) comp).getRef();
            
            if (force || config.getState() == ScreenBuffer.CHANGED)
            {
               buffer.putWidgetValue(list[i], 
                                     ((DataContainer) comp).getValue());
               
               if (comp instanceof ComboBox)
               {
                  buffer.putScreenValue(list[i], 
                                        ((ComboBox) comp).getScreenValue());
               }
               else if (comp instanceof SelectionList)
               {
                  buffer.putScreenValue(list[i], 
                                        ((SelectionList) comp).getScreenValue());
               }
            }
         }
      }

      // retrieve frame value/widget ID.
      
      Frame.FrameData dt = frame.getFrameValue(null);
      
      if (dt != null && dt.getValue() != null)
      {
         buffer.setFrameValue(dt.getValue());
         buffer.setValueWidgetId(dt.getId());
      }
      
      resetChanged(frameId);
      
      // non-redirected down frames need to switch screen buffers on
      // the server if the frame-line has changed during our execution
      // on the client
      if (frame.getFrameHelper().isDown() && !frame.isRedirected())
      {
         buffer.putLine(getFrameLine(frame, 0));
      }
      
      buffer.putKeyCode(lastKey);
      
      return buffer;
   }

   /**
    * Parses <code>ScreenBuffer</code> instance for current set of frames
    * and assigns new widget values.
    *
    * @param    frameBuf
    *           <code>ScreenBuffer</code> instance.
    */
   private void setScreenBuffer(ScreenBuffer frameBuf)
   {
      setScreenBuffer(frameBuf, false);
   }
   
   /**
    * Parses <code>ScreenBuffer</code> instance for current set of frames
    * and assigns new widget values.
    * <p>
    * Allows direct drawing of headers.
    *
    * @param    frameBuf
    *           <code>ScreenBuffer</code> instance.
    * @param    direct
    *           if set to <code>true</code>, header widgets draw directly
    *           instead of posting <code>PaintEvent</code>.
    */
   private void setScreenBuffer(final ScreenBuffer frameBuf, boolean direct)
   {
      if (frameBuf == null)
      {
         return;
      }
      
      int baseId = frameBuf.getBaseId();
      int size   = frameBuf.size();
      int state  = -1;
      int id     = -1;

      final Frame     frame  = getFrame(baseId);
      FrameHelper     fh     = frame == null ? null : frame.getFrameHelper();
      BaseDataType    value  = null;
      ComponentConfig config = null;
      
      if (fh != null)
      {
         if(fh.isDown())
         {
            if (fh.getCursor() < 0)
               fh.clear();

            fh.touchRow();
         }
         else
            frame.touchBody();
      }

      // do not do direct drawing if frame is hidden
      if (frame != null && !frame.isVisible())
         direct = false;
      
      // honor dynamic title if the frame is HIDDEN and a dynamic title was
      // specified (it is not null)
      String dynTitle = frameBuf.getTitle();
      
      if (frame != null && dynTitle != null)
      {
         frame.getRef().setTitle(dynTitle);
      }       
      
      // iterate through all widgets and push new values and state as needed      
      for (int i = 0; i < size; i ++)
      {
         id = baseId + i + 1;
         state = frameBuf.getState(id);

         // get the widget
         Component comp = screenCurrent.getComponent(id);
         if (fh != null && fh.isDown())
            comp = fh.getField(comp);

         config = ((AttributeContainer) comp).getRef();

         int oldStat = config.getState();
         if (oldStat == state && state != ScreenBuffer.CHANGED)
            continue;

         boolean enabled = comp.isEnabled();

         switch (state)
         {
            case ScreenBuffer.UNCHANGED:
               if (oldStat == ScreenBuffer.UNINITIALIZED && !enabled)
                  continue;
               // fall through               
               
            case ScreenBuffer.UNINITIALIZED:
               
               if ((config instanceof ControlTextConfig)  && 
                   ((ControlTextConfig)config).isStatic() &&
                   state == ScreenBuffer.UNINITIALIZED)
                  break;
                  
               if (fh == null)
                  config.setState(state);
               else
                  fh.setState(comp, state);
               
               if (config instanceof RadioSetConfig && frame == null)
                  ((RadioSet) comp).reset();
               else
                  if (fh != null)
                     fh.resetWidget(comp);

               break;

            case ScreenBuffer.CHANGED:
               if (comp instanceof DataContainer)
               {
                  String fmt = null;
                  Object wv  = frameBuf.getWidgetValue(new Integer(id));
                  
                  if (wv instanceof FormattedValue)
                  {
                     FormattedValue fv = (FormattedValue) wv;
                     value = fv.getValue();
                     fmt   = fv.getFormat();
                  }
                  else
                  {
                     value = (BaseDataType) wv;
                  }
                          
                  setChangedValue(frame, comp, config, value, fmt);
               }
               else
               {
                  System.err.println("Unexpected type (not DataContainer) " +
                                     "in setScreenBuffer: " + comp);
               }
               break;

            default:;
         }
      }
      
      // refresh frame headers
      if (direct)
      {
         eventBracket(true, new Runnable()
         {
            public void run()
            {
               ScreenBitmap bitmap = null;
               
               if (frame != null)
                  bitmap = widgetDrawingOn(frame);
               
               updateHeaders(frameBuf);
         
               if (frame != null)
                  widgetDrawingOff(bitmap);
            }
         });
      }
      else
      {
         updateHeaders(frameBuf);
      }
      
      if (frame != null && 
          frame.getFrameHelper().isDown() &&
          frame.getFrameHelper().getCursor() < 0)
      {
         // reset cursor to 0
         frame.getFrameHelper().clear();
      }
   }

   /**
    * Update form headers from given {@link ScreenBuffer} instance.
    * 
    * @param   frameBuf
    *          The {@link ScreenBuffer} instance to update headers from.
    *          
    * @return  List of IDs of updated header widgets. 
    */
   private ArrayList<Integer> updateHeaders(ScreenBuffer frameBuf)
   {
      Iterator<Integer> iter = frameBuf.getHeaderIDIterator();
      BaseDataType value;
      ArrayList<Integer> updated = new ArrayList<Integer>();
      
      while (iter.hasNext())
      {
         Integer hid = iter.next();
         
         Component comp = screenCurrent.getComponent(hid.intValue());
         
         if (comp instanceof DataContainer)
         {
            value = (BaseDataType) frameBuf.getHeaderValue(hid);
            
            if (!value.equals(((DataContainer) comp).getValue()))
            {
               ((DataContainer) comp).setValue(value);
               updated.add(hid);
            }
         }
      }
      
      return updated;
   }
   
   /**
    * Applies the encoded event to the widget currently in focus in the given
    * frame, or directly to the specified widget.
    * 
    * @param   event
    *          Progress event 
    */
   private void applyWorker(final KeyInput event)
   {
      eventDrawingBracket(false, false, new Runnable()
      {
         public void run()
         {
            Toolkit.postEvent(event);
         }
      });
   }
   
   /**
    * Clear specified frame widgets.
    * 
    * @param   frame
    *          Frame to clear.
    * @param   noPause
    *          Do not generate pause.
    * @param   clearAll
    *          If <code>true</code> then all widgets are cleared (the
    *          <b>CLEAR ALL</b> behavior).
    */
   void clearWorker(final Frame frame, boolean noPause, final boolean clearAll)
   {
      // must be called first!
      boolean res = honorConditionalUpDown(noPause);

      if (frame == null)
         return;

      // DOWN frame is wrapped, nothing to do, it is already
      // paused/wrapped as necessary.
      if (frame.getFrameHelper().isDown() && res)
         return;
      
      try
      {
         if (!noPause && frame.isNeedPause() && frame.isVisible())
            pauseBeforeHide();
      }
      
      catch (ConditionException ce)
      {
         // rebalance the invalidate mode
         tk.resetInvalidate();
         
         // let the original exception propagate
         throw ce;
      }
      
      finally
      {
         eventDrawingBracket(new Runnable()
         {
            public void run()
            {
               if (clearAll)
               {
                  // this represents a CLEAR FRAME ALL stmt call - in this case,
                  // we must clear the entire frame, not just the current iteration.
                  frame.getFrameHelper().clearWorker(clearAll);
               }
               else
               {
                  // this is the plain CLEAR FRAME stmt call - in this case, only
                  // the current iteration of a DOWN frame must be cleared
                  frame.getFrameHelper().clearCurrent();
               }
      
               if (!frame.isRedirected())
                  frame.repaint();     
            }
         });
      }
   }
     
   /**
    * Collect HIDDEN or VISIBLE flag to sent it to the server, if it changed.
    * This should always be called BEFORE the actual flag change.
    *  
    * @param    comp
    *           The component to be checked.
    * @param    state
    *           The new state of the flag.
    * @param    visFlag
    *           <code>true</code> if this is for the VISBILE flag.
    *           <code>false</code> if this is for the HIDDEN flag.
    */
   void collectFlag(Component comp, boolean state, boolean visFlag)
   {
      boolean oldState = visFlag ? comp.isVisible() : comp.isHidden();
      if (oldState != state)
      {
         Map<Integer, Integer> map = null;
         if (visFlag)
         {
            if (visibleWidgets == null)
            {
               visibleWidgets = new HashMap<Integer, Integer>(5);
            }
            map = visibleWidgets;
         }
         else
         {
            if (hiddenWidgets == null)
            {
               hiddenWidgets = new HashMap<Integer, Integer>(5);
            }
            map = hiddenWidgets;
         }
         
         map.put(getWidgetId(comp), state ? 1 : 0);
      }
   }
   
   /**
    * Execute the given code inside a new bracket of event processing which
    * starts the toolkit capturing invalidation rectangles, then starts a new
    * event bracket with a {@link #push}, executes the code and then ends with
    * a {@link #pop} and finally disables invalidation mode. The closing of the
    * event bracket and the disabling of invalidation mode (which can cause 
    * drawing) occur even if the executed code throws an exception. This ensures
    * that the event processing and drawing will never get out of balance.
    * <p>
    * This version forces flushing when the invalidation mode is disabled. In
    * addition, the invalidation counter is saved at the beginning and restored
    * at the end of this method.
    *
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>.
    */
   boolean independentEventDrawingBracket(Runnable code)
   {
      boolean result = false;
      int     save   = tk.saveInvalidate();
      
      try
      {
         result = eventDrawingBracket(false, true, code);
      }
      
      finally
      {
         tk.restoreInvalidate(save);
      }
      
      return result;
   }
   
   /**
    * Execute the given code inside a new bracket of event processing which
    * starts the toolkit capturing invalidation rectangles, then starts a new
    * event bracket with a {@link #push}, executes the code and then ends with
    * a {@link #pop} and finally disables invalidation mode. The closing of the
    * event bracket and the disabling of invalidation mode (which can cause 
    * drawing) occur even if the executed code throws an exception. This ensures
    * that the event processing and drawing will never get out of balance.
    * <p>
    * This version forces flushing when the invalidation mode is disabled.
    *
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>.
    */
   boolean eventDrawingBracket(Runnable code)
   {
      return eventDrawingBracket(false, true, code);
   }
   
   /**
    * Execute the given code inside a new bracket of event processing which
    * starts the toolkit capturing invalidation rectangles, then starts a new
    * event bracket with a {@link #push}, executes the code and then ends with
    * a {@link #pop} (or {@link #popTrash}) and finally disables invalidation
    * mode. The closing of the event bracket and the disabling of invalidation
    * mode (which can cause drawing) occur even if the executed code throws an
    * exception. This ensures that the event processing and drawing will never
    * get out of balance.
    * <p>
    * This version forces flushing when the invalidation mode is disabled.
    *
    * @param    discard
    *           <code>true</code> to use <code>popTrash</code>, 
    *           <code>false</code> to use <code>pop</code>.
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>. If 
    *           <code>popTrash</code> is used, the result will always be
    *           <code>false</code>.
    */
   boolean eventDrawingBracket(boolean discard, Runnable code)
   {
      return eventDrawingBracket(discard, true, code);
   }
   
   /**
    * Execute the given code inside a new bracket of event processing which
    * starts the toolkit capturing invalidation rectangles, then starts a new
    * event bracket with a {@link #push}, executes the code and then ends with
    * a {@link #pop} (or {@link #popTrash}) and finally disables invalidation
    * mode. The closing of the event bracket and the disabling of invalidation
    * mode (which can cause drawing) occur even if the executed code throws an
    * exception. This ensures that the event processing and drawing will never
    * get out of balance.
    *
    * @param    discard
    *           <code>true</code> to use <code>popTrash</code>, 
    *           <code>false</code> to use <code>pop</code>.
    * @param    flush
    *           Force flushing when invalidation mode is disabled.
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>. If 
    *           <code>popTrash</code> is used, the result will always be
    *           <code>false</code>.
    */
   boolean eventDrawingBracket(boolean discard, boolean flush, Runnable code)
   {
      boolean result = false;
      
      tk.setInvalidate(true);
      
      try
      {
         result = eventBracket(discard, code);
      }
      
      finally
      {
         tk.setInvalidate(false, flush);
      }
      
      return result;
   }
   
   /**
    * Execute the given code inside a new bracket of event processing which
    * starts with a {@link #push} and ends with a {@link #pop} even if the
    * executed code throws an exception. This ensures that the event processing
    * will never get out of balance.
    *
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>.
    */
   boolean eventBracket(Runnable code)
   {
      return eventBracket(false, code);
   }
   
   /**
    * Execute the given code inside a new bracket of event processing which
    * starts with a {@link #push} and ends with a {@link #pop} or
    * {@link #popTrash} even if the executed code throws an exception. This
    * ensures that the event processing will never get out of balance.
    *
    * @param    discard
    *           <code>true</code> to use <code>popTrash</code>, 
    *           <code>false</code> to use <code>pop</code>.
    * @param    code
    *           Logic to safely execute inside the event bracket.
    *
    * @return   The result as returned by <code>pop</code>. If 
    *           <code>popTrash</code> is used, the result will always be
    *           <code>false</code>.
    */
   boolean eventBracket(boolean discard, Runnable code)
   {
      boolean result = false;
      
      push();

      try
      {
         code.run();
      }
      
      catch (RuntimeException re)
      {
         // allow these to propagate
         throw re;
      }
      
      catch (Exception exc)
      {
         // log and ignore these
         exc.printStackTrace(System.err);
      }
      
      finally
      {
         if (discard)
         {
            popTrash();
         }
         else
         {
            result = pop();
         }
      }
      
      return result;
   }

   /**
    * Process events and restore previous event queue.
    *
    * @return <code>true</code> if at least one frame related 
    *         <code>PaintEvent</code> has been processed
    */
   private boolean pop()
   {
      try
      {
         return processEventsWorker();
      }
      finally
      {
         tk.popEventQueue();
      }
   }

   /**
    * Trash all events and restore previous event queue.
    */
   private void popTrash()
   {
      while (Toolkit.getEvent() != null)
      {
         // eat the events
      }
      tk.popEventQueue();
   }
   
   /**
    * Save current event queue.
    */
   private void push()
   {
      tk.pushEventQueue();
   }

   /**
    * Accessor for active {@link Screen} instance.
    * 
    * @return  {@link Screen} instance.
    */
   public static Screen getScreen()
   {
      return getInstance().screenCurrent;
   }
   
   /**
    * Get trigger nesting.
    * 
    * @return  current trigger nesting level.
    */
   public static int getTriggerNesting()
   {
      return getInstance().triggerNesting;
   }
   
   /**
    * Get current state of the needPause flag. It tracks recent input and
    * other operations to tell application that pause is required.
    * 
    * @return  Current state of the flag.
    */
   public static boolean needPause()
   {
      return getInstance().needPause;
   }

   /**
    * Perform pause before end of application.
    */
   public void pauseBeforeEnd()
   {
      if (isNeedPause() || messageNeedPause)
      {
         pause(timePBH, PBE_DEFAULT_TEXT, false);
      }
      else
      {
         hideMessage(false);
      }
   }
   
   /**
    * Set help text.
    * When the request comes from a widget which is part of an ongoing CHOOSE,
    * the help text is overridden with whatever was supplied for the CHOOSE
    * operation.
    * 
    * @param   text
    *          Text for help string or <code>null</code>, which indicates the
    *          default status message has to be restored
    */
   public static void setHelp(String text)
   {
      String helpText = text;
      ThinClient tc = getInstance();

      Component comp = getCurrentFocus();
      if (comp != null)
      {
         Frame frame = (Frame)locateFrame(comp);
         if (frame != null && frame.isInChoose(comp))
         {
            helpText = frame.getChooseHelp();
         }
      }
      
      if (helpText == null || helpText.isEmpty())
      {
         tc.helpText = null;
         tc.status(null);
      }
      else
      {
         tc.status(helpText);
         tc.helpText = helpText;
         tc.statusType = HELP;
      }
   }

   /**
    * Callback method that notifies the application about Ctrl-C keystroke
    * and returns the decision the application has made about the disposition
    * of the Ctrl-C keystroke.
    * <p>
    * This method is called in the context of the key reader thread.
    * <p>
    * This implementation's goal is to properly generate the STOP condition.
    *
    * @return Ctrl-C keystroke disposition value as one of the predefined
    *         action code constants.
    */
   public int watchCtrlC()
   {
      if (authMode)
         return Watcher.CLEAR_AND_QUEUE;

      pauseCleanup();
      setNeedPause(false);
      
      Control.interrupt();
      return Watcher.IGNORE;
   }

   /**
    * Set authentication mode flag. When authentication mode is enabled,
    * Ctrl-C keystroke is not passed to {@link Control} but processed as a
    * normal sequence.
    * 
    * @param    auth
    *           New state of the flag.
    */
   public void setAuthMode(boolean auth)
   {
      authMode = auth;
   }
   
   /**
    * Looks up the redirected terminal instance for a given stream.
    *
    * @param    out
    *           The stream being queried.
    * @return   instance of <code>RedirectedTerminal</code> associated with
    *           the stream.
    */
   public static RedirectedTerminal redirectedTerminalForStream(Stream out)
   {
      ThinClient tc = ThinClient.getInstance();
      int sid = tc.sd.idFromStream(out);
      Integer key = new Integer(sid);
      RedirectedTerminal rt = tc.redirections.get(key);
      return rt;
   }

   /**
    * Notify the registered object that the given stream is being closed.
    * <p>
    * This listener is used to catch the event of the stream closure to flush
    * any output that may remain sitting in the redirected terminal buffer.
    *
    * @param    out
    *           The stream being closed.
    */
   public void notifyClose(Stream out)
   {
      flushStream(out);
   }

   /**
    * Notify registered object that given stream reached end of life cycle and
    * all references to it can be safely discarded.
    * <p>
    * This notification is used to unbind frames which were switched from
    * interactive mode to redirected mode and restore their initial state.
    *
    * @param    out
    *           The stream being closed.
    */
   public void notifyUnbind(Stream out)
   {
      Component[] c = screenCurrent.getTopLevelComponents();

      int id = idFromStream(out);

      for (Component comp : c)
      {
         if (comp instanceof Frame)
         {
            ((Frame) comp).unbind(id);
         }
      }
   }
   
   /**
    * Notify the registered object that the given stream is advancing page.
    * <p>
    * This listener is used to catch the event to flush
    * any output that may remain sitting in the redirected terminal buffer.
    *
    * @param    out
    *           The stream being paged
    */
   public void notifyAdvance(Stream out)
   {
      int sid = sd.idFromStream(out);

      boolean switched = currentStream == sid;
      int tempOut = -1;

      if (!switched)
      {
         tempOut = switchOutput(sid, true);
      }

      switchFrames(sid, -1);
 
      if (!switched)
      {
         switchOutput(tempOut, false);
      }
   }

   /**
    * Notify the registered object that the given stream is flushing.
    * <p>
    * This listener is used to catch the event to flush
    * any output that may remain sitting in the redirected terminal buffer.
    *
    * @param    out
    *           The stream being flushed.
    */
   public void notifyFlush(Stream out)
   {
      flushStream(out);
   }

   /**
    * Returns the stream ID associated with a given stream.
    *
    * @param    s
    *           stream to operate on
    *
    * @return   The stream ID or <code>-1</code> if no stream ID is associated
    *           with the given stream.
    */
   public int idFromStream(Stream s)
   {
      return sd.idFromStream(s);
   }

   /**
    * Returns the stream associated with a given stream.ID.
    *
    * @param    sid
    *           stream ID
    *
    * @return   The stream instance or <code>null</code> if stream ID unknown.
    */
   public Stream streamFromId(int sid)
   {
      return sd.streamFromId(sid);
   }

   /**
    * Perform complete input operation: create fill-in widget, process user
    * input and return result.
    * 
    * @param   row
    *          Message row in which operation will be performed.
    * @param   var
    *          Source variable.
    * @param   format
    *          Format which will be used for variable.
    * @param   auto
    *          Fill-in AUTO-RETURN flag.
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    *          
    * @return  The result of the input operation if the operation finished
    *          successfully, otherwise <code>null</code>.
    */
   private MessageReturnValue tinyInput(int          row,
                                        BaseDataType var,
                                        String       format,
                                        boolean      auto,
                                        boolean      set)
   {
      // save the currently focused widget so it can be restored later
      // (this must be done before the frame is added otherwise the result
      // will yield the "fin" widget below)
      final Component oldFocus = getCurrentFocus();
      Frame saveLastViewed = lastViewed;
      
      String[] msg = Window.getMessageText();
      int msgl = msg[row].length();
      int id0 = INPUT_FRAME_ID;     
      int id1 = INPUT_FRAME_ID + 1; // must remain INPUT_FRAME_ID + 1
      
      FrameConfig cfg = new FrameConfig(ComponentConfig.FRAME, null);
      cfg.setID(id0);
      cfg.setBox(false);
      cfg.setRow(Window.getScreenDimension().height + row);
      cfg.setColumn(msg[row].length() + 1);
      cfg.setOverlay(true);
      cfg.setLabels(false);

      FillInConfig fcfg = new FillInConfig(ComponentConfig.FILL_IN, null);
      fcfg.setID(id1);
      fcfg.setFrameId(id0);
      fcfg.setAutoReturn(auto);
      fcfg.setDataType(var.getTypeName());
      
      if (format != null)
         fcfg.setFormat(format);
      
      Frame savedTinyInput = (Frame) getWidget(INPUT_FRAME_ID);
      String[] messages = Window.getMessageText();
      int[]    colors   = Window.getMessageColor();
      
      if (savedTinyInput != null)
      {
         screenCurrent.removeFrame(savedTinyInput);
         Window.getDefaultWindow().removeFrame(savedTinyInput);
      }
      
      screenCurrent.reconstructWidget(cfg);
      screenCurrent.reconstructWidget(fcfg);
      
      Frame frame = (Frame) screenCurrent.getComponent(id0); 
      FillIn fin = (FillIn) screenCurrent.getComponent(id1);
      fin.setForceLeft(true);
      
      int maxlen = screenSize.width - fin.getWidth() - 1;
      
      if (msgl > maxlen)
      {
         // message is too long to be accommodated with the update FILL-IN
         // and is to be truncated

         msg[row] = msg[row].substring(0, maxlen) +
                    StringHelper.repeatChar(' ', msgl - maxlen);

         Window.setMessageText(msg, null);
         
         cfg.setColumn(maxlen + 1);
         
         Point loc = frame.getLocation();
         loc.x = maxlen + 1;
         frame.setLocation(loc);
      }
      
      frame.addWidget(fin);
      Window.getDefaultWindow().addFrame(frame, 0, 0);

      frame.doLayout();
      
      // Frame adding/removal is performed manually,
      // so we must protect if from automated processing.
      frame.setProtected(true);
      
      ScreenBuffer sb = new ScreenBuffer(id0, 1);
      
      if (!var.isUnknown())
         sb.putWidgetValue(id1, var.duplicate());
      
      tk.setInvalidate(true);

      eventBracket(new Runnable()
      {
         public void run()
         {
            if (oldFocus != null)
               Toolkit.postEvent(new FocusEvent(AWTEvent.FOCUS_LOST, oldFocus));
         }
      });
      
      enable(id0, null, sb, true);
      clear(id0, true);
      sb.resetChanged();
      tk.setInvalidate(false);

      setNeedPause(false);
      
      ScreenBuffer sbf = null;
      
      tinyInputLine = row;
      
      lastViewed = frame;
      
      // save insert indicator state in order to restore after on leaving
      // waitForWorker()
      int savedInsState = saveInsState();
      String status     = Window.getStatusText();
      int type          = statusType;

      try
      {
         EventList list = new EventList(true);
         list.setOperation(set ? EventList.SET : EventList.UPDATE);

         sbf = waitForWorker(list, id1, -1, sb, false, true, false);
      }
      
      catch (ConditionException ce)
      {
         // rebalance the invalidate mode
         tk.resetInvalidate();

         // let the original exception propagate
         throw ce;
      }
         
      finally
      {
         // cleanup that must occur even if a condition is raised
         tinyInputLine = -1;
         enable(id0, null, sb, false);
         hide(id0, true);
         Frame tmpFrame = (Frame) screenCurrent.getComponent(id0);
         screenCurrent.removeFrame(tmpFrame);
         Window.getDefaultWindow().removeFrame(tmpFrame);

         if (savedTinyInput != null)
         {
            screenCurrent.addFrame(savedTinyInput);
            Window.getDefaultWindow().addFrame(savedTinyInput, 0, 0);
            Window.setMessageText(messages, colors, true);
         }
         else
            hideMessage(true);
         
         setNeedPause(false);
         
         statusType = type;
         status(status);
         
         if (oldFocus != null)
         {
            eventDrawingBracket(false, false, new Runnable()
            {
               public void run()
               {
                  setInitialFocus(oldFocus);
               }
            });
         }
         
         lastViewed = saveLastViewed;

         restoreInsState(savedInsState, true);
      }
      
      MessageReturnValue mrv = new MessageReturnValue();
      
      mrv.keycode      = sbf.getKeyCode();
      mrv.var          = (BaseDataType) sbf.getWidgetValue(id1);
      mrv.formattedVar = fin.getText();

      // no edit occurred
      if (mrv.var == null)
      {
         // return original variable
         mrv.var = var;
      }
      
      return mrv;
   }
   
   /**
    * Simulate the display of a varaible attached to a MESSAGE ... UPDATE/SET
    * statement, to be able to retrieve the correct formatted value.
    * <p>
    * TODO: check if there is a simpler way to do this. 
    * 
    * @param    var
    *           The varabile for which the formatted value is needed.
    * @param    format
    *           The format of the variable.
    * @param    auto
    *           The AUTO-RETURN state of the MESSAGE statement.
    *           
    * @return   the formatted value of the given variable.
    */
   private String viewFillInVariable(BaseDataType var, 
                                     String format,
                                     boolean auto)
   {
      int id0 = INPUT_FRAME_ID;
      int id1 = INPUT_FRAME_ID + 1; // must remain INPUT_FRAME_ID + 1

      FrameConfig cfg = new FrameConfig(ComponentConfig.FRAME, null);
      cfg.setID(id0);
      cfg.setBox(false);
      cfg.setOverlay(true);
      cfg.setLabels(false);

      FillInConfig fcfg = new FillInConfig(ComponentConfig.FILL_IN, null);
      fcfg.setID(id1);
      fcfg.setFrameId(id0);
      fcfg.setAutoReturn(auto);
      fcfg.setDataType(var.getTypeName());

      if (format != null)
         fcfg.setFormat(format);

      screenCurrent.reconstructWidget(cfg);
      screenCurrent.reconstructWidget(fcfg);

      Frame frame = (Frame) screenCurrent.getComponent(id0);
      FillIn fin = (FillIn) screenCurrent.getComponent(id1);
      fin.setForceLeft(true);
      fin.setValue(var);
      frame.addWidget(fin);

      try
      {
         return fin.getText();
      }
      finally
      {
         screenCurrent.removeFrame(frame);
      }
   }

   /**
    * Request indexes of selected items in RadioSet, ComboBox or SelectionList
    * widgets.
    * 
    * @param   widgetId
    *          Widget ID.
    * 
    * @return  array of 0-based indexes of selected items or <code>null</code>
    *          if specified widget does not exists or does not support such a
    *          request.
    */
   public int[] getSelection(int widgetId)
   {
      Component widget = getWidget(widgetId);
      
      if (!(widget instanceof AttributeContainer))
         return null;
      
      if (((AttributeContainer) widget).getRef() instanceof ControlSetConfig)
         return null;
      
      ControlSetConfig config = 
                     (ControlSetConfig)((AttributeContainer) widget).getRef();
      
      return config.getSelection();
   }

   /**
    * Request HELP application execution on the server.
    */
   private void requestHelp()
   {
      // not initialized
      if (server == null)
      {
         return;
      }
      
      // recursive help
      if (helpActive)
      {
         tk.beep();
         return;
      }
      
      // flag recursion
      helpActive = true;
      
      // the following code arranges the safe help application invocation
      // in a way that is similar to triggers
      
      Component inFocus = getCurrentFocus();

      // save state variables
      Frame focusedFrame  = (Frame) locateFrame(inFocus);
      EventList save = currentEventList;
      
      String status = Window.getStatusText();

      boolean needPause = false;
      boolean canHide   = false;
      boolean inScope   = false;
      
      boolean curvis = true;
      
      curvis = tk.setCursorStatus(true);

      if (focusedFrame != null)
      {
         needPause = focusedFrame.isNeedPause();
         canHide   = focusedFrame.isCanHide();
         inScope   = focusedFrame.isAlive();
      }

      // save visibility state of all existing frames
      int[] frames = screenCurrent.getFrameIDs();
      boolean[] framesV = new boolean[frames.length];
      
      for (int i = 0; i < frames.length; i ++)
      {
         Frame f = getFrame(frames[i]);
         framesV[i] = f.isVisible();
         f.setProtected(true);
      }
      
      // save message lines
      String[] messages = Window.getMessageText();
      int[] colors = Window.getMessageColor();
      
      // clear messages and leave only the message input text
      if (tinyInputLine >= 0)
      {
         int nlines = messages.length;
         
         for (int i = 0; i < nlines; i ++)
         {
            if (i != tinyInputLine)
            {
               messages[i] = null;
               colors[i] = -1;
            }
         }
      }

      // call the help application
      int saveLevel = tk.saveInvalidate();

      int val = triggerNestingByWaitFor.pop();
      triggerNestingByWaitFor.push(val + 1);

      triggerNesting++;

      server.help(lastKey);

      triggerNesting--;

      val = triggerNestingByWaitFor.pop();
      triggerNestingByWaitFor.push(val - 1);

      if (triggerNesting == 0)
         autoCleared = false;
      
      tk.restoreInvalidate(saveLevel);

      int[] framesNow = screenCurrent.getFrameIDs();

      // restore cursor visibility
      tk.setCursorStatus(curvis);

      // mark all HELP related frames for hiding
      int[] fresh = Utils.diffIntArrays(framesNow, frames);
      
      for (int i = 0; i < fresh.length; i ++)
      {
         Frame x = getFrame(fresh[i]);
         x.setCanHide(true);
         x.setAlive(false);
      }
      
      // restore existing frames visibility
      tk.setInvalidate(true);
      
      for (int i = 0; i < frames.length; i ++)
      {
         Frame f = getFrame(frames[i]);

         f.setProtected(false);
         
         if (f.isVisible() == false && framesV[i] == true)
            viewWorker(f, null, null, true);
      }
      
      // restore message lines
      Window.setMessageText(messages, colors);
      
      Frame currFrame = (Frame) locateFrame(getCurrentFocus());
      
      // restore state variables
      if (focusedFrame != null)
      {
         if (focusedFrame.isVisible())
         {
            if(currFrame == null || 
               !((FrameConfig) currFrame.getRef()).isOverlay())
            {
               placeFrameInt(focusedFrame, true);
            }
         }

         focusedFrame.setNeedPause(needPause); 
         focusedFrame.setCanHide(canHide); 
         focusedFrame.setAlive(inScope); 
      }
      
      currentEventList = save;
      
      // restore the focus
      if (inFocus != getCurrentFocus() && inFocus != null)
         inFocus.requestFocus();
      
      tk.setInvalidate(false);

      // restore the status line
      Window.status(status);
      
      // unflag recursion
      helpActive = false;
   }
   
   /**
    * Reports if the terminal is currently redirected to a stream.
    *
    * @return   <code>true</code> if the terminal is redirected to a stream,
    *           <code>false</code> if the terminal is interactive.
    */
   public boolean isRedirected()
   {
      return (currentStream != -1);
   }
   
   /**
    * Reports if the terminal is currently redirected to the specified stream.
    *
    * @param    s
    *           a stream to to check the redirection for
    * @return   <code>true</code> if the terminal is redirected to the stream,
    *           <code>false</code> if the terminal is interactive.or 
    *           redirected to a different stream
    */
   public boolean isRedirected(Stream s)
   {
      int sid = sd.idFromStream(s);

      return (sid != -1 && currentStream == sid);
   }

   /**
    * Get ID of the widget which has the input focus.
    *
    * @param   widget
    *          Event source.
    * @return  ID of the widget or <code>-1</code>
    */
   private int getWidgetId(Component widget)
   {
      int widgetId = -1;
      
      while (widget != null && widgetId == -1)
      {
         if (widget instanceof AttributeContainer)
            widgetId = ((AttributeContainer) widget).getRef().getID();
         
         widget = widget.getParent();
      }
      
      return widgetId;
   }
   
   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and close/
    * cleanup resources for the redirected terminal associated with
    * <code>oldId</code> (unless <code>oldId</code> is -1 in which case 
    * nothing needs to be closed).
    *
    * @param   oldId
    *          The previously redirected stream for which the redirected
    *          terminal is to be closed or -1 for no close/cleanup action.
    * @param   newId
    *          The stream for which a redirected terminal is to be setup for
    *          redirection or -1 to remove redirections and activate the
    *          interactive terminal.
    */
   public void redirectOutput(int oldId, int newId)
   {
      if (newId == oldId)
      {
         return;
      }
      
      // flush stream if switching from non-terminal stream to terminal
      if(oldId != -1 && newId == -1)
      {
         Stream out = sd.streamFromId(oldId);
         
         if (out != null)
            flushStream(out);
      }
      
      int old = switchWorker(newId);
      
      // all we need to do is delete it from the map
      if (oldId != -1)
      {
         if (oldId != old)
         {
            throw new IllegalArgumentException("Attempt to close non-active" +
                                               " terminal redirection.");
         }
         
         redirections.remove(new Integer(oldId));
      }      
   }
   
   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and return
    * the id of the current output destination.  This may be another
    * redirected terminal or it may be -1 if the interactive terminal is in 
    * use.
    * <p>
    * If redirection has not yet been established for the given stream, it
    * will be setup.
    *
    * @param    newId
    *           The stream for which a redirected terminal is to be setup for
    *           redirection or -1 to remove redirections and activate the
    *           interactive terminal.
    * @param    clean
    *           <code>true</code> if this stream will need deferred cleanup.
    *
    * @return   The id of the previous terminal or -1 if the previous 
    *           terminal was interactive.   
    */
   public int switchOutput(int newId, boolean clean)
   {
      return switchOutput(newId, clean, false);
   }
   
   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and return
    * the id of the current output destination.  This may be another
    * redirected terminal or it may be -1 if the interactive terminal is in 
    * use.
    * <p>
    * If redirection has not yet been established for the given stream, it
    * will be setup.
    *
    * @param    newId
    *           The stream for which a redirected terminal is to be setup for
    *           redirection or -1 to remove redirections and activate the
    *           interactive terminal.
    * @param    clean
    *           <code>true</code> if this stream will need deferred cleanup.
    * @param    endOut
    *           <code>true</code> to force any pending PUT or EXPORT output
    *           to be ended for the given stream (so long as that stream is
    *           not the terminal).
    *
    * @return   The id of the previous terminal or -1 if the previous 
    *           terminal was interactive.   
    */
   public int switchOutput(int newId, boolean clean, boolean endOut)
   {
      if (endOut)
         endOutput(newId);
      
      if (newId != -1 && clean)
      {
         // remember that this stream may need deferred cleanup
         deferredCleanup.add(newId);
      }
      
      return switchWorker(newId);
   }
   
   /**
    * Called when a stream is closed to ensure that any terminal resources
    * associated with that id are cleaned up.  While this will be called
    * for any output stream close, actual cleanup processing is only 
    * triggered when the given stream id was used at least once via the
    * {@link #switchOutput} interface.
    * <p>
    * This should never be called when the redirection is currently active.
    *
    * @param    id
    *           The stream needing cleanup.
    */
   public void cleanupPseudoTerminal(int id)
   {
      Integer key = new Integer(id);
      
      if (deferredCleanup.remove(key))
      {
         if (id == currentStream)
         {
            throw new IllegalArgumentException("Attempt to close currently " +
                                               "active terminal redirection.");
         }
         // we need cleanup for this id
         redirections.remove(key);
      }
   }   
   
   /**
    * Obtains the next widget in the frame's tab order from the client.
    *
    * @param    frameId
    *           frame where the widget is placed
    *
    * @param    widgetId
    *           target widget which is being queried
    *
    * @return   id of the widget or<code>-1</code>.
    */
   public int nextTabItem(int frameId, int widgetId)
   {
      Component widget = getWidget(widgetId);
      if (widget == null)
      {
         return -1;
      }
      
      Component[] list = widget.getParent().getComponents();
      
      for (int i = 0; i < list.length; i ++)
      {
         if (getWidgetId(list[i]) == widgetId)
         {
            if (i == list.length - 1)
            {
               return -1;
            }
            else
            {
               // TO-DO: doesn't it need a check for isFocusTraversable()?
               return getWidgetId(list[i + 1]);
            }
         }
      }
      
      return -1;
   }

   /**
    * Hide WINDOW widget.
    */
   public void hideWindow()
   {
      String[] content = Window.getMessageText();
		
		Component focus = getCurrentFocus();
            
      hideAll(true);

      if (focus != null && triggerNesting > 0)
      {
         Frame frame = (Frame) locateFrame(focus);

         if (frame != null)
            frame.setFocusReset(true);
      }

      if (content[0] == null && content[1] == null)
         return;

      // message lines were not empty, so simulate hiding of message lines
      // instead of wiping out their content
      if (content[0] != null)
         content[0] = "";
      if (content[1] != null)
         content[1] = "";

      Window.setMessageText(content, null);
   }

   /**
    * Reset buffer size manager to its initial state.
    */
   public void resetBufferSizeManager()
   {
      BufferSizeManager.getInstance().reset();
   }

   /**
    * Looks up the first focusable widget of the frame.
    * 
    * @param   frame
    *          frame to query
    *          
    * @return  widget ID of the first focusable widget or <code>-1</code>
    *          if no focusable widget is found
    */
   public static int getFirstFocusableId(Frame frame)
   {
      // get list of all widgets in this frame
      Component[] list  = frame.getContentPane().getComponents();
      
      // scan the list
      for (int i = 0; i < list.length; i++)
      {
         if (!(list[i] instanceof AttributeContainer))
         {
            continue;
         }

         if (!list[i].isEnabled())
         {
            continue;
         }

         if (!list[i].isFocusTraversable())
         {
            continue;
         }

         return ((AttributeContainer)list[i]).getRef().getID();
      }

      return -1;
   }

   /**
    * Initiates the focus change procedure by creating and handling a LEAVE
    * event and, if that event is not consumed by a trigger, proceeds with
    * the focus change.
    * 
    * @param   widget
    *          <code>Component</code> that is the candidate to receive focus.
    *          
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean tryFocusChange(Component widget)
   {
      return tryFocusChange(widget, true, true);
   }

   /**
    * Assign VALUE CHANGED event.
    * 
    * @param   input
    *          The event to assign.
    */
   public static void setValueChangedEvent(KeyInput input)
   {
      getInstance().postponedValueChanged  = input;
   }
   
   /**
    * Initiates the focus change procedure by creating and handling a LEAVE
    * event and, if that event is not consumed by a trigger, proceeds with
    * the focus change.
    * <p>
    * Conditionally issues ENTRY event that follows the focus change.
    * 
    * @param   widget
    *          <code>Component</code> that is the candidate to receive focus.
    *          
    * @param   doLeave
    *          <code>true</code> means LEAVE event will be issued
    *          
    * @param   doEntry
    *          <code>true</code> means ENTRY event will be issued
    *          
    * @return  <code>true</code> if focus change is successful.
    */
   private static boolean tryFocusChange(Component widget, boolean doLeave,
                                         boolean doEntry)
   {
      // safety check
      if (widget == null)
         return false;

      Component inFocus = getCurrentFocus();
      
      if (inFocus == widget)
         return false;
      
      if (inFocus != null && inFocus.isEnabled() && doLeave)
      {
         // create and process the LEAVE event first so that any trigger
         // defined for that event gets its chance to run
         KeyInput leave = new KeyInput(Keyboard.SE_LEAVE,
                                       KeyEvent.KEY_PRESSED,
                                       inFocus);

         ThinClient.getInstance().processProgressEvent(leave);
   
         // check to see if a trigger returned no-apply
         if (((KeyEvent) leave).isConsumed())
            return false;
      }

      // if there is no ENTRY sent, focus must not be adjusted
      if (!doEntry)
         return true;

      // change the focus temporarily now
      widget.requestFocus();
      
      // create and process the ENTRY event now
      KeyInput entry = new KeyInput(Keyboard.SE_ENTRY,
                                    KeyEvent.KEY_PRESSED,
                                    widget);
      ThinClient.getInstance().processProgressEvent(entry);

      // check to see if a trigger returned no-apply
      if (!((KeyEvent) entry).isConsumed())
         return true;

      // reverse the focus change
      if (inFocus != null)
         inFocus.requestFocus();
      else
         widget.getParent().setFocus(null);

      return false;
   }

   /**
    * Performs conditional focus change procedure to the next widget.
    * The focus change can be cancelled by a trigger.
    * 
    * @param   widget
    *          <code>Component</code> that is the reference widget, which is
    *          currently retaining focus.
    *          
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean nextFocus(Component widget)
   {
      return focusChange(widget, true);
   }

   /**
    * Performs conditional focus change procedure to the previous widget.
    * The focus change can be cancelled by a trigger.
    * 
    * @param   widget
    *          <code>Component</code> that is the reference widget, which is
    *          currently retaining focus.
    *          
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean previousFocus(Component widget)
   {
      return focusChange(widget, false);
   }

   /**
    * Remove frame from the registry. Here removed all local references to the
    * frame and frame is registered in the list of frames which should be
    * removed at server side. Actual removal of frames at server side is
    * performed when processing of server request is done and result of
    * processing is returned to server.
    * 
    * @param   frame
    *          The frame to remove.
    */
   public static void removeFrame(Frame frame)
   {
      if (frame == null)
         return;
    
      if (frame.getRef() != null && getInstance().server != null)
         getInstance().removedFrames.add(new Integer(frame.getRef().getID()));
      
      Window.getDefaultWindow().removeFrame(frame);
      getInstance().screenCurrent.removeFrame(frame);
      
      if (getInstance().lastExposedFrame == frame)
         getInstance().lastExposedFrame = null;
   }

   /**
    * Remove frame from the registry. Here removed all local references to the
    * frame and frame is registered in the list of frames which should be
    * removed at server side. Actual removal of frames at server side is
    * performed when processing of server request is done and result of
    * processing is returned to server.
    * 
    * @param   frame
    *          The frame to remove.
    */
   public static void wrapFrame(Frame frame)
   {
      if (frame == null)
         return;
    
      if (frame.getRef() != null && getInstance().server != null)
         getInstance().wrappedFrames.add(new Integer(frame.getRef().getID()));
   }
   
   /**
    * Performs conditional focus change procedure to the next or previous
    * widget in the tab order. The focus change can be cancelled by a
    * trigger.
    * 
    * @param   widget
    *          <code>Component</code> that is the reference widget, which is
    *          currently retaining focus.
    * @param   direction
    *          The direction of focus change. If this parameter is set to
    *          <code>true</code> then next in Tab order widget is selected.
    *          
    * @return  <code>true</code> if the focus change is successful.
    */
   private static boolean focusChange(Component widget, boolean direction)
   {
      if (widget == null)
         return false;
      
      Container parent = widget.getParent();
      
      // if there is no other focusable widget in the parent container, then
      // there won't be an actual focus change, just return before any events
      // are generated
      if (!parent.hasFocusable(widget))
         return false;

      // create and process the LEAVE event first so that any trigger
      // defined for that event gets its chance to run
      
      if (widget.isEnabled())
      {
         KeyInput leave = new KeyInput(Keyboard.SE_LEAVE,
                                       KeyEvent.KEY_PRESSED,
                                       widget);

         ThinClient.getInstance().processProgressEvent(leave);
         
         // check to see if the trigger returned no-apply
         if (((KeyEvent) leave).isConsumed())
            return false;
      }
      
      // change the focus temporarily now
      if (direction)
         parent.nextFocus();
      else
         parent.previousFocus();

      // this is the new widget that will receive focus
      ThinClient tc = getInstance();
      Component inFocus = getCurrentFocus();

      if (inFocus == widget && !widget.isFocusTraversable())
      {
         // if the current widget is left on focus and it is not focusable,
         // focus remains undefined. this is the case when focus LEAVE has
         // disabled current focus
         Window.getDefaultWindow().setFocus(null);
         return true;
      }

      // Text widgets grouping mode support - TAB/BACKTAB in this mode
      // switches groups, not individual FillIn widgets.
      if (inFocus instanceof FillIn)
      {
         FillIn fin = (FillIn) inFocus;
         
         fin = fin.getFirstInGroup();
         
         if (!fin.hasFocus())
            fin.requestFocus();
      }
      
      // restore the focus to the original widget
      tc.setFocusOn(widget);

      // create and process the ENTRY event now. the ENTRY event will be 
      // responsible for setting focus to the required widget, if it accepts
      // it
      KeyInput entry = new KeyInput(Keyboard.SE_ENTRY,
                                    KeyEvent.KEY_PRESSED,
                                    inFocus,
                                    true);
      tc.processProgressEvent(entry);

      // check to see if the event was consumed (due to no-apply or widget 
      // lost focusability - was disabled or hidden)
      if (!((KeyEvent) entry).isConsumed())
      {
         // pass event to widget
         inFocus.processKeyEvent(entry);
         return true;
      }
      
      // if a trigger returned no-apply, search and focus the next widget
      // in tab-order after it, without wrapping; this widget will NOT 
      // ignore no-apply
      if (tc.triggerResultNoApply.peek())
      {
         Frame frame = (Frame) locateFrame(inFocus);
         do 
         {
            inFocus = tc.getNextEnabledWidget(frame, inFocus, FOCUS_NOWRAP);
         }
         while (inFocus != null && !tc.sendEntry(inFocus));

         if (inFocus == null)
         {
            // in no widget found, initial widget has focus
            widget.requestFocus();
         }
      }
      // if the widget was disabled by the triggers, focus remains on 
      // initial focus
      else
      {
         // reverse the focus change
         widget.requestFocus();
      }

      return false;
   }
   
   /**
    * Resets the changed state for all widgets of the frame into unchanged.
    * 
    * @param   frameId
    *          ID of frame this request came for
    */
   private void resetChanged(int frameId)
   {
      getFrame(frameId).getFrameHelper().resetChanged();
   }
   
   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and return
    * the id of the current output destination.  This may be another
    * redirected terminal or it may be -1 if the interactive terminal is in 
    * use.
    * <p>
    * If redirection has not yet been established for the given stream, it
    * will be setup.
    *
    * @param    newId
    *           The stream for which a redirected terminal is to be setup for
    *           redirection or -1 to remove redirections and activate the
    *           interactive terminal.
    *
    * @return   The id of the previous terminal or -1 if the previous 
    *           terminal was interactive.   
    */
   private int switchWorker(int newId)
   {
      Integer key = new Integer(newId);
      
      if (newId == -1)
      {
         Window.enableSpecial();
         
         // switch to interactive mode
         tk.switchOutput(null);
         
         honorPBH = true;
      }
      else if (redirections.containsKey(key))
      {
         Window.disableSpecial();
         
         // pre-existing redirection target
         RedirectedTerminal rt = redirections.get(key);
         tk.switchOutput(rt);

         honorPBH = false;
      }
      else
      {
         Window.disableSpecial();
         
         // first time redirection
         Stream out = sd.streamFromId(newId);
         RedirectedTerminal rt = new RedirectedTerminal(out, 512, 128);
         
         redirections.put(key, rt);
         tk.switchOutput(rt);

         honorPBH = false;
      }            

      // remember the current stream id and return the old id
      int oldId = currentStream;
      currentStream = newId;

      return oldId;
   }
   
   /**
    * Switch the terminal into interactive mode if it is redirected. This is
    * just a convenience function.
    *
    * @return   The terminal id of the redirected terminal or -1 if the
    *           terminal is already interactive.  Used with 
    *           {@link #restoreRedirection} to restore the terminal when 
    *           done with interactive mode.
    */
   private int forceInteractive()
   {
      int tempTerm = -1;
      
      if (isRedirected())
      {
         // set up temporary interactive terminal
         tempTerm = switchOutput(-1, false);
      }
      
      return tempTerm;
   }
   
   /**
    * Restore the terminal to redirected mode from interactive mode, if the
    * terminal to switch to is not -1.  This is just a convenience function.
    *
    * @param    tempTerm
    *           The terminal id of the redirected terminal that was previously
    *           removed using {@link #forceInteractive} or -1 if there was no
    *           previous redirected terminal.
    */
   private void restoreRedirection(int tempTerm)
   {
      if (tempTerm != -1)
      {
         // take down temporary interactive terminal, restore redirection
         switchOutput(tempTerm, false);
      }                        
   }

   /**
    * Processes all outstanding events without blocking for user input.
    * <p>
    * This method corresponds to the Progress statement <br>
    * <b>PROCESS EVENTS </b>
    *
    * @return <code>true</code> if at least one frame related 
    *         <code>PaintEvent</code> has been processed
    */
   private boolean processEventsWorker()
   {
      AWTEvent event;
      
      boolean syncPending = false;
      boolean losingFocus = false;
      boolean framePaint  = false;
      boolean gainingFocus = false;
      
      while ((event = Toolkit.getEvent()) != null)
      {
         // synchronization service events
         if (event instanceof SyncEvent)
         {
            syncPending = true;
            continue;
         }
         
         Component src = (Component)event.getSource();

        // keyboard events
         if (event instanceof KeyEvent && currentEventList != null)
         {
            KeyInput ki = (KeyInput)event;

            processProgressEvent(ki);
            
            // block default processing for this event if 
            // the widget in focus is disabled
            // TO-DO: why not the event source???
            if (!ki.isConsumed())
            {
               Component inf = (Component) ki.getSource();
               Component infFrame = locateFrame(inf);

               // the source component may be the window, so null frame is
               // allowed
               if (inf != null &&
                   !(inf.isEnabled() &&
                     inf.isVisible() &&
                     (infFrame == null ||
                      infFrame.isVisible())))
               {
                  // Do not consume VALUE-CHANGED and ENTRY targeted to BROWSE,
                  // it receives them regardless from the state.
                  // Note that this has no side effects visible by application  
                  // since these events are used by browse internally.
                  
                  if (!(src instanceof Browse &&
                        (ki.getActionCode() == Keyboard.SE_VALUE_CHANGED ||
                         ki.getActionCode() == Keyboard.SE_ENTRY)))
                  {
                     // this consume doesn't mean the trigger returned 
                     // no-apply - it just stops further processing of the
                     // event
                     ki.consume();
                  }
               }
               
               if (ki.getActionCode() == Keyboard.SE_VALUE_CHANGED && 
                   !ki.isSpecial())
               {
                  Frame frame = (Frame) locateFrame(src);
                  
                  if (frame != null)
                     frame.setFocusAttractor(true);
               }
            }

            if (ki.getActionCode() == Keyboard.KA_GO && 
                !triggerResultNoApply.peek())
            {
               Component inf = (Component) ki.getSource();
               Component frame = locateFrame(inf);

               if (frame instanceof Frame && frame != ki.getSource())
               {
                  // the following widgets will trigger GO for frame, even if
                  // the event was consumed
                  boolean override = inf instanceof SelectionList     ||
                                     inf instanceof SelectionListBody ||
                                     inf instanceof RadioSet          ||
                                     inf instanceof Button            ||
                                     inf instanceof ComboBox          ||
                                     inf instanceof Editor            ||
                                     inf instanceof Browse;
                  
                  if (!ki.isConsumed() || override)
                  {
                     KeyEvent res = new KeyInput(ki.getKeyCode(), 
                                                 ki.getActionCode(), 
                                                 frame, 
                                                 true);
                     Toolkit.postEvent(res);
                  }
               }
            }

            // if this is a NEXT/PREV-FRAME event, generate a new event and
            // send it to the parent frame, if exists
            if (!triggerResultNoApply.peek() &&
                (ki.getActionCode() == Keyboard.SE_NEXT_FRAME ||
                 ki.getActionCode() == Keyboard.SE_PREV_FRAME))
            {
               Component inf = (Component) ki.getSource();
               Component frame = locateFrame(inf);

               if (frame instanceof Frame && frame != ki.getSource())
               {
                  KeyEvent res = new KeyInput(ki.getKeyCode(), 
                                              ki.getActionCode(), 
                                              frame, 
                                              true);
                  Toolkit.postEvent(res);
               }
            }

            // no default processing for consumed events
            if (ki.isConsumed())
            {
               continue;
            }
         }
         
         // generate CHOOSE event
         if (event instanceof ItemEvent)
         {
            ItemEvent evt = (ItemEvent) event;
            KeyEvent res = null;
            
            int state = evt.getStateChange();
            
            if (state == ItemEvent.SELECTED)
            {
               res = new KeyInput(Keyboard.SE_CHOOSE, KeyEvent.KEY_PRESSED, 
                                  (Component) evt.getSource());
               Toolkit.postEvent(res);
            }
         }

         // pre-processing of FOCUS_LOST and FOCUS_GAINED events
         if (event instanceof FocusEvent)
         {
            switch (event.getID())
            {
               case AWTEvent.FOCUS_LOST:
                  losingFocus = true;
                  break;
               case AWTEvent.FOCUS_GAINED:
                  gainingFocus = true;
                  break;
            }
         }
         
         // monitoring Paint events
         if (event instanceof PaintEvent)
         {
            if (src instanceof Frame)
               framePaint = true;
         }

         if (alertBoxActive == 0)
         {
            if (event instanceof KeyInput)
            {
               int actionCode = ((KeyInput) event).getActionCode();
               if (actionCode != Keyboard.SE_LEAVE &&
                   actionCode != Keyboard.SE_ENTRY)
               {
                  lastActionCode = actionCode;
               }
            }
            else if (!(event instanceof PaintEvent) &&
                     !(event instanceof FocusEvent))
            {
               lastActionCode = 0;
            }
         }

         // processing of all events by CHARVA
         try
         {
            tk.getTopWindow().processEvent(event);
         }
         
         catch (IllegalComponentStateException e)
         {
            // Do not print anything, since this is normal situation under
            // some circumstances. Removing dependency on CHARVA may make
            // this code redundant.
         }
         
         catch(ConditionException e)
         {
            // condition exceptions generated during event processing must
            // be handled
            throw e;
         }

         catch (Exception e)
         {
            e.printStackTrace(System.err);
            continue;
            // throw new RuntimeException("")
         }
         
         // post-processing of FOCUS_LOST event
         if (losingFocus)
         {
            losingFocus = false;
            
            // redraw the widget just lost the focus
            
            if (src.isVisible() && src.isValid())
            {
               src.setHighlighted(false);
               src.draw();
               src.requestSync();
            }
         }
         
         if (gainingFocus)
         {
            gainingFocus = false;
            if (src.isVisible() && src.isValid())
            {
               src.setHighlighted(true);
            }
         }

         if (immediateException != null)
         {
            throw immediateException;
         }
      }
      
      if (syncPending)
         tk.getTopWindow().processEvent(new SyncEvent(tk.getTopWindow()));
      
      return framePaint;
   }
   
   /**
    * Handle event as Progress event.
    * <p>
    * Precedence order for honoring Progress key events:
    * <ol>
    *    <li> In the accelerator list for the current window.
    *    <li> In the event list for an active user interface trigger.
    *    <li> An active pop-up menu key.
    *    <li> An active menu mnemonic key.
    *    <li> An active field-level (button) mnemonic key (Windows only).
    *    <li> A special Progress internal key (built-in key function, like 
    *         GO).
    *    <li> A key associated with a key function in the PROTERMCAP file
    *         (UNIX only).
    *    <li> A standard input key (A, B and so on).
    * </ol>
    * <p>
    * If any exit condition is found, the <code>isGoPending</code> flag is
    * set.
    * 
    * @param    evt
    *           Event to handle.
    */
   private void processProgressEvent(KeyInput evt)
   {
      Component src = (Component)evt.getSource();
      int key       = evt.getKeyCode();
      int action    = evt.getActionCode();

      if (currentEventList == null)
         return;

      int cop = currentEventList.getOperation();

      boolean foundCondExit = false;
      
      if(action == Keyboard.KA_STOP   ||
         action == Keyboard.KA_ENDKEY ||
         action == Keyboard.KA_ERROR  ||
         action == Keyboard.KA_END_ERROR)
      {
         condExit = true;
         foundCondExit = true;
      }
      
      if (authMode && key == KeyEvent.VK_CTRL_C)
      {
         condExit = true;
         throw new StopConditionException("STOP key function executed", 
                                          key);
      }
      
      TriggerResult tr = null;
      Component cur = getCurrentFocus();
      if (cur != null && !cur.isEnabled())
      {
         cur = null;
      }
      
      if (src instanceof Frame || src instanceof Window)
      {
         cur = src;
      }

      boolean foundGo = false;
      boolean exceptionalExit = false;

      // detect exit condition early 
      if (checkBreak(src, key, evt.isRealKey()) ||
          checkBreak(src, action, false))
      {
         isGoPending++;
         foundGo = true;
      }
      else if (foundCondExit)
      {
         exceptionalExit = true;
      }

      // low level Progress events:
      
      // in the accelerator list for the current window

      // implied LEAVE on exit
      if (isGoPending > 0 && !leaveSent && cur != null)
      {
         // clear pending GO temporarily for possible NO-APPLY and recursion
         int save = isGoPending;
         isGoPending = 0;

         // mark LEAVE as sent so there is no duplicate sent from WAIT-FOR
         leaveSent = true;

         // generate LEAVE and check if accepted, but only if the current
         // focus is valid
         if (getFocusState(cur) == FOCUS_VALID)
         { 
            if (!sendLeave(cur))
            {
               // suppress further processing of this event
               evt.consume();
               
               // if LEAVE was not accepted, then the flag must be reset to
               // allow LEAVE to be sent again
               leaveSent = false;
               return;
            }
            
            if (checkBreak(cur, Keyboard.SE_LEAVE, false))
            {
               isGoPending ++;
               return;
            }
         }

         // LEAVE accepted; resume exit processing
         isGoPending = save;
      }
      
      // in the event list for an active user interface trigger
      if (key > 0 && key != KeyEvent.CHAR_UNDEFINED)
      {
         tr = invokeTriggers(src, key, false);
         
         if (tr.consume)
         {
            evt.consume();

            if (foundGo)
            {
               isGoPending--;
               leaveSent = false;
            }
            return;
         }
      }
      
      // other high level events
      if ((tr == null || !tr.executed) && key != action &&
          action != Keyboard.SE_ENTRY)
      {
         tr = invokeTriggers(src, action, false);
         
         // let widget perform "hard" check if necessary
         if (action == Keyboard.SE_LEAVE && src instanceof FillIn)
         {
            src.processKeyEvent(evt);
            
            if (evt.isConsumed())
               tr.consume = true;
         }
         
         if (tr.consume)
         {
            evt.consume();

            if (foundGo)
            {
               isGoPending--;
               leaveSent = false;
            }
            return;
         }
      }
      
      // check for possible ANY-KEY
      if ((tr == null || !tr.executed) && evt.isRealKey())
      {
         tr = invokeTriggers(src, key, true);

         if (tr.consume)
         {
            evt.consume();

            if (foundGo)
            {
               isGoPending--;
               leaveSent = false;
            }
            return;
         }
      }
      
      // disable all following logic if this is a drop-down portion of a
      // combo-box
      if (src instanceof List)
      {
         return;
      }
      
      // an active pop-up menu key
      
      // an active menu mnemonic key
      
      // an active field-level (button) mnemonic key (Windows only)
      
      // translate not consumed CHOOSE to ENDKEY/GO for some buttons 
      if (action == Keyboard.SE_CHOOSE &&  src instanceof Button)
      {
         ButtonConfig btc = (ButtonConfig) ((Button) src).getRef();
         
         int code = -1;
         
         if (btc.autoEndKey)
            code = Keyboard.KA_ENDKEY;
         
         if (btc.autoGo)
            code = Keyboard.KA_GO;
         
         if (code != -1)
         {
            KeyInput ki;
            
            ki = new KeyInput(code, KeyEvent.KEY_PRESSED, locateFrame(src));
            Toolkit.postEvent(ki);
            return;
         }
      }
      
      // BELL action
      if (action == Keyboard.KA_BELL)
      {
         tk.beep();
      }

      // Internal browse events (currently just END-ERROR under certain
      // circumstances)
      if (processBrowseEvent(evt))
      {
         return;
      }
      
      // LEAVE event: do widget validation if defined
      if (action == Keyboard.SE_LEAVE)
      {
         if (src instanceof DataContainer)
         {
            DataContainer dc = (DataContainer) src;
            
            if (dc.isModified() && !validateWidget(src))
            {
               evt.consume();
               return;
            }
         }

         // remove the highlight from the widget
         src.setHighlighted(false);
      }
      
      // ENTRY event: do widget validation if defined
      if (action == Keyboard.SE_ENTRY)
      {
         Component inFocus = getCurrentFocus();
         boolean fire = inFocus != src && !(src instanceof Frame);
         Frame frame = (Frame) locateFrame(src);
         
         boolean pushed = false;
         boolean failed = false;

         try
         {
            if (fire)
            {
               if (!src.isVisible() ||
                   !src.isEnabled() ||
                   (frame != null &&
                    !frame.isVisible()))
               {
                  // reject entry, source is not valid to receive such an event
                  evt.consume();
                  return;
               }
               
               // since the attempted focus change produces related CHARVA events
               // we have to shield ourselves with another event queue
               push();
               
               // force pop()/popTrash() cleanup in finally block
               pushed = true;
   
               if (!tryFocusChange(src, !evt.isSpecial(), false))
               {
                  // a trigger blocked LEAVE event
                  // trash the focus change events and rebalance the queue
                  failed = true;
   
                  // no go
                  evt.consume();
                  return;
               }
   
               if (!evt.isSpecial())
               {
                  // this must be done only if we are in an APPLY call
                  // if focus and target frame are different, send ENTRY to
                  // frame first
                  if (locateFrame(inFocus) != frame)
                  {
                     // if NO-APPLY, focus is undefined and no trigger is
                     // invoked
                     if (sendEntry(frame) == false)
                     {
                        // TODO: we use popTrash() here but is that correct? 
                        failed = true;
                        
                        Window.getDefaultWindow().setFocus(null);
                        return;
                     }
                  }
               }
            }
            
            // delayed trigger processing for the ENTRY event
            tr = invokeTriggers(src, action, false);
            
            if (tr.consume)
            {
               if (fire)
               {
                  // disable deferred cleanup because we will cleanup here
                  pushed = false;
                  popTrash();
               }
   
               // a trigger for ENTRY returned NO-APPLY: must adjust focus to 
               // another widget in frame
               evt.consume();
   
               if (evt.isSpecial() == false)
               {
                  // this must be done only if we are in an APPLY call
                  Component next = getNextEnabledWidget(frame, src, FOCUS_NOWRAP);
                  if (next != null)
                  {
                     applyWorker(new KeyInput(evt.isRealKey() ? key : action,
                                              evt.getID(),
                                              next,
                                              evt.isSpecial(),
                                              evt.isRealKey()));
                  }
               }
   
               return;
            }
   
            // allow the focus change events and rebalance the queue
            if (fire)
            {
               if (triggerNesting == 0)
               {
                  // save target widget, so it can be used in subsequent root
                  // level wait-for loops
                  pendingFocus = src;
               }
               else
               {
                  pendingFocus = null;
               }
   
               // ENTRY was accepted, set focus to source only if the widget
               // is still enabled, visible and within a nested trigger
               if (src.isEnabled() &&
                   src.isVisible() &&
                   (frame == null ||
                    frame.isVisible()))
               {
                  if (evt.isSpecial() || triggerNesting > 0)
                  {
                     src.requestFocus();
   
                     // if a CHOOSE is active, notify the frame
                     if (frame != null && frame.isChooseActive())
                     {
                        // CHOOSE will have no highlight if src is not part of it
                        frame.setChooseAt(src);
                     }
                  }
               }
               else
               {
                  // if focus was accepted and source is disabled or hidden, 
                  // focus must not be changed -> but this doesn't mean the
                  // ENTRY was not accepted
                  return;
               }
            }
            else
            {
               // highlight the widget
               if (inputActive > 0)
               {
                  src.setHighlighted(true);
               }
            }
         }
         
         finally
         {
            if (pushed)
            {
               if (failed)
               {
                  popTrash();
               }
               else
               {
                  pop();
               }
            }
         }
      }

      // check additional exit conditions
      if (isGoPending == 0 && cop != EventList.WAIT_FOR)
      {
         if (currentEventList.isGoEvent(Keyboard.eventName(key)) ||
             currentEventList.isGoEvent(Keyboard.eventName(action)))
             isGoPending ++;
      }
      
      // GO action
      if ((isGoPending > 0 || action == Keyboard.KA_GO) && cur != null && 
         cop != EventList.WAIT_FOR)
      {
         Frame frame = null;

         if (src instanceof Frame)
         {
            frame = (Frame)src;
         }
         else
         {
            frame = (Frame)locateFrame(src);
         }

         // validate all validateable widgets before accepting GO
         Component[] list = null;

         if (frame == null)
         {
            list = new Component[0];
         }
         else
         {
            Container container = frame.getFrameHelper().getContentPane();
            list = container.getComponents();
         }
         
         for (int i = 0; i < list.length; i++)
         {
            // no need to validate headers
            if ((list[i] instanceof AttributeContainer) && 
                ((AttributeContainer) list[i]).getRef().isHeader())
            {
               continue;
            }
            
            // only certain cases get validated
            if (list[i] instanceof DataContainer && list[i].isEnabled())
            {
               // validate
               if (!validateWidget(list[i]))
               {
                  evt.consume();
                  isGoPending --;
                  return;
               }
            }
         }
         
         isGoPending ++;
      }                        

      // exit condition for current-window
      if (isGoPending == 0 && cop == EventList.WAIT_FOR)
      {
         src = Window.getDefaultWindow();
         
         if (checkBreak(src, key, evt.isRealKey()) || 
             checkBreak(src, action, false))
         {
            isGoPending ++;
         }
      }

      // check if we are exiting a choose
      if (isGoPending > 0 && cop == EventList.CHOOSE)
      {
         // choose does no further processing once a GO-ON has been reached 
         evt.consume();
         return;
      }

      // a key associated with a key function in the PROTERMCAP file
      // (UNIX only)

      // a standard input key (A, B, and so on)

      // HELP action
      if (isGoPending == 0 && action == Keyboard.KA_HELP)
      {
         requestHelp();
      }

      // STOP, ENDKEY and END-ERROR actions
      if (isGoPending == 0 || exceptionalExit)
      {
         processCondition(action, cop, false);
      }
   }
   
   /**
    * Give a browse widget, if it is in this event's component hierarchy, a
    * chance to process and consume the END-ERROR event.  This is only
    * necessary if the browse widget is in edit mode and is currently set to
    * handle an internal, row-level undo.  Otherwise the event is not handled
    * or consumed.
    * 
    * @param   evt
    *          Current event.
    *          
    * @return  <code>true</code> if the event was an END-ERROR event and was
    *          handled and consumed by a browse widget in the event's
    *          component hierarchy;  else <code>false</code>.
    */
   private boolean processBrowseEvent(KeyInput evt)
   {
      if (evt.getActionCode() == Keyboard.KA_END_ERROR)
      {
         // Give containing browse a chance to consume END-ERROR.
         Component comp = (Component) evt.getSource();
         do
         {
            if (comp instanceof Browse)
            {
               ((Browse) comp).keyPressed(evt);
               if (evt.isConsumed())
               {
                  return true;
               }
               
               break;
            }
            comp = comp.getParent();
         } while (comp != null);
      }
      
      return false;
   }

   /**
    * Check the {@link #rowDisplayEvents} map for a ROW-DISPLAY trigger
    * associated with this widget. If this is a browse widget and such a
    * trigger is found, it will be executed.
    *
    * @param   widgetId
    *          event source.
    *
    * @return  Execution and event consume flags for an executed trigger.
    */
   TriggerResult invokeRowDisplayTrigger(int widgetId)
   {
      TriggerResult tr = new TriggerResult();
      if (rowDisplayEvents == null)
      {
         return tr;
      }

      Component widget = getWidget(widgetId);
      // allow only BROWSE widgets to trigger ROW-DISPLAY events
      if (widget instanceof Browse == false)
      {
         return tr;
      }

      // search for triggers
      int triggerId = -1;

      if (rowDisplayEvents.containsKey(widgetId))
      {
         triggerId = rowDisplayEvents.get(widgetId);
      }
      else
      {
         // search for ANYWHERE trigger
         if (rowDisplayEvents.containsKey(-1))
         {
            triggerId = rowDisplayEvents.get(-1);
         }
      }

      // activate trigger
      if (triggerId != -1)
      {
         // obtain its home frame ID
         int frameId  = -1;
         Component frame = locateFrame(widget);
         int eventCode = Keyboard.eventCode(Keyboard.ROW_DISPLAY);

         if (frame != null && frame instanceof AttributeContainer)
            frameId = ((AttributeContainer)frame).getRef().getID();

         tr.executed = true;

         // save state variables
         EventList save = currentEventList;
         boolean state = inRowDisplayTrigger;

         inRowDisplayTrigger = true;

         // call the trigger
         try
         {
            tr.consume = trigger(triggerId,
                                 eventCode,
                                 widgetId,
                                 getScreenBuffer(frameId, -1));
         }

         finally
         {
            // restore state variables
            currentEventList = save;
            inRowDisplayTrigger = state;
         }
      }

      return tr;
   }
   
   /**
    * Check active <code>EventList</code> instance for triggers defined
    * for this event and widget pair and activate the first one found.
    * <p>
    * Triggers are searched either for a specific widget, or, if no widget is
    * enabled (in focus), for the session window. The latter is also performed
    * if the specific widget happens to be the session window.
    * <p>
    * The specific search is done in the following order:
    * <ul>
    *  <li>trigger for the specified widget and event;
    *  <li>trigger for the specified event anywhere within the frame;
    *  <li>trigger for the specified event anywhere globally.
    * </ul>
    * <p>
    * The search for the session window is a one step process.
    * <p>
    * The first found trigger, if any, gets activated.
    * 
    * @param   widget
    *          event source.
    * @param   evtCode
    *          event code.
    * @param   isKey
    *          This flags marks reals keys typed by user.
    *          
    * @return  Execution and event consume flags for an executed trigger.
    */
   private TriggerResult invokeTriggers(Component widget, 
                                        int       evtCode, 
                                        boolean   isKey)
   {
      TriggerResult tr = new TriggerResult();
      
      // no wait list, this is a combo drop-down list widget or this is an
      // APPLY with implicit target, and no focus exists - nothing to do
      if (currentEventList == null ||
          widget instanceof List   ||
          implicitApplyNoTrigger)
         return tr;

      // obtain the widget ID and its home frame ID
      int frameId  = -1;
      int widgetId = getWidgetId(widget);
      Component frame = locateFrame(widget);

      // Progress watches the focus during event processing and restores it
      // whenever triggers move it away but then allow the default event
      // processing to occur. I assume here that it should only happen if
      // the event source is also the widget in focus, although it may leave
      // other cases out

      final Component inFocus = getCurrentFocus();

      // use the session window if no specific widget was specified AND there
      // is nothing in focus
      if (widgetId == -1 && inFocus == null)
      {
         widgetId = WindowConfig.SESSION_WINDOW_ID;
      }
      else
      {
         // otherwise try to initialize the frame ID
         if (frame != null && frame instanceof AttributeContainer)
            frameId = ((AttributeContainer)frame).getRef().getID();
      }

      // do the search for triggers
      int triggerId = currentEventList.lookup(frameId, 
                                              widgetId,
                                              evtCode,
                                              true, 
                                              isKey);

      // activate trigger
      if (triggerId != -1)
      {
         tr.executed = true;

         // save state variables
         EventList save = currentEventList;

         // call the trigger
         try
         {
            triggerResultNoApply.push(false);
            triggerHadWaitFor.push(false);

            tr.consume = trigger(triggerId,
                                 evtCode,
                                 widgetId,
                                 getScreenBuffer(frameId, -1));
         }

         finally
         {
            triggerResultNoApply.pop();
            boolean hadWaitFor = triggerHadWaitFor.pop();

            // save last result, on this level 
            triggerResultNoApply.pop();
            triggerResultNoApply.push(tr.consume);

            triggerHadWaitFor.pop();
            triggerHadWaitFor.push(hadWaitFor);

            // restore state variables
            currentEventList = save;
                    
            Component focusNow = getCurrentFocus();

            // restore focus under certain circumstances
            // the exact circumstances are to be determined (work in progress)
            // - event is not consumed
            // - event is not VALUE-CHANGED
            // - widget is FILL-IN
            if (!tr.consume              && 
                focusNow != inFocus      && 
                inFocus != null          && 
                widget == inFocus        &&
                widget instanceof FillIn &&
                evtCode != Keyboard.SE_VALUE_CHANGED)
            {
               setFocusSilent(inFocus);
            }
            else if (tr.consume          &&
                     focusNow != inFocus &&
                     inFocus != null     && 
                     widget == inFocus   &&
                     widget instanceof Button)
            {
               setFocusSilent(inFocus);
               focusNow = inFocus;
            }
            else if (frame != null && ((Frame) frame).isFocusReset())
            {
               focusNow = getFirstEnabledWidget((Frame) frame);
               
               if (focusNow != null)
                  setFocusSilent(focusNow);
            }
            
            if (focusNow != null && focusNow == inFocus)
            {
               checkHelpText(focusNow);
            }
         }
      }
      else
      {
         // no trigger, reset the result to false 
         triggerResultNoApply.pop();
         triggerResultNoApply.push(false);
         triggerHadWaitFor.pop();
         triggerHadWaitFor.push(false);
      }
      return tr;
   }

   /**
    * Silently restore and redraw the focus. No LEAVE/ENTRY events should be
    * generated here.
    * 
    * @param    inFocus
    *           Widget to focus.
    */
   private void setFocusSilent(final Component inFocus)
   {
      eventBracket(new Runnable()
      {
         public void run()
         {
            inFocus.requestFocus();
            
            if (inFocus instanceof FillIn)
            {
               ((FillIn) inFocus).enableZapMode();
            }
         }
      });
   }
   
   /**
    * Check help text and restore it if necessary.
    * 
    * @param    focus
    *           Widget in focus.
    */
   private void checkHelpText(Component focus)
   {
      if (!(focus instanceof AttributeContainer))
         return;

      ComponentConfig cfg = ((AttributeContainer) focus).getRef();
      String help = cfg.getHelp();
      String status = Window.getStatusText();

      if (help == null || help.equals(status))
         return;
      
      setHelp(help);
   }

   /**
    * Check active <code>EventList</code> instance for exit conditions
    * for this event and widget pair.
    * <p>
    * Events are searched either for a specific widget, or, if no widget is
    * enabled (in focus), for the session window. The latter is also performed
    * if the specific widget happens to be the session window.
    * <p>
    * The specific search is done in the following order:
    * <ul>
    *  <li>exact match for the specified widget and event;
    *  <li>match for the specified event anywhere within the frame;
    *  <li>match for the specified event anywhere globally.
    * </ul>
    * <p>
    * The search for the session window is a one step process.
    * 
    * @param   widget
    *          event source.
    * @param   evtCode
    *          event code.
    * @param   isKey
    *          Mark real keys typed by user.
    *          
    * @return  <code>true</code> if a matching event exists
    */
   private boolean checkBreak(Component widget, int evtCode, boolean isKey)
   {
      // no wait list; nothing to do
      if (currentEventList == null)
         return false;
      
      // obtain the widget ID and its home frame ID
      int frameId  = -1;
      int widgetId = getWidgetId(widget);
      Component frame = locateFrame(widget);
      Component inFocus = getCurrentFocus();

      // use the session window if no specific widget was specified AND there
      // is nothing in focus
      if (widgetId == -1 && inFocus == null)
      {
         widgetId = WindowConfig.SESSION_WINDOW_ID;
      }
      else
      {
         // otherwise try to initialize the frame ID
         if (frame != null && frame instanceof AttributeContainer)
            frameId = ((AttributeContainer)frame).getRef().getID();
      }
      
      // do the search for exit condition
      int eventId = currentEventList.lookup(frameId, widgetId, evtCode,
                                            false, isKey);

      return eventId >= 0;
   }

   /**
    * Handles processing of ENDKEY, END-ERROR and STOP conditions in a
    * uniform manner.  The given action is checked for a match with the
    * list of conditions, if a match exists then the associated condition
    * is raised (an exception is thrown).
    *
    * @param    action
    *           The key action to check against the list of possible
    *           conditions.
    * @param    opcode
    *           Specifies which operation is in process (for example, a
    *           {@link EventList#WAIT_FOR}).
    * @param    inPause
    *           Specifies whether the method is called from a pausing 
    *           function.
    *
    * @throws   StopConditionException
    *           If action is {@link Keyboard#KA_STOP}.
    * @throws   EndConditionException
    *           If action is {@link Keyboard#KA_ENDKEY} or on
    *           {@link Keyboard#KA_END_ERROR} on the first interaction in the
    *           current block or when opcode is 
    *           {@link EventList#WAIT_FOR}.
    * @throws   ErrorConditionException
    *           If action is {@link Keyboard#KA_END_ERROR} and one of the
    *           above conditions does not hold.
    */
   private void processCondition(int action, int opcode, boolean inPause)
   throws StopConditionException,
          EndConditionException,
          ErrorConditionException
   {
      // Note that the opcode == EventList.UPDATE, pause(), message(),
      // choose() and other cases on the server need to know if the cause of
      // a condition was a specific key (e.g. END-ERROR).  For this reason
      // there is an extra flag passed to the constructor for all of these
      // cases.

      boolean store = (lastAppliedCode == Keyboard.KA_STOP ||
                       lastAppliedCode == Keyboard.KA_END_ERROR ||
                       inPause &&
                       triggerNestingByWaitFor.peek() > 1) &&
                      currentEventList != null &&
                      currentEventList.getOperation() == EventList.WAIT_FOR &&
                      !inEditingBlock;

      // STOP action
      if (action == Keyboard.KA_STOP)
      {
         ConditionException exception =
            new StopConditionException("STOP key function executed", lastKey);

         if (store)
            immediateException = exception;
         throw exception;
      }
      
      // ENDKEY action
      if (action == Keyboard.KA_ENDKEY)
         throw new EndConditionException("END key function executed",
                                         lastKey);
      
      // ERROR action
      if (action == Keyboard.KA_ERROR)
         throw new ErrorConditionException("ERROR key function executed",
                                           lastKey);
      
      // END-ERROR action
      if (action == Keyboard.KA_END_ERROR)
      {
         int effective = interactions;
         
         // pause has a special behavior in that it does not change the
         // interactions counter BUT if it executes when the interactions
         // counter is 1 (but not when it is 0), then an END-ERROR will be
         // translated into ERROR (as if the real interactions counter was
         // 2 or more)
         if (inPause && effective == 1)
         {
            effective = 2;
         }
         
         // no interactions in this block iteration, 1st interaction in this
         // block iteration or a wait-for both always get an ENDKEY instead
         // of an error
         if (effective < 2 ||
             (opcode == EventList.WAIT_FOR && !inEditingBlock))
         {
            ConditionException exception =
               new EndConditionException(END_ERR_MSG, lastKey);

            if (store)
               immediateException = exception;
            throw exception;
         }
         
         // the opcode == EventList.UPDATE on the server needs even one
         // more flag... to note that this was an END-ERROR special case
         ConditionException exception =
            new ErrorConditionException(END_ERR_MSG, lastKey, true);

         if (store)
            immediateException = exception;
         throw exception;
      }
   }
   
   /**
    * Validates the widget's current value and displays the validation error
    * message if any.
    * 
    * @param   widget
    *          widget to validate
    * @return  The result of the validation.
    */
   private boolean validateWidget(Component widget)
   {
      int wid = getWidgetId(widget);
      if (currentEventList.isValidatable(wid))
      {
         // hide previously displayed error message, if any
         if (validationError != null)
            hideMessage(true);
         
         // call the validation expression evaluator
         if (!validate(wid, getScreenBuffer(widget)))
         {
            // validation failed: display message, beep and stay in
            message(validationError);
            
            // what widget is in focus right now?
            Component comp = getCurrentFocus();
            
            if (widget == comp)
            {
               // enable auto-zap and reset insert mode for fill-ins
               if (widget instanceof FillIn)
               {
                  FillIn fin = (FillIn) widget;
                  fin.enableZapMode();
                  fin.setInsertMode(false);
               }

               // request focus will silently fail here, but we do need to
               // reset the cursor and state as if we just entered the field
               // so force activation here
               Toolkit.postEvent(new FocusEvent(AWTEvent.FOCUS_GAINED, comp));
               comp.repaint();
            }
            else
            {
               // focus had already changed, change the focus back
               widget.requestFocus();
            }

            bell();
            return false;
         }
      }
      
      return true;
   }
   
      
   /**
    * Validates the widget's ID by comparing with array of allowed IDs.
    * 
    * @param   id
    *          widget id to validate
    * @param   ids
    *          array of allowed ids or <code>null</code
    * @return  equals the input id or <code>-1</code>
    */
   private int validateWidget(int id, int[] ids)
   {
      if (ids == null)
      {
         return id;
      }

      for (int i = 0; i < ids.length; i ++)
      {
         if (id == ids[i])
         {
            return id;
         }
      }

      return -1;
   }

   /**
    * Initiates direct widget drawing.
    * Initiating drawing includes calculation and setup of a proper clipping
    * rectangle that defines the scroll viewport.
    * 
    * @param   frame
    *          the frame where widgets being drawn belong
    *
    * @return  the previously established screen state
    */
   private ScreenBitmap widgetDrawingOn(Frame frame)
   {
      Component content = frame.getContentPane();
      Point origin = content.getLocationOnScreen();
      Component viewport = content.getParent();
      Dimension extent = ((JViewport)viewport).getExtentSize();
      Rectangle r  = new Rectangle(origin, extent);
      
      // frame bounds are (practically) in screen coordinates
      Rectangle bounds = frame.getBounds();
      r = r.intersection(bounds);
      return tk.getAndSetBitmap(r);
   }

   /**
    * Terminates direct widget drawing.
    * Terminating drawing includes restoring the clipping rectangle and
    * syncing the terminal so that the drawing becomes visible.
    * 
    * @param   bitmap
    *          the previously established screen state, returned from the
    *          <code>widgetDrawingOn()</code> call
    */
   private void widgetDrawingOff(ScreenBitmap bitmap)
   {
      tk.setBitmap(bitmap);
      tk.sync();
   }

   /**
    * Changes the highlighted state of the widget and repaints it.
    *
    * @param    widget
    *           The widget to modify.
    * @param    state
    *           The new state of the widget's highlight.
    */
   private void highlight(final Component widget, final boolean state)
   {
      if (widget != null)
      {
         // change the highlighted state of the widget
         eventDrawingBracket(false, false, new Runnable()
         {
            public void run()
            {
               widget.setHighlighted(state);
               getCurrentFocus().repaint();
            }
         });
      }
   }

   /**
    * Save insert mode before using a prompting function or a trigger.
    *
    * @return  Insert mode that was saved.
    */
   private int saveInsState()
   {
      int savedInsState = Window.getInsertMode();

      // save insert mode in order to make the next focused fill-in accept it
      savedInsStates.push(savedInsState);

      return savedInsState;
   }

   /**
    * Restore insert mode after a prompting function or a trigger.
    *
    * @param   savedInsState
    *          Insert mode that was before the call of the corresponding
    *          prompting function or trigger.
    * @param   force
    *          Whether we have to repaint the indicator even if the new mode
    *          is the same as the old one.
    */
   private void restoreInsState(int savedInsState, boolean force)
   {
      // restore insert indicator state
      if (savedInsState != -1)
      {
         // skip insert mode adjustment for editors
         Component focus = getCurrentFocus();
         if (focus == null || !(focus instanceof Editor))
         {
            Window.setInsertMode(savedInsState, force);
         }
      }
      savedInsStates.pop();

      // replace the top stack value with the restored insert mode in order to
      // make the next focused fill-in accept it
      savedInsStates.pop();
      savedInsStates.push(savedInsState);
   }

   /**
    * Set new value for the <code>needPause</code> flag.
    * 
    * @param   needPause
    *          New value for the flag.
    */
   void setNeedPause(boolean needPause)
   {
      if (needPause == false)
      {
         screenCurrent.resetNeedPause();
         messageNeedPause = false;
      }

      this.needPause = needPause;
   }

   /**
    * Get current value of the <code>needPause</code> flag.
    * 
    * @return  Returns the needPause.
    */
   boolean isNeedPause()
   {
      return needPause || screenCurrent.isNeedPause();
   }

   /**
    * Returns <code>true<code> if <code>enable(..)<code> function is being
    * run.
    *
    * @return  See above.
    */
   boolean isInEnable()
   {
      return inEnable;
   }
   
   /**
    * Returns saved insert indicator state that should be applied to the next
    * focused fill-in, or <code>-1</code> if there is no state to apply.
    *
    * @return  See above.
    */
   int getSavedInsState()
   {
      return savedInsStates.peek();
   }

   /**
    * Clears saved insert mode.
    */
   void clearSavedInsState()
   {
      savedInsStates.pop();
      savedInsStates.push(-1);
   }

   /**
    * Get the stream to which the terminal is redirected or <code>null</code>
    * if no redirection is currently in effect.
    *
    * @return   The current redirection or <code>null</code> if not
    *           redirected.
    */
   Stream getCurrentRedirection()
   {
      if (!isRedirected())
         return null;
      
      return sd.streamFromId(currentStream);
   }
   
   /**
    * Display frame content in redirected mode by page element.
    * 
    * @param   frameId
    *          Frame ID.
    */
   void view(int frameId)
   {
      if (getWidget(frameId) == null)
         return;
      
      ScreenBuffer frameBuf = server.refresh(frameId);
      
      Frame topFrameIn = screenCurrent.getTopFrame(true); 
      
      // this is an equivalent of view call for page headers/footers
      view_(frameId, frameBuf, null, true, false);
      
      Frame topFrameOut = screenCurrent.getTopFrame(true); 
      
      if (topFrameIn != topFrameOut)
         topFrameOut.moveToBottom();
   }
   
   /**
    * Execute the NEXT-FRAME or PREV-FRAME event, depending on flag. Frame
    * search starts from the specified frame. 
    * 
    * @param   frame
    *          The frame from which search will start.
    * @param   nextFrame
    *          <code>true</code> if this is the NEXT-FRAME event, <code>false
    *          </code> if this is the PREV-FRAME event.
    */
   void handleNextPrevFrameEvents(Frame frame, boolean nextFrame)
   {
      Frame newFrame = null;

      // handle NEXT/PREV frame events
      if (nextFrame)
      {
         newFrame = screenCurrent.getNextEnabledFrame(frame);
         if (newFrame == null)
         {
            // maybe no other enabled frames in this direction, get first
            // enabled and visible frame, in z-order
            newFrame = screenCurrent.getNextEnabledFrame(null);
         }
      }
      else
      {
         newFrame = screenCurrent.getPrevEnabledFrame(frame);
         if (newFrame == null)
         {
            // maybe no other enabled frames in this direction, get last
            // enabled and visible frame, in z-order
            newFrame = screenCurrent.getPrevEnabledFrame(null);
         }
      }
      
      
      if (newFrame != null && newFrame != frame)
      {
         // first enabled widget in new frame
         Component comp  = getWidget(getFirstFocusableId(newFrame));
         // get current focus
         Component focus = getCurrentFocus();
         // get next enabled widget in frame
         Component next  = getNextEnabledWidget(frame, focus, FOCUS_WRAP);

         if (!focus.isEnabled())
         {
            // if focus is disabled and there is another enabled widget in
            // its parent frame, set focus on it; else, focus first enabled
            // widget in new frame
            if (next != null)
            {
               setFocusOn(next);
            }
            else
            {
               setFocusOn(comp);
            }

            focus = next;
         }
         
         if (focus != null)
         {
            // current focus receives leave first
            sendLeave(focus);
   
            // send LEAVE to frame and stop processing, if it doesn't accept 
            // LEAVE
            if (frame != focus)
            {
               sendLeave(frame);
               // if NO-APPLY, stop further processing
               if (triggerResultNoApply.peek())
               {
                  return;
               }
            }
         }

         // order of events is: ENTRY to frame first, ENTRY to widget second
         sendEntry(locateFrame(comp));
         sendEntry(comp);
      }
   }
   
   /**
    * Checks whether the frame needs switching streams and does switching.
    * The last used stream for the frame is compared with the current stream
    * in effect. If these are different streams, the switching will occur.
    *  
    * @param   frame
    *          the target frame for the operation
    * @return  <code>true</code> if the stream switch has happened and it was
    *          not the very first switch for the frame
    */
   private boolean switchStreams(Frame frame)
   {
      int oldid = -1;
      int newid = -1;
      Stream olds = null;

      if (isRedirected())
      {
         newid = currentStream;
      }

      if (frame.isRedirected())
      {
         oldid = frame.getStreamId();
         olds = sd.streamFromId(oldid);
      }

      if (oldid == newid)
      {
         return false;
      }

      // flush the old stream
      if (oldid != -1)
      {
         Integer key = new Integer(oldid);
         RedirectedTerminal rt = redirections.get(key);

         // is the stream still around?
         if (rt != null)
         {
            // flush the terminal and stream buffers
            if (rt.flush())
            {
               try
               {
                  // only done if there was pending RT output so make sure
                  // that headers are bypassed
                  olds.flushControlled(true);
               }
               
               catch (IOException ioe)
               {
                  ErrorManager.showErrorAndAbend(140,
                                                 Stream.PIPE_BROKEN_MESSAGE);
               }
            }
         }
      }

      // note the new stream
      if (newid == -1)
      {
         frame.setRedirected(false, -1);
      }
      else
      {
         frame.setRedirected(true, newid);
      }

      // normally, at this point, the method should return true meaning
      // there was a real stream switch but the very first switch for the
      // frame should be excluded and this what the use of the streamed flag
      // provides
      return frame.isStreamed();
   }

   /**
    * Checks whether the stream and its redirected terminal need to switch to
    * another frame and does switching if so.
    * <p>
    * Switching the stream between frames causes flushing the buffers of the
    * redirected terminal to the stream. Otherwise the redirected terminal
    * receives continuous output from the same frame.
    *  
    * @param   streamId
    *          ID of the output stream the current output is redirected to.
    * @param   frameId
    *          ID of the frame about to produce some output
    * @return  <code>true</code> when the frame switch has happened or
    *          <code>false</code> for the case of continuous output.
    */
   private boolean switchFrames(int streamId, int frameId)
   {
      Stream             stream = sd.streamFromId(streamId);
      RedirectedTerminal rt     = redirections.get(streamId);
      Frame              frame  = (Frame) getWidget(frameId);

      if (stream == null || rt == null || (frame == null && frameId != -1))
      {
         return false;
      }

      // save the fact of whether there is something to flush or not since
      // the targetFrame() call will cause internal flushing and does not
      // return an indication of whether flushing was needed or not;
      boolean bypass = !rt.isFresh();
      
      int flushId = rt.targetFrame(frameId);
      
      boolean flushed = flushId != frameId && flushId != -1;
      
      if (flushed)
      {
         // redirected terminal has changed ownership and flushed buffers
         // flush the stream now
         flushRedirected(null, bypass);

         // the previous owner of the redirected terminal needs to be notified
         // this notification removes the stream from the list of streams that
         // have frame headers output and also hides the frame.
         // these are exactly the actions that are taken when the page break
         // occurs and that't why the notifyPageBreak() is reused here 
         Frame gone = (Frame) getWidget(flushId);
         
         if (gone != null)
         {
            FrameConfig cfg = (FrameConfig) gone.getRef();
            
            boolean top = cfg.getPageTop();
            boolean bot = cfg.getPageBottom();
            
            if (!top && !bot)
            {
               gone.getFrameHelper().notifyPageBreak(stream);
            }
         }
      }
      else if (flushId == frameId)
      {
         return false;
      }

      return true;
   }
   
   /**
    * Push any pending output from the redirected terminal to the stream and
    * then flush the stream.
    *
    * @param    out
    *           The stream being flushed.
    */
   public void flushStream(Stream out)
   {
      int sid = sd.idFromStream(out);

      boolean switched = currentStream == sid;
      int tempOut = -1;

      if (!switched)
      {
         tempOut = switchOutput(sid, true);
      }

      RedirectedTerminal rt = redirectedTerminalForStream(out);

      // flush the redirected terminal buffer to the stream 
      boolean bypass = rt.flush();

      // flush the stream buffer
      try
      {
         out.flushControlled(bypass);
      }
      
      catch (IOException ioe)
      {
         ErrorManager.showErrorAndAbend(140, Stream.PIPE_BROKEN_MESSAGE);
      }

      finally
      {
         if (!switched)
         {
            switchOutput(tempOut, false);
         }
      }
   }
   
   /**
    * Force the given stream to output any buffered PUT or EXPORT content, so
    * long as the stream is not the terminal.
    *
    * @param    sid
    *           The stream on which to call <code>endOutput</code>.
    */
   private void endOutput(int sid)
   {
      if (sid != -1)
      {
         Stream stream = sd.streamFromId(sid);
         
         if (stream != null)
         {
            // when ending output, a NL must be added for the PUT statement
            // the EXPORT statement has this NL added after its execution
            stream.endOutput(true);
         }
      }
   }

   /**
    * Flushes all buffered output to the redirected output stream (does
    * nothing if output is not redirected).  This call may cause recursive
    * calls to {@link #view} in order to write page headers or footers.
    *
    * @param    stream
    *           The stream to flush or <code>null</code> to flush the
    *           current redirection destination.
    * @param    bypass
    *           <code>true</code> if page header processing must be bypassed
    *           when flushing the output. This should be <code>true</code>
    *           if the caller has pushed pending redirected terminal output
    *           into the stream since this will have already processed the
    *           page headers and this will avoid duplicates.
    *
    * @return   The next line number AFTER all buffered lines are output but
    *           BEFORE any final page break processing has occurred. This can
    *           be compared with the next line number as queried after this
    *           method returns to determine if a page break was output or
    *           not (among other uses). If there was no output or if the
    *           terminal is not redirected, then -1 will be returned.
    */
   private int flushRedirected(Stream stream, boolean bypass)
   {
      int rc = -1;
      
      if (isRedirected())
      {
         if (stream == null)
         {
            stream = sd.streamFromId(currentStream);
         }
         
         if (stream == null)
         {
            throw new IllegalStateException("Invalid stream id " +
                                            currentStream + ".");
         }
         
         try
         {
            rc = stream.flushControlled(bypass);
         }
         
         catch (IOException ioe)
         {
            ErrorManager.showErrorAndAbend(140, Stream.PIPE_BROKEN_MESSAGE);
         }
      }
      
      return rc;
   }
   
   /**
    * Changes the value of a widget in a frame.
    *
    * @param   frame
    *          <code>Frame</code> where the widget belongs
    * @param   comp
    *          <code>Component</code> for the widget
    * @param   config
    *          <code>ComponentConfig</code> for the widget
    * @param   value
    *          new value
    * @param   fmt
    *          format if formatted value
    */
   private void setChangedValue(Frame           frame,
                                Component       comp, 
                                ComponentConfig config,
                                BaseDataType    value,
                                String          fmt)
   {
      // force the new value into the widget
      if (frame != null)
      {
         // handles any down frame processing as needed
         frame.getFrameHelper().setValue(comp, value, fmt);
      }
      else
      {
         if (comp instanceof FillIn)
         {
            // honor any format string override
            FillIn fi = (FillIn) comp;
            fi.setValue(value, fmt);
         }
         else
         {
            ((DataContainer) comp).setValue(value);
         }
         
         config.setState(ScreenBuffer.UNCHANGED);
      }
   }

   /**
    * Get reference to active frame.
    * 
    * @return  Reference to active frame or <code>null</code> if there is no
    *          such frame.
    */
   private Frame getActiveFrameInt()
   {
      Frame frame = screenCurrent.getTopFrame(false);
      
      if (frame != null)
         return frame;

      if (lastViewed == null      || 
          !lastViewed.isVisible() || 
          !lastViewed.isAlive())
      {
         return null;
      }
      
      return lastViewed;
   }

   /**
    * Stop the WAIT-FOR loop, as it determined there are no available widgets,
    * in the following cases:
    * <ul>
    *    <li>when the terminal is reset, inside an editing block</li>
    *    <li>
    *       when the WAIT-FOR loop is NOT waiting for global events and there
    *       are no enabled components registered for exit events
    *     </li>
    * </ul>
    * 
    * @throws  StopConditionException
    *          always thrown - stop all other processing
    */
   private void stopNoAvailableWidgets()
   {
      String err = "None of the widgets used in the WAIT-FOR " + 
                   "statement are in a state (such as SENSITIVE) " + 
                   "such that the specified event can occur.  " +
                   "WAIT-FOR terminated. (4123)";
      messageNeedPause = false;
      message(err);
      
      try
      {
         pause(-1, textPBH, false);
      }
      catch (Throwable t)
      {
         // do nothing
      }
      
      // all other processing fails
      throw new StopConditionException(err);      
   }
   
   /**
    * Outputs text as the result of PUT SCREEN operation.
    * 
    */
   private void putScreenDirect(String text, double row, double column, 
                                int color)
   {
      int effRow = (int)(row - 1);
      int effCol = (int)(column - 1);

      boolean curvis = true;
      int cpx = 0;
      int cpy = 0;
      
      if (inputActive > 0)
      {
         // the cursor is active in an edit operation, remember the current
         // cursor position (and make the cursor invisible to avoid flashing)
         cpx = directOutput.getX();
         cpy = directOutput.getY();
         curvis = tk.setCursorStatus(false);
      }
         
      // honor the default color
      color = (color == -1) ? ColorSpec.COLOR_NORMAL : color;
      
      P2JColor clr = ColorTable.getColor(color);

      boolean res = directOutput.setCursor(effCol, effRow);
      
      if (res)
      {
         directOutput.addString(text, clr.getAttribute(), clr.getColor());
      }

      if (inputActive > 0)
      {
         directOutput.setCursor(cpx, cpy);
      }
         
      directOutput.sync();

      // set the cursor visibility
      tk.setCursorStatus(curvis);
   }

   /**
    * Positions the real cursor before pausing according to the focused widget
    * "directions".
    */
   private void positionCursor() 
   {
      if (isRedirected())
      {
         return;
      }

      // cursor parking position
      Point p = new Point(screenSize.width - 2, screenSize.height - 1);

      Component comp = getCurrentFocus();
      if (comp != null && !comp.isEnabled())
      {
         comp = null;
      }

      // reposition the cursor before blocking (interactive mode only)
      if (comp != null)
      {
         Frame frame = (Frame)locateFrame(comp);

         if ((frame != null && !frame.isInChoose(comp)) ||
             comp instanceof List) 
         {
            if (!skipCaretDraw)
            {
               if (comp instanceof P2JComponent) 
               {
                  ((P2JComponent)comp).drawCaret();
               }
               else if (comp instanceof ComboBox)
               {
                  ((ComboBox)comp).drawCaret();
               }
               else if (comp instanceof RadioButton)
               {
                  ((RadioButton)comp).drawCaret();
               }
               else if (comp instanceof List)
               {
                  ((List)comp).drawCaret();
               }
               else
               {
                  return;
               }
            }
            else
            {
               // reset the flag
               skipCaretDraw = false;
            }

            // get the last cursor position as drawCaret() dispatched it
            tk.getCursor(p);
         }
      }

      // make the cursor stay at this position
      tk.setCursorStay(p.x, p.y);
   }

   /**
    * Restore current status type and text from stack.
    * 
    * @param    cnt
    *           How many elements should be popped from stack.
    */
   private void popStatus(int cnt)
   {
      StatusData storedStatus = null;
      
      for (int i = 0; i < cnt; i++)
      {
         if (statusStack.isEmpty())
            continue;
         
         storedStatus = statusStack.pop();
      }
      
      if (storedStatus == null)
         return;

      final int type = storedStatus.type; 
      
      independentEventDrawingBracket(new Runnable()
      {
         public void run()
         {
            // the whole status line will lose PUT SCREENs
            tk.refresh(Window.getStatusLineRectangle());
      
            String msg = "";
      
            switch (type)
            {
               case DEFAULT:
                  msg = defaultMessage;
                  break;
               case INPUT:
                  msg = inputMessage;
                  break;
               case HELP:
                  msg = helpText;
                  break;
            }
      
            if (msg == null)
            {
               // TODO: Check if this is correct, probably there is more complicated
               // fallback logic in cases when some status text is null. 
               msg = "";
            }
            
            Window.status(msg);
         }
      });
      
      statusType = storedStatus.type;
      inputActive = storedStatus.input;
   }

   /**
    * Save current status type and text in stack.
    * 
    * @param    cnt
    *           How many times state should be saved in stack.
    */
   private void pushStatus(int cnt)
   {
      for (int i = 0; i < cnt; i++)
      {
         statusStack.push(new StatusData(statusType, inputActive));
         inputActive = 0;
      }
      
      if (statusType != DEFAULT)
         status(null);
   }
   
   /**
    * Container for the flags related to trigger processing.
    */
   private static class TriggerResult
   {
      /** <code>true</code> if a trigger was found and executed. */
      public boolean executed = false;
      
      /** <code>true</code> sets <code>NO-APPLY</code> for the trigger. */
      public boolean consume = false;
   }

   /**
    * Status data container.
    */
   private static class StatusData
   {
      final int type;
      final int input;
      
      /**
       * Constructor.
       * 
       * @param type
       * @param input
       */
      public StatusData(int type, int input)
      {
         this.type = type;
         this.input = input;
      }
   }
}
